;; @(#) tiny-gnus-sig.el -- change sig according to GNUS Article content

;; @(#) $Id$
;; @(#) $Keywords: news $

;; Copyright (C) 1995,96 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@ntc.nokia.com>
;; Maintainer:   Jari Aalto <jari.aalto@ntc.nokia.com>
;; Created:      Jun 18 1995
;;

;; This file is *NOT* part of GNU emacs
;; This code is free software in terms of GNU Gen. pub. Lic. v.2 or later

;;; Commentary:

;; Feb 10 1996
;; - This file is not currently under development. I just answered to
;;   someone's post where he asked how to create custom signatures
;;   inside GNUS.
;;
;; - Offered as is, haven't looked this for long time. Send me a
;;   mail if this doesn't work.

;;; Change Log:

;;; Code:

(require 'advice)
(require 'gnus)				;variable gnus-signature-file

(if (not (boundp 'gnus-signature-file))
    (error "Only works in GNUS 4.x, Emacs 18.28 or lower."))

(defvar my-gnus-summary-post-news-hook nil
  "*advice added: hook that runs after gnus-summary-post-news
has stopped composing article to be posted.")

(defvar my-gnus-signature-file-default "~/sig/gnus-default.sig"
  "*Default signature file")



;; - Note: The word "newsgroups" is on the  message only if you
;;   use "fF" to compose message, if you use "rR" there will not be
;;   any references --> we peek "*Article* buffer
;;
;;
(defvar  my-gnus-signature-list
  '(("groups.*emacs"              "~/sig/emacs.sig")
	("groups.*lang[.]c[+]"    "~/sig/c++.sig")
	("groups.*lang[.]c"       "~/sig/c.sig")
	)
  "*List of \(\(REGEXP SIGNATURE-FILE\) ..\) The buffer is searched from
the beginning of file for RE and first one matched is picked.")


;;; ----------------------------------------------------------------------
;;;
(defadvice gnus-summary-post-news (after my-gnus-post-news act)
  "Run my-gnus-summary-post-news-hook after  gnus-summary-post-news has
finished composing message."
  (run-hooks 'my-gnus-summary-post-news-hook))



;;; ----------------------------------------------------------------------
;;;
(defun my-gnus-sig-set ()
  "Changes signature file according to my-gnus-signature-list if none
matches uses my-gnus-signature-file-default

References
  gnus-signature-file
"
  (let* ((ptr	    my-gnus-signature-list)
	 (def	    my-gnus-signature-file-default)
	 (file-var  'gnus-signature-file)	;ByteComp silencer
	 el re file
	 sig
	 )

	;;  The "Article" will tell use what newsgroup we're reading
	(save-excursion
	  (if (null (set-buffer "*Article*")) nil
		(goto-char (point-min))
		(while ptr
		  (setq el   (car ptr)
				re   (nth 0 el)
				file (nth 1 el))
		  (if (null (re-search-forward re nil t))
			  (setq ptr (cdr ptr))
			(setq ptr nil  sig file)))))

	(if sig nil
	  (setq sig def))

	(if (null file-var)
	    (setq file-var file-var)) ;ByteComp silencer, no-op

	(set file-var sig)
	))


;;; ----------------------------------------------------------------------
;;;
(defun my-gnus-setup ()
  "Gnus settings when composing message"
  (my-gnus-sig-set)
  )


;;; ----------------------------------------------------------------------
;;; - installing the package
;;;
;;; - note: 'setq' clears all other funcs
;;; - if you have something special here, use add-hook instead.
;;;
;;; - There exists 'news-inews-hook' which is run just before message is
;;;   is sent, but it's undocumented. Either way you could replace all
;;;   these settings with that hook. The problem is that you can't
;;;   experiment with that hook so easily: it may send your message, while
;;;   these hooks will set things after message is _prepared_ --> you have
;;;   plenty of time to check that all new thigs went ok.
;;;

(add-hook 'mail-setup-hook                'my-gnus-setup)
(add-hook 'news-reply-header-hook         'my-gnus-setup)
(add-hook 'my-gnus-summary-post-news-hook 'my-gnus-setup)

(provide 'tiny-gnus-sig)

;;; tiny-gnus-sig.el ends here
