;; @(#) tinybm.el -- Bookmark package, keep your code organized

;; @(#) $Id: tinybm.el,v 1.18 1997/06/19 10:37:34 jaalto NotReleased $
;; @(#) $Keywords: X , bookmarks , organize buffer content $
;; $KnownCompatibility: 19.28, XE 19.14 $
;; $OutlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1995,1996,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Feb 22 1995
;;
;; To get information on this program use ident(1) or do M-x tibm-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinybm|Jari Aalto|jari.aalto@poboxes.com|
;; Add handy section breaks, keep your code organized, imenu compatible|
;; 19-Jun-1997|$Revision: 1.18 $|~/misc/tinybm.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Installation

;; ........................................................ &t-install ...
;; Put this file on your Emacs-Lisp load path, add following into your
;; ~/.emacs startup file
;;
;;* _
;;*      	(require 'tinybm)
;;
;; or use autoload, your emacs starts up faster, prefered.
;; All these commands can easily be ripped of with
;; tinylib.el / ti::m-pkg-rip-magic. Select area and call it.
;;
;;* _
;;* 	(autoload 'tibm-insert		"tinybm" t t)
;;* 	(autoload 'tibm-repeat		"tinybm" t t)
;;*     (autoload 'tibm-parse		"tinybm" t t)
;;* 	(autoload 'tibm-forward		"tinybm" t t)
;;* 	(autoload 'tibm-backward	"tinybm" t t)
;;*
;;* 	(cond
;;* 	 ((ti::xe-window-system)
;;*	  (autoload 'tibm-mouse		"tinybm" t t)
;;*       (autoload 'tibm-mouse-parse	"tinybm" t t)
;;*       )
;;*      (t
;;*       (autoload 'tibm-keyboard	 "tinybm" t t)
;;*       (autoload 'tibm-keyboard-parse "tinybm" t t)
;;* 	  ))
;;* _
;;
;; Probably you want to use 'M-x bm' for quick bookmark command, so
;; add this too:
;;
;;* _
;;* 	(defalias 'bm  'tibm-insert)
;;*     (autoload 'bm  "tinybm" t t)
;;* _
;;
;; And suggested keybindings
;;
;;* _
;;*     ;;  This is for X. It brings up nice popup menu
;;*     ;;
;;*     (global-set-key [?\e C-mouse-1]	 'tibm-mouse)
;;*     (global-set-key [?\e C-S-mouse-1] 'tibm-mouse-parse)
;;*
;;*     ;; And for keyboard call if you don't have X
;;*     ;; 's' for selections
;;*     ;; 'p' for parse
;;*     ;; 'b' for inserting (b)ookmark
;;*     ;;
;;*     (use-prefix-key global-map "\C-z")
;;*	(global-set-key "\C-zbs"	    'tibm-keyboard)
;;*	(global-set-key "\C-zbp"	    'tibm-keyboard-parse)
;;*	(global-set-key "\C-zbb"	    'tibm-insert)
;;*
;;* _
;;*     ;;  Keyboard users can move between bookmarks fast
;;*     ;;  with these
;;*     (global-set-key [S-left]	'tibm-backward)
;;*	(global-set-key [S-right]	'tibm-forward)
;;* _
;;
;; BE SURE THAT
;;
;;      you have defined comment syntax, otw the inserted field
;;      won't have proper prefix + endings
;;      [for transparent comment handling, see tinycom.el]
;;
;; If you have any questions, use this function
;;
;;      M-x tibm-submit-bug-report



;;}}}
;;{{{ Documentation

;;; .................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "tibm-" in front of
;;   every function & variable. It stands for '(ti)ny (b)ook (m)arks.
;; - variable names contain letters 'tibm-:', except version/hook vars.

;;  Preface, feb 1995
;;
;;	I started programming with 18.57 emacs and it took long time
;;	to get used to 19.28. I think that was just because of all changes
;;	needed to be done; converting my huge .emacs.XXX files for 19.28
;;	and to learn new features.
;;
;;	Back then I used little function I wrote for myself that inserted
;;	section breaks, those that I call 'bookmarks'.  I also used
;;	folding.el to keep the code in separate sections. Findings things
;;	was easy when you just searched either bookmarks or jumped
;;	between folds.
;;
;;	Recently I found imenu.el which provided X-popup for my bookmarks
;;	and I immediately started writing this .el
;;
;;  Overview of features
;;
;;	o   Provides some 'setting bookmarks' functions: adding
;;	    repeated characters and sequences up till end of line with
;;	    named identifier. (like breaks in this file)
;;	o   Automatically parses bookmarks from file, if it contains
;;	    RCS identifier 'bookMarkRegexp' which defines bookmark syntax for
;;	    the file. Uses X-popup [imenu] for showing those bookmarks and
;;	    moving between them.
;;
;;  Read.me -- how to keep your files organized
;;
;;	There are several tools to keep your code organized and they are
;;	at their best if you think how they can co-operate. For
;;	example I have chosen to use folding.el and tinybm.el, which
;;	might seem to do double job, since they both divide code into
;;	more easily manageable sections.
;;
;;	The key point is that when I use folding, I work _within_ some
;;	special section and I want to hide all the rest of the code. But
;;	when I want to jump easily back and forth on the buffer, I
;;	usually 'unfold' all and use tinybm to find bookmarks.
;;
;;	Now, to confse you more, there is also 'imenu.el' which I use too
;;	for jumping inside code. I have configured it so that it will
;;	pick all function names inside list, and when I want to go to
;;	specific function, I just pick one from imenu.
;;
;;	To summarize:
;;
;;	o   folding.el     -- for hiding unneeded code,
;;			      clear view on the structure
;;	o   tinybm.el      -- Jumping between/finding  _large_ code sections
;;	o   imenu.el       -- finding specific funtion, more detailed control.
;;	o   tinyhotlist.el -- Add/remove files from permanent X-popup list
;;
;;
;;  How to use this package
;;
;;	There is following function that inserts bookmark on the current line
;;
;;	    tibm-insert
;;
;;	There is also normal repeat function, that fills line with your
;;	pattern:
;;
;;	    tibm-repeat
;;
;;	Normally the most usual bookmark separator is the "." <dot> , which
;;	isn't so 'noisy' as continuous '-' line. Normally I add some unused
;;	ID character, like '&' at front of real bookmark, like this:
;;
;;	    ;;; .................................. &How-to-use ...
;;	    (defun test ()
;;	     (progn
;;	      ..
;;            (goto-char ..
;;            ;; ^^^^^^^^^^^^^^^^^^^^^^^ sepratorInsideCode ^^^
;;
;;	The 'How-to-use' is bookmark, because it has '&' on it, whilst
;;	the latter isn't -- it is used inside code to make it more
;;	readable and I don't want to include it into imenu.
;;
;;
;;  About the bookmark identifier naming
;;
;;	When you name the breaks, keep in mind that when identifiers are
;;	sorted, the ones that start with big letters A-Z show up first,
;;	and a-z next. Allthougt it would be convenient to have all
;;	subwords in capital, it is usually better to start with lowercase
;;	letter, because it's easily unintentinally mix up/down case
;;	letters. Besides you have to reah out for shift to have uppercase :-/
;;
;;	    ............. breakName ...         ;prefered, starting low
;;          ............. BreakName ...         ;watch out for mixed case!
;;
;;	it is also adviced that you choose some common beginning for the
;;	identifier, so that they get sorted nicely. If you define
;;	variables at the beginning of file it might be good idea to
;;	attach beginning letter like 'v-' for varibles before the real
;;	identifier name begins, like:
;;
;;	    ............. v-globals ...
;;	    ............... v-hooks ...
;;
;;	Of course, we can now use the uppercase letter trick to have them
;;	sorted fist in the list :-/, just change 'v-' with 'V-'.
;;	Generally you should think which ones do you use most, do you
;;	leave the variables alone when you have defined them and mostly
;;	work with new functions ? Then the variables can stay at the end
;;	of list and there is no need for 'V-' trick. but if you need to
;;	access variables often, then you might want to see variables
;;	first in the list. It's up to your decision how you name the
;;	variables and how you want to see them listed.
;;
;;  Breaks and sub-break naming
;;
;;	If you have very large file, you'll propably need major breaks,
;;	level one breaks and possibly level 2 breaks too. To keep the
;;	list well sorted, put the functions into bigger groups and name
;;	the sub-level breaks so that they have some common beginning in
;;	respect to the major break they belong to. Let's see an example
;;	where you're dealing with mail handling. Notice the CAPITAL letter.
;;
;;	    ;; ################################# &h-Header ###
;;	    ;;  this is beginning block of header handling
;;
;;	    ;; ..................................... &h-cc ...
;;	    ;;  Some special function here to handle CC
;;	    ;;  field, killing all recipients, or only
;;	    ;;  some of them
;;
;;	    ;; .. .. . .. . .. . .. . .. . .. . .. . .. . .. .
;;	    ;;  More detailled functions under h-cc, Not
;;	    ;;  named, because there is only 2 small funcs
;;	    ;;  easily found.
;;
;;	Again there is couple of point's to follow here. All the tricks
;;	are discussed already: the 'Big' letter trick put's major break
;;	to the top of imenu list, common beginning keeps the subsections
;;	together.
;;
;;	If you want to see very complex, live example, download my mail
;;	reply package which works with gnus and Ding and should work with
;;	any emacs mailer with couple of adjustments. [tinyreply.el]
;;
;;  Example breaks
;;
;;	Some bokmark breaks are proposed here, but you can use
;;	whatever you like. Thumb of rule: be consistent, always use
;;	same convention in your files and consider the 'noisy level' of your
;;	breaks, so that they build up nicely and the code is easy to
;;	read. Too many _different_ breaks is not good idea, because they
;;	clutter the view fast, instead use variations on a theme: same break
;;	character but varying spaces and continuous character lengths.
;;
;;	Thumb rule: select 1-3 break chars, and never change them in you
;;	files; your files look alike. Vary the spacing, not the
;;	break chars.
;;
;;      These are 'noisy breaks' , Major section separators, pick only one
;;      and use it in your files, do not use all three!
;;
;;	    ##############################################################
;;	    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
;;	    ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
;;
;;      less noisy breaks
;;
;;	    .`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`
;;
;;	    .^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^
;;
;;	    .:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:.:
;;	    .~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~
;;
;;
;;	This is subSection break
;;
;;	    ................................................................
;;
;;
;;      This is even lighter subsection break (varying spacing)
;;
;;	    ...  ...  ...  ...  ...  ...  ...  ...  ...  ...  ...  ...  ...
;;
;;      'Draw one's attention' break: something special in this section
;;
;;
;;	    --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++--
;;
;;      Internal break 1, inside function, long C++-'switch' begin etc.
;;
;;	    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;;
;;	Internal break 2, to separate long case elements etc.
;;
;;
;;	    ^^^  ^^^  ^^^  ^^^  ^^^  ^^^  ^^^  ^^^  ^^^  ^^^  ^^^  ^^^  ^^^
;;
;;	Bookmark cache
;;
;;	So that imenu works fast, it is not desirable that the breaks are
;;	always parsed from scratch, because it takes time to scan the
;;	file for possible bookmarks. That's why the information is
;;	cached. If the break cache is empty, the breaks are gathered from
;;	buffer and stored to the cache and when you call the imenu, the
;;	cache is offered to it --> fast response time.
;;
;;	When you add new breaks to the buffer [especially at the
;;	beginning of code development], you may want to call function
;;
;;	    tibm-parse
;;
;;	which will empty the cache and re-read all bookmarks.
;;
;;	If you write lot of code the points that were cached do no longer
;;	represent exact points of bookmarks, because they have been
;;	sliding off their places.
;;
;;	If you want *always* have updated bookmark points, there is variable
;;
;;	    tibm-cache-update
;;
;;	which you can set to 'always, if you want the cache to be updated
;;	always prior showing X-menu. In large buffer this remarkably
;;	slows down the menu appering. See variable for more choices.
;;
;;  Automatic bookmark detection
;;
;;	In order bookmarks to be detected in file, you must define following
;;	RCS identifier [see ident(1)] preferably at the
;;	beginning of your file:
;;
;;	    $BookMarkRegexp:<space>'REGEXP'<space>$
;;
;;	Be careful so that the identifier is _exactly_ in this form:
;;	pay attention to spaces and (') around the REGEXP. The regular
;;	expression tells what line can be considered as bookmark and
;;	the bookmark name is indicated in subexpression 1 [\\(.*\\)] ,
;;	look at this file, how it is constructed.
;;
;;	In order to find all bookmarks and build up the cache, it needs
;;	to widen the buffer in case the file is narrowed with some
;;	folding or outline editor. When the cache has been built the
;;	buffer's narrowing is restored, so you shouldn't even notice
;;	this. Of course you don't want to find bookmarks from your RMAIL
;;	file, don't you? :(
;;
;;	One word about the regexp contruction, let's see my regexp that
;;	matches the identifier:
;;
;;	    &+\\([^ ]+\\)
;;
;;	Pay attention to using exclusive regexp, not just '.*'
;;	construction. When you use folding or outline editor the '.*'
;;	form is very ill behaving, because if the line being scanned
;;	is currently folded, IT WILL MATCH WHOLE folded section --> your
;;	identifier surely isn't that one. We can't unfold the sections
;;	during scanning, because if there are subfolds, what editor is on
;;	use .. it's too complex/slow to handle such situations. But using the
;;	exclusive list [^ ] will surely match the identifier, because it
;;	stops when it can find first space. This means that
;;	you can't use _spaces_ inside the identifiers. Cat the words together.
;;
;;  If the BookMarkRegexp isn't defined in file
;;
;;	Then the programs tries to search for the default bookmarks.
;;	See function tibm-re-default for more.
;;
;; Empty cache. Building...
;;
;;	Do you wonder why you get this message displayed, while you were
;;	sure that you the buffer had cache already? Don't be suprised.
;;	This is totally normal behaviour: whenever you switch mode for the
;;	buffer the new mode _kills_ all local variables, including this
;;	cache information. Obviously the information must be restored
;;	when you call the hotlist again.
;;
;;	I could have programmed the cache to be intern variable with
;;	buffer name added to it, but what happens when you kill the
;;	buffer? The variable lies around in emacs with no use. [oh yeah,
;;	there is buffer kill hooks to take care of that... but let's be
;;	simple here]. Keeping track of the interned variables is not what
;;	I had in mind: the more simpler, the better and the easier to
;;	maintain. You just have to live with those messages :-/
;;
;;  About imenu
;;
;;	You definitely want to look at the documentation of imenu
;;	to find many more usages for it. It makes your day shine in
;;	X-display. You should also configure few variables for it, like:
;;
;;	    (setq imenu-max-items 20)
;;
;;  Test run
;;
;;	Load this file and set those keybindings mentioned. Hit the mouse
;;	bindings and you're running bookmark package.
;;
;;	Since the break marks are used in commentary also, the list of
;;	bookmarks are not in their most informative form,
;;	I use following convention to name bookmarks;
;;
;;	    'v-'     variable topic
;;	    't-'     text topic
;;
;;  Design thoughts
;;
;;	Sooner or later someone wonder's: "Can't we have sub-breaks listed
;;	nicely with indentation in front lines in X-popup?"
;;
;;	My answer is "No", since it would require keeping track
;;	of the 'Main break' and then seeing if there exist sub-breaks.
;;	Immediately this leads to question "What is the main break?", and
;;	if we say main breaks start with "#|/%" charset we limit the use
;;	of breaks. Besides deciding what are sub-breaks, main-breaks with
;;	regexp may be too slow.
;;
;;	I wanted to have *general* view of the buffer. Please put imenu
;;	finding single functions if you don't feel like tapping couple of
;;	pgUp/pgDown afer you have jumped to bookmark.


;;}}}
;;{{{ history

;;; Change Log:
;; ........................................................ &t-history ...
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; Jun	19	1997	[jari]		19.28	v1.16-1.18	NotReleased
;; - Internal changes.
;;
;; Mar	19	1997	[jari]		19.28	v1.14-1.15	NotReleased
;; - XE20 CHAR changes. Now works in non-windowed emacs too.
;; - Installation instructions changes (function names changed)
;;
;; Feb	22	1997	[jari]		19.28	v1.12-1.13	NotReleased
;; - only text changes. Library function name changes.
;;
;; Sep	25	1996	[jari]		19.28	v1.11		NotReleased
;; - Rearranged loading. No wuses autoloads.
;;
;; Aug	20	1996	[jari]		19.28	v1.10		NotReleased
;; - Wiped away Xe 19.14 reported errors. Removed tibm-occur function
;;   because it was not needed.
;;
;; Jun	28	1996	[jari]		19.28	v1.9		NotReleased
;; - Some textual changes only.
;; - The ache was wiped awy whenever the major mode changed. Added
;;   permanent-local property.
;; - Some "tdi" prefix corrected to "tibm". Version function now has ARG.
;; - Got rid of last compiler error "fold-show" not defined. Used 'sfuncall'
;;
;; Apr	15	1996	[jari]		19.28	v1.8		NotReleased
;; - Requires tinylib.el version 1.40, where some serious errors
;;   hanged the insert function.
;;   tibm-repeat			:! small fix when extar spaces on line
;;
;; Feb	5	1996	[jari]		19.28	v1.7		NotReleased
;; - Changed one variable into function, so that the changes in variables
;;   take effect immediately.
;;   tibm-:re-default-chars		-> tibm-re-default-chars
;;
;; Feb	5	1996	[jari]		19.28	v1.6		NotReleased
;; - Added movement functions and reformatted the installation to use
;;   automatic code ripping. New local variable holds the
;;   movement regexp --> fast response
;;   tibm-forward			:+ move easily
;;   tibm-backward			:+
;;
;; Feb	5	1996	[jari]		19.28	v1.6		NotReleased
;; - Corrected error in tibm-insert. lisp-mnt lm-verify checked.
;; - Rewrote version information.
;;
;; Sep	1	1995	[jari]		19.28	v1.5		NotReleased
;; - When Inserting bookmark section break after a initial TAB, the
;;   comment positioning to the beginning of section was broke. Now it
;;   inserts to the point user sits on.
;;
;; Aug	21	1995	[jari]		19.28	v1.4		NotReleased
;; - Added check if there were no bookmarks --> message displayed.
;; - Added extra prefix arg to tibm-insert, so that strict fill is
;;   possible. Added tibm-:insert-strict user variable for default
;;   setting if arg is not given.
;; - Corrected tibm-repeat bug: reference to non-existing function
;;   ti::..-column.
;; - Corrected bookmark insertion failure, when point was after several
;;   tabs. Like heavy indented lisp code.
;; - Added more documentation.
;;
;; May	16	1995	[jari]		19.28	v1.3		NotReleased
;; - Text corrections. Added tibm-:re-default variable if the
;;   BookMarkRegexp is not defined in the file itself.
;; - Threshold option corrected. Previously it stored MARKER into the
;;   end value, but it actually should have stored the VALUE of marker.
;;   Now. threshold option working ok.
;;
;; May	15	1995	[jari]		19.28	v1.2		NotReleased
;; - Hehe, too early to be pleased. Corrected inconsistent variable
;;   names. Cache didn't get buffer local, corrected.
;;
;; May	12	1995	[jari]		19.28	v1.1		NotReleased
;; - Write whole code in one night and it's been working gorgiously!
;;
;; Feb	22	1995	[jari]	19.28	v0.0			an vision only


;;}}}

;;; Code:

;;{{{ setup: require

;;; ......................................................... &require ...

(require 'tinylibm)

(eval-and-compile
    (autoload 'imenu--mouse-menu    "imenu"	t t)
    (autoload 'fold-show	    "folding"	t t)
    )

(eval-when-compile (use-dynamic-compilation))

;;}}}
;;{{{ setup: -- hooks

;;; ......................................................... &v-hooks ...

(defvar tibm-:parse-before-hook nil
  "*Hook that is run just before the buffer is scanned for bookmarks.")

(defvar tibm-:load-hook nil
  "*Hook run when file is loaded.")

;;}}}
;;{{{ setup: -- private vars

;;; ....................................................... &v-private ...


(defvar tibm-:cache nil
  "Private. Cache where bookmarks are stored in a
list \(bookMarkName . point\)")
(make-variable-buffer-local 'tibm-:cache)

;;  We don't want cache to be wiped away when major mode changes
(put 'tibm-:cache           'permanent-local t)


(defvar tibm-:cache-char-count nil
  "Private. Totals characters in buffer.")
(make-variable-buffer-local 'tibm-:cache-char-count)


(defvar tibm-:bookmark-regexp nil
  "Private. Holds buffers bookmark regexp.")
(make-variable-buffer-local 'tibm-:bookmark-regexp)


(defconst tibm-:package 'tibm
  "Private. Package name.")

(defvar tibm-:msg-id (symbol-name tibm-:package)
  "Private. Identifier for package's messages.")

;;}}}
;;{{{ setup: user conf

;;; ........................................................ &v-public ...

(defvar tibm-:cache-update 'threshold
  "*Method when to update cache.
nil         ,manual update -- you have to call tibm-parse
'always     ,always update cache when menu displayed.
'threshold  ,update happens when buffer's total character change
             exceeds previous value of tibm-:cache-threshold-val.
")

(defvar tibm-:cache-threshold-val 100
  "*When cache is constructed, the total character count is saved.
When user adds more code, the total count changes accordingly. If the
difference between current count and last saved count gets bigger than
this value the cache is re-read.")


(defconst tibm-:re-default-chars "[-~+=*%/|#.,'`^]"
  "*Default bookmark repeat chars.")

(defvar tibm-:max-col '(progn  (tibm-calc-max-col))
  "*Last column to extend the break. This can be FORM which evaluates
to column number")

(defvar tibm-:trailer-space-len 3
  "*How much space is left to the right before the bookmark ID ends.")

(defvar tibm-:comment-start-func 'tibm-comment-start
  "*Function that returns appropriate start comment. Must return empty
string if comment not defined.")

(defvar tibm-:comment-end-func 'tibm-comment-end
  "*Function that returns appropriate end comment. Must return empty
string if comment not defined.")


(defvar tibm-:scan-filter-func 'tibm-scan-filter
  "*When building up the bookmarks from file, there may be false hits,
or you may look at special lines only. This function accepts three arguments:
- current line string
- line beginning point
- identifier found from line

If the function return nil the line is not added to the cache.
")


(defvar tibm-:goto-func 'tibm-goto
  "*Function that handles moving to the point. If you have folding in effect
around that point you may wish to open it in your custom function.

This function receives one argument: POINT
")


(defvar tibm-:insert-strict t
  "*Controls if the bookMark insertion is strict when no argument is given.
See tibm-insert")

;;}}}
;;{{{ setup: -- version

;;; ....................................................... &v-version ...

(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinybm.el"
 "tibm"
 tibm-:version-id
 "$Id: tinybm.el,v 1.18 1997/06/19 10:37:34 jaalto NotReleased $"
 '(tibm-:version-id

   tibm-:parse-before-hook
   tibm-:load-hook
   tibm-:cache
   tibm-:cache-char-count
   tibm-:bookmark-regexp
   tibm-:msg-id
   tibm-:cache-update
   tibm-:cache-threshold-val
   tibm-:max-col
   tibm-:trailer-space-len
   tibm-:comment-start-func
   tibm-:comment-end-func
   tibm-:scan-filter-func
   tibm-:goto-func
   tibm-:insert-strict

   tibm-:package
   tibm-:re-default-chars
   )))

;;}}}

;;; ########################################################## &macros ###

;;{{{ Macros

;;; ----------------------------------------------------------------------
;;;
(defsubst tibm-get-bm-re ()
  "Returns buffer's bookmark regexp. If the local value where the
regexp is stored is nil, the rescan buffer.

References:
  tibm-:bookmark-regexp
"
  (or tibm-:bookmark-regexp		;changing mode kill local vars
      (setq tibm-:bookmark-regexp
	    (tibm-search-bm-re))))


;;; ----------------------------------------------------------------------
;;; Default bookmark syntax that is used if file does not contain
;;; it's own definition of bookmark syntax.
;;;
(defsubst tibm-re-default  ()
  "Returns default bookmark regexp.
References:
  tibm-:re-default-chars
"
  (concat
    tibm-:re-default-chars
    tibm-:re-default-chars "+"
    " &+\\([^ \t]+\\) "
    tibm-:re-default-chars "+"
    ))

;;}}}

;;; ########################################################### &Funcs ###

;;{{{ movement funcs

;;; ........................................................ &movement ...

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tibm-backward (&optional back)
  "Searches bookmark line backward."
  (interactive)
  (tibm-forward 'back))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tibm-forward (&optional back)
  "Searches bookmark line forward."
  (interactive)
  (let* ((func (if back 're-search-backward 're-search-forward))
	 (re   (concat "^" (or comment-start "") "+ "
		       (tibm-get-bm-re)))
	 )
    (funcall func re nil t)))

;;}}}

;;{{{ misc funcs

;;; ............................................................ &misc ...

;;; ----------------------------------------------------------------------
;;; LISP column
;;;  - I can hear you saying: "Why 74? why not 70 or 75 ?..."
;;;  - Well, I uasually add bookmark section to my elisp code and while
;;;    I did them by hand I added ';;; ' comment at the beginning of
;;;    line and fed 70 continuous charactes with ESC 70 '-'after
;;;    comment  --> totals 4 + 70 chars :-/
;;;
;;;  - The idea of this calculation is that when you hit separator,
;;;    like this: COMMENT-SPACE-70_CHAR_SEPARATOR, this will calculate
;;;    the column so, that when tibm-insert is called, the last
;;;    char lines up with yours.
;;;
;;;    Eg. in shell mode:
;;;
;;;		# ---------------, 70 chars long sep, last col is 2 + 70
;;;             # ..............., tibm-insert lines up to col 72
;;;
;;;    But in lisp
;;;
;;;		;;; -------------, again 70 chars long sep, 4 + 70
;;;		;;; ............., tibm-insert lines up to col 74
;;;
;;;    Now you can hit 70 line separator in any mode and to be sure the
;;;    tibm-insert lines up with you.
;;;
(defun tibm-calc-max-col ()
  "Calculates column for mode."
  (let* ((mode		(symbol-name major-mode))
	 (cs		(or comment-start ""))
	 (def-len	70)		; like "# " + 70 chars
	 )
    (cond
     ((string-match "lisp" mode)
      74)
     (t
      (if (string-match "[ \t]+" cs)	;does it already have space "# "
	  (+ def-len (length cs))	;no it does not "#", add room for it.
	(1+ (+ def-len (length cs))))
      ))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tibm-goto (point)
  "Goes to the selected point."
  (let* ((re  ".*{{{")
	 )
    (cond
     ((and (featurep 'folding)
	   (symbol-value 'folding-mode))
      (goto-char point)
      (save-excursion
	(beginning-of-line)
	(if (looking-at re)
	    (fold-show)))
      )
     (t
      (goto-char point)
      ))
    ))


;;; ----------------------------------------------------------------------
;;; - include all lines
;;;
(defun tibm-scan-filter (full-line pos id)
  "Returns always t, so all matched lines are cached."
  t)

;;; ----------------------------------------------------------------------
;;;
(defun tibm-comment-end ()
  "Returns appropriate comment end, according to mode"
 (let* ((str (or comment-end ""))
	)
  (if (equal "" str)
   nil				;no comments
   (setq str (ti::s-add-space str)))
  str
  ))


;;; ----------------------------------------------------------------------
;;;
(defun tibm-comment-start ()
  "Returns appropriate comment, according to mode"
  (let* ((m   major-mode)
	 (str (or comment-start ""))	;comment
	 )
    (cond
     ;;   Lisp is special, due to it's comment usage
     ((memq m  '(lisp-mode emacs-lisp-mode lisp-interaction-mode))
      (if (bolp)
	  (setq str ";;;")
	(setq str ";;")))
     )
    (if (equal "" str)
	nil				;no comments
      (setq str (ti::s-add-space str t)))
    str
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tibm-cache-update ()
  "Determines when and how to update cache.
References: tibm-cache-update
"
  (let* ((mode       tibm-:cache-update)
	 (end        (marker-position (point-max-marker)))
	 (cache-end  (or tibm-:cache-char-count  end))
	 (limit      tibm-:cache-threshold-val)
	 diff
	 )
    (cond
     ((eq mode nil)
      nil)				;manual

     ((eq mode 'always)
      (tibm-parse))

     ((eq mode 'threshold)
      (setq diff (abs (- end cache-end)))
;;;         (d! diff limit end cache-end )
      (if (< diff limit) nil
	;; Hmm, should we print a message "threshold exceeded, reparsing..?"
	;; Let's be transparent this time: no messages.
;;;	(d! "reparsing..")
	(tibm-parse)
	))
     )
    ))


;;}}}
;;{{{ bookmark line insert

;;; ##################################################### &bookmarkAdd ###

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tibm-repeat (str count &optional col strict)
  "Repeats character or string sequence STR COUNT times.

COUNT can be:
  0       ,repeat until position 79 or COL , or if the STR is not single
	   character, until fits below COL
  \"\"    ,interactive insert, as long as user presses RET or SPACE.

STRICT has effect only if COL is given:
  nil     ,insert as long as STR fits below COL
  t       ,insert strictly up till COL and cut away portion
           of STR if necessary
"
  (interactive "sString:\nsCount[0=eol]: ")

  (let* ((swide (or col 79))		;screen width
	 (i	0)
	 (ok	t)
	 ch c len p
	 )
    (if (or (not (stringp str))	;it's not string
	    (eq 0 (length str)))	;string is empty
	(throw 'done nil))


    (cond
     ((stringp count)
      (if (equal "" count)
	  (setq c -1)			;interactive
	(setq c (string-to-int count))))
     ((numberp count)
      (setq c count))
     (t
      (error "Invalid count arg" count)))

;;;    (d! "c-val" c)

    ;; ... ... ... ... ... ... ... ... ... ... ... ... ... ... ... ... ...
    (cond
     ((eq c -1)			;ask confirmation every time...

      (while ok
	(message "insert? <spc,enter> ") (setq ch (read-char))
	(cond
	 ((or (char= ch ?\C-m ) (char= ch ?\ ))
	  (insert str))
	 (t (setq ok nil))))

      (message "")			;clear
      )

     ;; ... ... ... ... ... ... ... ... ... ... ... ... ... ... ... ... ...
     ((eq c 0)

      (setq len		(length str)
	    p		(current-column))

      ;; we have to remove tabs from this line to get count right
      (untabify (BOLP) (EOLP))
      (move-to-column p)		;restore position

      ;; the added str must not move point further than COL

      (while (<= (+ len (current-column)) swide)
	(insert str))

;;;      (d!  "c-val 0" (current-column) swide)

      ;;   Check if it was multicharacter and we didn't get to last position
      ;;   Insert the last string and cut after COL

      (if (null strict) nil
	(if (= (current-column) swide) nil
	  (insert str)
	  (ti::m-move-to-col swide)
	  (delete-region (point) (progn (end-of-line) (point)))))
      )

     ;; ... ... ... ... ... ... ... ... ... ... ... ... ... ... ... ... ...
     (t					;straight number !
      (while (< i c)
	(insert str) (setq i (1+ i)))
      ))

    ))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tibm-insert (txt sep &optional strict)
  "Adds bookmark till the end of line.
Normally line is filled as long as the pattern fits below max column,
but if the optional argument is given, it will be filled in _full_ ,
truncating if necessary. To see an example, try with some _long_
pattern.

Optional prefix arg STRICT can have values:
  0       strict is nil in spite of tibm-:insert-strict
  1       strict is t   in spite of tibm-:insert-strict
  nil     use default value in tibm-:insert-strict

References:
	tibm-:insert-strict
"
  (interactive "sBookmark: \nsSep: \nP")
  (let* (
	 (strict-def  tibm-:insert-strict)
	 (cs-func     tibm-:comment-start-func)
	 (ce-func     tibm-:comment-end-func)
	 (line-col    (eval tibm-:max-col))
	 (trail-len   tibm-:trailer-space-len) ;how much to leave
	 (bolp        (BOLP))
	 (cur-col     (current-column))
	 (orig-point  (+ bolp cur-col))
	 col
	 cs ce
	)

    (cond
     ((eq nil strict)			;use default
      (setq strict strict-def))

     ((eq 0 strict)
      (setq strict nil))

     ((eq 1 strict)
      (setq strict t)))



    (if (= 0 (length sep))
	(error "No separator"))


    ;;  Add surrounding spaces if they are not included
    (if (equal "" txt)
	nil				;no text
      (setq txt (ti::s-add-space txt)
	    txt (ti::s-add-space txt t)))


    (setq cs (funcall cs-func)		;Set comments
          ce (funcall ce-func))

    ;;  I tried to turn overwrite-mode on, but SUPRISE!
    ;;  - While it was on, and I tried to do 'insert',
    ;;    it didn't _overwrite_; emacs why can't you behave as expected ?
    ;;  - So I have to hack with delete-region instead.

    ;;  - skip lenght of comment start
    (ti::m-move-to-col (+ cur-col (length cs)))

    ;; We must clear the rest of line
    ;;
    (if (not (looking-at "$"))
	(delete-region (point) (EOLPP)))

    ;;  - draw the line until max col
    (setq col line-col)			;maximum repeat column


    (tibm-repeat sep 0 col strict)	;insert full separator

    ;;  - Now cut room for identifier
    (backward-char (+ (length txt) trail-len))	;leave trailer space
    (delete-region (point) (+ (point) (length txt)))

    ;;  - write the identifier
    (insert txt)
    (end-of-line)
    (insert ce)				;comment end

    ;;  - delete spaces at front and replace it with comment start
    (goto-char orig-point)
    (delete-region (point) (+ (point) (length cs)))
    (insert cs)

    (goto-char orig-point)
    ))

;;}}}
;;{{{ Bookmark find, cacheing

;;; #################################################### &bookmarkScan ###

;;; ----------------------------------------------------------------------
;;;
(defun tibm-scan (re)
  "Gathers all bookmarks from current point forward. Returns list:
\(id . beginning-of-line-point\).

References
	tibm-:scan-filter-func
"
  (let* ((func tibm-:scan-filter-func)	;should we filter something ?
	 id
	 p
	 list
	 )

    (while (re-search-forward re nil t)
;;;      (d! (read-match 1) (read-full-line))
      (if (null (setq id (read-match 1))) nil ;No identifier at subexpression 1
	(setq p (BOLP))

	;;  Is this line allowed to add ?
	(if (funcall func (read-full-line) id p)
	    ;;  Nothing magic in this expression, just build list
	    ;;  '((id point) (id .point) ..)
	    (setq list (append list (list (cons id p)))))
	))
    list
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tibm-search-bm-re ()
  "Searches buffer for automatic bookmark identifier 'BookMarkRegexp'
and returns regexp defined in it. if is doesn't exist returns default
bookmark regexp.
"
  (let* ((id          "BookMarkRegexp")
	 (re-default  (tibm-re-default))
	 id-val
	 fixed-val
	 ret
	 )
    (setq id-val (ti::m-rcs-str-find-buffer id t))
    ;;  while reading from buffer the \ doubles, convert it back to \
    (setq fixed-val (ti::c-str2re id-val))

    (if (or  (null fixed-val)
	     ;;  We must find the '' marks
	     (not (string-match "'\\(.+\\)'" fixed-val)))
	(setq ret re-default)
      (setq ret (read-match 1 fixed-val)))
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tibm-parse ()
  "Builds up bookmark list and stores them into cache.
Return:
  t     .cache build
  nil   ,bookmarks not found or error happened. Cache untouched.
"
  (interactive)
  (let* ((op             (point))	;user's original point
	 (beg            (point-min-marker))
	 (end            (point-max-marker))
	 (end-pos        (marker-position (point-max-marker)))
	 (end-max        (point-max))
	 end-wmax
	 re
	 ret
	 list
	 )

    (run-hooks 'tibm-:parse-before-hook)
    (setq tibm-:cache-char-count end-pos)	;<< GLOBAL

    (if (null (setq re (tibm-get-bm-re)))
	(message "No bookmark syntax Identifier found.")
      (unwind-protect			;handle narrowed buffers too
	  (progn
	    (widen)
	    (setq end-wmax (point-max))
	    (PMIN)
	    (setq list (tibm-scan re))
	    (if (null list)
		(message "No bookmarks found.")
	      (setq ret t)
	      (setq tibm-:cache list))	;<< GLOBAL
	    )
	(save-excursion
	  (set-buffer (marker-buffer beg))
	  ;; what about after widen ? Were we in narrow mode ?
	  (if (not (= end-wmax end-max))
	      (narrow-to-region beg end)))
	))
    ;; only reasonable way to return to current point
    (goto-char op)
    ret
    ))



;;}}}
;;{{{ mouse

;;; ################################################### &mouseHandling ###

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tibm-keyboard-parse ()
  "Reparse bookmarks."
  (tibm-mouse-parse nil (interactive-p)))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tibm-mouse-parse (&optional event verb)
  "Reparse bookmarks. This fucntion is called from mouse binding."
  (interactive "e")
  (VERB)
  (if (and verb (tibm-parse))
      (message "Bookmarks cached.")))

;;; ----------------------------------------------------------------------
;;;
(defun tibm-selection (event)
  "Displays cache menu."
  (interactive "e")
  (let* ((go-func   tibm-:goto-func)
	 cache
	 data
	 )
    (tibm-cache-update)
    (setq cache     tibm-:cache)
    (if (null cache)
	(message "No bookmarks found.")
      (cond
       ((ti::xe-window-system)
	(setq data (imenu--mouse-menu cache event)))
       (t
	(setq data (completing-read "Select: " cache))
	(setq data (assoc data cache))
	))
      (if data
	  (funcall go-func (cdr data)))
      )
    ))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tibm-keyboard (&optional arg)
  "See tibm-mouse."
  (interactive "P")
  (tibm-mouse nil arg))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tibm-mouse (event &optional arg)
  "Pops up bookmark menu. Optional arg:
  list        ,C-u reparses list
"
  (interactive "e\nP")
  (let* ((cache-ok   tibm-:cache)
	 )
    (if (and cache-ok
	     (not (equal arg '(4))))
	nil				;something on the cache ok..
      (message "building cache...")
      (sleep-for 1) (message "")
      (setq cache-ok (tibm-parse)))
    (tibm-selection event)
    ))

;;}}}

(provide   'tinybm)
(run-hooks 'tibm-:load-hook)

;;; tinybm.el ends here
