;; @(#) tinydired.el -- Dired enchancements. Backgroud Ange ftp support

;; @(#) $Id: tinydired.el,v 1.64 1997/06/19 09:34:47 jaalto NotReleased $
;; @(#) $Keywords: extensions $
;;
;; $KnownCompatibility: 19.28 $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of GNU emacs

;;{{{ Id

;; Copyright (C) 1996,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Jan 31 1996
;;
;; To get information on this program use ident(1) or do M-x tdd-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinydired|Jari Aalto|jari.aalto@poboxes.com|
;; Dired enchancements. Bg ange ftp for marked files, More ange funcs...|
;; 19-Jun-1997|$Revision: 1.64 $|~/misc/tinydired.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ....................................................... &t-install ...
;;  Put this file on your Emacs-Lisp load path, add following into your
;;  ~/.emacs startup file. It gives you the default installation
;;  Code can be easily ripped with tinylib.el / ti::m-pkg-rip-magic
;;
;;* _
;;*  	(autoload 'tdd-hook-control			"tinydired" t t)
;;*  	(autoload 'tdd-switch-to-some-ange-ftp-buffer	"tinydired" t t)
;;*  	(add-hook 'tdd-:load-hook			'tdd-hook-control)
;;*  	(add-hook 'dired-mode-hook '(lambda () (require 'tinydired) nil))
;;* _
;;* _   ;;  Make sure the keys are defined.
;;*     (setq     tdd-:force-add-keys-flag 'override)
;;
;;  For more personal setup, you have to look at the calls in
;;  tdd-hook-control and put your own initializations into
;;  dired-mode-hook and dired-after-readin-hook.
;;
;;  To select/kill ange buffers, use these bindings
;;
;;* _
;;* 	(global-set-key "\C-cab" 'tdd-switch-to-some-ange-ftp-buffer)
;;* 	(global-set-key "\C-cak" 'tdd-kill-all-ange-buffers)
;;* 	(global-set-key "\C-caK" 'tdd-kill-all-ange-and-dired-buffers)
;;
;;  If you don't want default keybindings, modify variable
;;
;;	tdd-:bind-hook
;;
;;  Help about keys -- do this in dired buffer after you've loaded this file
;;
;;	t C-h		,enchanced dired commands
;;	a C-h		,enchanced ange-ftp related commands
;;
;;  If you have any questions, always use function
;;
;;      M-x tdd-submit-bug-report

;;}}}
;;{{{ Documentation
;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "tdd-" in front of
;;   every function & variable. It stands for '(t)iny (di)re(d) tool'
;; - variable names contain letters 'tdd-:', excecpt version/hook/mode vars.

;;  Preface, Jan 1996
;;
;;	This package started evolving, when I wasn't satisfied the
;;	ange-ftp'ing in dired. Why did I have to ftp file by file? And I
;;	certainly didn't want to get popped to .zip or .tar.gz buffer after
;;	loading! I just wanted to download the files somewhere other than
;;	inside emacs.
;;
;;	Many times I was annoyed on the fact that I couldn't just select
;;	the files in dired buffer which held remote files, and say
;;
;;	    "Hey, I want to download these files into this directory."
;;
;;	Now you can do this with dired.  Sometimes I also wanted to go
;;	associated ange-ftp buffer and give commands directly there. I also
;;	wished there would have been some way to switch between the
;;	ange-ftp buffer and dired.
;;
;;	All these wishes are now fullfilled. hope you enjoy your ange ftp
;;	as much as I do now.
;;
;;  Overview of features
;;
;;	o   Connecting to VAX host supported
;;	o   Connection to DOS host supported (experimental)
;;	o   Few enchancements to dired mode. Eg. keep only one
;;	    dired buffer when ascending to directory. Shorten symlinks.
;;	o   User can mark and put files into STORE and start a backgroud
;;	    ange-ftp session to get STORED files into download directory
;;	o   Easy switching between ange-ftp session buffer and dired buffer
;;	o   Dealing with ange ftp buffers in general
;;	    (x)  killing all ange buffers at once
;;	    (x)  killing all ange + dired ange buffers at once.
;;	    (x)  switching to ange buffers with completion
;;	o   Special VC dired commands (experimental)
;;	o   Run "!" on ange ftp dired buffer (operate on local copy)
;;	o   customizable backup file flagging.
;;	o   other handy dired commands, like "pop to this file in emacs."
;;	    "find all marked files"...
;;
;;  Xemacs note
;;
;;	The ange-ftp implementation is different in XEmacs, that what it
;;	is in Emacs. The original developer has continued to develop
;;	ange-ftp in XEmacs and nowadays we expect to see its successor *efs*
;;	to be included in later XEmacs releases.
;;
;;	There is no traps to detect if you're in XEmacs and try to use the ange
;;	extension. Expect to get weir results/error messages/failures.
;;
;;	The dired is also very different in XEmacs, so I don't guarrantee
;;	that this package works ar all in XEmacs. Use at your own risk. I may
;;	improve XEmacs support if it's easy, but this package will
;;	always be crippled in XEmacs.
;;
;;  General dired additions
;;
;;	In simplest form. This module installs some functions in your
;;	dired hooks. Their purpose is
;;
;;	o   To keep your dired buffer sorted so that directories are
;;	    always put first.
;;	o   Delete unwanted files from dired buffer automatically.
;;	o   Shorten the symlink references, so that they don't spread
;;	    multiple lines and ruin your view.
;;
;;	It also changes one dired function with defadvice, so that
;;	you can control if you want to have only one dired buffer
;;	when ascending to another directory. See variable:
;;
;;	    tdd-:use-only-one-buffer-flag
;;
;;  Dired and ange-ftp additions
;;
;;	In general, when you want to start ftp session in emacs you just do
;;
;;	    C-x C-f /login@site:/dir/dir/file
;;
;;	Let's take an example: To see what new things has arrived
;;	to GNU site, you'd do this:
;;
;;	    C-x C-f /ftp@prep.ai.mit.edu:/pub/gnu/
;;
;;	After that you are put into the dired listing, where you
;;	can mark files with dired-mark command
;;
;;	    "m"
;;
;;	Now you have files ready. Next put files into batch STORAGE.
;;	There is "a" prefix for ange-ftp related commands.
;;
;;	    "a" "S"	,big S put selected files into storage
;;	    "a" "q"	,to check what files you have batched
;;	    "a" "c"	,to clear the batch storage
;;
;;	Now start ftp'ding the files in background. You're prompted
;;	for the download directory.
;;
;;	    "a" "g"	,"get" files
;;
;;	If you want to operate on the associated ftp buffer
;;	directly, there is command
;;
;;	    "a" "b"	,b for "buffer change"
;;
;;	that puts you into ftp, where the dired buffer refers. When
;;	you're in the ftp buffer you have some keybinding available.
;;
;;	    C-c f	,insert stored files on the line
;;	    C-d d	,minsert directory name
;;	    C-c b	,back to dired window
;;
;;	It's sometimes handy that you can give direct ftp commands.
;;
;;
;;  Setting up ange ftp
;;
;;	Here is my settings, which you can use as a reference so that
;;	you get the ange running. For more details, see the
;;	ange-ftp.el's source code. These settings include my
;;	firewall "ftpgw.poboxes.com"
;;
;;* ;; (setq ange-ftp-generate-anonymous-password t)
;;* (setq ange-ftp-dumb-unix-host-regexp  "tntpc") ;PC hosts
;;* (setq ange-ftp-gateway-host "ftpgw.poboxes.com")
;;* (setq ange-ftp-smart-gateway t)
;;* (setq ange-ftp-local-host-regexp "\\.nokia\\.\\(com|fi\\)|^[^.]*$")
;;* ;;  Always use binary
;;* (setq ange-ftp-binary-file-name-regexp ".")
;;* (autoload 'ange-ftp-set-passwd "ange-ftp" t t)
;;* (setq ange-ftp-generate-anonymous-password "jaalto@tre.tele.nokia.fi")
;;
;;
;;  How to use this modeule 3 -- special vc
;;
;;	There are some extra commands that you may want to use too.
;;	See the bind function
;;
;;	    tdd-default-other-bindings
;;
;;	What additional commands you get when loading this module.
;;
;;	The VC special commands were programmed, because I felt that the
;;	C-x v v in dired mode didn't quite do what I wanted. I wanted
;;	simple ci/co/revert commands for files that were in VC control.
;;	And I wanted to handle them individually, expecially when ci'ing.
;;
;;	This VC part of the package is highly experimental.
;;	I'm not sure if I support it in further releases.
;;
;;
;;  Important ange-ftp interface note
;;
;;	The ange ftp batch interface used here may cause unpredictable
;;	problems. Sometimes the 'get' or 'put' process doesn't start at all
;;	although you see message saying it started the job. I have had
;;	several occurrances where 'lcd' cmd succeeded, but then nothing
;;	happened. Repeating the 'put' or 'get' command cleared the problem
;;	whatever it was.
;;
;;	So, never trust the message 'completed' when you're putting files
;;	at the bg to the remote site. Ange FTP that comes with Emacs
;;	doesn't have any direct support for bg jobs. And it may be that the
;;	layer on top of ange-ftp, that I wrote, may not always function right.
;;
;;	Always go to the ange buffer and _check_ yourself that the ftp went
;;	ok. Try again if it didn't. Another way to clear the problem: kill
;;	the ange ftp buffer and try the command from dired again. It
;;	automatically opens session to the site.
;;
;; Advertise -- other usefull packages
;;
;;	If you really are heavy remote file user, get this module too:
;;
;;	    tinyfh.el
;;
;;	It allows point and shoot loading a file on the line. This
;;	is not quite same module as ffap.el.
;;
;;	o   You can even sit on elisp file, eg. put cursor over
;;	    the word tinyfh and it'll get the file from OHIO
;;	    automagically. Many other file ref formats are rocognized.
;;	o   You can be in Archie listing and shoot the file into emacs.
;;
;;	There are exellent dired extensions around, please consider getting
;;	these packages too:
;;
;;	o   dired-sort.el (requires date-parse.el)
;;	o   dired-tar.el
;;
;;
;;  Bugs
;;
;;  Troubles with ange-ftp -- repeated
;;
;;	You mark a file, or files, and you saw message "completed...
;;	but not all files were loaded? Don't trust it! The ange-ftp
;;	interface is a little bit tricky and it sometimes does't
;;	work as expected.
;;
;;	The cure for all ange problems is: Go and kill the
;;	*internal ftp* buffer for current site and try again.
;;
;;  Slow autoload
;;
;;	When you have added the autoloads into your .emacs, the first time
;;	you bring up dired buffer may be quite slow. This is normal, Emacs
;;	just need to load some additional files that this package uses.
;;
;;  Refreshing the view takes long time / point isn't exatly the same
;;
;;	This is normal, dired is just slow and program has to do lot of
;;	work to maintain the "view". Eg. save view, save marks, delete
;;	marks, revert, sort, restore marks... Only the current line
;;	position is preserved where user was, not point.
;;
;;  Note: Code
;;
;;	At least 19.28 ships with package dired-x.el, which seems to offer
;;	some more goodies to dired. I just noticed it the other day and I
;;	already had coded 2 similar functions that was found from it.
;;
;;	Currently, if the dired-x is detected the appropriate functions
;;	in this package are diabled, to prevent overlapping behavior.
;;	However, if the function behaves differently than the one in
;;	some dired extension package, then the function isn't disabled.
;;	Eg. see tdd-load-all-marked-files, which can turn off marks.
;;
;;  Note: limited vax directory support
;;
;;	You can connect to VAX host with reguler ange-ftp find-file
;;	syntax and this package provides modified functions to allow you
;;	to use "f" key to browse listings and getting files.
;;	Any other tinydired commands are disabled.
;;
;;	See function tdd-remove-bindings. You must add your own
;;	keyboard disable function if you have added tdd functions to
;;	other bindings than the default to disable this package in
;;	VAX buffers. Add your disable funcion to
;;
;;	    tdd-:readin-hook
;;
;;  Note: limited dos directory support
;;
;;	You can  connect PC hosts that print 'dir' into the dired buffer.
;;	Your only command available is "f", so please don't even try
;;	any other choices.
;;
;;	The PC support is experimentary and is based on UNIX --> LAN
;;	PC which is running pctcp software: ctlapp.exe ftp server.
;;	Do not mail me about other PC support, since I won't
;;	add anything else. Contact Andy Norman and get full EFS distribution
;;	which is complete rewrite of ange-ftp and will have all you need.
;;
;;  Thanks
;;
;;	Some ideas I got from people in the gnu news groups.
;;	I Have also written a pointer to person, whose work I
;;	have possibly used as a base.
;;
;;	Ideas presented in the net by:
;;
;;	<sof@dcs.glasgow.ac.uk> (Sigbjorn Finne), comp.emacs
;;	Kevin Rodgers <kevin.rodgers@ihs.com>, gnu.emacs.help


;;}}}
;;{{{ history

;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; Jun	19	1997	[jari]		19.28	v1.60-1.64	NotReleased
;; - Better XEmacs warning at compile time. Checkdoc 1.42 errors corrected.
;; - defcustom support. Renames many varaibles. re --> regexp, boolean
;;   varaibles has -flag suffix etc.
;; - Library function changes.
;;
;; May	14	1997	[jari]		19.28	v1.57-1.59	NotReleased
;; - internal changes.
;; - Changed all non standard boolean tests to use when and unless
;;
;; Mar	24	1997	[jari]		19.28	v1.54-1.56	NotReleased
;; - added tdd-advice-control. XE20 CHAR handling corrected.
;; - Added mput history variable.
;;
;; Feb	14	1997	[jari]		19.28	v1.53		NotReleased
;; - I don't know how reliable the DOS code support is. It caused many
;;   unpredictable results when used in normal ange-ftp. I have disabled
;;   the dos support for now and maybe I look it later.
;; - new user command tdd-store-delete-filename "a d".
;;
;; Jan	15	1997	[jari]		19.28	v1.51-1.52	NotReleased
;; - Added advices to DOS host connection: WfW that is reachable by LAN
;;   connection running pctcp: ctlapp.exe ftp server.
;;
;; Dec	7	1996	[jari]		19.28	v1.50		NotReleased
;; - Renamed some library functions. Moved VAX ange-ftp
;;   patches from tinypatch.el to this package. Now you canb use "f" to
;;   find a file from VAX listing after you do C-x C-f to vax host
;;   The version number reflects this support.
;;   tdd-switch-to-ange-ftp-buffer	:! bug corrected
;;   tdd-kill-marked-lines		:! bug corrected
;;
;; Dec	5	1996	[jari]		19.28	v1.42-1.43	NotReleased
;; - Removed 'require ange-ftp' and put it into autoload to speed up
;;   compilation.
;; - Corrected "funtion" --> "function" thanks to
;;   Geoffroy VILLE <ville@cena.dgac.fr>
;;
;; Oct	8	1996	[jari]		19.28	v1.41		NotReleased
;; - Changed all '(lambda forms to use (function (lambda. Minor
;;   text changes.
;;
;; Sep	25	1996	[jari]		19.28	v1.38-1.39	NotReleased
;; - Examined the code and saw that I could make this XEmacs compatible:
;;   Added defalis definitions to refer to XEmacs equivalents and
;;   removed all other Emacs dependecy functions.
;; - Only the FTP will never work in XEmacs. Rewote the section that
;;   talks about XEmacs compatibiliy.
;; - Rearranged loading. Now uses autoloads.
;;
;; Sep	24	1996	[jari]		19.28	v1.37		NotReleased
;;   tdd-mark-opposite-toggle		:+ new used cmd "tmo"
;;   tdd-map-over-re			:! serious bug, beg-of-line missed
;;
;; Sep	12	1996	[jari]		19.28	v1.36		NotReleased
;;   tdd-kill-lines			:! key changed to "tkk"
;;   tdd-kill-marked-lines		:+ new used cmd   "tkm"
;;   tdd-kill-unmarked-lines		:+ new user cmd   "tkM"
;;   tdd-mark-writable-files		:+ new user cmd   "tmw"
;;   tdd-mark-read-only-files		:+ new user cmd   "tmr"
;;   tdd-leave-marked-files		:- deleted        "tml"
;;   tdd-map-over-unmarked		:+
;;
;;
;; Aug	30	1996	[jari]		19.28	v1.35		NotReleased
;; - Wrote new 'xemacs note' section and most errors reported by
;;   XEmacs 19.14 byteCompiler, which is much better than the Emacs's.
;;
;; Jun	28	1996	[jari]		19.28	v1.34		NotReleased
;; - There was severe problem in this package when using many dired
;;   commands for one file only. Dired runs dired-add-entry to update the
;;   line, and this function runs the dired-after-readin-hook -->
;;   effectively running unnecessary function from this package: they
;;   reorder the directory (eg. sorting dir)
;; - I added tests tdd-normal-buffer-p for the functions that are
;;   added to the hook tdd-:readin-hook. If you add some other function
;;   to that hook. it should use similar test if it modifies dired
;;   content.
;; - Many functions touched. Not listed here.
;;
;; Jun	28	1996	[jari]		19.28	v1.31-1.33	NotReleased
;;    tdd-one-dir-up			:+ New "t-" command
;;    tdd-store-ftp-mput		:! now offers old mput site
;;    tdd-sort-dir			:! now order dirs/links/files
;;
;; Jun	27	1996	[jari]		19.28	v1.30		NotReleased
;;    tdd-kill-dired-and-ange-session	:+ New "abk" command, also in "q"
;;
;; Jun	19	1996	[jari]		19.28	v1.29		NotReleased
;; - Added "tG" keybinding, which reads the directory without any
;;   filltering.
;; - Replaced standard dired function dired-flag-backup-files:
;;   now you can customize qhat you rehard as backup files, when
;;   flagging them with "~"
;;   tdd-:backup-file-regexp		:+
;;   dired-flag-backup-files		:+
;;   tdd-read-dir-as-is			:+
;;
;; Jun	17	1996	[jari]		19.28	v1.28		NotReleased
;; - Removed extra debug call, that was visible to user.
;;
;; Jun	13	1996	[jari]		19.28	v1.27		NotReleased
;; - The "tg" refresh view didn't quite work as expected. I couldn't
;;   determine the right cause of the problem, so I rewrote the method
;;   of matching files. now it's more restrictive and case sensitive.
;;   tdd-view-get			:! changed
;;   tdd-refresh-view			:! rewritten
;;
;; Jun	13	1996	[jari]		19.28	v1.26		NotReleased
;; - Thanks to Alexis Kotte <kotte@radth.ruu.nl>, the ange-ftp dired
;;   buffers are also detected by ":", not just "@" in dired-directory.
;;   tdd-:dired-directory-ange-regexp		:+
;;   tdd-kill-all-ange-and-dired-buffers	:! corrected
;;
;; Jun	12	1996	[jari]		19.28	v1.25		NotReleased
;; - Added new function to kill both ange and associated dired buffers.
;;   With it you can clean your emacs in a second.
;;   tdd-kill-all-ange-and-dired-buffers	:+
;;   tdd-ange-ftp-buffer-for-this-dired		:+
;;   tdd-switch-to-mput-ange-ftp-buffer		:! rewrote
;;
;; Jun	7	1996	[jari]		19.28	v1.23-1.24	NotReleased
;; - More reasonable bytecompiler setup used. Now uses autoload,
;;   and if 19.30+'s packages support lazy-load, this should reduce
;;   memory radically.
;;
;; Jun	7	1996	[jari]		19.28	v1.22		NotReleased
;; - Martin Steppler <steppler@comnets.rwth-aachen.de> proposed
;;   to add Dick Knowles <knowles@inmet.com>'s dired-ediff command
;;   to this package.
;; - Thank you. Added as is. Great code Dick! Bound to key "te".
;;
;; May	13	1996	[jari]		19.28	v1.21		NotReleased
;; - Mistake in key definitions, There is no function tdd-last-file.
;;   It should have read tdd-last-line.
;;   Thanks Martin Steppler steppler@comnets.rwth-aachen.de
;; - Corrected some documentation spelling errors.
;;
;; May	13	1996	[jari]		19.28	v1.20		NotReleased
;;   tdd-mark-vc-files-in-emacs		:! "tmvv"
;;   tdd-mark-vc-has-diffs		:+ "tmvd"
;;   tdd-get-marked-files		:+ no errors
;;   tdd-get-marked-files		-> tdd-get-marked-files-no-dir
;;
;; May	13	1996	[jari]		19.28	v1.19		NotReleased
;;   tdd-marked-revert-files		:+ "tfr" binding
;;   tdd-mark-files-in-emacs		:+ "tme" binding
;;
;; Apr	26	1996	[jari]		19.28	v1.18		NotReleased
;; - The tdd-kill-lines now saves cursor position and only kills filenames.
;;   tdd-kill-lines			:!
;;
;; Apr	19	1996	[jari]		19.28	v1.17		NotReleased
;; - Noticed, that this package should disable itself in VAX dired listing.
;;   tdd-remove-bindings		:+
;;   tdd-vax-p				:+
;;   tdd-handle-vax			:+
;;   tdd-first-file			:! error corr.
;;   tdd-next-file			:! error corr.
;;
;; Apr	15	1996	[jari]		19.28	v1.16		NotReleased
;; - This requires tinylib 1.39, ange ftp correction
;; - Jan Braun <Jan.Braun@tu-bs.de> said that he wished that he could use
;;   "!" command in ange ftp buffers too. After thinking it a while
;;   I programmed a partial support. This doesn't allow transparent
;;   actions for remote files, but rather running, say "xdvi" on _local_copy
;;   of the file.
;;   tdd-:tmp-dir			:+
;;   tdd-:tmp-dir-function		:+
;;   tdd-create-tmp-dir			:+
;;   tdd-get-tmp-dir			:+
;;   tdd-dired-do-shell-command		:+
;;
;; Mar	29	1996	[jari]		19.28	v1.15		NotReleased
;; - Now the expand/collapse links are complete. Now the you can do
;;   the opposite of tdd-shorten-links (when it's in readin-hook) to see
;;   what were the original directory.
;;   tdd-dir-original			:+
;;   tdd-dir-original-get-line		:+
;;   tdd-:file-store			:! made buffer local
;;   tdd-lenghten-links			:! totally rewritten
;;
;; Mar	29	1996	[jari]		19.28	v1.14		NotReleased
;; - Added support for mput ange buffers. Now easy switching to them
;;   from dired to see that all went ok. Commands "abb" "abp".
;;   tdd-:mput-last-ftp			:+
;;   tdd-:previous-buffer		:+
;;   tdd-switch-to-mput-ange-ftp-buffer	:+
;;
;; Mar	8	1996	[jari]		19.28	v1.13		NotReleased
;; - Changed version variables not to have ":"
;;
;; Feb	22	1996	[jari]		19.28	v1.12		NotReleased
;; - vc-co advice problem corrected. Renamed 'my-' to 'tdd-'
;; - Realized that I needed function that would marks all VC files I
;;   had loaded into emacs, so that I could start dired-vc-mode
;;   for them...
;;   tdd-mark-file-list			:+
;;   tdd-mark-vc-files-in-emacs		:+ user func "tmv"
;;   tdd-load-all-marked-files		:! rewritten
;;
;; Feb	18	1996	[jari]		19.28	v1.11		NotReleased
;; - Rewrote the hook-control to use advanced functions from lib 'M'.
;; - Got annoyed to the link references and made function to shorten them.
;;   tdd-:readin-hook			:+ user var
;;   tdd-shorten-links			:+ user func, keys "t<"
;;   tdd-lenghten-links			:+ user func, keys "t>"
;;   tdd-map-over-re			:+ renamed, changed
;;   tdd-map-over-files			:! rewritten
;;   tdd-first-line-point		:! bug, moved point
;;   tdd-last-line-point	pppp	:! bug, moved point
;;   tdd-save-mark-1			-> tdd-remember-marks
;;
;; Feb	18	1996	[jari]		19.28	v1.10		NotReleased
;; - Serious ByteComp errors corrected. Added more require statements
;; - Added 'how to use 3', 'Notes', and many VC related commands.
;; - Bug found. If dired did internal narrowing the sort code went
;;   broke. (eg. "l" to redisplay line). Added new function ti lib 'M'
;;   to check narrow.
;; - Installlation 'marks' corrected. Now they can be ripped correctly.
;; - Added 'ho to use..3', 'Notes'  sections
;; - Rewrote the whole keybinding thing to be more carefull. Also
;;   took care of some dired-x things to prevent overlapping.
;;   tdd-:force-add-keys-flag		:+
;;   tdd-feature-p			:+
;;   tdd-normal-buffer-p		:+ user fun
;;   tdd-map-over-files			:+
;;   tdd-remember-marks			:+
;;   tdd-first-line-point		:+
;;   tdd-last-line-point		:+
;;   tdd-feature-p			:+
;;   tdd-normal-buffer-p		:+
;;   tdd-save-marks			:+ user func
;;   tdd-restore-marks			:+ user func
;;   tdd-first-file			:+ user func
;;   tdd-mark-today-files		:+ user func
;;   tdd-leave-marked-files		:+ user func
;;   tdd-marked-vc-ci			:+ user func, VC, experimental
;;   tdd-marked-vc-co			:+ user func, VC, experimental
;;   tdd-marked-vc-revert		:+ user func, VC, experimental
;;   tdd-leave-only-lines		:! rewritten
;;   tdd-dir-sort			:! rewritten
;;   tdd-default-other-bindings		:!
;;   tdd-kill-files			:! rewritten dired-x
;;   tdd-dir-sort			:! rewritten, narrow problem
;;
;;   vc-finish-logentry			:+ adviced
;;   dired-find-file			:! adv, dired-x handled
;;
;; Feb	16	1996	[jari]		19.28	v1.9		NotReleased
;; - Inpired by recenbt post in news while I was going to bed. New feature
;;   tdd-load-all-marked-files		:+
;;
;; Feb	16	1996	[jari]		19.28	v1.8		NotReleased
;; - Reassigned the keys: now uses "t" prefix for non-ange keys.
;; - I wanted some more features to dired, but they turned out to be quite
;;   hard to implement. Many hours went struggling with them. Eg the sorting
;;   didn't work correctly if there were any marks.
;; - I'm just experimenting with DIRED and VC combinbation,
;;   I have plan for smart ci and co commads over marked files.. :-)
;; - Now there is 2 new user functions: refresh and 'pop'
;;   tdd-first-line		        :! rewritten
;;   tdd-get-mark			:+
;;   tdd-first-line-point		:+
;;   tdd-last-line-point		:+
;;   tdd-view-get			:+
;;   tdd-pop-to-buffer			:+
;;   tdd-refresh-view			:+
;;
;; Feb	16	1996	[jari]		19.28	v1.7		NotReleased
;; - Error in variable tdd-:show-storage-function caused 19.29+ emacs
;;   to freeze. Funny, that the same code worked in 19.28 :-)
;;   There was "" used inside doc string --> changed to \"\"
;; - Setup: one error, the hook name should be tdd-:load-hook
;; - Added 'bugs' and 'thanks' sections.
;;
;; Feb	16	1996	[jari]		19.28	v1.6		NotReleased
;; - There were wrong version functions, copied from tdi. Moved hook
;;   functions to their own place, among the rest of the code.
;;   tdd-kill-all-ange-buffers)		:+
;;
;; Feb	12	1996	[jari]		19.28	v1.5		NotReleased
;; - Corrected some nasty ByteCompiler errors which I hadn't noticed.
;; - Rewrote the whole installation. Much cleaner now.
;; - Added 'important' section. User can now use rsz-mini.el easily for
;;   showing the full storage.
;;   tdd-:bind-hook			:+
;;   tdd-default-other-bindings		:+
;;   tdd-:show-storage-function		:+
;;   tdd-:page-step			:+
;;   tdd-leave-only-lines		:+
;;   tdd-pgup, tdd-pgdown, tdd-first-line , tdd-last-line :+
;;   tdd-hook-control			:! rewritten
;;   tdd-switch-to-some-ange-ftp-buffer	:! bug in promt when big listing
;;   tdd-store-ftp-mput			:! bug in promt when big listing
;;
;; Feb	11	1996	[jari]		19.28	v1.4		NotReleased
;; - Now yo can select files in local dir and _upload_ then to
;;   remote site. Great, just what I needed fro my cs.uta.fi ftp site
;; - New key bindings "ap" and "ag", the "af was removed. ByteComp ok.
;;   tdd-:download-dir			:+
;;   tdd-store-ftp-mget			:+
;;   tdd-switch-to-some-ange-ftp-buffer	:+ easy ange buffer select
;;   tdd-switch-to-ange-ftp-buffer	:! corrected
;;
;; Jan	31	1996	[jari]		19.28	v1.3		NotReleased
;; - Textual changes
;;
;; Jan	31	1996	[jari]		19.28	v1.2		NotReleased
;; - Cleared all code and throwed away dired-nstd.el functions and idea
;;   to make them work in 19.xx. Too much trouble and not much to gain.
;;   The same can be done with the "f" -- ascending to
;;   directories.
;; - Started creating whole new package that would ease
;;   downloading files with ange-ftp.
;;
;; Jan	31	1996	[jari]		19.28	v1.1		Created

;;}}}

;;; Code:

;;{{{ setup: require

;;; ......................................................... &require ...


(require 'backquote)
(require 'dired)
(require 'advice)

(require 'tinylibm)


(eval-and-compile
  (if (xemacs-p)			;Some dired emulations
    (require 'tinylibxe))

  ;; We really don't need to load full packages, so use these..
  ;;
  (defvar   vc-dired-mode			nil)
  (autoload 'vc-dired-mode			"vc")
  (autoload 'vc-finish-logentry			"vc")
  (autoload 'vc-next-action-on-file		"vc")
  (autoload 'vc-workfile-unchanged-p		"vc")

  ;; Too bad that can't autoload this one...
  ;;
  (defvar    vc-dired-mode			nil)
  (autoload 'vc-registered			"vc-hooks")

  ;; The ange interface in this package is based on Emacs only
  (if (emacs-p)
      (autoload 'ange-ftp-ftp-name			"ange-ftp"))

  (autoload 'dired-bunch-files			"dired-aux")
  (autoload 'dired-run-shell-command		"dired-aux")
  (autoload 'dired-shell-stuff-it		"dired-aux")

  (defvar   diff-switches			nil) ;in diff.el
  (autoload 'ediff-files			"ediff"	t t)


  (autoload 'ti::s-verify-ends			"tinylib")
  (autoload 'ti::b-buffer-list			"tinylib")
  (autoload 'ti::b-kill-line			"tinylib")
  (autoload 'ti::c-time-list			"tinylib")
  (autoload 'ti::b-find-ange-to-dired-buffer	"tinylib")
  (autoload 'ti::b-get-ange-buffer-list		"tinylib")
  (autoload 'ti::s-index-substring		"tinylib")
  (autoload 'ti::f-ange-file-handle		"tinylib")
  )

(eval-when-compile
  (use-dynamic-compilation)
  (when (xemacs-p)
    (message "\n\
tinydired.el: ** This package is for Emacs 19.28+ only.\n\
	      ** Dired and ange-ftp interfaces are incompatible between\n\
	      ** Emacs and XEmacs. Use at your own risk.")
    ))


(defgroup-tiny TinyDired tdd-: extensions
  "Dired enchancements. Backgroud Ange ftp support.
  Overview of features

	o   Connecting to VAX host supported
	o   Connection to DOS host supported (experimental)
	o   Few enchancements to dired mode. Eg. keep only one
	    dired buffer when ascending to directory. Shorten symlinks.
	o   User can mark and put files into STORE and start a backgroud
	    ange-ftp session to get STORED files into download directory
	o   Easy switching between ange-ftp session buffer and dired buffer
	o   Dealing with ange ftp buffers in general
	    (x)  killing all ange buffers at once
	    (x)  killing all ange + dired ange buffers at once.
	    (x)  switching to ange buffers with completion
	o   Special VC dired commands (experimental)
	o   Run ! on ange ftp dired buffer (operate on local copy)
	o   customizable backup file flagging.
	o   other handy dired commands, like: pop to this file in emacs.
	    find all marked files ..."
  )

;;}}}
;;{{{ setup: vars


;;; .......................................................... &v-bind ...
;;; handle extra keybindings.


(defcustom tdd-:bind-hook
  '(tdd-default-ange-bindings
    tdd-default-other-bindings
    )
  "*Single function or list of functions to bind keys.
These are installed to `dired-mode-hook' automatically when this package
is loaded."
  :type 'hook :group 'Tinydired
  )


;;; ......................................................... &v-hooks ...

(defcustom tdd-:readin-hook
  '(tdd-handle-vax
    tdd-sort-dir
    tdd-kill-files
    tdd-shorten-links
    )
  "*List of functions that should be run after each dired call.
These are inserted into `dired-after-readin-hook' when package
installs itself. Do not remove 'tdd-sort-dir' or it paralyses package."
  :type 'hook :group 'Tinydired
  )



(defcustom tdd-:load-hook 'tdd-xemacs-note
  "*Hook run when package has been loaded."
  :type 'hook :group 'Tinydired
  )


;;; ....................................................... &v-private ...

(defvar tdd-:file-store nil
  "Private. Storage of filenames.")
(make-variable-buffer-local 'tdd-:file-store)


(defvar tdd-:directory nil
  "Private. Directory name.")


(defvar tdd-:mark-list  nil
  "Private. Saved filename mark list.")

(defvar tdd-:mput-last-ftp  nil
  "Private. Last ftp mput site string.")
(make-variable-buffer-local 'tdd-:mput-last-ftp)


(defvar tdd-:mput-history  nil
  "Private. History variable.")


(defvar tdd-:previous-buffer  nil
  "Private. Recorded buffer, before leaping in another.")


(defvar tdd-:dir-copy-buffer  "*tdd-dir*"
  "Private. Copy of current directory. Created every time when needed.")


(defvar tdd-:dired-directory-ange-regexp "[@:]"
  "Regexp to match `dired-directory' to find ange-ftp buffers.")




;;}}}
;;{{{ setup: User vars

;;; ........................................................ &v-public ...
;;; User configurable

(defcustom tdd-:tmp-dir  "~/tmp/dired/"
  "*Temporary directory where store ange ftp files.
This should be user's private directory, and _must_not_ not be
/tmp,  because someone else may be running tinydired too and using
same filenames."
  :type 'directory :group 'Tinydired
  )

(defcustom tdd-:download-dir "~/ftp/"
  "*Directory where to down load selected files in dired listing."
  :type 'directory :group 'Tinydired
  )


(defcustom tdd-:force-add-keys-flag nil
  "*Non-nil means to install and override default keys to dired.
Normally the keys are defined _only_ if the prefix key is in state
'undefined"
  :type 'boolean :group 'Tinydired
  )


(defcustom tdd-:use-only-one-buffer-flag t
  "*Non-nil means the previous dired buffer is killed when ascending to next.
This makes sure you have only one dired buffer for each dired session.
This feature is not used if dired-x is present."
  :type 'boolean :group 'Tinydired
  )


(defcustom tdd-:page-step  10
  "*Page Up step size in lines."
  :type 'integer :group 'Tinydired
  )

(defcustom tdd-:enable-dos-support-flag  nil
  "*Non-nil means that DOS dired support is enables.
Once the dos support is it, it cannot be cancelled."
  :type 'boolean :group 'Tinydired
  )


(defcustom tdd-:unwanted-files-regexp "\\.o$\\|~$"
  "*Regexp to match files that should not be shown in dired buffer.
Set to nil, if you want to see all files.
This feature is not used if dired-x is present."
  :type '(string :tag "Regexp") :group 'Tinydired
  )


(defcustom tdd-:backup-file-regexp "\\(\\.bak\\|\\.backup\\|[~#]\\)$"
  "*Backup filename regexp, used by advised `dired-flag-backup-files'."
  :type '(string :tag "Regexp") :group 'Tinydired
  )



(defcustom tdd-:mput-sites nil
  "*List of ange-ftp style site location strings, where user can upload files.

Format '(\"ANGE-FTP-REF\" ..),  ange-ftp-ref is like /login@site:dir/dir/"
  :type '(repeat (string :tag "Ange-Ftp"))
  :group 'Tinydired
  )



(defcustom tdd-:tmp-dir-function 'tdd-create-tmp-dir
  "*Create directory for `tdd-:tmp-dir'."
  :type 'function :group 'Tinydired
  )


(defcustom tdd-:show-storage-function 'message
  "*How to show the storage to user. Default is to use 'message' function.
The function is called with _one_ string argument containing
all files separated by spaces.

If you have rsz-mini.el package [std emacs 19.2x], you can set this
to function

         (function
	  (lambda (args)
	    (cond
	     ((< (length args) 75)	;fits in minibuffer
	      (message args))
	     (t				;too long, show them all
	      (read-from-minibuffer \"\" args)
              )))

Which activates the minibuffer resizing automatically and you can see all
files in storage."
  :type 'function :group 'Tinydired
  )


;;}}}
;;{{{ version

;;; ....................................................... &v-version ...

(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinydired.el"
 "tdd"
 tdd-:version-id
 "$Id: tinydired.el,v 1.64 1997/06/19 09:34:47 jaalto NotReleased $"
 '(tdd-:version-id
   tdd-:bind-hook
   tdd-:readin-hook
   tdd-:load-hook

   tdd-:file-store
   tdd-:mark-list
   tdd-:mput-last-ftp
   tdd-:mput-last-ftp
   tdd-:previous-buffer
   tdd-:dir-copy-buffer

   tdd-:tmp-dir
   tdd-:tmp-dir-function
   tdd-:force-add-keys-flag
   tdd-:use-only-one-buffer-flag
   tdd-:unwanted-files-regexp
   tdd-:download-dir
   tdd-:mput-sites
   tdd-:show-storage-function
   tdd-:page-step

   ;;  This tells if used has dired-x loaded
   dired-find-subdir
   )))

;;}}}

;;; ########################################################### &Funcs ###

;;{{{ code: bind, hook control

;;; ----------------------------------------------------------------------
;;;
(defun tdd-default-ange-bindings ()
  "Extra dired bindings."

  ;;  "a" for Ange ftp related commands, since the file information
  ;;  stored is best used in *ftp* buffer itself.
  ;;
  (cond
   ((or (eq 'undefined (lookup-key  dired-mode-map "a"))
	tdd-:force-add-keys-flag)

    ;; clear this only it the map is not in our use.
    (if (not (keymapp (lookup-key  dired-mode-map "a")))
	(define-key dired-mode-map "a" nil))

    ;;  "b"  for buffer handling
    ;;
    (define-key dired-mode-map "abb" 'tdd-switch-to-ange-ftp-buffer)
    (define-key dired-mode-map "abp" 'tdd-switch-to-mput-ange-ftp-buffer)
    (define-key dired-mode-map "abk" 'tdd-kill-dired-and-ange-session)

    ;; we redefine the key "q" too. Was 'dired-delete-and-exit'
    ;;
    (define-key dired-mode-map "q" 'tdd-kill-dired-and-ange-session)

    (when (emacs-p)			;Sorry, not for XEmacs

      (define-key dired-mode-map "as" 'tdd-store-filename)
      (define-key dired-mode-map "ad" 'tdd-store-delete-filename)
      (define-key dired-mode-map "aS" 'tdd-store-add-marked)

      (define-key dired-mode-map "ar" 'tdd-store-remove-file)
      (define-key dired-mode-map "aR" 'tdd-store-delete-marked)

      (define-key dired-mode-map "ac" 'tdd-store-clear)

      ;;  the "q" is just close to "a" key, no other particular logic used.
      ;;
      (define-key dired-mode-map "aq" 'tdd-store-show)

      ;;  "g"  for "get"
      ;;
      (define-key dired-mode-map "ag" 'tdd-store-ftp-mget)
      (define-key dired-mode-map "ap" 'tdd-store-ftp-mput)
      )

    ))

  nil
  )


;;; ----------------------------------------------------------------------
;;;
(defun tdd-default-other-bindings ()
  "Extra dired bindings."

  ;;  Use prefix
  ;;  "t"   -- Base for "tiny extensions"
  ;;  "tg"  -- "g"  mimics dired "g" behavior
  ;;

  (cond
   ((or (eq 'undefined (lookup-key  dired-mode-map "t"))
	tdd-:force-add-keys-flag)

    ;;  make prefix key available for us.
    (if (not (keymapp (lookup-key  dired-mode-map "t")))
	(define-key dired-mode-map "t" nil))

    ;;  You propably want to do also
    ;;  (define-key  dired-mode-map "!"	'tdd-dired-do-shell-command)
    ;;
    (define-key  dired-mode-map "t!"	'tdd-dired-do-shell-command)
    (define-key  dired-mode-map "t-"	'tdd-one-dir-up)

    ;;  "f" for find-file related
    (define-key  dired-mode-map "tff"	'tdd-load-all-marked-files)
    (define-key  dired-mode-map "tfr"	'tdd-marked-revert-files)

    (define-key  dired-mode-map "tg"	'tdd-refresh-view)
    (define-key  dired-mode-map "tG"	'tdd-read-dir-as-is)

    (define-key  dired-mode-map "tkk"	'tdd-kill-lines)
    (define-key  dired-mode-map "tkm"	'tdd-kill-marked-lines)
    (define-key  dired-mode-map "tkM"	'tdd-kill-unmarked-lines)

    (define-key  dired-mode-map "tl"	'tdd-leave-only-lines)
    (define-key  dired-mode-map "tp"	'tdd-pop-to-buffer)
    (define-key  dired-mode-map "te"	'tdd-ediff)

    (define-key  dired-mode-map "t<"    'tdd-shorten-links)
    (define-key  dired-mode-map "t>"    'tdd-lenghten-links)

    ;;  Mark related commands in "m" map
    ;;
    (define-key  dired-mode-map "tme"	'tdd-mark-files-in-emacs)
    (define-key  dired-mode-map "tmd"	'tdd-mark-today-files)
    (define-key  dired-mode-map "tmo"	'tdd-mark-opposite-toggle)
    (define-key  dired-mode-map "tmr"	'tdd-mark-read-only-files)
    (define-key  dired-mode-map "tms"	'tdd-marks-save)
    (define-key  dired-mode-map "tmS"	'tdd-marks-restore)
    (define-key  dired-mode-map "tmw"	'tdd-mark-writable-files)

    (define-key  dired-mode-map "tmvv"	'tdd-mark-vc-files-in-emacs)
    (define-key  dired-mode-map "tmvd"  'tdd-mark-vc-has-diffs)

    ;;  some special VC functions for marked files in "v" map
    ;;
    (define-key dired-mode-map "tvi"	'tdd-marked-vc-ci)
    (define-key dired-mode-map "tvo"	'tdd-marked-vc-co)
    (define-key dired-mode-map "tvu"	'tdd-marked-vc-revert)

    ;;  Override some Emacs default bindings to better follow
    ;;  this buffer's content.
    ;;
    (define-key dired-mode-map "\M-<"	'tdd-first-line)
    (define-key dired-mode-map "\M->"	'tdd-last-line)
    (define-key dired-mode-map [home]	'tdd-first-file)
    (define-key dired-mode-map [end]	'tdd-last-line)
    ))

  ;;  This is HP-UX settings, check compile flag of this emacs.
  ;;
  (cond
   ((string-match "hpux" (emacs-version))
    (define-key dired-mode-map [select]	'tdd-last-line) ;; 'end' key
    (define-key dired-mode-map [prior]  'tdd-pgup)
    (define-key dired-mode-map [next]   'tdd-pgdown)
    ))

  nil
  )


;;; ----------------------------------------------------------------------
;;; - I notices that it's not good idea to have tdd active in VAX
;;;   dired listings..
;;; - If user has relocated some keys...well, we don't handle those.
;;;
(defun tdd-remove-bindings ()
  "Remove bindings from this dired session.
User must be in Dired buffer Makes the `dired-mode-map' local to current buffer."
  (interactive)
  (let* ((list
	  '(
	    "abb" "abp" "as" "aS" "ar" "aR" "ac" "aq" "ag" "ap"
	    "t!" "tf" "tg" "tk" "tl" "tp" "t<" "t>"
	    "tmd" "tml" "tms" "tmS" "tmv"
	    "tvi" "tvo" "tvu"
	    ))
	 )
    (cond
     ((and (memq major-mode '(dired-mode))
	   dired-mode-map)
      (make-local-variable 'dired-mode-map)

      (while list
	(define-key dired-mode-map (car list) 'tdd-ignore)
	(NEXTP list))

      ;;  And the rest
      (define-key dired-mode-map "\M-<"	'beginning-of-buffer)
      (define-key dired-mode-map "\M->"	'end-of-buffer)
      (define-key dired-mode-map [home]	'beginning-of-buffer)
      (define-key dired-mode-map [end]	'end-of-buffer)
      ))
    ))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-hook-control (&optional remove)
  "Add hooks to dired mode. Optional REMOVE all hooks inserted by package."
  (interactive "P")
  (let* ((list (list-make tdd-:bind-hook))
	 x
	 )
    (cond
     (remove
      (add-hooks 'dired-after-readin-hook tdd-:readin-hook	'remove)
      (add-hooks 'dired-mode-hook		tdd-:bind-hook	'remove)
      )
     (t
      ;;  Now, install the package
      ;;
      (add-hooks 'dired-after-readin-hook tdd-:readin-hook)

      (list-loop list x			;bind the keys
	(add-hook 'dired-mode-hook x)

	;;  This is due to autoload: while the package is beeing loaded,
	;;  it should also set the bindings immediately
	;;
	(if (boundp 'dired-mode-map)
	    (funcall x))
	)
      ))
    nil					;clean ending
    ))



;;}}}
;;{{{ XEmacs compatibility

;;; .............................................................. &xe ...
;;; Some functions are not found from XEmacs, mimic them


;;; ----------------------------------------------------------------------
;;;
(defun tdd-dired-unmark-all-files-no-query  ()
  "XEmacs compatibility."
  (let* (
         )
    (if (fboundp 'dired-unmark-all-files-no-query)
	(sfuncall 'dired-unmark-all-files-no-query)
      (save-line-col nil nil
	(tdd-first-line)
	(while (or (not (eobp))
		   (not (looking-at "^[ \t]*$")))
	  ;;  Just use brute force for all lines.
	  (dired-unmark 1))
	))))


;;; ----------------------------------------------------------------------
;;;
(defun tdd-xemacs-note  ()
  "Warn that tinydired.el may work improperly in XEmacs."
  (if (xemacs-p)
      (read-char-safe-until
       "\
You do realize that all TinyDired's features won't work with XEmacs? [press]"
       )))


;;}}}

;;{{{ code: ange-ftp.el

;;; ............................................................ &ange ...

;;; ----------------------------------------------------------------------
;;;
(defun tdd-advice-control  (&optional disable verb)
  "Activate all advises. Use extra argument to DISABLE all. VERB."
  (interactive "P")
  (let* ((re    "^tdd")
         (doit  t)
         msg
         )
    (VERB)
    ;;  This is rough! Be sure...
    (if verb
        (setq
         doit
         (y-or-n-p (format "advices %s: No mistake here? "
                           (if disable "off" "on")))))
    (when doit
      (cond
       (disable
        (ad-disable-regexp re)          ;only sets flag
        (setq msg "All advices deactivated")
        )
       (t
        (ad-enable-regexp re)           ;only sets flag
        (setq msg "All advices activated")
        ))
      (ad-update-regexp re)
      (if verb  (message msg))
      )
    ))



;;; ----------------------------------------------------------------------
;;;
;; - 19.28 has error , it calls apply with two parameters
;;   apply(ange-ftp-file-name-sans-versions ("/jaalto@...tt-vise.txt" nil))
;;
;; - we add '&rest args' here

(if (eq emacs-minor-version 28)
    (defun ange-ftp-vms-sans-version (name &rest args)
      "Patched."
      (save-match-data
	(if (string-match ";[0-9]+$" name)
	    (substring name 0 (match-beginning 0))
	  name))))


;;; ----------------------------------------------------------------------
;;;
(defadvice ange-ftp-set-binary-mode  (before tdd-error-prevent-fix  act)
  "Sometimes you can get error:
ash(nil -4)
  ange-ftp-set-binary-mode(\"ftp.uit.no\" \"ftp\")

Which is due to missing variables
  ange-ftp-ascii-hash-mark-size
  ange-ftp-binary-hash-mark-size

This advice resets them to some default values, so that you don't get
errors."
  (save-excursion
    (set-buffer (ange-ftp-ftp-process-buffer host user))
    (if (null ange-ftp-ascii-hash-mark-size)
	(setq ange-ftp-ascii-hash-mark-size 1024))
    (if (null ange-ftp-binary-hash-mark-size)
	(setq ange-ftp-binary-hash-mark-size 1024))
    ))




;;}}}
;;{{{ code: VAX/PC  support (advice)

;;; ........................................................... &dired ...
;;; If ange ftp creates VAX directory, user can't load file on the line
;;; --> fixed.

;;;  This is same as default 19.28 disrtib regexp + VAX extension
;;;  Do not touch there unless with good reason
;;;
(defconst tdd-dired-move-to-filename-regexp-unix
  (concat
   "\\(Jan\\|Feb\\|Mar\\|Apr\\|May\\|Jun\\|Jul\\|Aug\\|Sep"
   "\\|Oct\\|Nov\\|Dec\\)"
   "[ ]+[0-9]+ [ 0-9][0-9][:0-9][0-9][ 0-9] "
   )
  "Filename position beginning in Unix.
Do not touch this unless you know what you're doing.")

(defconst tdd-dired-move-to-filename-regexp-vax
   ;; FTP_SERVER.LOG;8 0  31-AUG-1995 15:10 [JAALTO] (RWED,RWED,RE,RE)
   ;; The "D" is delete flag, "*" is mark flag.
   ;;
   ;; Also suitable for PC.
   ;;
   "^[D*]?[ \t]+"
  "Filename position beginning in VAX/VMS.
o not touch this unless you know what you're doing.")


;;; ----------------------------------------------------------------------
;;;
;;;
(defun tdd-dired-vax-p ()
  "Check if dired listing is vax/Dos .
Changes dired variable `dired-move-to-filename-regexp' if needed."
  (save-excursion
    (PMIN)
    (if (or (re-search-forward "\\[.*\\]" nil t) ;; VAX like directory
	    ;;
	    ;;  COMMAND.COM        54619  09-30-1993  07:20
	    ;;  MSOFFICE      <DIR>       05-19-95  08:55
	    ;;
	    (re-search-forward "^[ \t]+[^ \t]+[ \t]+[0-9][0-9]-" nil t)
	    (re-search-forward "^[ \t]+[^ \t]+[ \t]+<" nil t)
	    )
        (setq dired-move-to-filename-regexp
              tdd-dired-move-to-filename-regexp-vax)
      (setq dired-move-to-filename-regexp
            tdd-dired-move-to-filename-regexp-unix)
      nil)
    ))

;;; ----------------------------------------------------------------------
;;;
(defadvice dired-move-to-end-of-filename (around tdd-vax-files act)
  "Replace function. Handles vax file Directory."
  (if (null (tdd-dired-vax-p))
      ad-do-it
    (beginning-of-line)
    (cond
     ;;  .RHOSTS;2  1  21-JUN-1995 10:57 [JAALTO] (RWED,RWED,,)
     ((re-search-forward ";[0-9]+" nil t))
     ;;
     ;; PC
     ;;
     ((re-search-forward "^[ \t]+[^ \t\n]+" nil t))
     ((error "Can't find vax/Dos file end" (point) )
      ))
    (setq ad-return-value (point))
    )
  ad-return-value
  )

;;; ----------------------------------------------------------------------
;;;
(defadvice dired-get-filename (around tdd-vax-get-filename act)
  "Replace function. Handle vax directory."
  (let* (fb                             ;file begin, file end
         fe
         ffile
         )
    (if (null (tdd-dired-vax-p))
        ad-do-it
      (if (not
           (and
            (setq fb (dired-move-to-filename))
            (setq fe (dired-move-to-end-of-filename))))
          (error "Cannot determine vax/Dos file")
        (setq ffile (buffer-substring fb fe))

        (cond
         ((eq 'no-dir localp)
          (setq ad-return-value ffile))
         (t
          (setq ad-return-value
                (concat (dired-current-directory localp) ffile))))
        ))
    )
  ad-return-value
  )


;;; ..................................................... &Dos-support ...
;;; You should set this variable, so that commnd "dir" is used
;;; to read the listing in PC hosts.
;;;
;;; (setq ange-ftp-dumb-unix-host-regexp  "tntpc") ;PC hosts


(defun tdd-dos-dired-fix  ()
  "If the `dired-directory' entry is not on a line by itself, add newline.
This affects only DOS buffers.

  /root@tntpc53:/  COMMAND.COM        54619  09-30-1993  07:20
  MSOFFICE      <DIR>       05-19-95  08:55

-->
  /root@tntpc53:/
  COMMAND.COM        54619  09-30-1993  07:20
  MSOFFICE      <DIR>       05-19-95  08:55"
  (let* (buffer-read-only
         )
    (save-excursion
      (PMIN)
      (forward-char (+ 2 (length dired-directory)))
      (when (and (tdd-dos-dir-p)
		 (not (looking-at "[ \t]*$")))
	(insert "\n")
	))))

;;; ----------------------------------------------------------------------
;;;
(defun tdd-dos-dir-p  ()
  "Check if current line is DOS directory,.
Return 'back if the entry is '..' and t if it is other directory."
    (cond
     ((and (eq major-mode 'dired-mode) ;Handle PC dired directory line
	   (save-excursion
	     (beginning-of-line)
	     (and (looking-at ".*[ \t]+[<]DIR[>][ \t]+")
		  (if (looking-at "^[ \t]+\\.\\.")
		      'back t))
	     )))))


(when tdd-:enable-dos-support-flag
;;; --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++--

(add-hook 'dired-after-readin-hook 'tdd-dos-dired-fix)
(remove-hook 'dired-after-readin-hook 'tdd-dos-dired-fix)

;;; ----------------------------------------------------------------------
;;; (ad-unadvise 'ange-ftp-get-pwd)
;;;
;;; (defun ange-ftp-get-pwd (host user)
;;;
(defadvice ange-ftp-get-pwd  (around tdd-dos  act)
  "Change DOS C:\dir\dir into  /dir/dir. Replace function."
  (let* ((result (ange-ftp-send-cmd host user '(pwd) "Getting PWD"))
         (line (cdr result))
         dir)

     (when (car result)
       (save-match-data
	 (cond
	  ((string-match "^.:\\(\.*\\)" (cdr result))
	   (setq dir (read-match 1 (cdr result))
		 dir (subst-char dir ?\ ?/))
	   )
	  (t
	   (and (or (string-match "\"\\([^\"]*\\)\"" line)
		    (string-match " \\([^ ]+\\) " line))
		;;  stone-age VMS servers!
		(setq dir (substring line
				   (match-beginning 1)
				   (match-end 1))))
	   ))))
     (setq ad-return-value (cons dir line))
     ))

;;; ----------------------------------------------------------------------
;;; (ad-unadvise 'ange-ftp-expand-file-name)
;;;
;;;(defun ange-ftp-expand-file-name (name &optional default)
;;;
(defadvice ange-ftp-expand-file-name (around tdd-dos  act)
  "Handle '..' directory in DOS dired listing."
  (let* ((stat (tdd-dos-dir-p))
	 file host dir
	 )
    ;;  In DOs listing, hitting the "f" key in dired over ".." returns only
    ;;  The site address like "/root@tntpc53:/"
    ;;
    ;;  But if you hit "f" over real directory name, you get correct entry
    ;;  /root@tntpc53:/MSOFFICE/SETUP
    ;;
    ;;  /root@tntpc53:/MSOFFICE:
    ;;  ..            <DIR>       05-19-95  08:55
    ;;  SETUP         <DIR>       05-19-95  08:55
    ;;
    ;;  I have no clue why this happens. For now we just fix the ".."
    ;;  special case. The 'default' arg holds the dired root dir.

    (setq
     ad-return-value

     (cond
      ((eq stat 'back)
       (setq file (if default
		      default
		    dired-directory))
       (if (string-match "^\\(.*:\\)\\(.*\\)" file)
	   (setq host (read-match 1 file)  file (read-match 2 file))
	 (error "Cannot parse remote DOS dired."))

       ;;   /MSOFFICE/ -->   "/"
       (if (string-match "\\(.*/\\)\\(.*/\\)" file)
	   (setq file (read-match 1 file))
	 ;;  No directories
	 (setq file "/"))

       (concat host file)
       )

      (t
       (save-match-data
	 (if (char= (string-to-char name) ?/)
	     ;;  don't upset Apollo users
	     (while (cond ((string-match "[^:]+//" name)
			   (setq name (substring name (1- (match-end 0)))))
			  ((string-match "/~" name)
			   (setq name (substring name (1- (match-end 0))))))))
	 (cond ((char= (string-to-char name) ?~)
		(ange-ftp-real-expand-file-name name))
	       ((char=  (string-to-char name) ?/)
		(ange-ftp-canonize-filename name))
	       ((zerop (length name))
		(ange-ftp-canonize-filename (or default default-directory)))
	       ((ange-ftp-canonize-filename
		 (concat (file-name-as-directory
			  (or default default-directory))
			 name))))))

      ))
    ))



;;; ----------------------------------------------------------------------
;;; (ad-unadvise 'ange-ftp-get-file-entry)
;;;
(defadvice ange-ftp-get-file-entry (around tdd-dos act)
  "Detect DOS directory line in `dired-mode' listing.
Replaces original function"
  (setq
   ad-return-value

     (cond
      ((tdd-dos-dir-p)
       t)

      (t
       (let* ((name (directory-file-name name))
	      (dir (file-name-directory name))
	      (ent (ange-ftp-get-hash-entry dir ange-ftp-files-hashtable))
	      (file (ange-ftp-get-file-part name)))
	 (if ent
	     (ange-ftp-get-hash-entry file ent)
	   (or (and (ange-ftp-allow-child-lookup dir file)
		    (setq ent (ange-ftp-get-files name t))
		    (ange-ftp-get-hash-entry "." ent))
	       ;; i.e. it's a directory by child lookup
	       (ange-ftp-get-hash-entry file
					(ange-ftp-get-files dir)))))
       )))
  )

;;; Dos support ends here
;;; --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++--
)


;;}}}
;;{{{ code: misc

;;; .......................................................... &macros ...


;;; ----------------------------------------------------------------------
;;; - See dired-repeat-over-lines
;;;
(defmacro tdd-map-over-re  (re &rest body)
  "If '(looking-at RE)' then do BODY over all lines matching.
Start from current point. The point is positioned at the beginning of line.
Buffer read-only is removed.

The BODY should move the pointer to next file and bol, until eob reached."
  (`
   (let* ((end	    (tdd-last-line-point))
	  buffer-read-only
	  )
     (beginning-of-line)
     (while (and (not (eobp))
		 (< (point) end))
       (beginning-of-line)
;;;       (d! "RE" (looking-at (, re)) (, re)  (str-right (read-full-line) 50))
       (if (looking-at (, re))
	   (progn
	     (,@ body))
	 (forward-line 1))
       ))))


;;; ----------------------------------------------------------------------
;;;
(put 'tdd-map-over-files 'lisp-indent-function 0)
(defmacro tdd-map-over-files  (&rest body)
  "Map over files. No links, no dirs are included.
You must advance the cursor inn the BODY. See `tdd-map-over-re'."
  (`
   (progn
     (tdd-first-file)
     (tdd-map-over-re "^. +[^dl]" (,@ body)))))


;;; ----------------------------------------------------------------------
;;;
(put 'tdd-map-over-unmarked 'lisp-indent-function 0)
(defmacro tdd-map-over-unmarked  (&rest body)
  "Map over unmarked lines and execute BODY at the beginning of line.
The calling BODY should position the cursor for next search so
that current line is skipped when BODY finishes.

The buffer is writable during mapping."
  (`
   (let* (buffer-read-only
	  (ReGexp  (dired-marker-regexp))
	  )
     (progn
       (tdd-map-over-files
	 (if (looking-at ReGexp)
	     (forward-line 1)
	   (beginning-of-line)
	   (,@ body))
	 )))))




;;; ----------------------------------------------------------------------
;;;
(defmacro tdd-remember-marks (var-sym &optional beg end)
  "Save mark list to variable VAR-SYM between points BEG and END.
START and END defaults to all files"
  (`
   (setq (, var-sym)
	 (dired-remember-marks
	  (or (, beg)   (tdd-first-line-point))
	  (or (, end)   (tdd-last-line-point))
	  ))))


;;; ............................................................ &misc ...


;;; ----------------------------------------------------------------------
;;;
(defun tdd-vax-p  ()
  "Check if dired line is VAX listing."
  ;;  We use just simple regexp test
  (and (tdd-normal-buffer-p)
       (save-excursion
	 (PMIN)
	 ;; directory name:                USER115:[JAALTO]
	 ;;
	 (or (re-search-forward ":\\[.*\\]$" nil t)
	     (re-search-forward "[A-Z];[0-9] +" nil t)  ;; version numbers
	     ))))

;;; ----------------------------------------------------------------------
;;;
(defun tdd-handle-vax  ()
  "If dired buffer contain VAX directory structure, disable TDD."
  (if (tdd-vax-p)
      (tdd-remove-bindings)))

;;; ----------------------------------------------------------------------
;;;
(defun tdd-ignore  ()
  "Ignore message."
  (interactive)
  (message "Function is not available in this dired buffer.")
  )


;;; ----------------------------------------------------------------------
;;;
(defun tdd-create-tmp-dir  ()
  "Create directory `tdd-:tmp-dir' if possible."
  (make-directory (expand-file-name tdd-:tmp-dir)))



;;; ----------------------------------------------------------------------
;;;
(defun tdd-get-tmp-dir ()
  "Return temp directory with slash at the end."
  (let* ((dir   tdd-:tmp-dir)
	 (func  tdd-:tmp-dir-function)
	 )
    (cond
     ((not (file-exists-p dir))
      (setq dir (funcall func))
      ))

    (setq dir (expand-file-name dir))

    (if (not (file-exists-p   dir))
	(error "Directory not exist %s" dir))

    (ti::s-verify-ends dir "/")
    ))

;;; ----------------------------------------------------------------------
;;;
(defsubst tdd-get-filename ()
  "Return only filename without directory."
  ;;  The (dired-get-filename t) almos does the same, but it _may_
  ;;  contains slahes.. docs say so.
  ;;
  (strmatget "\\([^/]+\\)$" 1 (dired-get-filename)))


;;; ----------------------------------------------------------------------
;;;
(defsubst tdd-get-mark ()
  "Return first char to the left. Point is not preserved."
  (beginning-of-line)
  (following-char))



;;; ----------------------------------------------------------------------
;;;
(defsubst tdd-get-marked-files ()
  "Signal no errors."
  (safe-do (dired-get-marked-files)))


;;; ----------------------------------------------------------------------
;;;
(defun tdd-get-marked-files-no-dir ()
  "Return LIST of marked files."
  (dired-map-over-marks
   (tdd-get-filename)
   nil))


;;; ----------------------------------------------------------------------
;;;
(defsubst tdd-first-line-point ()
  "Return first file point."
  (save-excursion (tdd-first-line) (BOLP)))


;;; ----------------------------------------------------------------------
;;;
(defsubst tdd-last-line-point ()
  "Return last file point."
  (save-excursion (tdd-last-line) (EOLP)))



;;; ----------------------------------------------------------------------
;;;
(defsubst tdd-mark-re  (re &optional unmark)
  "Mark files matching RE. Give prefix argument to UNMARK."
  (save-excursion
    (PMIN)
    (while (re-search-forward re nil t)
      (if unmark
	  (dired-unmark 1)
	(dired-mark 1)
	))))


;;; ----------------------------------------------------------------------
;;;
(defsubst tdd-mark-file-list (list &optional unmark)
  "Mark files in LIST. Give prefix argument to UNMARK."
  (let* ((list  (list-make list))
	 )
    (while list
      (tdd-mark-re (concat (regexp-quote (car list)) "$") unmark)
      (NEXTP list)
      )))



;;; ----------------------------------------------------------------------
;;;
(defun tdd-feature-p  (arg)
  "Check if we already have this functionality in dired. See ARG from code."

  ;;  19.28 ships with dired-x, which I just noticed had some of
  ;;  the same functionality. We don't use TDD if those
  ;;  are present in some cases.
  (cond
   ((eq arg 'auto-delete)
    (and (featurep 'dired-x)
	 (> emacs-minor-version 27))
    )					;see dired-omit-files-p
  (t
   nil
   )))


;;; ----------------------------------------------------------------------
;;;
;;;
(defun tdd-normal-buffer-p  ()
  "Check if dired  buffer's first two characteristic lines are visible."
  (interactive)
  (and (not (buffer-narrowed-p))
       (save-excursion
	 (PMIN)
	 (and (looking-at "^[ \t]+/")
	      (forward-line 1)
	      (looking-at "^[ \t]+total[ \t]+[0-9]"))
	 )))


;;; ----------------------------------------------------------------------
;;;
(defun tdd-kill-files ()
  "After each dired read, remove unwanted files."
  (let* ((re     tdd-:unwanted-files-regexp)
	 buffer-read-only
	 )
    (cond
     ((not (tdd-feature-p 'auto-delete))
      ;;  Is this new directory buffer ..
      (if (and (eq major-mode 'dired-mode)
	       (stringp re))
	  (flush-lines re))			;don't wanna see these
      nil
      ))))


;;; ----------------------------------------------------------------------
;;;
(defun tdd-view-get (&optional arg mode)
  "Gets all files/dir entries in the view.
The ARG is `dired-get-filename' parameter.

Input:
  ARG		,no absolute names
  MODE		,if 're then make regexp out of files.

Return list:
  (re re ..)		,depends on mode
  ((mark file) ..)	,depends on mode

The 'mark' is first character in the left for file or dir."
  (let* (last-point
	 list
	 file
	 )
    (save-excursion
      (setq last-point (save-excursion (tdd-last-line) (point)))
      (tdd-first-line)

      (if (setq file (safe-do (dired-get-filename arg)))
	  (cond
	   ((eq mode 're)
	    (beginning-of-line)
	    (if (looking-at dired-re-sym)
		(nconcc list (concat (regexp-quote file) " +->"))
	      (nconcc list (format " %s$" (regexp-quote file))))
	    )
	   (t
	    (nconcc list (list (tdd-get-mark) file)))
	    ))

      (while (< (point) last-point)
	(dired-next-line 1)
	(if (setq file (safe-do (dired-get-filename arg)))
	    (cond
	     ((eq mode 're)
	      (beginning-of-line)
	      (if (looking-at dired-re-sym)
		  (nconcc list (concat (regexp-quote file) " +->"))
		(nconcc list (format " %s$" (regexp-quote file))))
	      )
	     (t
	      (nconcc list (list (tdd-get-mark) file)))
	     )))

      )
    list
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tdd-sort-dir ()
  "Puts directories first in dired listing."
  (let (buffer-read-only
	marks
	p1 p2				;points
	region
	)
    ;;  - Buffer gets narrowed in some dired internal operations, like
    ;;    pressing "l", dired-do-redisplay
    ;;  - We do nothing in these cases
    ;;
    ;;  - We have to save position, because eg. pressing "Z" to
    ;;    compress file, causes reading the whole dir --> point moved.

    (when (tdd-normal-buffer-p)
      (save-with-marker
	(tdd-first-line)

	(tdd-remember-marks marks (point))
	(tdd-dired-unmark-all-files-no-query) ; sort goes nuts otherwise
	(message "")			; stupid message from dired-un...

	;; Sort regexp by
	;; 19 Nov 1995, sof@dcs.glasgow.ac.uk (Sigbjorn Finne), comp.emacs
	;;
	(tdd-first-line) (beginning-of-line)
	(sort-regexp-fields t "^.*$" "[ ]*." (point) (point-max))


	;;  now, We prefer to have dirs first, and then links, allthough
	;;  some links may be dirs (we can't know anything about links)
	;;
	(PMIN)
	(when (re-search-forward "^[ \t]+lr" nil t)
	  (setq p1 (BOLPP))
	  ;;  We know that dirs are after links, because the listing is
	  ;;  sorted.
	  (re-search-forward "^[ \t]+d" nil t)
	  (setq p2 (BOLPP))

	  (setq region (buffer-substring p1 p2))
	  (delete-region p1 p2)

	  (re-search-forward "^[ \t]+-" nil t) ;go after dirs
	  (beginning-of-line)
;;;	  (d! p1 p2 (point))
	  (insert region)
	  )


	(dired-mark-remembered marks)
	(set-buffer-modified-p nil)

	))))


;;; ----------------------------------------------------------------------
;;;
(defun tdd-dir-original (dir &optional buffer)
  "Do same as `dired-insert-directory'.
Insert DIR to BUFFER, which defaults to `tdd-:dir-copy-buffer'"
  (save-excursion
    ;;  See dired.el dired-readin-insert
    (use-tmp-buffer (or buffer tdd-:dir-copy-buffer) 'clear)
    (set-buffer (or buffer tdd-:dir-copy-buffer))
    (insert-directory (expand-file-name dir) dired-listing-switches nil t)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tdd-dir-original-get-line (file)
  "Return original line for FILE.
Be sure you have called `tdd-dir-original' first.
Signal no error. Use `regexp-quote' for FILE if it contains unusual characters.

Return:
  line
  nil	,no line was found"
  (let* (line
	 )
    (save-excursion
      (set-buffer tdd-:dir-copy-buffer) (PMIN)

      ;;  Pick first match
      (if (re-search-forward (concat " " file) nil t)
	  (setq line (read-full-line)))
      line
      )))


;;}}}


;;{{{ code: interactive

;;; ..................................................... &interactive ...


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-ediff (file &optional switches)
  "Compare file at point with file FILE using `ediff'.
FILE defaults to the file at the mark.
The prompted-for file is the first file given to `ediff'.
With prefix arg, prompt for second argument SWITCHES,
 which is options for `diff'."
  (interactive
   (let ((default (if (mark t)
                      (save-excursion (goto-char (mark t))
                                      (dired-get-filename t t)))))
     (list

      (read-file-name			;ARG 1
       (format "Ediff %s with: %s"
	       (dired-get-filename t)
	       (if default
		   (concat "(default " default ") ")
		 ""))
       (dired-current-directory) default t)


      (if current-prefix-arg		;ARG 2
	  (read-string
	   "Options for diff: "
	   (if (stringp diff-switches)
	       diff-switches
	     (mapconcat 'identity diff-switches " "))))
      )))
  ;; Interactive part end
  ;;
  (ediff-files file (dired-get-filename t) switches))



;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-read-dir-as-is ()
  "Read the directory without any filtering."
  (interactive)
  (let* (dired-after-readin-hook
         )
    (revert-buffer)
    ))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-mark-files-in-emacs  ()
  "Mark all files in current directory that are in Emacs."
  (interactive)
  (let* ((dir (expand-file-name dired-directory)) ;get rid of "~"
	 (list (ti::b-buffer-list
		(quote
		 (and (buffer-file-name)
		      (string-match (regexp-quote dir) (buffer-file-name))
		      ))))
	 elt
	 )
    (if (null dir) (setq dir dir))	;Shut up XEmacs 19.14 byteCompiler

      (list-loop list elt
	(save-excursion
	  (tdd-first-file)
	  (if (re-search-forward elt nil t)
	      (dired-mark 1))
	  ))
      ))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-marked-revert-files (&optional arg)
  "Revert ie. replace files in Emacs with true copies in directory.
If ARG is non-nil, remove any marks if file was loaded.

Exceptions:
  Only reload files in emacs whose modify flag is non-nil.
  If file does not exist in emacs, do nothing."
  (interactive "P")
  (let* ((list (tdd-get-marked-files))
	 buffer
	 file
	 )
    (while list
      (setq file (car list))

      (cond
       ((setq buffer (get-file-buffer file))
	(save-excursion
	  (set-buffer buffer)

	  (cond
	   ((null (buffer-modified-p))
	    (revert-buffer nil t)	;no confirmation
	    (setq buffer 'done)
	    )))))

      (cond
       ((and arg  (eq 'done buffer))
	(tdd-mark-re
	 (regexp-quote (file-name-nondirectory file)) 'unmark)
	))
      (NEXTP list))
      ))


;;; ----------------------------------------------------------------------
;;; - It's lot faster to use direct command, than search the buffer
;;;   for ".." and use "f" or click mouse over it.
;;;
;;;###autoload
(defun tdd-one-dir-up ()
  "Goes up one directory."
  (interactive)
  (find-file (concat dired-directory "..")))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-dired-do-shell-command  (command &optional arg)
  "Like `dired-do-shell-command', but run running command in dired ange-ftp.
This is not remote shell, but instead it
transfers the file to your local system and then executes the dired
command on the file.

Remember: Every time you run this command this files are copied _blindly_
to your local directory. No file cache information is kept.

Input:
  COMMAND
  ARG

References:
  `tdd-:tmp-dir'"
  (interactive
   (list
    (dired-read-shell-command
     (concat "! on "
	     "%s: ")
     current-prefix-arg
     (dired-get-marked-files
      t current-prefix-arg))

    current-prefix-arg))

;;;  (require 'dired-aux)

  (let* ((to-dir  (tdd-get-tmp-dir))
	 (ange	  (ange-ftp-ftp-name dired-directory))
	 (on-each (not (string-match "\\*" command)))
	 host user dir
	 file-list list
	 )
    (cond
     ((null ange)
      ;; Simple local dired.
      ;;
      (dired-do-shell-command command arg))
     (t
      (setq host  (nth 0 ange)
	    user  (nth 1 ange)
	    dir   (nth 2 ange)
	    )
      (setq file-list (dired-get-marked-files t))
      (ti::f-ange-file-handle 'get user host dir to-dir file-list 'foreground)


      (while file-list			; All directory to every filename
	(push (concat to-dir (car file-list))  list)
	(NEXTP file-list))
      (setq file-list list)


      ;; ......................................... copy from dired-aux ...
      (if on-each
	  (dired-bunch-files
	   (- 10000 (length command))
	   (function
	    (lambda (&rest files)
	      (dired-run-shell-command
	       (dired-shell-stuff-it command files t arg))))
	   nil
	   file-list)
        ;; ... ... ... ... ... ... ... ... ... ... ... ... ... .. then ...
	;; execute the shell command
	(dired-run-shell-command
	 (dired-shell-stuff-it command file-list nil arg)))
      ))

    ))



;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-lenghten-links  ()
  "Opposite to `tdd-shorten-links'.
This may take a while, because the whole directory structure must
be read again."
  (interactive)
  (let* ((line   (LINE))
	 file
	 marks
	 buffer-read-only
	 )
    (when (tdd-normal-buffer-p)
      ;;	Now create copy of original directory.
      ;;
      (tdd-dir-original dired-directory)


      (tdd-remember-marks marks)
      (tdd-dired-unmark-all-files-no-query)
      (message "")

      (dired-mark-symlinks nil)

      ;;   This didn't update full line, only the data part, not the
      ;;   linked name portion "->"
      ;;      (dired-do-redisplay)

      (dired-map-over-marks
       (progn
	 (setq file (dired-get-filename 'no-dir))
	 (setq line (tdd-dir-original-get-line (regexp-quote file)))

	 ;;  now, delete line and relace it with original entry.
	 ;;
	 (cond
	  (line
	   (beginning-of-line)
	   (re-search-forward " l")
	   (backward-char 1)
	   (delete-region (point) (EOLPP))
	   (insert line)
	   ))
	 )
       nil)

      (dired-mark-symlinks 'unmark)

      (if marks
	  (dired-mark-remembered marks))
      (set-buffer-modified-p nil)
      )))


;;; ----------------------------------------------------------------------
;;; - It's awfull to see 30 linked files whyen they don't fit on one line...
;;;
;;;###autoload
(defun tdd-shorten-links  ()
  "Shortens all linked files. The link part is removed."
  (interactive)
  (let* ((line   (LINE))
	 buffer-read-only

	 )
    (when (tdd-normal-buffer-p)
      (PMIN)
      (while (not (eobp))
	(if (looking-at ".* +->\\([^\n]+\\)")
	    (kill-match 1))
	(forward-line 1)
	)
      (goto-line line)
      (dired-move-to-filename)
      )))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-marks-save  ()
  "Save mark list to private storage.
Use this function if you know next operation will remove the marks.
You can get the marks back with `tdd-marks-restore'."
  (interactive)
  (save-excursion			;due to next command
    (tdd-remember-marks tdd-:mark-list)
    (message "Marks saved.")))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-marks-restore  ()
  "Restore mark list saved by `tdd-marks-save'."
  (interactive)
  (if (null tdd-:mark-list)
      (message
       (substitute-command-keys
	"No marks saved. Use '\\[tdd-marks-save]' first."))
    (dired-mark-remembered tdd-:mark-list)))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-pgup ()
  "Moves cursor to _last_ file in dired mode."
  (interactive)
  (dired-next-line (- tdd-:page-step))
  (if (bobp) (tdd-first-line)))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-pgdown ()
  "Moves cursor up."
  (interactive)
  (dired-next-line tdd-:page-step)
  (if (eobp) (tdd-last-line)))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-first-line ()
  "Moves to fist _line_ in dired."
  (interactive)
  (let* (point
	 )
    (save-excursion
      (PMIN) (forward-line)
      (cond
       ((re-search-forward "total +[0-9]" nil t)
	(forward-line 1)
	(dired-move-to-filename)
	(setq point (point))
	)))
    (if point
	(goto-char point)
      ;; Then,  it's some strange non-unix propably ...
      )
    ))


;;; ----------------------------------------------------------------------
;;; - Supposing the directory is in order...dirs first then files...
;;;
;;;###autoload
(defun tdd-first-file ()
  "Moves to fist _file_ in dired."
  (interactive)
  (let* (point
	 )
    (save-excursion
      (PMIN)
      (cond
       ((re-search-forward "^. +-" nil t)
	(dired-move-to-filename)
	(setq point (point))
	)))
    (if point
	(goto-char point)
      ;; Then, it's some strange non-unix propably ...
      )
    ))




;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-last-line ()
  "Moves to last file in dired."
  (interactive)
  (PMAX)
  (while (not (looking-at ".*[0-9]"))
    (forward-line -1))
  (dired-move-to-filename))

;;; ----------------------------------------------------------------------
;;;
(defun tdd-kill-marked-lines ()
  "Remove lines that are unmarked."
  (interactive)
  (let (buffer-read-only
	list
	re
	)
    (dired-map-over-marks
     (push (regexp-quote (read-full-line)) list)
     nil
     )

    (list-loop list re
      (PMIN)
      (if (re-search-forward re nil t)
	  (ti::b-kill-line)
	))
    ))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-kill-unmarked-lines  ()
  "Remove unmarked lines. Ignore directories and symlinks."
  (interactive)
  (tdd-map-over-unmarked
   (let* (char
	  )
     ;; We're at the beginning of line, suppose std unix 'ls'
     ;; drwx--x--x
     (setq char (buffer-substring (+ 2 (point)) (+ 3 (point))))
     (if (not (or (string= char "d")
		  (string= char "l")))
	 (ti::b-kill-line)
       (end-of-line)			;Continue mapping
       )))
  (tdd-first-file)
  )


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-kill-lines (re)
  "Delete lines matching RE."
  (interactive "sKill files re: ")
  (let* (buffer-read-only
	 )
    (cond
     ((not (var-nil-p re))
      (save-line-col
       (tdd-first-file) (dired-move-to-filename) ;cursor pos forms

       (tdd-first-file)			;do this in save-line-col
       (flush-lines re))
      ))))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-leave-only-lines (re)
  "Leave only lines matching RE. Directory lines are skipped.
You can easily undo this with reverting the buffer (dired \"g\")."
  (interactive "sLeave re: ")
  (if (var-nil-p re)
      nil
    (PMIN)
    (tdd-map-over-files
     (if (string-match re (read-full-line))
	 (forward-line 1)
       (ti::b-kill-line)))
    (tdd-first-file)
    ))




;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-pop-to-buffer ()
  "Pops to buffer if it exists in emacs."
  (interactive)
  (let* ((file (safe-do (dired-get-filename)))
	 buffer
	 )
    (cond
     ((and (stringp file)
	   (setq buffer (get-file-buffer file)))
      (pop-to-buffer buffer))
     (t
      (message (format "Can't pop ... Not in emacs. [%s]" file))
      ))
    ))


;;; ----------------------------------------------------------------------
;;; - This behaves differently than dired-x.el dired-do-find-marked-files
;;;
;;;###autoload
(defun tdd-mark-today-files ()
  "Mark all files, not dirs, that are created today.
Point sits on first today file. If no today's files are found, point stays
on current filename."
  (interactive)
  (let* ((list   (ti::c-time-list))
         (line   (LINE))
	 ;;      1024 Oct  3
	 (re     (concat ".*[0-9] " (nth 5 list) " +"
			 (int-to-string (nth 0 list))
			 " +"
			 ;;  This year's file have time in this field
			 "[0-9]+:"))
	 )
    (tdd-map-over-files
     (if (not (looking-at re))
	 (forward-line)
       (dired-mark 1)
       ))

    (tdd-first-file)
    (if (re-search-forward re nil t)
	(dired-move-to-filename)
      (goto-line line)
      (dired-move-to-filename))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tdd-mark-writable-files  ()
  "Mark Your files that have writable flag set."
  (interactive)
  (let* ((re    ".*.w..[-w]..[-w]. ")
         )
    (tdd-map-over-files
      (if (not (looking-at re))
	  (forward-line)
	(dired-mark 1)
       ))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tdd-mark-read-only-files  ()
  "Mark Your files that have writable flag set."
  (interactive)
  (let* ((re    ".*r-.[-r]..[-r].. ")
         )
    (tdd-map-over-files
      (if (not (looking-at re))
	  (forward-line)
	(dired-mark 1)
       ))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tdd-mark-opposite-toggle  ()
  "Mark opposite files.
Ie. if you have marked some files, unmark those and mark all other files."
  (interactive)
  (let* ((re  (dired-marker-regexp))
         )
    (save-line-col nil nil
      (tdd-map-over-files
	(beginning-of-line)
	(if (looking-at re)
	    (dired-unmark 1)
	  (dired-mark 1))
	))))

;;; ----------------------------------------------------------------------
;;; - This behaves differently than dired-x.el dired-do-find-marked-files
;;;
;;;###autoload
(defun tdd-mark-vc-files-in-emacs (&optional unmark)
  "Mark all files in the current _view_ that are in emacs _and_ in VC control.
Optionally UNMARK."
  (interactive)
  (let* ((dir		(expand-file-name dired-directory))
	 (verb		(interactive-p))
	 (msg		(if unmark "Unmarking..." "Marking..."))
	 list
	 )

    (if (null dir) (setq dir dir))	;Shut up XEmacs 19.14 ByteComp
;;;    (require 'vc)
    (setq list
	  (ti::b-buffer-list
	   '(and buffer-file-name
		 (string-match dir buffer-file-name)
		 (vc-registered buffer-file-name)
		 )))

    (if verb  (message msg))

    (cond
     ((and (null list)   verb)
      (message "No VC files of this dir in emacs."))
     (t
      (tdd-mark-file-list list unmark)
      (if verb  (message (concat msg "Done")))
      ))))



;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-refresh-view ()
  "Refresh current dired view.
If you have used `tdd-leave-only-lines' and have done some changes to
the files. You can use this function to re-read the current view.

The dired \"g\" will load full view back. This instead caches the
current view, executes read, and deletes lines that weren't in the
cache --> you get refreshed view. All this may take a while...

Return:
 t		,if refreshed
 nil"
  (interactive)
  (let* ((cache		(tdd-view-get 'no-path-names 're))
	 (verb          (interactive-p))
	 (line          (LINE))		;save user position
	 (re		"")
	 buffer-read-only		;allow write
	 marks
	 )
    (cond
     (cache
      (setq re (mapconcat 'concat cache "\\|"))

      (setq marks
	    (dired-remember-marks
	     (tdd-first-line-point) (tdd-last-line-point)))

      (tdd-dired-unmark-all-files-no-query)	; sort goes nuts otherwise
      (message "")			; stupid message from dired-un...

      (revert-buffer) (PMIN) (forward-line 2)		;leave headers

      (tdd-first-file)
      (beginning-of-line)
      (let (( case-fold-search nil))	;case sensitive
	(delete-non-matching-lines re))

      (dired-mark-remembered marks)

      (goto-line line)  (dired-move-to-filename)
      (if verb (message "Refresh done."))
      t
      )
     (t
      (if verb (message "Can't cache view."))
      nil))
    ))


;;; ----------------------------------------------------------------------
;;; - This behaves differently than dired-x.el dired-do-find-marked-files
;;;
;;;###autoload
(defun tdd-load-all-marked-files (&optional arg)
  "Load all marked files into emacs.
Does not load files which are already in emacs.
If ARG is non-nil, remove mark if file was loaded."
  (interactive "P")
  (let* ((verb		(interactive-p))
	 (files         (tdd-get-marked-files))
	 (loaded	0)
	 (not-loaded	0)
	 (all		0)
	 file
	 )
    (cond
     ((and verb (null files))
      (message "No marked files."))

     ((y-or-n-p "Load all marked files, No kidding? ")
      (list-loop files file
	(inc all)

	(if (get-file-buffer file)
	    (inc not-loaded)
	  (inc loaded)
	  (find-file-noselect file))

	(if arg (save-excursion (dired-unmark 1)))
	)))


    (if verb
	(cond
	 ((eq all not-loaded)
	  (message "Hmm, all files are in emacs already.."))
	 (t
	  (message "%s files loaded." loaded))))
      ))

;;}}}
;;{{{ code: vc special

;;; .............................................................. &vc ...

(defun tdd-mark-vc-has-diffs  (&optional arg)
  "Leave mark to files: VC controlled, have diffs and are in emacs.
If ARG is non-nil, examine file whether it was in emacs or not.

Note:
  Please be patient, taking diffs may be slow per file."
  (interactive)
  (let* ((list  (tdd-get-marked-files))
	 file
	 fn
	 buffer
	 vc-reg-stat
	 diff-no-stat
	 )

    (list-loop list file
      (setq fn		(file-name-nondirectory file)
	    buffer	(get-file-buffer file)
	    vc-reg-stat	(vc-registered file)
	    )

      (cond
       ((or (not vc-reg-stat)
	    ;;  Not exist in emacs, do not bother looking
	    (and (null arg) (null buffer)))
	(tdd-mark-re (regexp-quote fn) 'unmark))
       (t
	(setq diff-no-stat (vc-workfile-unchanged-p file 'get-diffs))
	(if diff-no-stat
	    (tdd-mark-re (regexp-quote fn) 'unmark))
	))
      )

    ))



;;; ----------------------------------------------------------------------
;;;
(defun tdd-marked-vc-revert  (&optional arg)
  "Revert all version controlled/no changed/marked files. Ignore ARG."
  (interactive "P")
  (let* ((list		(tdd-get-marked-files))
	 (display       (if list t))
	 (vc-dired-mode nil)		;turn mode off
	 (verb		(interactive-p))
	 (count         0)
	 (handled       0)
	 load
	 file
	 buffer
	 vc-reg-stat
	 diff-no-stat
	 )

    (list-loop list file
      (setq buffer	 (get-file-buffer file)
	    vc-reg-stat  (vc-registered file)
	    load         nil
	    diff-no-stat nil
	    )

      (inc count)


      ;; ... ... ... ... ... ... ... ... ... ... ... ... possible load . .

      (cond
       ((and (null buffer)
	     (file-writable-p file)
	     vc-reg-stat
	     )
	(setq buffer (find-file-noselect file)
	      load   t)
	))

      (if buffer
	  (setq diff-no-stat (vc-workfile-unchanged-p file 'get-diffs)))

      ;; ... ... ... ... ... ... ... ... ... ... ... ... ... .. handle . .
      (cond
       ((null buffer)
	nil)				;no file, no vc controlled

       (diff-no-stat
	(inc handled)
	(save-window-excursion
	  (vc-next-action-on-file file 'verbose)
	  (if load
	      (kill-buffer buffer))
	  )
	))

      ;; ... ... ... ... ... ... ... ... ... ... ... ... ... while end . .
      )

    (if display
	(dired-do-redisplay))

    (if verb
	(message "VC revert:  %s/%s handled "  handled count))
    ))


;;; ----------------------------------------------------------------------
;;;
;;;
(defun tdd-marked-vc-co  (&optional arg)
  "Check Out all marked files and load them inside emacs.
Do some checking, before doing co.
o  if file is writable, skip over.
o  if file is not in RCS, skip over.

Optional ARG skips all load confirmations.

Marks are left only to files which were loaded into emacs."
  (interactive "P")
  (let* ((list		(tdd-get-marked-files))
	 (dired-vc-mode nil)		;turn mode off
	 (count         0)
	 (loaded        0)
	 (handled       0)
	 file fn
	 buffer
	 load
	 vc-reg-stat
	 modify-stat
	 read-stat
	 )
;;;    (require 'vc) (require 'vc-hooks)

    (list-loop list file
      (setq fn		(file-name-nondirectory file)
	    buffer	(get-file-buffer file)
	    vc-reg-stat (vc-registered file)
	    load        nil
	    )


      (if buffer			;read stat only if it's in emacs
	  (save-excursion
	    (set-buffer buffer)
	    (setq modify-stat (buffer-modified-p)
		  read-stat   buffer-read-only
		  )))

      (inc count)

      ;; ... ... ... ... ... ... ... ... ... ... ... ... possible load . .

      (cond
       ((and (null buffer)
	     vc-reg-stat			;; in VC
	     (not (file-writable-p file))	;; -r--r--r--
	     (or arg
		 (y-or-n-p (concat "file " fn " not in emacs. Load? " )))
	     )
	(inc loaded)
	(setq buffer (find-file-noselect file)
	      load   t)
	))

      ;; ... ... ... ... ... ... ... ... ... ... ... ... ... .. handle . .
      (cond
       (load
	)
       ((and buffer				;; in Emacs, be extra carefull
	     vc-reg-stat			;; in VC
	     (not (file-writable-p file))	;; -r--
	     (null modify-stat)			;; %*
	     read-stat)				;; %%
	)					;; --> no-op, valid state

       (t					;; User has modified it!
;;;	(d! buffer vc-reg-stat
;;;	    (not (file-writable-p file)) (null modify-stat) read-stat)
	;;  This situation may occur very easily
	;;  - You load -r-- file in emacs that's in VC
	;;  - you want to temporary play with it, like changing one
	;;    flag in .mak temporarily
	;;  - you go and M-x toggle-read-only, change it, C-x C-s
	;;  ...
	;;  Now you have modified the read-only file !
	(setq buffer nil)
	))


      (cond
       ((null buffer)
	;;     18 15:09 test3.txt
	(tdd-mark-re (concat "[0-9] +" fn) 'unmark)
;;;	(setq RE   (concat "[0-9] +" fn))  (d! "<un>" fn)
	)
       (t
	(save-window-excursion
;;;	  (d! ">>" buffer)
	  (inc handled)
	  (vc-next-action-on-file file 'verbose))
	))

      ;; ... ... ... ... ... ... ... ... ... ... ... ... ... while end . .
      )

    (if (not (eq 0 handled))
	(dired-do-redisplay))

    (message (format "VC co: %s/%s handled, loaded %s"
		     handled count loaded))
    ))







;;; ----------------------------------------------------------------------
;;; - This is vastly different than C-x v v in dired mode
;;;
;;;
(defun tdd-marked-vc-ci  (&optional arg)
  "Check In all marked files and load them inside emacs. Ignore ARG.
Do some heavy checking, before doing ci.
o  if file is not writable, skip over
o  if file is not in emacs, load it first
o  if file is in emacs, but read only, suppose no diffs
o  if file is in emacs, check rcsdiff, --> do nothing if no diffs
o  if file is in emacs, check rcsdiff, if file not saved, offer save

Notice, that this function enters `recursive-edit' if it thinks file should
be Checked In. Use \\[exit-recursive-edit] to get back to this function
and continue with rest of the files.

Recursive edit is shown with those [ ] marks in the modeline.

Note
  There is plenty of messages for each file in marked, because
  used should know if the marked file couldn't be processed with ci.

  Marks are removed from handled files.

Bugs:
  This function automatically removes marks from files where user has
  used recursive edit. If user didn't ci the file, this program
  can't know that.

  Anyway, the mark is gone."
  (interactive "P")
  (let* ((list		(tdd-get-marked-files))
	 (verb		(interactive-p))
	 (count         0)
	 (handled       0)
	 (loaded        0)
	 file fn
	 buffer
	 load				;flag
	 diff-no-stat
	 modify-stat
	 read-stat
	 vc-reg-stat
	 )
;;;    (require 'vc) (require 'vc-hooks)
    (if (and (null vc-dired-mode)
	     (y-or-n-p "Buffer must be in VC dired mode. Turn it on? "))
	(vc-dired-mode)
      (error "Aborted."))

    (list-loop list file
      (setq fn		(file-name-nondirectory file)
	    buffer	(get-file-buffer file)
	    vc-reg-stat (vc-registered file)
	    load nil
	    )

      (inc count)

      ;; ... ... ... ... ... ... ... ... ... ... ... ... possible load . .

      (cond
       ((and (null buffer)
	     vc-reg-stat
	     (file-writable-p file)	; "-r--r--r--" , not ci'able file
	     (y-or-n-p (concat "file " fn " not in emacs. Load? " )))
	(setq buffer (find-file-noselect file)
	      load   t)
	(inc loaded)
	))

      ;; ... ... ... ... ... ... ... ... ... ... ... ... ... ...  stat . .
      (cond
       ((setq buffer (get-file-buffer file))
	(save-excursion
	  (set-buffer buffer)
	  (setq modify-stat (buffer-modified-p)
		read-stat   buffer-read-only
		)
	  ;;  Can't ask stat if not in VC control
	  (and vc-reg-stat
	       (setq diff-no-stat (vc-workfile-unchanged-p file 'get-diffs)))

	  ))
	)



      ;; ... ... ... ... ... ... ... ... ... ... ... ... set diff stat . .
      (cond
       ((and buffer
	     (null vc-reg-stat))
;;;	(read-char-safe-until (concat fn " _not_ vc registered. (ok)" ))
	)

       ;; ... ... ... ... ... ... ... ... ... ... ... ... ... ... ... ...
       ((and buffer
	     read-stat			;is file in CheckOut state ?
	     )
	(read-char-safe-until
	 (concat fn " in VC, but _buffer_ is read-only. (ok)" ))

	(and load
	     (y-or-n-p (concat "Unload " fn " ? "))
	     (kill-buffer buffer)
	     (dec loaded))
	)

       ;; ... ... ... ... ... ... ... ... ... ... ... ... ... ... ... ...
       ((and buffer
	     vc-reg-stat
	     diff-no-stat
	     (null modify-stat)
	     )
	(if (null load)
	    (read-char-safe-until
	     (concat fn " contains NO changes. (ok) "))
	  (if (y-or-n-p (concat fn " contains NO changes, unload NOW? "))
	      (kill-buffer buffer)))
	)

       ((and buffer
	     vc-reg-stat)
	(inc handled)
;;;	(read-char-safe-until (concat "Ready for.." file))

	(save-excursion
	  (save-window-excursion
	    (unwind-protect
		(progn
		  (tdd-mark-re (concat "[0-9] +" file) 'unmark)
		  (pop-to-buffer buffer)
		  (call-interactively 'vc-next-action)

		  (ad-enable-advice 'vc-finish-logentry
				    'after 'tdd-recursive-edit)
		  (ad-activate 'vc-finish-logentry)
		  (recursive-edit)
		  (message
		   (substitute-command-keys
		    (concat
		     "Use \\[exit-recursive-edit] to abort action."
		     "to next file")))
		  (sleep-for 1)
 		  )
	      (ad-disable-advice  'vc-finish-logentry
				  'after 'tdd-recursive-edit)
	      (ad-activate 'vc-finish-logentry)
	      )))
	(PMIN)				;remove file after VC
	(if (re-search-forward fn nil t)
	    (dired-unmark 1))
	))

      ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. ..  list-loop  ..
      )

    (if verb
	(message (format "VC ci: %s/%s handled, loaded %s"
			 handled count loaded)))
    ))


;;}}}
;;{{{ code: advice


;;; ----------------------------------------------------------------------
;;; - Until someone fixes dired to honor the  backup-file-name-p
;;;   this stays replaced...
;;; - This is copy from 19.30 dired.el
;;;
(defadvice dired-flag-backup-files (around tdd act)
  "Replace original function.
This function honours the `backup-file-name-p' function and
additionally flag files that match regexp `tdd-:backup-file-regexp'."
  (let ((dired-marker-char	(if unflag-p ?\040 dired-del-marker))
	(re			tdd-:backup-file-regexp)
	file
	)
    (dired-mark-if
     (progn
       (beginning-of-line)
       (when (not (looking-at dired-re-dir))
	 (setq file  (dired-get-filename t t))

	 (if (stringp file)
	     (or (backup-file-name-p file)
		 (and re
		      (string-match re file))))
	 ))
     "backup file")
    ))

;;; ----------------------------------------------------------------------
;;;
(defadvice vc-finish-logentry (after tdd-recursive-edit dis)
  "When this advice is enabled, it call `exit-recursive-edit'.
Only if f recursive edit is in effect.

This advice is controlled by function `tdd-marked-vc-ci' and it is never
enabled outside of that function."
  (safe-do
   (exit-recursive-edit)
   ))


;;; ----------------------------------------------------------------------
;;; 18 Oct 1995, Kevin Rodgers <kevin.rodgers@ihs.com>, gnu.emacs.help
;;; - When using "f" it loads directory to same buffer.
;;; - only kills the Dired buffer if a prefix arg is given
;;;
(defadvice dired-find-file (around tdd-kill-dired-buffer last act)
  "If a prefix argument is given, kill the Dired buffer.

If you have loaded dired-x and it contains variable
`dired-find-subdir', this advice does nothing."
  (let* ((dired-buffer (current-buffer))
	 )
    (prog1
	ad-do-it
      (if (and (eq major-mode 'dired-mode)
	       (not (eq (current-buffer) dired-buffer))
	       (or current-prefix-arg
		   tdd-:use-only-one-buffer-flag)
	       (or (not (featurep 'dired-x)) ;not loaded
		   (and (featurep 'dired-x) ;is loaded, but this var not exist
			(not (boundp 'dired-find-subdir))))
	       )
          (kill-buffer dired-buffer)
	))))



;;}}}
;;{{{ code: store

;;; .......................................................... storing ...


;;; ----------------------------------------------------------------------
;;;
(defsubst tdd-store-get-string ()
  "Return content of storage as string."
  (list-to-string tdd-:file-store))

;;; ----------------------------------------------------------------------
;;;
(defun tdd-store-show ()
  "Show filenames in storage."
  (interactive)
  (if (null tdd-:file-store)
      (message "Store is empty.")
    (funcall tdd-:show-storage-function (tdd-store-get-string))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tdd-store-filename ()
  "Save current filename into variable."
  (interactive)
  (let* ((file (tdd-get-filename))
	 )
    (if (member file tdd-:file-store)
	(tdd-store-show)
      (nconcc  tdd-:file-store file)
      (if (interactive-p)
	  (tdd-store-show))
      )))

;;; ----------------------------------------------------------------------
;;;
(defun tdd-store-delete-filename ()
  "Remove filename from store."
  (interactive)
  (let* ((file (tdd-get-filename))
	 )
    (setq tdd-:file-store (delete file tdd-:file-store))
    (if (interactive-p)
	(tdd-store-show))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tdd-store-clear ()
  "Clear variable holding files."
  (interactive)
  (setq tdd-:file-store nil)
  (if (interactive-p)
      (message "Storage cleared."))
  )


;;; ----------------------------------------------------------------------
;;;
(defun tdd-store-remove-file ()
  "Delete current filename from storage."
  (interactive)
  (let* ((file   (tdd-get-filename))
	 (verb   (interactive-p))
	 (store  tdd-:file-store)
	 list
	 x
	 )
    (if (null store)
	(if verb (message "Storage is empty."))
      (list-loop store x
	  (if (not (string= x file))
	      (push x list))
	  )
      (setq tdd-:file-store list)
      (if verb
	  (tdd-store-show))
      )))



;;; ----------------------------------------------------------------------
;;;
(defun tdd-store-delete-marked ()
  "Delete marked files from store."
  (interactive)
  (tdd-store-add-marked 'delete))



;;; ----------------------------------------------------------------------
;;;
(defun tdd-store-add-marked (&optional delete)
  "Add marked files into store. No duplicates are inserted.
If parameter DELETE is non-nil, removes marked files from store."
  (interactive)
  (let* ((list    tdd-:file-store)
	 (marked  (tdd-get-marked-files-no-dir))
	 x
	 )
    (if (null delete)
	(list-loop marked x
	  (if (not (member x list))
	      (push x tdd-:file-store)))
      (list-loop marked x
	  (if (member x list)
	      (setq tdd-:file-store (delete tdd-:file-store x)))
	  ))
    (if (interactive-p)
	(tdd-store-show))
    ))

;;}}}
;;{{{ code: ange ftp

;;; ......................................................... ange-ftp ...



;;; ----------------------------------------------------------------------
;;;
(defun tdd-store-ftp-message (&rest args)
  "Show Message from ange ftp after finishing the mget. Ange ARGS."
  (message-quick "Store: ftp completed.")
  )

;;; ----------------------------------------------------------------------
;;;
(defun tdd-back-to-dired-buffer ()
  "Switch back to dired buffer, which is associated with ange-ftp buffer.
If no such buffer is found, do nothing."
  (interactive)
  (let* ((buffer (ti::b-find-ange-to-dired-buffer))
	 )
    (if buffer
	(pop-to-buffer (car buffer))
      (message "No dired buffer found."))
    ))



;;; ----------------------------------------------------------------------
;;; - If I have 2-3 dired ftp sessions and I want to close the current
;;;   one, this is a handy command.
;;;
;;;###autoload
(defun tdd-kill-dired-and-ange-session (&optional verb)
  "Kill the current dired buffer and possible ange-ftp buffer. VERB.
This is like `dired-delete-and-exit'."
  (interactive)
  (let* ((buffer  (tdd-ange-ftp-buffer-for-this-dired))
	 (verb    (or verb (interactive-p)))
         )
    (if buffer
	(kill-buffer buffer))
    (kill-buffer (current-buffer))

    (if verb
	(message
	 (if buffer
	     "Ange buffer killed too."
	   "No ange buffer associated with dired.")))
    ))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-kill-all-ange-and-dired-buffers  (&optional verb)
  "Kill all ange-ftp buffers _and_ all remote dired buffers. VERB."
  (interactive)
  (let* ((verb  (or verb (interactive-p)))
	 (ange	(ti::b-get-ange-buffer-list))
	 (dired (ti::b-buffer-list
		 (quote
		  (and (eq major-mode 'dired-mode)
		       (string-match tdd-:dired-directory-ange-regexp
				     dired-directory)
		       ))))
	 (ange-count  0)
	 (dired-count 0)
	 elt
         )
;;;    (d! dired)

    (list-loop ange elt
      (kill-buffer elt) (inc ange-count))

    (list-loop dired elt
      (kill-buffer elt) (inc dired-count))

    (if verb
	(message "Killed %s ange, %s dired buffers."
		 ange-count dired-count))
    ))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-kill-all-ange-buffers ()
  "Kill all ange-ftp process buffers.
If you want to kill one buffer at a time, use
`tdd-switch-to-some-ange-ftp-buffer' to switch to individual buffer
and use \\[kill-buffer] to kill session.

This function is primarily used for cleanups. After a while
you may end up with many ftp session and it's nice if
you can get rid of them fast.

Don't worry about the dired buffers, Ange will automatically
create connection, if you use \"g\" -- rever-buffer, in a dired
that is associated with ange-ftp."
  (interactive)
  (let* ((list	(ti::b-get-ange-buffer-list))
	 (i	0)
	 elt
	 )
    (list-loop list elt
      (inc i) (kill-buffer elt))

    (if (> i 0 )
	(message (concat "Ange buffers killed: " i))
      (message "No ange buffers found."))
    ))



;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tdd-switch-to-some-ange-ftp-buffer ()
  "Gather all ange FTP buffers and offer completion menu.
If there is only one Ange buffer, switches to it without asking."
  (interactive)
  (let* ((list  (ti::b-get-ange-buffer-list))
	 buffer go
	 )
    (if (null list)
	(message "no Ange-ftp sessions at the moment.")

      (if (eq 1 (length list))
	  (setq buffer (car list))
	(setq buffer
	      (completing-read "go ange: " (list-to-assoc-menu  list))))

      (if (setq go (get-buffer buffer))
	  (switch-to-buffer go)
	(message (concat "No ange buffer: " buffer))
	))
    ))



;;; ----------------------------------------------------------------------
;;; - This is handy, when you want to check that the mput went ok.
;;;
(defun tdd-switch-to-mput-ange-ftp-buffer  ()
  "Switch to ange buffer where last mput was made.
Does nothing if no mput were recorded or such ange buffer does not exist.

Binds local keys to ftp buffer
  C - c b		,switch back to previous buffer

References:
  `tdd-:previous-buffer'
  `tdd-:mput-last-ftp'

Return
  nil			,no action taken.
  t"
  (interactive)
  (let* ((file  tdd-:mput-last-ftp)
	 list host buffer
	 ret
	 )
    (cond
     ((null tdd-:mput-last-ftp)
      (message "Sorry, No mput information."))

     ((not (string-match "/.*@.*:" file))
      (message "Sorry, No ange reference in tdd-:mput-last-ftp"))

     (t
      ;;  This return 3 member list: SITE LOGIN DIRECTORY/FILE
      ;;
      (setq list  (ange-ftp-ftp-name file)
	    host  (nth 0 list))

      (setq tdd-:previous-buffer (current-buffer))

      ;;  Try to find buffer , ange uses SITE name for buffer names
      ;;  *ftp omc@venus*
      (cond
       ((and list
	     (setq buffer (car (ti::b-buffer-list
				(concat "[*]ftp.*" (regexp-quote host))
			      'temp-buffers))))
	(switch-to-buffer-other-window buffer)
	(PMAX)

	;; Switching back to previous (b)uffer
	(local-set-key "\C-cb"
		       (function
			(lambda ()
			  "TDD: mput ange, back to previous buffer"
			  (interactive)
			  (pop-to-buffer tdd-:previous-buffer))))
	(setq ret t)
	)

       (t
	(message (concat "Sorry, can't find ange buffer for " host))
	))
      ))
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tdd-ange-ftp-buffer-for-this-dired  (&optional file)
  "Return ange ftp buffer-name-string for current dired or FILE, or nil."
  (let* (host
	 buffer
	 list
         )
    (setq file (or file (dired-get-filename)))

    ;;  This return 3 member list: SITE LOGIN DIRECTORY/FILE
    ;;
    (setq list  (ange-ftp-ftp-name file)
	  host  (nth 0 list))


    (when list				;This dired is not in remote site

      ;;  Remove that ange-ftp site information from the string.
      (setq tdd-:directory
	    (ti::s-index-substring dired-directory ?: nil 'right))

      ;;  Try to find buffer , ange uses SITE name for buffer names
      ;;  *ftp omc@venus*
      (when list
	(setq buffer
	      (car (ti::b-buffer-list
		    (concat "[*]ftp.*" (regexp-quote host))
		    'temp-buffers)))
	(unless (get-buffer buffer)
	  (setq buffer nil))
	))

    buffer
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tdd-switch-to-ange-ftp-buffer (&optional verb)
  "If the dired is ange ftp buffer, switch to the real ftp buffer. VERB.

Sets global
 `tdd-:directory'		,filename for current line

Binds local keys to ftp buffer
 C - c af			,insert files stored in current point
 C - c ad			,insert directory name
 C - c ab			,switch back to dired buffer"
  (interactive)
  (let* (buffer
	 dir
	 )
    (setq verb (or verb (interactive-p)))
    ;;  1.  try normal ange ftp
    ;;  2.  did user used 'put' to remove site ?
    ;;
    (setq buffer (tdd-ange-ftp-buffer-for-this-dired))

    (cond
     ((and (null buffer)
	   tdd-:file-store
	   tdd-:mput-last-ftp
	   (null (tdd-switch-to-mput-ange-ftp-buffer))
	   )
      (if verb (message "Sorry, can't locate associated ftp buffer."))
      )
     (buffer
      (switch-to-buffer-other-window buffer)
      (set (make-local-variable 'tdd-:directory) dir)

      (PMAX)
      ;;  "f"  for file information
      (local-set-key
       "\C-caf"
       (function
	(lambda (&optional arg)
	  "TDD: Inserts file storage string."
	  (interactive)
	  (setq arg (tdd-store-get-string))
	  (if (var-nil-p arg)
	      (message "No files in storage.")
	    (insert (tdd-store-get-string))))))

      ;; "d" for directory information
      (local-set-key
       "\C-cad"
       (function
	(lambda ()
	  "TDD: Inserts dired's directory string."
	  (interactive)
	  (insert tdd-:directory))))

      ;; Switching back to dired (b)uffer
      (local-set-key
       "\C-cab"
       (function
	(lambda ()
	  "TDD: Back to dired buffer"
	  (interactive)
	  (tdd-back-to-dired-buffer))))
      ))
    ))



;;; ----------------------------------------------------------------------
;;; mget = multiple get
;;;
(defun tdd-store-ftp-mget ()
  "Send command to ange to fetch all files in store."
  (interactive)
  (let* ((files		tdd-:file-store)
	 (down          tdd-:download-dir)
	 (store		(tdd-store-get-string))
	 (ange		(ange-ftp-ftp-name dired-directory))
	 to-dir
	 host user dir
	 )
    (cond
     ((null ange)
      (message "Can't find ftp process. Start one first."))

     ((var-nil-p files)
      (message "No files in store."))

     (t
      (if (or (not
	       (y-or-n-p (concat "Really get files: "
				 ;;   Get nicer prompt
				 (if (> (length store) 50)
				   (concat (substring store 0 50 )
					   "...")
				 (concat store " ")
				 ))))
	      (var-nil-p
	       (setq to-dir
		     ;;  Hack to read directory easily
		     (let ((default-directory down))
		       (call-interactively
			(function
			 (lambda (dir)
			   (interactive "Ddownload dir: ")  dir))))
		   )))
	  (message "Cancelled.")

        ;; ................................................. then part ...
        ;; - First update the value, so that user gets the old selection
        ;;
	(setq tdd-:download-dir to-dir)

	;; Next, get all needed parameters
	;;
	(setq host  (nth 0 ange)
	      user  (nth 1 ange)
	      dir   (nth 2 ange)
	      to-dir (expand-file-name to-dir)
	      )

	(ti::f-ange-file-handle 'get user host dir to-dir files)
	)))
    ))




;;; ----------------------------------------------------------------------
;;; - remember to be in DIRED before you call this
;;; mput = multiple put
;;;
(defun tdd-store-ftp-mput (ange-ref-to)
  "Send all files in store to remote site ANGE-REF-TO."
  (interactive
   (list
    (completing-read
     "mput site: "
     (list-to-assoc-menu tdd-:mput-sites)
     nil nil tdd-:mput-last-ftp
     'tdd-:mput-history
     )))

  (if (null dired-directory)
      (error "Must execute command in dired buffer."))

  ;;	Record the site name where the mput was made
  (setq tdd-:mput-last-ftp ange-ref-to)

  (let* ((files		tdd-:file-store)
	 (store		(tdd-store-get-string))
	 (dir           dired-directory)
	 ange
	 to-dir
	 host user
	 )

    ;;  If user is in remote dired buffer, signal error
    ;;  We don't support this. At least not now.
    ;;
    (if (string-match "@" dired-directory)
	(error "Sorry, load files first to your site."))


    (if (not (var-nil-p ange-ref-to))
	(setq ange   (ange-ftp-ftp-name ange-ref-to) ;crack it
	      host   (nth 0 ange)
	      user   (nth 1 ange)
	      to-dir (nth 2 ange)
	      ))


    (cond
     ((var-nil-p ange-ref-to)
      (message "No site given"))

     ((var-nil-p files)
      (message "No files in store."))

     ((var-nil-p to-dir)
      (message "No destination download directory given"))

     (t
      (if (not (y-or-n-p (concat "Put " host ": "
				 ;;   Get nicer prompt
				 (if (> (length store) 50)
				     (concat (substring store 0 50)
					     "...")
				   (concat store " ")))))
	  (message "Cancelled.")
	;; ................................................. then part ...

;;;	(d! "~~,put" user host "to" to-dir "source"  dir files)
	;; (mode user host dir lcd file-list &optional not-bg msg-func)
	(ti::f-ange-file-handle 'put user host to-dir dir files)
	)))
    ))



;;}}}


(provide   'tinydired)
(run-hooks 'tdd-:load-hook)

;;; tinydired.el ends here
