;; @(#) tinylisp.el -- Emacs elisp programming help grabbag

;; @(#) $Id: tinylisp.el,v 1.40 1997/07/01 09:08:02 jaalto NotReleased $
;; @(#) $Keywords: lisp programming maint tools $
;;
;; $KnownCompatibility: 19.28+ XEmacs19.14+ $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1997 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Feb 20 1997
;;
;; To get information on this program use ident(1) or do M-x tili-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinylisp|Jari Aalto|jari.aalto@poboxes.com|
;; Emacs elisp programming help: elp profiling, checkdoc, eval commands ...|
;; 01-Jul-1997|$Revision: 1.40 $|~/misc/|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ....................................................... &t-install ...
;; Put this file on your Emacs-Lisp load path, add following into your
;; ~/.emacs startup file. Rip code with tinylib.el / ti::m-pkg-rip-magic
;;
;;*     ;;  Select some unused, non-shifted, fast prefix key
;;*     ;;  my lbd accesses $ without shif and it is seldom used
;;*     ;;  in lisp. Other alternatives: "!", "_" ":"
;;*     ;;
;;*     (setq tili-:mode-prefix-key  "$")
;;*     (require 'tinylisp)
;;* _
;;
;; Or prefer following autoload: your emacs loads this package only when
;; you need it. This assumes that you want the package to be installed
;; behind your "\C-zmE" key, if you want to relocate the menus somewhere
;; else, please make necessary modifications using tili-:load-hook and
;; change variable tili-:menu-elp accordingly.
;;
;;* _
;;*     (autoload 'tili-mode "tinylisp"         t t)
;;*     (add-hook 'lisp-mode-hook               'turn-on-tili-mode)
;;*     (add-hook 'emacs-lisp-mode-hook         'turn-on-tili-mode)
;;*     (add-hook 'lisp-interaction-mode-hook   'turn-on-tili-mode)
;;* _
;;*     ;;  Put all minor modes to 'm' map.
;;*     ;;  I use small 'e' for calling eldoc.el::eldoc-mode.
;;*     ;;  Elp I call much less often, that's why big 'E' key.
;;*     ;;
;;*     ;;  If this is occupied, choose another binding.
;;*     ;;
;;*     (global-set-key "\C-cme" 'tili-mode)
;;*
;;*     ;; get eldoc.el too from Noah Friedman <friedman@prep.ai.mit.edu>
;;*     (global-set-key "\C-cmE" 'eldoc-mode)
;;* _
;;
;; If you don't want to use the echo-menu, but regular emacs keymap
;; call instead, put following into your .emacs. This must be before any
;; other TILI settings. You must reload package every time if you change this
;; this setting.
;;
;;*     (setq tili-:menu-use-flag nil)
;;* _
;;
;; If you have any questions, use this function
;;
;;      M-x tili-submit-bug-report

;;}}}

;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "tili-" in front of
;;   every function & variable. It stands for '(ti)ny (li)sp help
;; - variable names contain letters 'tili-:'

;;  Preface, Feb 1997
;;
;;      My private lisp helper functions were scattered around my
;;      keysetting files and in other private libraries. One day
;;      I decided that I wanted a minor mode to access all those
;;      tools I had written for lisp. I didn't want to continue
;;      stacking up my lisp-mode-hook for all the growing features.
;;
;;      So, if you're programming in elisp, this minor mode make your
;;      program development much easier.
;;
;;  Overview of features
;;
;;      Lisp coding help
;;
;;      o   Create list of all variables from the buffer.
;;          Uou can use the list in your bug report function or just
;;          to get an overview of the variables: names and the order how
;;          you have used them (The order is important if you use defcustom)
;;      o   Create function list (or occur menu)
;;      o   Create autoload list (or occur menu)
;;      o   Evaluate current definition under point (reparse function,
;;          reset defvar or defcustom variable)
;;      o   Print variable's value under point, set variable's value
;;          under point.
;;      o   Call function under point (to test it immediately)
;;      o   Indent function/variable around point.
;;      o   Two tun on Modes: Show underlying  properties/overlays or
;;          charcter symbol information.
;;      o   FINDS LISP CODE ERROR POINT.
;;      o   defcustom.el variable detection and evaluation support.
;;
;;      Elp support -- Lisp code profiling
;;
;;      o   Accessing elp commands from command menu
;;      o   Get the latest from http://www.python.org/ftp/emacs/
;;      o   Profiling your package or buffer functions very easily.
;;          Sit somewhere in function and un/instrument it with one command.
;;          Un/instrument all functions in buffer with one command.
;;
;;      Checkdoc support --  Check documentation strings for style requirements
;;
;;	o   Easy interface to checkdoc commands.
;;	o   ftp://ftp.ultranet.com/pub/zappo
;;	o   A tool that makes sure your package follows the guidelines
;;	    presented in File: elisp, Node: Documentation Tips.
;;
;;  Checklist, how you would check your own package
;;
;;	o   $ f  Take a look at your function names: are they well named,
;;	    so that same categories begin with same words. Below it would be
;;	    a mistake to have latter as xxx-clear-buffer, because then
;;	    you cant find all common function with lisp-complete-symbol
;;	    command on xxx-buffer-*
;;
;;		xxx-buffer-handling
;;		xxx-buffer-clear
;;
;;	    Check also that your macros are defined first before functions.
;;
;;		defsubst
;;		defmacro
;;		defun
;;
;;	o   C-u $ v Check variable names as the function names above,
;;	    but also see that you have defined right user variables.
;;	    The extra C-u argument will print this information.
;;	o   $ x  Check the lisp package layout: first line and footer must
;;	    be in proper format and that Author etc. tags are in
;;	    their places.
;;	o   Check the documentation strings: Checkdoc.
;;	    To get overview of error do: $ c -  and $ c RET
;;	    which a) turned off query b) checked whole buffer for errors.
;;	    When you have got a clear look, then start correcting mistakes
;;	    and do $ c a (semiautomatic correction) $ c BACKSPACE to correct
;;	    full buffer.
;;
;;  Defcustom.el and evaluating an `defcustom' variable
;;
;;	If you don't know what defcustom is, you if you don't use it, you
;;	can ignore this section. The defcustom variables are evaluated
;;	pretending like they were `defconst'.
;;
;;	If you got thrown to error during evaluation, pay attention now,
;;	CALL COMMAND $ Z  or M-x tili-emergency IMMEDIATELY. For full
;;	details, see function documentation strings in the source file
;;	for these:
;;
;;	    (defadvice defconst
;;	    (defun     tili-eval-at-point
;;
;;  Find lisp code error position
;;
;;      The most usefull functions in this package are two error finding
;;      functions which try their best to put you on a line that generates
;;      the lisp error. You can use this feature to eg. check your
;;      ~/.emacs starup files and find the spot where eval-current-buffer
;;      fails.
;;
;;      The best function is tili-error-find-1 was programmed by
;;      Mikael Djurfeldt <mdj@sanscalc.nada.kth.se> and is included
;;      here with his permission. Thanks Mikael, the funstion has
;;      saved from me from lot of debugging!
;;
;;  Following lisp code call chain
;;
;;	The traditional way to follow lisp code is to use TAGS file
;;	(See 'etags' or 'ctags' shell binary and C-h a "tags") which
;;	reads bunch of *el files and builds up internal representation
;;	of all defined symbols and their locations.
;;
;;	But using tags is not very flexible if you write the code yourself,
;;	because when you add new fucntion or new varaible, the TAGS file is
;;	immediately out of date. Hm. The TAGS is general tool for many
;;	programming languages, but in Emacs lisp, we can take advantage
;;	of the fact that Emacs already knows where the symbols are defined.
;;	The information is stored to `load-history' wheneved you run
;;	`load' `require' `load'file' or `load-library'.
;;
;;	In this package, there exists two functions that make use of
;;	`load-history' and if the symbol is not there, it  tries to
;;	find definition from the current buffer. You see, if you
;;	do `eval-current-buffer' the definition information is not
;;	stored to `load-history'. With these commands you can browse
;;	almost any package without any extra TAGS file.
;;
;;	  [The only assumption is that you have `loaded' the file !!]
;;
;;	    $ '	    tili-jump-to-definition (do not record call chain)
;;	    $ +     tili-jump-to-definition-chain (record call chain)
;;	    $ \177  tili-back-to-definition (propably your backspace key)
;;		    This returns to previously saved call-chain point
;;
;;	The jump command also know followin prefix arguments
;;
;;	    M-0 $ ' tili-jump-to-definition (empty call chain)
;;	    C-u $ ' tili-jump-to-definition (record call-chain)
;;
;;  Examining text properties and overlays in buffer
;;
;;      If you have ever played with text properties or overlays
;;      (called extents in XEmacs), you know how hard it is to examine
;;      buffer's characters and debug where the properties are.
;;
;;      In this package there is "constant char browing mode" where
;;      every time you move your cursor, the face info and/or
;;      overlay info is displayed in the echo-area. Ig you supply
;;      3 C-u arguments, the information is also recored to the separate
;;      buffer. This is the most easiest way to examone some chacter
;;      properties in arbitrary buffer positions. See function:
;;
;;          $ p     tili-property-show-mode
;;
;;  Examining charcter syntax
;;
;;      Major modes define syntax tables for characters and sometimes you
;;	want to see the syntax class of a character under cursor. This
;;	mode behaves in the same manner as text propery display, just turn
;;	it on and it will constantly show char info.
;;
;;	    $ y	    tli-syntax-show-mode
;;
;;  Snooping interesting variables
;;
;;      Has is happeed to you that your're debugging package and it
;;      installs many hooks and and sets many different variables and
;;      then you suddenly realize that it went all wrong? You may have
;;      noticed that some ill behaving package keeps preventing
;;      file writing!
;;
;;      No problem, you can define interesting variable sets to peek
;;      their contents, eg. checking all file related hooks for problems.
;;      And if you supply C-u prefix arg, your editings are updated to
;;      the variables. With any other non-nil arg, the contents of
;;      the variables are recorded (kinda before install -- after install
;;      snooping) See function:
;;
;;          $ s     tili-snoop-variables
;;
;;      And additional prefix arguments: You can save variables states,
;;      modify them as you like, and go back to restores values.
;;
;;  Elp: notes
;;
;;	  [excerpt from Barry's elp.el]
;;        ...Elp can instrument byte-compiled functions just as easily as
;;        interpreted functions, but it cannot instrument macros.  However,
;;        when you redefine a function (e.g. with eval-defun), you'll need to
;;        re-instrument it with M-x elp-instrument-function.  This will also
;;        reset profiling information for that function.  Elp can handle
;;        interactive functions (i.e. commands), but of course any time spent
;;        idling for user prompts will show up in the timing results.
;;
;;      To elp functions right, follow these steps. _*important*_
;;      '(defun" must be left flushed in order the function to be
;;      found. If there is any leading spaces before the '(' or
;;      'defun', then function won't be found and (un)instrumented.
;;
;;      o   $ e A Restore (a)ll elp'd functions
;;      o   $ -   Eval buffer containing functions (or eval single function)
;;      o   $ I   Instrument all functions in buffer (or single function)
;;      o   $ h   Run the harness test that calls the functions
;;
;;  Elp: Summary mode's sort capabilities
;;
;;	When you call $ e s to show the elp result(s), the results buffer
;;	is put into tili-elp-summary-mode where you can sort the columns
;;	with simple keystrokes. The sort keys correspond to the column names.
;;
;;	f)unction Name  c)all Count  e)lapsed Time  a)verage Time
;;	 =============   ==========   ============   ============
;;
;;  Elp: customizations
;;
;;      You should be aware of this variable in elp; which resets the
;;      list every time you display it.
;;
;;          elp-reset-after-results
;;
;;      You can toggle it's value from the elp command menu.
;;
;;  Lisp references (by Erik Naggum)
;;
;;     http://www.cs.cmu.edu/Groups/AI/html/faqs/lang/lisp/part1/faq-doc-3.html
;;
;;	contains item 1-2 of the comp.lang.lisp FAQ: Lisp books,
;;	introductions, ;; documentation, periodicals, journals, and
;;	conference proceedings.
;;
;;	I recommend Paul Graham: ANSI Common Lisp; Prentice Hall, 1996.
;;	ISBN 0-13-370875-6.  still available from www.books.com,
;;	probably others -- it seems to have gone into a second printing,
;;	but the most recent copies I got my hands on were xerographic
;;	copies, not offset imprints. Could be print on demand.

;;}}}
;;{{{ history

;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; Jun	30	1997	[jari]		19.28	v1.37-1.40	NotReleased
;; - Added new fast forward jump command "$+".
;; - Corrected elp-summary mode: didn't define keymap or menu.
;;
;; Jun	19	1997	[jari]		19.28	v1.35-1.36	NotReleased
;; - Wrote section 'Following lisp code call chain'.
;; - Reordered functions in the file, renamed functions to be more
;;   in line with the rest of the name space.
;; - Now there is CALL CHAIN follow commands, yo can now easily trace
;;   the code flow, go forward and backward in the chain. See prefix
;;   argument to tili-jump-to-definition and new fucntion
;;   tili-back-to-definition*
;; - The tili-find-func-list now print ?! warnings
;; - The tili-find-symbol-load-info now prints AD advice info.
;; - New user function $ * tili-eval-current-buffer-from-file.
;;
;; Jun	4	1997	[jari]		19.28	v1.32-1.34	NotReleased
;; - checkdoc 1.42 corrections; and checkdoc menu changes (function
;;   calls too)
;; - checkdoc 1.42 spell variables can now be set from echo-menu.
;;
;; May	30	1997	[jari]		19.28	v1.31		NotReleased
;;   tili-byte-compile-sexp		:+
;;   tili-byte-compile-display-call-tree :+
;;   tili-jump-to-definition		:! corrected
;;   tili-read-symbol-around-point	!1 corrected
;;
;; May	30	1997	[jari]		19.28	v1.30		NotReleased
;; - New user commands $ { and $ } to find user varaibles and user functions.
;;   defadvice dofconst			:+ `defcustom' eval support.
;;   tili-backward-user-option		:+
;;   tili-forward-user-option		:+
;;   tili-emergency			:+
;;
;; May	28	1997	[jari]		19.28	v1.28-1.29	NotReleased
;; - checkdoc 1.29 corrections. Now supports defcustom too.
;; - XEmacs menu display error corrected.
;; - Eval at 'defcustom' variable is not functional yet. Added advice, but
;;   that's not working right.
;;   tili-forward-user-option		:+
;;   tili-backward-user-option		:+
;;
;; May	21	1997	[jari]		19.28	v1.27		NotReleased
;; - Now the $ ' function jumpt to definitions in external files too
;;   instead of searching only current buffer. This is quite nice
;;   TAGS like navigation as long as load-history entry is found.
;; - New command $ l which shows the package and path from where
;;   the symbol was defined.
;; - Added checkdoc.el support.
;;
;; May	14	1997	[jari]		19.28	v1.24-1.26	NotReleased
;; - now uses cl-prettyexpand or macroexpand.
;; - Changed all non standard boolean tests to use when and unless
;; - macroexpand for single function call without arguments corrected.
;;
;; Apr	21	1997	[jari]		19.28	v1.21-1.23	NotReleased
;; - Now the minor mode wizard problems have been solved.
;;   tili-macroexpand			: New $ @ user command.
;; - Problem in the occur command fixed: broke, because could run forward-sexp
;;   -1 in the comment area.
;;
;; Mar	27	1997	[jari]		19.28	v1.20		NotReleased
;; - THIS Is NOT WORKING VERSION. Tried to figure out how use the wizard
;;   macro. Haven't succeeded yet. The package is wrecked now...
;;
;; Mar	26	1997	[jari]		19.28	v1.19		NotReleased
;; - Rearranged menus, rewrote the fmacro's to use new format.
;;
;; Mar	25	1997	[jari]		19.28	v1.18		NotReleased
;; - Wrote section 'Checklist, how you would check your own package fast'
;; - renamed many variables and some functions. Macros at the top of file.
;; - Rewrote the variable listing function, $ v now accepts prefix arg.
;;
;; Mar	22	1997	[jari]		19.28	v1.17		NotReleased
;; - Adjusted elp functions a bit: harness is bit better. Indented functions
;;   are also now found, if there are no regular left flushed ones.
;; - The overlay property show mode works in XEmacs too.
;;
;; Mar	21	1997	[jari]		19.28	v1.15-1.16	NotReleased
;; - Added elp summary minor mode. Added Elp "e" command to reparse
;;   the current buffer and do reinstrumentations.
;;
;; Mar	20	1997	[jari]		19.28	v1.10-1.14	NotReleased
;; - Added more hooks to default list. Added RESET argument to snoop
;;   function.
;; - Fixed variable display (If bigger that 80, now uses rsz-mini to
;;   display full content. Corrected occur function which deletes
;;   comment lines (wrong regexp)
;; - XE20 CHAR changes. Uses cl function remprop to remove property.
;; - Added char syntax printing.
;;
;; Mar	12	1997	[jari]		19.28	v1.9		NotReleased
;; - Corrected the occur functions. Improved the snoop function so that
;;   you can backup the original values; before you make modifications.
;;   Later you can restore saved values when your experiments are over
;; - Added jump to definition function.
;; - Added symbol *desc* scanning from Emacs obarray. Added lm-verify.
;;
;; Feb	28	1997	[jari]		19.28	v1.8		NotReleased
;; - more improvement to elp interface. Now there is harness function
;;   that eases profiling individual functions in buffer real easily
;; - So many changes, that I don't care to list here.
;;
;; Feb  27      1997    [jari]          19.28   v1.7            NotReleased
;; - Added nice text property/overlay browsing mode.
;; - Added hook snooping function.
;;
;; Feb  25      1997    [jari]          19.28   v1.6            NotReleased
;; - Added error finding functions. Corrected eval command.
;; - Added eval buffer as "defconst". Corrected call at point.
;;
;; Feb  25      1997    [jari]          19.28   v1.4-1.5        NotReleased
;; - menu-bar menu added. ELP support added. many new functions.
;; - Rewrote the installation instructions. Wrote the documentation
;; - Added widen/indent functions and many more.Fixed all bytecompiler errors.
;; - Rearranged package loding to use autoloads
;; - Now package installs itself to lisp modes when it is loaded.
;;
;; Feb  23      1997    [jari]          19.28   v1.3            NotReleased
;; - Okay. Now I have added basic functions here and the menus work great
;;   Next step is to extend this package to what comes tomy mind more.
;; - elp is not yest supported.
;;
;; Feb  21      1997    [jari]          19.28   v1.2            NotReleased
;; - Added buch of newfunctions and rewrote the menus and added more
;;   mode functions.
;;
;; Feb  20      1997    [jari]          19.28   v1.1            Created

;;}}}

;;; Code:

;;{{{ require

;;; ......................................................... &require ...

(require 'tinylibm)
(eval-when-compile (require 'advice))

(eval-and-compile
  (use-dynamic-compilation)

  (autoload 'remprop "cl-extra")

  ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. eldoc ..
  ;; Hmmm, we don't need this in this package, but I like to give
  ;; user a warning: If he's serious at all in programming lip, he should
  ;; get this

  (unless (safe-do (require 'eldoc))
    (message "\n\
tinylisp.el: ** Hm, no eldoc.el? If you're doing lisp at all, please
	     ** get this package fast. It shows function syntax in echo-area.
	     ** ftp://alpha.gnu.ai.mit.edu/friedman/emacs-lisp"))

  ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. . rsz ..
  ;; In XEmacs this package was renamed.
  ;;
  (unless (if (xemacs-p)
	      (safe-do (require 'rsz-minibuf))
	    (or (safe-do (require 'rsz-mini))
		(safe-do (require 'rsz-minibuf))
		))
    (message "\n\
tinylisp.el: ** Try to find rsz-mini.el from some ftp site in order to \n\
             ** be able to view data which spreads multiple lines in\n\
             ** minibuffer more easily.\n\
	     ** Without rsz-mini it may be difficult.
             ** ftp://alpha.gnu.ai.mit.edu/friedman/emacs-lisp (27 May 1997)"
	     ))

  ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. ..  emacs distrib ..
  (autoload 'lm-verify			"lisp-mnt.el" t t)

  ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. checkdoc ..

  (unless (safe-do (require 'checkdoc))
    (message "\n\
tinylisp.el: ** Hm, no checkdoc.el found. Please get it from
	     ** ftp://ftp.ultranet.com/pub/zappo (May 22 1997)")

    (autoload 'checkdoc-eval-current-buffer         "checkdoc" nil t)
    (autoload 'checkdoc-current-buffer              "checkdoc" nil t)
    (autoload 'checkdoc                             "checkdoc" nil t)
    (autoload 'checkdoc-comments                    "checkdoc" nil t)
    (autoload 'checkdoc-rogue-spaces                "checkdoc" nil t)
    (autoload 'checkdoc-eval-defun                  "checkdoc" nil t)
    (autoload 'checkdoc-defun                       "checkdoc" nil t)
    (autoload 'checkdoc-minor-mode                  "checkdoc" nil t)
    (autoload 'checkdoc-find-error-mouse            "checkdoc" nil t)
    (autoload 'checkdoc-find-error                  "checkdoc" nil t)
    )

  ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. . elp ..

  (unless (safe-do (require 'elp))
    (message "\n\
tinylisp.el: ** Hm, no ELP.EL get it from http://www.python.org/ftp/emacs/
	     ** profiling functions are not available. Calling them from
	     ** package will create run time error.")

    ;;  This pretends the functions exist and avoids byte compiler errors.
    ;;
    (defvar   elp-all-instrumented-list	nil)
    (defvar   elp-function-list		nil)

    (defvar   elp-master			nil)
    (defvar   elp-results-buffer		"*ELP Profiling Results*")
    (defvar   elp-reset-after-results	nil)

    (autoload 'elp-instrument-function              "elp" nil t)
    (autoload 'elp-restore-function                 "elp" nil t)
    (autoload 'elp-instrument-list                  "elp" nil t)
    (autoload 'elp-instrument-package               "elp" nil t)
    (autoload 'elp-restore-list                     "elp" nil t)
    (autoload 'elp-restore-all                      "elp" nil t)
    (autoload 'elp-reset-function                   "elp" nil t)
    (autoload 'elp-reset-list                       "elp" nil t)
    (autoload 'elp-reset-all                        "elp" nil t)
    (autoload 'elp-set-master                       "elp" nil t)
    (autoload 'elp-unset-master                     "elp" nil )
    (autoload 'elp-wrapper                          "elp" nil )
    (autoload 'elp-sort-by-call-count               "elp" nil )
    (autoload 'elp-sort-by-total-time               "elp" nil )
    (autoload 'elp-sort-by-average-time             "elp" nil )
    (autoload 'elp-output-result                    "elp" nil )
    (autoload 'elp-results                          "elp" nil t)
    (autoload 'elp-submit-bug-report                "elp" nil t)
    )
  )

(defgroup-tiny TinyLisp tili-: extensions
  "Lisp programming help module.
  Overview of features.

      Lisp coding help

      o   Create list of all variables from the buffer.
          Uou can use the list in your bug report function or just
          to get an overview of the variables: names and the order how
          you have used them (The order is important if you use defcustom)
      o   Create function list (or occur menu)
      o   Create autoload list (or occur menu)
      o   Evaluate current definition under point (reparse function,
          reset defvar or defcustom variable)
      o   Print variable's value under point, set variable's value
          under point.
      o   Call function under point (to test it immediately)
      o   Indent function/variable around point.
      o   Two tun on Modes: Show underlying  properties/overlays or
          charcter symbol information.
      o   FINDS LISP CODE ERROR POINT.
      "
  )

;;}}}
;;{{{ setup: mode definition

;;; ............................................................ &mode ...


(defcustom tili-:menu-use-flag t
  "*Non-nil means to use echo-area menu."
  :type 'boolean :group 'TinyLisp
  )

;;
;;  Creating minor mode
;;
;;	This macro creates the full minor mode and all needed variables!
;;
;;  Mode name E for mode name
;;
;;	general lisp helper mode; please see these too !
;;
;;	(e)lisp
;;	(e)lp
;;	(e)ldoc.el
;;	(e)xpand
;;
;;  Prefix variable "$"
;;
;;	You seldom use end anchor $ in lisp. Use $$ to self insert it
;;	Another characters you could choose: "!", "_"
;;	If possible, select some character that is non-shifted
;;	for maximum accessibility of this minor mode.
;;

(eval-and-compile

;;;###autoload (autoload 'tili-mode "tinylisp" t t)
(ti::m-fmacro-minor-mode-wizard
 "tili-" " E" "$" "E" 'TinyLisp "tili-:"  ;1-6

 "This minor mode is used along with the lisp major modes. You can
evaluate expressions, reread functions, check your lisp packages
syntaax, create autoloads and do many more things.

Defined keys:
\\{tili-:mode-map}
"

 "Lisp help"				;7
 nil					;8


 "Elisp programming help menu."		;9
 (list					;arg 10
   tili-:mode-easymenu-name
   ["Eval whole buffer"		     tili-eval-current-buffer	    t]
   ["Eval whole buffer, `load'"	     tili-eval-current-buffer-from-file  t]
   ["Eval whole buffer as defconst"  tili-eval-current-buffer-defconst t]
   ["Eval statement at point"	     tili-eval-at-point		    t]
   ["Eval reverse statement at point" tili-eval-reverse		    t]
   ["Eval and edit line "	     tili-eval-and-edit		    t]
   ["Macroexpand macro funcall"	     tili-macroexpand		    t]
   "----"
   ["Call statement at point"	     tili-call-at-point		    t]
   ["Set value at point"	     tili-set-value-at-point	    t]
   ["Jump to definiton"		     tili-jump-to-definition	    t]
   ["Jump to definiton (call-chain)" tili-jump-to-definition-chain  t]
   ["Back to definiton (call-chain)" tili-back-to-definition	    t]
   "----"
   ["Forward user var-func"	     tili-forward-user-option	    t]
   ["Backward user var-func"	     tili-backward-user-option	    t]
   "----"


   (list
    "Modes, find error"
    ["Mode, property show"	      tili-property-show-mode	    t]
    ["Mode, char syntax show"	      tili-syntax-show-mode	    t]
    ["Find lisp error, method 1"      tili-error-find-1		    t]
    ["Find lisp error, method 2"      tili-error-find-2		    t]
    )

   (list
    "Variables and symbols"
    ["Occur"			      tili-occur-verbose	    t]
    ["Occur, select next"	      tili-occur-select-forward	    t]
    ["Collect variable list"	      tili-find-var-list	    t]
    ["Collect variable list, occur"   tili-find-var-list-occur	    t]
    ["Collect function list"	      tili-find-func-list	    t]
    ["Collect function list, occur"   tili-find-func-list-occur	    t]
    ["Info, show symbol load path"    tili-find-symbol-load-info    t]
    ["Info, show loaded libraries"    tili-find-library-info	    t]
    "----"
    ["Construct autoloads from buffer" tili-autoload-generate	    t]
    ["Construct autoloads from file"  ti::m-pkg-autoload-create-on-file t]
    ["Describe symbols"		      ti::y-describe-symbols	    t]
    ["Describe loaded package"	      ti::y-get-file-documentation  t]
    ["Snoop variables"		      tili-snoop-variables	    t]
    )

   (list
    "Miscellaneous"
    ["Emergency - defcustom"          tili-emergency		    t]
    ["Indent function or variable"    tili-indent-around-point	    t]
    ["Narrow to function"	      tili-narrow-to-function	    t]
    ["Widen"			      widen			    t]
    ["Convert word to defmacro var."  tili-defmacro-surround-word   t]
    ["Check package's layout syntax"  tili-lm-verify		    t]
    ["Byte compile current function." tili-byte-compile-sexp	    t]
    ["Show Call tree for file"        tili-byte-compile-display-call-tree t]
    )

   (list
    "Documentation string check."
    ["Check forward"		    tili-checkdoc		t]
    ["Check buffer, take notes"	    tili-checkdoc-notes		t]
    ["Check comments"		    checkdoc-comments		t]
    ["Check comments, take notes"   tili-checkdoc-comment-notes	t]
    ["Check defun, current point"   checkdoc-eval-defun		t]
    ["Checkdoc minor mode."	    checkdoc-minor-mode		t]
    )

   (list
    "Elp lisp profiling menu"
    ["Instrument function"           tili-elp-instrument-function    t]
    ["Instrument Buffer"             tili-elp-instrument-buffer      t]
    ["Uninstrument function"         tili-elp-restore-function       t]
    ["Uninstrument Buffer"           tili-elp-restore-buffer         t]
    ["Uninstrument All"              tili-elp-restore-all            t]
    ["Reparse instrumentation"       tili-reparse-instrumentation    t]
    "----"
    ["List instrumented functions"   tili-elp-function-list          t]
    ["List All instrumented functions" tili-elp-function-list-partial t]
    ["Harness test (eval from point)" tili-elp-harness               t]
    ["Reset timing list"             tili-elp-reset-list             t]
    ["Show  timing list"             tili-elp-results                t]
    )
   )

 (progn				;arg 11
    (cond
     (tili-:menu-use-flag
      ;;  Using menu to remeber commands is easier if you don't use
      ;;  menu bar at all.
      ;;
      (define-key map p 'tili-menu)
      )

  (t
   (if (eq 1 (length p))
       (define-key map (concat p p) 'self-insert-command))

   (define-key map (concat p "Z") 'tili-emergency)


   (define-key map (concat p "-")   'tili-eval-current-buffer)
   (define-key map (concat p "*")   'tili-eval-current-buffer-from-file)
   (define-key map (concat p "=")   'tili-eval-current-buffer-defconst)
   (define-key map (concat p ".")   'tili-eval-at-point)
   (define-key map (concat p "\\")  'tili-eval-reverse)
   (define-key map (concat p "@")   'tili-macroexpand)
   (define-key map (concat p "B")   'tili-byte-compile-sexp)
   (define-key map (concat p "C")   'tili-byte-compile-display-call-tree)

   (define-key map (concat p ",")   'tili-call-at-point)
   (define-key map (concat p ";")   'tili-set-value-at-point)
   (define-key map (concat p "!")   'tili-error-find-1)
   (define-key map (concat p "#")   'tili-error-find-2)
   (define-key map (concat p "'")   'tili-jump-to-definition)
   (define-key map (concat p "+")   'tili-jump-to-definition)
   (define-key map (concat p "'\177")   'tili-back-to-definition)
   (define-key map (concat p "`")   'tili-defmacro-surround-word)

   ;; No special reason yo pick these keys, My keyboard just has tem
   ;; unshifed -- Choose your favourites.
   ;;
   (define-key map (concat p "{")   'tili-backward-user-option)
   (define-key map (concat p "}")   'tili-forward-user-option)

   (define-key map (concat p "a")   'tili-autoload-generate)
   (define-key map (concat p "A")   'ti::m-pkg-autoload-create-on-file)

   (define-key map (concat p "d")   'ti::y-describe-symbols)
   (define-key map (concat p "D")   'ti::y-get-file-documentation)

   (define-key map (concat p "e")   'tili-eval-and-edit)
   (define-key map (concat p "f")   'tili-find-func-list)
   (define-key map (concat p "F")   'tili-find-func-list-occur)
   (define-key map (concat p "l")   'tili-find-symbol-load-info)
   (define-key map (concat p "L")   'tili-find-library-info)

   (define-key map (concat p "i")   'tili-indent-around-point)
   (define-key map (concat p "n")   'tili-narrow-to-function)

   (define-key map (concat p "o")   'tili-occur-verbose)
   (define-key map (concat p "+")   'tili-occur-select-forward)


   (define-key map (concat p "p")   'tili-property-show-mode)
   (define-key map (concat p "s")   'tili-snoop-variables)


   (define-key map (concat p "v")   'tili-find-var-list)
   (define-key map (concat p "V")   'tili-find-var-list-occur)

   (define-key map (concat p "w")   'widen)
   (define-key map (concat p "x")   'tili-checkdoc)
   (define-key map (concat p "X")   'tili-lm-verify)
   (define-key map (concat p "y")   'tili-syntax-show-mode)

   (define-key map (concat p "bv")  'tili-b-variables)
   (define-key map (concat p "bf")  'tili-b-functions)
   (define-key map (concat p "br")  'tili-b-record)
   (define-key map (concat p "bR")  'tili-b-record-empty)
   (define-key map (concat p "bt")  'tili-b-eval)

   (define-key map (concat p "ei")  'tili-elp-instrument-function)
   (define-key map (concat p "eI")  'tili-elp-instrument-buffer)
   (define-key map (concat p "eu")  'tili-elp-restore-function)
   (define-key map (concat p "eU")  'tili-elp-restore-buffer)
   (define-key map (concat p "eA")  'tili-elp-restore-all)
   (define-key map (concat p "ee")  'tili-reparse-instrumentation)
   (define-key map (concat p "eh")  'tili-elp-harness)

   (define-key map (concat p "ef")  'tili-elp-function-list)
   (define-key map (concat p "eF")  'tili-elp-function-list-partial)
   (define-key map (concat p "er")  'tili-elp-reset-list)
   (define-key map (concat p "es")  'tili-elp-results)
   )))
 )
)

;;; ................................................... &&mode-summary ...

;;;###autoload (autoload 'tili-elp-summary-mode "tinylisp" t t)

(eval-and-compile

(ti::m-fmacro-minor-mode-wizard
 "tili-elp-summary-" " Elp-sum" nil " Elp-sum" 'TinyLisp
 "tili-:elp-summary-" ;1-6


 "Sorting elp summary buffers.
Defined keys:
\\{tili-:elp-summary-mode-map}
"

 "Elp summary sort"			;7
 nil					;8


 "Elp summary sort menu."		;9
 (list					;arg 10
   tili-:elp-summary-mode-easymenu-name
   ["Sort by function name"     tili-elp-summary-sort-column-1 t]
   ["Sort by call count"	tili-elp-summary-sort-column-2 t]
   ["Sort by elapsed time"	tili-elp-summary-sort-column-3 t]
   ["Sort by average time"	tili-elp-summary-sort-column-4 t]
   )

 (progn				;arg 11
   ;; Function Name  Call Count  Elapsed Time  Average Time
   ;; =============  ==========  ============  ============
   ;;
   (define-key map "f"  'tili-elp-summary-sort-column-1)
   (define-key map "c"  'tili-elp-summary-sort-column-2)
   (define-key map "e"  'tili-elp-summary-sort-column-3)
   (define-key map "a"  'tili-elp-summary-sort-column-4)
   )
 )

)

;;}}}
;;{{{ setup: hooks

;;; ......................................................... &v-hooks ...

(defcustom tili-:load-hook nil
  "*Hook that is run when package is loaded."
  :type 'hook :group 'TinyLisp
  )

(defcustom tili-:find-func-list-hook 'tili-highlight-default
  "*Hook run when tili-find-func-list-hook has displayed the list."
  :type 'hook :group 'TinyLisp
  )

(defcustom tili-:find-var-list-hook 'tili-highlight-default
  "*Hook run when `tili-find-func-list' has displayed the list."
  :type 'hook :group 'TinyLisp
  )


;;}}}
;;{{{ setup: private

;;; ....................................................... &v-private ...
;;; Private variables


(defvar HARNESS  nil
  "Described in function `tili-elp-harness'.
This variable is set to t when harness is on going and set to
nil when harness test is over.")


(defvar tili-:call-chain  nil
  "List of buffers and buffer positions. '(mark mark ..)
Whenever you call `tili-jump-to-definition' the current positions
is recoded and one more element to the _beginning_ of list is added.
You can navigate back with `tili-back-to-definition' and the first
element from the list is removed.")


(defvar tili-:buffer-elp  "*tili-elp*"
  "Temporary elp info buffer.")

(defvar tili-:buffer-autoload  "*tili-autoloads*"
  "Temporary buffer.")

(defvar tili-:buffer-variables "*tili-variables*"
  "Temporary buffer.")

(defvar tili-:buffer-functions "*tili-functions*"
  "Temporary buffer.")

(defvar tili-:buffer-record  "*tili-record*"
  "Record variable contents to this buffer.")


(defvar tili-:buffer-tmp "*tili-tmp*"
  "Temporary buffer.")

(defvar tili-:buffer-macro "*tili-macroexpand*"
  "Temporary buffer.")

(defvar tili-:buffer-eval " *tili-eval*"
  "Temporary buffer.")


(defconst tili-:regexp-macro-definition
  "^defcustom\\|^defgroup\\|^defadvice"
  "Regexp for commands that define macros, like `defcustom' `defgroup'.")


(defconst tili-:regexp-function
  (concat
   ;; see nntp.el for deffoo
   ;;
   "^(\\(defun\\|defsubst\\|defmacro\\|deffoo\\|defadv\\)"
   "[ \t]+\\([^ \t\n]+\\)"
   )
  "Regexp to match functions.
This smust have SUBMATCH1 and SUBMATCH2 which represent function
type and name.")

(defconst tili-:regexp-variable
  (concat
   "^(\\("

   ;;  Normal lisp variables
   "defvar\\|defconst"

   ;; Custom.el defined variables in 19.35
   "\\|defgroup\\|defcustom"

   "\\)[ \t]+\\([^ \t\n]+\\)"
   )
  "Regexp to match variables.
This smust have SUBMATCH1 and SUBMATCH2 which represent
variable type and name.")

(defvar tili-:variable-not-charset  "^()'`\", \t\n"
  "When reading variable from buffer, this is Charset to match.
Notice that ^ at the beginning of charset reverses it.")

(defvar tili-:find-error  nil
  "'Find error' functon's data.")

(defvar tili-:history-occur  nil
  "History.")

;;  Too bad this is hard coded in emacs..
;;
(defvar tili-:occur-buffer-name "*Occur*"
  "Emacs Occur buffer.")

;;}}}
;;{{{ setup: private, mode

;;; .................................................. &v-mode-private ...
;;; mode and map don't have ':' due to emacs convention of mode
;;; definition.


(defvar tili-:property-show-mode  nil
  "Property show mode (flag).")

(defvar tili-:syntax-show-mode  nil
  "Property show mode (flag).")


;;}}}
;;{{{ setup: public, user

;;; ........................................................ &v-public ...
;;; User configurable


(defcustom tili-:register  ?\'
  "*An Emacs register that is used eg. for saving point or copying text."
  :type 'character :group 'TinyLisp
  )


(defcustom tili-:macroexpand-function-list
  '("macroexpand" "cl-prettyexpand")
  "*Completion list of function STRINGS to expand macro call.
You can use commands `cl-prettyexpand', which sometimes does
good fromatting, but does not necessarily expand to what you want to see.
The default command is `macroexpand'."
  :type '(repeat string) :group 'TinyLisp
  )


(defcustom tili-:table-reverse-eval-alist
  '((add-hook       . remove-hook)
    (add-hook-local . remove-hook)	;defined in 'm' lib
    (remove-hook    . add-hook)
    )
  "*Table of reverse commands. Format '((ORIG-FSYM . REVERSE-FSYM) ..)."
  :type 'list :group 'TinyLisp
  )


(defcustom tili-:table-snoop-variables
  '(
    ("hook-command"
     (
      pre-command-hook
      post-command-hook
      post-command-idle-hook
      ))
    ("hook-file"
     (
      write-file-hooks
      find-file-hooks
      after-save-hook
      ))
    ("hook-mail"
     (
      mail-mode-hook
      mail-setup-hook
      mail-citation-hook
      mail-yank-hooks
      mail-send-hook
      ))

    ("hook-message"
     (
      message-mode-hook
      message-setup-hook
      message-signature-setup-hook
      message-header-setup-hook
      message-header-hook
      message-send-hook
      message-sent-hook
      ))

    ("hook-basic"
     (
      pre-command-hook
      post-command-hook
      post-command-idle-hook
      write-file-hooks
      find-file-hooks
      after-save-hook
      after-init-hook
      ))
    )
  "*List of interesting variables printed from `tili-snoop-variables'.
Non existing variables can also be listed but they are not checked.

Format:
 '((\"LIST-NAME\" (var var var ..))
   (L (v v v ...))
   )"

  :type '(repeat
	  (list
	   (string :tag "Completion name")
	   (repeat (symbol :tag "Var"))
	   ))
  :group 'TinyLisp
  )

;;}}}
;;{{{ setup: menu

;;; .......................................................... &v-menu ...

(defconst tili-:menu-main		;bookmark -- &menu
  (list
   ;;  All commands do not fit to echo menu, but here are at least
   ;;  the most used ones.
   ;;
   "\
Lisp: -=*.\\@E)val ,;')call wn)ar py)mode vVfFl i) +o)cc aA)load xXsdD >bec"
    (list

     (cons ?Z  (list '(call-interactively 'tili-emergency)))

     (cons ?-  (list '(call-interactively 'tili-eval-current-buffer)))
     (cons ?*  (list '(call-interactively
		       'tili-eval-current-buffer-from-file)))
     (cons ?=  (list '(call-interactively 'tili-eval-current-buffer-defconst)))
     (cons ?.  (list '(call-interactively 'tili-eval-at-point)))
     (cons ?,  (list '(tili-call-at-point current-prefix-arg)))
     (cons ?\\  (list '(call-interactively 'tili-eval-reverse)))
     (cons ?@  (list '(call-interactively 'tili-macroexpand)))
     (cons ?B  (list '(call-interactively 'tili-byte-compile-sexp)))
     (cons ?C  (list
		'(call-interactively 'tili-byte-compile-display-call-tree)))

     (cons ?\; (list '(call-interactively 'tili-set-value-at-point)))
     (cons ?!  (list '(call-interactively 'tili-error-find-1)))
     (cons ?#  (list '(call-interactively 'tili-error-find-2)))
     (cons ?+  (list '(tili-jump-to-definition-chain)))
     (cons ?'  (list '(tili-jump-to-definition current-prefix-arg 'verb)))
     (cons ?\177  (list '(tili-back-to-definition)))

     (cons ?{  (list '(call-interactively 'tili-backward-user-option)))
     (cons ?}  (list '(call-interactively 'tili-forward-user-option)))



     ;;  This is not used often
     ;;
     (cons ?` (list '(call-interactively 'tili-defmacro-surround-word)))

     (cons ?a  (list '(call-interactively 'tili-autoload-generate)))
     (cons ?A  (list '(call-interactively 'ti::m-pkg-autoload-create-on-file)))

     (cons ?d  (list '(call-interactively 'ti::y-describe-symbols)))
     (cons ?D  (list '(call-interactively 'ti::y-get-file-documentation)))

     (cons ?E  (list '(call-interactively 'tili-eval-and-edit)))

     (cons ?f  (list '(call-interactively 'tili-find-func-list)))
     (cons ?F  (list '(call-interactively 'tili-find-func-list-occur)))
     (cons ?l  (list '(call-interactively 'tili-find-symbol-load-info)))
     (cons ?L  (list '(call-interactively 'tili-find-library-info)))

     (cons ?i  (list '(call-interactively 'tili-indent-around-point)))
     (cons ?n  (list '(call-interactively 'tili-narrow-to-function)))

     (cons ?o  (list '(tili-occur-verbose current-prefix-arg)))
     (cons ?+  (list '(tili-occur-select-forward current-prefix-arg)))

     (cons ?p  (list '(tili-property-show-mode current-prefix-arg 'verb)))
     (cons ?s  (list '(let* ((i (tili-snoop-variables-i-args)))
                        (tili-snoop-variables (nth 0 i) (nth 1 i))
                        )))

     (cons ?v  (list '(tili-find-var-list current-prefix-arg)))
     (cons ?V  (list '(call-interactively 'tili-find-var-list-occur)))

     (cons ?w  (list '(call-interactively 'widen)))
     (cons ?x  (list '(call-interactively 'tili-lm-verify)))
     (cons ?y  (list '(tili-syntax-show-mode current-prefix-arg 'verb)))

     (cons ?e 'tili-:menu-elp)
     (cons ?b 'tili-:menu-buffers)
     (cons ?c 'tili-:menu-checkdoc)

     ;; Self insert command
     ;;     User may have defined multichararcter minor map entry
     ;;     like C-cE, we only do self insert if it is NOT
     ;;     multicharacter.
     ;;
     (cons (string-to-char              ;get first char
            (substring tili-:mode-prefix-key 0 1))

           (list
            '(progn
               (if (eq 1 (length tili-:mode-prefix-key))
                   (insert tili-:mode-prefix-key)
                 (message "The prefix key isnt one charcracter.")))
            ))
     ))
  "Lisp code helper menu.

Eval commands:

    -  Eval whole buffer
    *  Eval whole buffer, but with load command. This gas the effect that
       tat the function and variable definitions are recorded to load
       history and you can use \\[tili-jump-to-definition] command.
    =  Treat all variables as defconst and eval buffer. (With this
       you can read the defaults if you're in package buffer)
    .  Eval current statement. If you have made changes to the function or
       variable, which can be also defvar, this command evaluates it again
       so that it gets the new definition. (defvar is treated as defconst)
    \\ Reverse the command around point and eval the statement. See
       variable `tili-:table-reverse-eval-alist'.

    @  Macroexpand current call to macro.
    E  Read current line and allow editing it, then eval the statement.

    !  Find lisp errors. Go to `point-min' and evaluate buffer portions
       until error occurs.
    #  Find lisp error with method 2, if previous failed.
    Z  Emergency! If you evaled `defcustom' variable and you were thrown
       to error buffer, call this command immediately.


Byte compilation and function flow

    B  Byte compile 'defun' around point. With prefix arg DISSASSEBMLE.
    C  Display call tree for file.

    '  Jump to definition of variable or function in the same file.
       With \\[universal-argument], save the call-chain point.
       With non-nil prefix argument, clear the call chain. Use
       BACKSPACE or \177 key to go back the saved call chain.
    +  Record position to call chain before jump to definition. This
       like calling previous command with prefix arg.
    DEL Back to previous definition and remove mark from call chain.

    }  Go to next user option; a star mark, or to user function; interactive.
    {  Same as above, but backward.

Symbol manipulation

    ,  Call current word around point. If the word is variable, that propably
       prints its value. If word is function, call function.
    ;  Set new value for variable at point. If the read word is not existing
       variable, then this only prints warning messages.
    `  Surround current word with defmacro statement (, WORD)


Function commands:

    n  Narrow to current lisp function  (w to widen; \\[widen])
    w  widen
    i  Indent current function or variable around the point


Listing and occur commands:

    o  Run occur for full buffer and filter out comments. Prefix arg says
       _not_ to fil ter out full comment lines.
    +  Go to next occur line in buffer. With C - u backward.

    f  Find all functions from the buffer
    F  Find function and create occur menu.
    l  Show symbol Load information (file where is was defined)
    L  Library information, examine all packages in emacs.

    v  Find all variables from he buffer. Prefix args classifies variables.
    V  Find variables and create occur menu.


Misc commands

    a  Create autoloads by reading the buffer's file
    A  Create autoloads out of some file

    d  Describe symbols. This scans whole emacs obarray to find all the
       matching symbols.
    D  Describe loaded package. You can rip all the documentation from
       a file by doing this 1) load file into emacs 2) eval it and finally
       3) call this function and give file path. It collects all variable
       and function documentation to single list.


Modes and utilities

    p  Property show mode. With 3 C - u turns on recording.
    y  syntax mode, Show syntax of charcter under cursor.

    s  Snoop variables. Select from predefined variable sets.
       1           With prefix arg, variables values are recorded to
		   to buffer `tili-:buffer-record' and
       0           Save variables values.
       9           Restore variables from the saved copies.
       8           Kill saved variable state
       5           Set all variables to nil in list
       C - u       allows editing the variables.

    x  Check variables and functions documentation strings ao that they follow
       Emacs lisp code guidelines. File: elisp,  Node: Documentation Tip.
    X  Chech package layout synta(x). When you're making a package this ensures
       that tags and keywords are in proper format.


Additional menus

    e  elp.el, Emacs lisp profiler menu
    b  Buffer menu")


(defconst tili-:menu-elp
  '(
    (let (val
          )
      (format
       "%s%s[/]elp: (iIuUAe)instrument (fFrsS%s)list (h)arness"

       ;;  Is there functions instrumented?
       ;;
       (if elp-all-instrumented-list
           (if (eq 0 (setq val (length elp-all-instrumented-list)))
               ""
             (format "%d " val)
             )
         "")

       (if (null elp-master) "" (symbol-name val))

       (if elp-reset-after-results ":t" "")
       ))
    (
     (?i  . ( (call-interactively 'tili-elp-instrument-function)))
     (?I  . ( (call-interactively 'tili-elp-instrument-buffer)))

     (?u  . ( (call-interactively 'tili-elp-restore-function)))
     (?U  . ( (call-interactively 'tili-elp-restore-buffer)))
     (?A  . ( (call-interactively 'tili-elp-restore-all)))
     (?h  . ( (tili-elp-harness current-prefix-arg 'verb)))
     (?e  . ( (call-interactively 'tili-reparse-instrumentation)))

     (?f  . ( (tili-elp-function-list current-prefix-arg 'verb)))
     (?F  . ( (tili-elp-function-list-partial current-prefix-arg 'verb)))
     (?r  . ( (call-interactively 'tili-elp-reset-list)))
     (?s  . ( (tili-elp-results current-prefix-arg)))
     (?S  . (t (tili-elp-reset-after-results)))

     (?/  . tili:menu-main)

     ))
  "Elp help menu.
The menu shows some status parameters in the echo area.

  '[FCOUNT] elp:  (:t)list'
   |               |
   |               See 'S' key when this is shown
   Count of currently instrumented functions

Basic elp commands:
/   Back to root menu.

i   instrument current function at point
I   Instrument whole buffer's functions
u   uninstrument function at point
U   Uninstrument buffer's functions.
A   Uninstrument all function in elp list (reastore all)
e   r(e)parse instrumentation: forget all instrumented functions,
    eval buffer to read new function definitions, and instrument those
    functions.

Misc:
h   Harness test. Eval everything 3 times from current point forward
    and record results. See `tili-elp-harness' for full explanation.
    Prefix arg determines hasness rounds.

Function information:
f   List _all_ functions instrumented. Prefix arg print the function to
    separate buffer.
F   Same as above, but list all only specific functions in
    `elp-function-list'.

Timing information:
s   Show timing results. With prefix arg save results to RECORD buffer.
S   rese(:t) flag, Toggle setting of variable `elp-reset-after-results'.
r   Reset timing list.")


(defconst tili-:menu-buffers
  '(
    "go buffer: (rR)ecord (v)variables (f)unctions (e)val"
    (
     (?r  . ( (tili-b-record)))
     (?R  . ( (tili-b-record-empty)))
     (?v  . ( (tili-b-variables)))
     (?f  . ( (tili-b-functions)))
     (?e  . ( (tili-b-eval)))
     ))
  "Display TinyLisp buffers.

/   Back to root menu.
r   Display the record buffer where the variable contents
    are stored when you call C - u `tili-call-at-point'
R   Kill record buffer.
v   variables buffer
f   functions buffer
e   eval buffer")



(defconst tili-:menu-checkdoc
  '(
    (let ((spell checkdoc-spellcheck-documentation-flag)
	  val
          )
      (cond
       ((not (featurep 'checkdoc))
	(error "No checkdoc available (not loaded)."))
       ((not (boundp 'checkdoc-verb-check-experimental-flag))
	(error "You have old checkdoc.el version.")
	))

      (format
       (concat
	"%s%s%s%s%s checkdoc: "
	"SPC)point RET)notes DEL)buf cC)om  m)ode aA~- bB sS"
	)

       (cond
	((eq checkdoc-autofix-flag 'automatic)	    "Auto")
	((eq checkdoc-autofix-flag 'semiautomatic)  "Semi")
	((eq checkdoc-autofix-flag 'query)	    "Query")
	((null checkdoc-autofix-flag)		    "")
	(t "??"))

       (cond
	((null checkdoc-bouncy-flag)		    "")
	((eq   checkdoc-bouncy-flag 'never)	    "")
	(t "B"))

       (cond
	((null checkdoc-arguments-in-order-flag)    "")
	(t "G"))

       (cond
	((null checkdoc-verb-check-experimental-flag)    "")
	(t "V"))

       (cond
	((eq spell  'defun)	    "sD")
	((eq spell  'buffer)        "sB")
	((eq spell  'interactive)   "sI")
	((eq spell  t)		    "S")
	((null spell)		    "")
	(t			    "s?"))

       )
      )
    (
     (?\    . ( (checkdoc-eval-defun)))
     (?\177 . ( (tili-checkdoc)))
     (?\b   . ( (tili-checkdoc)))
     (?\C-m . ( (tili-checkdoc-notes)))
     (?\C-j . ( (tili-checkdoc-notes)))
     (?m    . ( (call-interactively 'checkdoc-minor-mode)))
     (?c    . ( (checkdoc-comments)))
     (?C    . ( (tili-checkdoc-comment-notes)))

     (?a    . (t (progn (setq  checkdoc-autofix-flag 'semiautomatic))))
     (?A    . (t (progn (setq  checkdoc-autofix-flag 'automatic))))
     (?~    . (t (progn (setq  checkdoc-autofix-flag 'query))))
     (?-    . (t (progn (setq  checkdoc-autofix-flag nil))))


     (?B    . (t (progn (setq  checkdoc-bouncy-flag t))))
     (?b    . (t (progn (setq  checkdoc-bouncy-flag nil))))

     (?G    . (t (progn (setq  checkdoc-arguments-in-order-flag t))))
     (?g    . (t (progn (setq  checkdoc-arguments-in-order-flag nil))))

     (?V    . (t (progn (setq  checkdoc-verb-check-experimental-flag t))))
     (?v    . (t (progn (setq  checkdoc-verb-check-experimental-flag nil))))

     (?S    . (t (progn (setq  checkdoc-spellcheck-documentation-flag t))))
     (?s    . (t (progn (setq  checkdoc-spellcheck-documentation-flag nil))))
     (?d    . (t (progn (setq  checkdoc-spellcheck-documentation-flag
			       'defun))))
     (?r    . (t (progn (setq  checkdoc-spellcheck-documentation-flag
			       'buffer))))
     ))
  "According to checkdoc manual:
...The emacs lisp manual has a nice chapter on how to write
documentation strings.  Many stylistic suggestions are fairly
deterministic and easy to check for programatically, but also easy
to forget. The main checkdoc engine will perform the stylistic
checks needed to make sure these styles are remembered.

The echo area menu shows following status information

  [FIX|BOUNCE|G|V|S] checkdoc:
   |   |      | | |
   |   |      | | `checkdoc-spellcheck-documentation-flag'
   |   |      | `checkdoc-verb-check-experimental-flag'
   |   |      `checkdoc-arguments-in-order-flag'
   |   `checkdoc-bouncy-flag' state
   `checkdoc-autofix-flag'

Commands:
/	Back to root menu.
SPC	`checkdoc-eval-defun'
DEL	Check code from current point forward.
RET	`tili-checkdoc-notes'
c	`checkdoc-comments'
C	`tili-checkdoc-comment-notes'

Misc:
m	Turn on minor mode which checks docstring while you write them
aA~-	Change `checkdoc-autofix-flag' a=semi A=auto ~=query -=nil
bB	Change `checkdoc-bouncy-flag'			b=nil B=t
gG	Change `checkdoc-arguments-in-order-flag'	g=nil G=t
vV	Change `checkdoc-verb-check-experimental-flag'	v=nil V=t
sSdr    Change `checkdoc-verb-check-experimental-flag'  s=nil S=interactive
        d=defun r=buffer")





;;}}}
;;{{{ version

;;; ....................................................... &v-version ...

(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinylisp.el"
 "tili"
 tili-:version-id
 "$Id: tinylisp.el,v 1.40 1997/07/01 09:08:02 jaalto NotReleased $"
 '(tili-:version-id

   tili-:load-hook
   tili-:find-func-list-hook
   tili-:find-var-list-hook
   tili-:menu-use-flag

   tili-:macroexpand-function-list
   tili-:table-reverse-eval-alist
   tili-:table-snoop-variables
   )))

;;}}}

;;; ########################################################## &macros ###

;;{{{ macros

;;; ----------------------------------------------------------------------
;;;
(defsubst tili-read-word  ()
  "Read word under point."
  (ti::b-read-word tili-:variable-not-charset))


;;; ----------------------------------------------------------------------
;;;
(defsubst tili-get-symbol  (string)
  "Return symbol from STRING.
If function does not exist or is string cannot be read, then return nil

  \"(function arg1\"  --> 'function
  \"(defvar xx\"      --> 'xx"
    (let* ((re-f  (substring tili-:regexp-function
			     1 (length tili-:regexp-function)))

	   (re-v	(substring tili-:regexp-variable
				   1 (length tili-:regexp-variable)))
	   sym
	   )
      (cond
       ((and (or (string-match re-f string)
		 (string-match re-v string))
	     (setq sym (intern-soft  (read-match 2 string)))
	     ))
       ;;  Read first word then
       ((setq sym (strmatget "[^() \t\n\]+" 0 string))
	(setq sym (intern-soft sym))
	))
      sym
      ))


;;; ----------------------------------------------------------------------
;;;
(defun tili-where-is-package  (package)
  "Try to locate PACKAGE. It one used in `load` command.
Nil parameter is also accepted."
  (let* (ret
         )
    (cond
     ((null package))			;Skip right away
     ((string-match "^/" package)
      (setq ret package))

     ((string-match "\\.el$\\|\\.elc$" package)
      (setq ret (ti::f-get-load-path package load-path)))

     (t
      (setq ret
	    (ti::f-get-load-path
	     (ti::s-verify-ends package ".el$" ".el")
	     load-path))
      ))
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defsubst tili-where-is-symbol  (sym)
  "Find file where SYM is defined. Return nil if cannot find definition."
  (let* (file
	 tmp
	 )
    (setq file (car-safe (ti::y-lh-where sym)))
    (when (stringp file)
      (cond
       ((string-match "/" file)
	;;  load history entry may also be ~/elisp/xxx but we need real
	;;  file name.
	;;
	(if (file-exists-p (setq tmp (ti::s-verify-ends file ".el" ".el")))
	    tmp
	  (error "%s.el does not exist." file))
	)
       (t
	(ti::f-get-load-path (ti::s-verify-ends file ".el$" ".el") load-path)
	)))))


;;; ----------------------------------------------------------------------
;;;
(defsubst tili-push-call-chain (&optional pop data verb)
  "Push current point to call chain.
Input:

  POP	    flag, instead of push, do pop to last saved positions
  DATA	    push DATA to chain.
  VERB	    print verbose messages.

Optionally POP. VERB prints message."
  (let* (mark
	 )
    (if (null pop)
	(push data tili-:call-chain)
      (if (null tili-:call-chain)
	  (error "tili-:call-chain is empty, nothing to pop.")
	(setq mark (car tili-:call-chain))
	(NEXTP tili-:call-chain)
	(goto-char mark)
	))
    ))



;;; ----------------------------------------------------------------------
;;;
(put 'tili-symbol-do-macro 'lisp-indent-function 2)
(defmacro tili-symbol-do-macro (string noerr &rest body)
  "Execute body if string is interned.
Input:
  STRING    function or variable name
  NOERR	    If nil, then call error. if Non-nil then print message if
	    STRING was not interned.
  BODY."
  (`
   (if (intern-soft (, string))
       (progn
         (setq (, string) (intern-soft (, string)))
         (,@ body)
         )
     (if (, noerr)
         (message "No symbol in obarray: %s" (, string))
       (error "No symbol in obarray: %s" (, string))
       ))))


;;; ----------------------------------------------------------------------
;;;
(put 'tili-record-macro 'lisp-indent-function 1)
(defmacro tili-record-macro  (flag &rest body)
  "If FLAG is non-nil execute BODY in record buffer."
  (`
   (if (, flag)
       (job-buffer (use-tmp-buffer tili-:buffer-record)
         (PMAX)
         (,@ body)
         ))))



;;; ----------------------------------------------------------------------
;;;
(put 'tili-defun-macro 'lisp-indent-function 0)
(defmacro tili-defun-macro (&rest body)
  "(&rest body) Determine sexp bounds and execute BODY.
Uses `end-of-defun' `forward-sexp' to determine sexp.

Bound variables in macro:
'beg' 'end'	sexp bounds.
'str'		full line read from 'beg' point
'buffer'	points to the current buffer

You use this macro to bounds of lisp defun,defvar,defconst
structures."
  (`
   (let* ((buffer  (current-buffer))
	  str
	  beg end
	  )
     (if (null buffer) (setq buffer nil)) ;No-op, byteComp silencer

     (save-excursion
      (end-of-defun)
      (setq end (point))
      (forward-sexp -1)

      ;;  If no used, ByteComp nags -- silence it so that this macro
      ;;  can be used
      (setq beg (point))
      (if (null beg) (setq beg nil))

      (setq str (ti::b-read-line 70))
      (goto-char end)
      (,@ body)
      ))))


;;; ----------------------------------------------------------------------
;;;
(put 'tili-defcustom-macro 'lisp-indent-function 0)
(defmacro tili-defcustom-macro (&rest body)
  "Activate advice 'tili' for `defconst' _only_ during BODY."
  (`
   (unwind-protect
       (progn
	 (ad-enable-advice 'defconst 'around 'tili)
	 (ad-activate 'defconst)

	 (,@ body)
	 )
     ;;  Make sure this is always executed.
     ;;
     (tili-emergency)
     )))



;;}}}

;;; ########################################################### &Funcs ###


;;{{{ Install

;;; ......................................................... &install ...

;;; ----------------------------------------------------------------------
;;;
(defun tili-menu  (&optional arg)
  "Show echo area menu and pass ARG to `timu-menu'."
  (interactive "P")
  (timu-menu 'tili-:menu-main arg))

;;}}}
;;{{{ advice

;;; ----------------------------------------------------------------------
;;;
(defadvice byte-compile-file  (around tili act)
  "Change interactive prompt and offer current buffer for compiling(.el)."
  ;;
  ;; byte-compile-file (filename &optional load)
  (interactive
   (list
    (read-file-name
     (if current-prefix-arg
	 "(tili) Byte compile and load file: "
       "(tili) byte compile file: ")

     (if (and buffer-file-name
	      (string-match "\\.el$" buffer-file-name)
	      )
	 buffer-file-name
       (file-name-directory (or (buffer-file-name)
				default-directory))
       ))
    current-prefix-arg
    ))
  ad-do-it
  )

;;; ----------------------------------------------------------------------
;;;
(defadvice defconst (around tili (sym val doc &rest args) dis)
  "This advice is only used in TinyLisp and elswhere unactivated.
It ignores any extra arguments passed to defconst. In order to
evaluate following statement

  (defcustom my nil \"docs\" :type 'string :group my)

TinyLisp first converts it to

  (defconst my nil \"docs\" :type 'string :group my)

And turns on this advice to ignore additional :type and :group arguments.
This all is needed, because defcustom defines the variable as defvar
and it cannot be re-evaluated/reset without this trick.

After the eval has been done, this advice is turned off.
If you see this message when calling following, there is bug in TinyLisp.

  (describe-function 'defconst)"
  (ad-with-originals (defconst)
    ;;  advice prior 19.36 will not work properly with special forms
    ;;  like defconst. Hans explained is as follows to me:
    ;;
    ;;  | >   (ad-with-originals (defconst)
    ;;  | >     (defconst sym val doc)    ;; Nothing happens?
    ;;
    ;; The reason nothing happens here, is that 'sym' does not get evaluated
    ;; (since 'defconst' is a special form), instead it actually assigns the
    ;; value to the constant with the name "sym".  What you would need to do
    ;; is use `eval', e.g.,
    ;;
    ;;    (ad-with-originals (defconst)
    ;;      (eval `(defconst ,sym ,val ,doc)))
    ;;
    ;; Hans Chalupsky <hans@ISI.EDU>
    ;;
    (eval (` (defconst (, sym) (, val) (, doc))))
    ))

;;}}}
;;{{{ misc

;;; ............................................................ &misc ...

;;; ----------------------------------------------------------------------
;;;
(defun tili-highlight-default ()
  "Highlight functions and variables, see tili-*[func,var}*-hook."
  (when (face-supported-p)		;does it make sense to show colors?
    (save-excursion (tima-re-search-forward "defmacro" 0 'highlight))
    (save-excursion (tima-re-search-forward "defsubst" 0 'bold))
    (save-excursion (tima-re-search-forward "defconst" 0 'highlight))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tili-show-register-message  (&optional msg)
  "Show what to do with register and show optional MSG."
  (message
   (or msg
	(substitute-command-keys
	 (format
	  (concat
	   "Jump back to previous positon with "
	   "\\[jump-to-register-compatibility-binding] %s"
	   )
	  (char-to-string tili-:register)
	  )))
   ))

;;; ----------------------------------------------------------------------
;;;
(defun tili-symbol-type  (symbol &optional noerr)
  "Return 'var or 'func according to SYMBOL.
If NOERR is non-nil, do not call error if symbol type isn't known.
That usually means that symbol is not yet defined to obarray."
  (cond
   ((and (fboundp symbol)
	 (boundp symbol))
    (if (y-or-n-p (format "select %s: Y = variable, N = Function "
			  (symbol-name symbol)))
	'var 'func))

   ((fboundp symbol) 'func)
   ((boundp symbol)  'var)
   (t
    (unless noerr
      (error "Don't know symbol type; not a variable or function %s" symbol)
      ))))


;;; ----------------------------------------------------------------------
;;;
(defun tili-backward-opening-paren  ()
  "Go backward until paren found."
  (if (char= ?\( (following-char))
      (point)
    (re-search-backward "(" nil t)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tili-read-symbol-around-point  ()
  "Read function name around point.
o  Check if cursor is at the beginning of line whitespace
   and sees ' +(', then valuate next statement
o  Go backward to opening paren and evaluate command.

Return:
 (point function-name-string statement)"
  (let* ((opoint   (point))
	 (word     (save-excursion (tili-read-word)))
         point
         func
         statement
         )
    (save-excursion
      (cond
       ((and (stringp word) (intern-soft word))
	(skip-chars-backward "^ \t")
	)

       ((eolp)                          ;move to opening paren in this line
        (re-search-backward "(" (BOLP) t))

       (t
        ;;   if there is whitespace  '^      (autoload 'tili-mode...'
        ;;   Then go to first opening paren in the line.
        ;;
        ;;   - there must be whitespace between bol and opoint
        ;;   - next we must see '(' in the current line (eol)
        ;;
        (beginning-of-line)
        (if (not (and (re-search-forward "^[ \t]*" opoint t)
                      (re-search-forward "(" (EOLP) t)))
            (goto-char opoint))         ;restore
        ))

      (when (and (tili-backward-opening-paren)
                 (setq point (point))
                 (re-search-forward "[^ \t\n(]" nil t))

        (setq func (or word (tili-read-word)))
        (goto-char point)
        (safe-do			;In comment; this breaks.
	  (forward-sexp 1)
	  (setq statement (buffer-substring point (point)))
	  )
	(if statement
	    (list point func statement))
        ))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tili-find-package-prefix  ()
  "Read function from the beginning of file and first word from the name.

    (defun XXX-do-it-like-this ()

Return:
  string    ,The XXX
  nil       ,can't find one."
  (let* (prefix
         )
    (save-excursion
      (PMIN)
      (if (re-search-forward "^(defun[ \t]+\\([^ \t]+-\\)" nil t)
          (setq prefix (read-match 1))))
    prefix
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tili-eval (str1 str2 type &optional arg1 arg2 arg3)
  "Substiture STR1 with STR2 in string and eval all in temporary buffer..

If TYPE is nil
  Read string from buffer ARG1, position ARG2 and ARG3.

If TYPE is non-nil
  ARG1 contains string

References:
 `tili-:buffer-eval'"
  (job-buffer (use-tmp-buffer tili-:buffer-eval 'clear)
    (if type
        (insert arg1)
      (if (not (get-buffer arg1))
          (error "arg1 must be (existing) buffer")
        (insert-buffer-substring arg1 arg2 arg3)
        ))

    (PMIN)
    (replace-string str1 str2)
    (tili-eval-current-buffer)
;;;    (erase-buffer)			;May be big
    ))

;;}}}
;;{{{ Internally used buffers

;;; --------------------------------------------------------- &buffers ---
;;;
(defun tili-b-display (buffer point-min)
  "Display BUFFER(must be string) if it exists and go to optional POINT-MIN.
Shrink and print message if not exist."
  (let* ((win   (get-buffer-window buffer))
	 (owin  (selected-window))
	 )

    (if (killed-buffer-p buffer)
	(message "tili: Buffer does not exist, %s" buffer)

      ;;  Do nothing special if window is already visible
      ;;
      (if win
	  (when point-min
	    (select-window win) (PMIN)
	    (select-window owin)
	    )
	(display-buffer buffer)
	(job-buffer buffer
	  (shrink-window-if-larger-than-buffer)
	  (if point-min (PMIN))
	  )))
    ))


;;; ----------------------------------------------------------------------
;;; (defun tili-b-eval (&optional pmin)
;;;     (interactive) (tili-b-display tili-:buffer-eval pmin))
;;;
;;; This is just byteComp forward declaration
(defun tili-b-record (&rest args)
  "Ignore ARGS."
  nil)

;;; Real functions are defined here.
;;;
(mapcar
 (function
  (lambda (x)
    (let ((sym (intern (format "tili-b-%s" x)))
	  (var (intern (format "tili-:buffer-%s" x)))
          def
          )
      (setq def
            (` (defun (, sym) (&optional pmin)
                 (interactive "P")
                 (tili-b-display (, var) pmin))))
      (eval def)
      )))
 '("eval" "record" "variables" "functions"))


;;; ----------------------------------------------------------------------
;;;
(defun tili-b-record-empty   (&optional verb)
  "Empty buffer `tili-:buffer-record'. VERB."
  (interactive)
  (VERB)
  (if (not (killed-buffer-p tili-:buffer-record))
      (clear-buffer tili-:buffer-record))
  (if verb (message "tili: record buffer emptied."))
  )

;;}}}

;;{{{ elp

;;; ................................................... &elp-functions ...

;;; ----------------------------------------------------------------------
;;;
(defun tili-elp-function-list-partial  (&optional arg verb)
  "Call `tili-elp-function-list'. See ARG and VERB parameters there."
  (interactive "P")
  (VERB)
  (tili-elp-function-list arg elp-function-list verb)
  )

;;; ----------------------------------------------------------------------
;;;
(defun tili-elp-function-list  (arg &optional list verb)
  "Print list of functions that are currently beeing elp'd.
If functions can fit in echo area they are printed there unless
prefix ARG is given.

LIST defults to `elp-all-instrumented-list`. VERB."
  (interactive "P")
  (let* (str
         elt
         )
    (VERB)
    (setq list (or list elp-all-instrumented-list)
          str  (if list (prin1-to-string list)))

    (if (null list)
        (progn
          (if verb (message "No functions elp'd"))
          nil                           ;function return code
          )
      (if (and (null arg)
               (< (length str) 80))
          (message str)
        (job-buffer (use-tmp-buffer tili-:buffer-elp 'clear)
          (list-loop list elt
            (insert (symbol-name elt) "\n"))
          (sort-lines nil (point-min) (point-max))
          (pop-to-buffer (current-buffer))
          (message "tili: %d functions have been elp'd"
                   (length elp-all-instrumented-list)
                   )
          ))
      t
      )))


;;; ----------------------------------------------------------------------
;;;
(defun tili-elp-reset-after-results  (&optional arg)
  "Toggle variable `elp-reset-after-results' accorging to ARG."
  (interactive "P")
  (var-toggle elp-reset-after-results))


;;; ----------------------------------------------------------------------
;;;
(defun tili-elp-restore-all  (&optional verb)
  "Removse all instrumented functions. VERB."
  (interactive)
  (VERB)
  (elp-restore-all)
  (if verb (message "ELP: all functions restored.")))

;;; ----------------------------------------------------------------------
;;;
(defun tili-elp-reset-list  (&optional verb)
  "Reset timing list. VERB."
  (interactive)
  (VERB)
  (elp-reset-all)
  (if (get-buffer-window elp-results-buffer)
      (tili-elp-results))               ;Clear the window
  (if verb (message "ELP: Timing list cleared."))
  )

;;; ----------------------------------------------------------------------
;;;
(defun tili-elp-results  (&optional record string verb)
  "Show results, but do not change window.
if RECORD is non-nil; then copy timing to record buffer.
Insert STRING after the record stamp. VERB."
  (interactive)
  (let ((obuffer (current-buffer))
        )
    (VERB)
    (elp-results)
    (PMIN)
    (tili-elp-summary-mode 1)
    (tili-record-macro record
      (insert "\nELP: "  (ti::d-standard-date) " " (buffer-name)
	      (if string string "\n")
	      )
      (insert-buffer elp-results-buffer)
      (if verb (message "Results RECORDED."))
      )
    (pop-to-buffer obuffer)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tili-elp-instrument-buffer-i-args  (pfx-arg iact)
  "Ask args for `tili-elp-instrument-buffer'.
PFX-ARG is usually `current-prefix-arg' if you know that alreadfy.
IACT signifies interactive spec."
  (let* (pfx
         )
    (setq pfx
          (read-from-minibuffer
           (format
            "%sInstrument using package prefix[empty=examine functions]: "
            (if pfx-arg "Un)" ""))

           (or (tili-find-package-prefix) "")))

    (if (var-nil-p pfx)
        (list nil pfx-arg 'find iact)
      (list pfx pfx-arg nil iact)
      )))

;;; ----------------------------------------------------------------------
;;;
(defun tili-elp-instrument-buffer  (prefix &optional remove type verb)
  "Instrument all functions in the current buffer.

There are two possibilities when you run this in the buffer
o  Buffer contains elisp package; where each function is prefixed
   with some unique identifier.
o  You're in scratch like buffer or badly formed package where
   there may be any function names.

Interactive call note:
   The choice how to instrument functions is asked.

Input:
   PREFIX   ,can be nil if type is non-nil.
   REMOVE   ,uninstrument functions. (Interactive call's prefix arg)
   TYPE     ,if nil then instrument using PREFIX
            ,if non-nil, force finding all function names and
             instrument them
   VERB     ,verbose mode"
  (interactive
   (tili-elp-instrument-buffer-i-args current-prefix-arg 'iact))

  (let* ((str (if remove "un" ""))
         list
         func
         count
         )
    (VERB)
    (cond
     (type
      (if (null (setq list (tili-find-func-list 'no-show 'alternative)))
          (if verb (message "tili: Can't find function list"))
        (setq count (length list))

        (list-loop list func
          (setq func (cdr func))        ;("defun" . "t1")
          (tili-symbol-do-macro func nil
            (elp-restore-function func)        ;do this first
            (if (null remove)
                (elp-instrument-function func))
            ))
        (if verb (message "tili: %sinstrumented %d functions" str count))
        ))
     (t
      (if remove
          (elp-restore-all)
        (elp-instrument-package prefix))
      (if verb (message
                "%sinstrumented package '%s'. Function count unknown."
                str prefix))

      ))

    ))



;;; ----------------------------------------------------------------------
;;;
(defun tili-elp-instrument-function  ()
  "Instrument current function. Search the function name."
  (interactive)
  (let* ((func (ti::m-find-function-name))
         )
    (if (not func)
	(message "Cna't find function name.")
      ;;  This evaluates the function prior elp'ing it.
      (tili-eval-at-point)
      (tili-symbol-do-macro func nil
        (elp-restore-function func)        ;do this first
        (elp-instrument-function func)
        )
      (message (format "elp instrumented [%s]" func))
      )))


;;; ----------------------------------------------------------------------
;;;
(defun tili-elp-restore-buffer ()
  "Read functions from the buffer and cancel elp for them."
  (interactive)
  (let* ((args
          (tili-elp-instrument-buffer-i-args 'pfx 'iact))
         )
    (tili-elp-instrument-buffer
     (nth 0 args) (nth 1 args) (nth 2 args) (nth 3 args))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tili-elp-restore-function  ()
  "Remove elp code from current function. Search the function name."
  (interactive)
  (let* ((func (ti::m-find-function-name))
         )
    (if (not func)
        (message "Sorry, can't find function name.")
      (tili-symbol-do-macro func nil
        (elp-restore-function func))

      (message (format "elp restored [%s]" func))
      )))



;;; ----------------------------------------------------------------------
;;;
(defun tili-reparse-instrumentation  (&optional verb)
  "Uninstrument all currently instrumented functions.
Then eval current buffer (to get new function definitions) and last instrument
all found functions in the buffer.

In short: remove previous instrumentation and do new one. VERB."
  (interactive)
  (let* (
         )
    (VERB)
    (tili-elp-restore-all)
    (tili-eval-current-buffer)
    (tili-elp-instrument-buffer nil nil 'find verb)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tili-elp-harness (&optional count verb)
  "Ccall elp multiple times to get reliable results.
Default is call count is 3,but you can supply numeric prefix COUNT. VERB.

   ** You must have instrumented the functions before you call this function

This is abit exotic function and it requires that you have written
following test setup in the clear lisp buffer. Let's say were're
interested if 'let*' is slowers that 'let'.

            (defun t-1 () (let* () ))
            (defun t-2 () (let  () ))
            (defun t-3 () )

            [* point/cursor is before this statement]
            ;; The trick here is that when you instument whole
            ;; buffer and eval all the functions with '$ -' ,
            ;; the when forms are bypassed
            ;;
            ;; When you have Evaled/instrumented buffer, then chage
            ;; it to 'when t' and call the harness function.
            ;;
            ;; The variable HARNESS is set to t when you cann this
            ;; function and set to nil when this function finishes.
            ;;
            (when HARNESS
              (loop-for 1 500           ; in tinylibm.el, run 500 times
                (t-1)
                (t-2)
                (t-3)))

This function evals everything from current point forward ARG times.
If there is word HARNESS in the buffer, the current point is not
used but the eval is stanrted from the beginning of that line forward.

After each eval round it records the elp result to `tili-:buffer-record'.
In the above setup, this means that we repeat the test setup 3 times
to get 3 elp timing results. Since using elp only once for small functions,
doesn't give reliable results; we have to repeat the test at least 3 times.

The `tili-:buffer-record' buffer is displayed after the harness run is over."
  (interactive "P")
  (let* (case-fold-search
         beg
         h-found
         rounds
         )
    (VERB)
    (setq count  (or count 3)
          rounds count
          )

    ;;  See if there this word in the buffer
    ;;
    (save-excursion
      (PMIN)
      (when (re-search-forward "HARNESS" nil t) ;case sensitive, see let
        (setq beg (BOLPP)  h-found t)))

    (or beg                             ;we already found it
        (setq beg (point)))             ;nope, use current point

    (if (null elp-all-instrumented-list)
        (error "No functions in elp list"))

    (if (and verb
             (null
              (y-or-n-p
               (format
                (if h-found
                    "HARNESS %s times, ok? "
                  "Harness %s times, from current point forward, ok? "
                  )
                count
                ))))
        (error "Abort."))

    (if (and verb
             (y-or-n-p "Do you want to clear RECORD buffer first? "))
        (tili-b-record-empty))

    (unwind-protect                     ;make sure HARNESS is set to nil
        (progn
          (setq HARNESS t)
          (loop-for 0 count
            (tili-elp-reset-list)               ;wipe timings
            (if verb (message "tili: Eval round %d/%d ... "
                              (1+ COUNTER) rounds))
            (eval-region beg (point-max))
            (tili-elp-results
	     'record (format " -- %d/%d\n" (1+ COUNTER) rounds))
            )
          (if verb (message "tili: Eval rounds done."))
          (tili-b-record 'pmin)
          )
      (setq HARNESS nil)
      )

    ))

;;}}}
;;{{{ elp results

;;; ..................................................... &elp-results ...



;;; ----------------------------------------------------------------------
;;;
(mapcar
 (function
  (lambda (x)
    (let ((sym (intern (format "tili-elp-summary-sort-column-%d" x)))
          def
          )
      (setq def
            (` (defun (, sym) ()
                 (interactive)
                 (tili-elp-summary-sort-column (, x) ))))
      (eval def)
      )))
 '(1 2 3 4 5 6 7 8 9))

;;; ----------------------------------------------------------------------
;;;
(defun tili-elp-summary-sort-column  (nbr)
  "Sort cloumn NBR."
  (interactive "p")

  (untabify (point-min) (point-max))
  (save-with-marker
    (PMIN)
    (forward-line 2)			;Skip header.
    (cond
     ((memq nbr '(2 3 4))
      (sort-numeric-fields nbr (point) (point-max)))
     (t
      (sort-fields nbr (point) (point-max))
      ))
    ))


;;}}}

;;{{{ code help: find-error

;;; ...................................................... &find-error ...




;;; ----------------------------------------------------------------------
;;; jari's simple solution
;;;
(defun tili-error-find-2  ()
  "Start from point min and evel region at time until error occurs."
  (interactive)
  (let* ((p         -1)
         (opoint    (point))
         last-p
         )
    (PMIN) (setq last-p (point))

    (while (not (eq p (point)))
      (setq p (point))
      (eval-region last-p (point))
      (setq last-p (point))  (end-of-defun)
      )
    ;; The while loop never finishes if there was error
    (message "No lisp errors found.")
    (goto-char opoint)
    ))

;;; ----------------------------------------------------------------------
;;;  Originally, 24 Aug 1995, gnu.emacs.sources,
;;;
;;;  Mikael Djurfeldt <mdj@sanscalc.nada.kth.se>
;;;  All code is used with permission from Mikael
;;;
(defun tili-error-find-1 ()
  "Find lisp code error position and put point near the error."
  (interactive)
  (let ((lower-bound 1)
        )
    (setq tili-:find-error nil)
    (save-excursion
      (let (half
            (low 1)
            (high (tili-error-count-sexps))
            )
        (if tili-:find-error            ;See tili-error-count-sexps
            (setq lower-bound (point))
          (setq high (1+ high))
          (while (< low high)
            (if (tili-error-try-parse lower-bound
                              (tili-error-sexp-position
                               (setq half (/ (+ low high) 2))))
                (progn (setq low (1+ half))
                       (forward-sexp 2)
                       (backward-sexp)
                       (while (not (bolp))
                         (backward-sexp))
                       (setq lower-bound (point)))
              (setq high half)))
          (backward-sexp)
          (setq lower-bound (point)))))

    (if (not tili-:find-error)
        (message "No errors found.")
      (goto-char lower-bound)
      (message "%s" tili-:find-error))
    ))


;;; ----------------------------------------------------------------------
;;; Mikael Djurfeldt <mdj@sanscalc.nada.kth.se>
;;;
(defun tili-error-try-parse (from to)
  "Eval regions and try to find error in FROM TO."
  (condition-case err
      (progn (eval-region from to) t)
    (error (progn (setq tili-:find-error err) nil))))

;;; ----------------------------------------------------------------------
;;; Mikael Djurfeldt <mdj@sanscalc.nada.kth.se>
;;;
(defun tili-error-count-sexps ()
  "Eval regions and try to find error."
  (goto-char (point-max))
  (condition-case err
      (let ((n 0))
        (while (not (bobp))
          (backward-sexp)
          (setq n (1+ n)))
        n)
    (error (setq tili-:find-error err))))


;;; ----------------------------------------------------------------------
;;; Mikael Djurfeldt <mdj@sanscalc.nada.kth.se>
;;;
(defun tili-error-sexp-position (n)
  "Find sexp N."
  (goto-char 1)
  (forward-sexp n)
  (if (or (not (eobp))
          (save-excursion
            (goto-char 1)
            (forward-sexp (1- n))
            (skip-chars-forward " \t\n")
            (not (eobp))))
      (backward-sexp))
  (point))


;;}}}
;;{{{ code help: jump, eval

;;; ............................................................ &eval ...


;;; ----------------------------------------------------------------------
;;;
(defun tili-back-to-definition ()
  "Jump back to last call chain point in `tili-:call-chain'."
  (interactive)
  (tili-push-call-chain 'pop)
  (message "tili: Call chain %d" (length tili-:call-chain))
  )


;;; ----------------------------------------------------------------------
;;;
(defun tili-jump-to-definition-chain  (&optional verb)
  "Save position to call chain and jump to definition.
See `tili-jump-to-definition'."
  (interactive)
  (VERB)
  (tili-jump-to-definition '(4) verb)
  )

;;; ----------------------------------------------------------------------
;;;
(defun tili-jump-to-definition  (&optional save verb)
  "Search function or variable definition in the same file or from outside.
This function relies on the `load-history' and if there is no such
symbol, this fucntion can't jump to definition. If you have evaled buffer
or function inside emacs and not used the load* commands then the
definition information is not in `load-history'.

if optinal prefix arg SAVE is

 \\[universal-argument]     then save the point so
         that you can build call-chain and use \\[tili-back-to-definition] to
         return this point.

 non-nil then clear the call chain, save point, and jump to definition.
         This lets you start building call chain again.

VERB allows displaying verbose messages.


References:
  `tili-:call-chain'"
  (interactive "P")
  (let* ((word (tili-read-word))
	 (f-re "^(\\(defun\\|defmacro\\|defsubst\\)[ \t]+%s[ \t\n]")
	 (v-re "^(\\(defvar\\|defconst\\|defcustom\\)[ \t]+%s[ \t\n]")
	 (reg  tili-:register)
	 (call-chain-data  (point-marker))
	 re
	 type
	 point
	 file
	 sym
         )
    (VERB)
    (if (var-nil-p word)
	(message "Can't read word on current point.")
      (if (setq sym (intern-soft word))
	  (setq type (tili-symbol-type sym 'noerr)))

      (save-excursion
	(PMIN)
	(cond
	 ((eq type 'func)
	  (setq re (format f-re word)))

	 ((eq type 'var)
	  (setq re (format v-re word)))

	 (t
	  ;;  since the symbol is not defined in emacs we can't
	  ;;  know which one to search, variable or function.
	  ;;  Try anything.
	  (setq re (concat (format f-re word) "\\|" (format v-re word)))
	  ))

	(when (re-search-forward re nil t)
	  (setq point (BOLPP)))
	)

      (cond
       ((and (null point)
	     (null (intern-soft word)))
	(message "tili: Can't find definition for %s (undef)" word)
	)

       ((null point)
	;; Can't find from this file, does load history entry say
	;; from which file was loaded ?
	;;
	(if (null (setq file (tili-where-is-symbol (intern-soft word))))
	    (message "tili: Can't find load-history definition for %s" word)
	  (when
	      (or (find-buffer-visiting file) ;Already loaded
		  (y-or-n-p		;Confirm loading first
		   (format "Jump to external file: %s ?" file)))
	    (pop-to-buffer (find-file-noselect file))
	    (PMIN)
	    (unless (re-search-forward re nil t)
	      (message "(tili)Strange... %s was supposed to be here? " word)
	      (sit-for 2))

	    (when save
	      (if (and save (not (equal save '(4))))
		  (setq tili-:call-chain nil)
		(tili-push-call-chain nil call-chain-data verb)
		(message "tili: Call chain %d" (length tili-:call-chain))
		))

	    ))
	)
       (t

	(when save
	  (if (and save (not (equal save '(4))))
	      (setq tili-:call-chain nil)
	    (tili-push-call-chain nil call-chain-data verb)
	    (message "tili: Call chain %d" (length tili-:call-chain))
	    ))

	(point-to-register reg)
	(goto-char point)
	(when (null type)
	  (message "tili: Warning, this symbol is not in obarray.")
	  (sit-for 1))
	(tili-show-register-message)
	))
      )
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tili-backward-user-option  ()
  "See `tili-forward-user-option'."
  (interactive)
  (tili-forward-user-option 'back (interactive-p)))


;;; ----------------------------------------------------------------------
;;;
(defun tili-forward-user-option  (&optional back verb)
  "Search forward or BACK a user variable or user callable function. VERB.
Note: we suppose that variables are _always_ defined at the beginning
of file and then comes the functions. This means that you should
write files like this

   defvar1     * point here
   defun1

   defvar2
   defun2

Because next search will go from (*)point to defvar2 and ignore possible
user function defun1."
  (interactive)
  (let* ((re-func   (if back 're-search-backward    're-search-forward))
	 (opoint    (point))
	 type sym
	 point
	 beg end
         )
    (VERB)
    (while (and (null point)
		(or
		 (and (funcall re-func tili-:regexp-variable nil t)
		      (prog1 (setq type (read-match 1))
			(setq sym (intern-soft (read-match 2)))))

		 (and (funcall re-func "^(defun[ \t]+\\([^ \t]+\\)" nil t)
		      (prog1 (setq type "defun")
			(setq sym (intern-soft (read-match 1)))))
		 ))
      ;; ..................................................... examine ...
      ;; Okay we're somewhere at the beginning of variable of
      ;;
      (cond
       ((looking-at "defcustom")	;Yes, this is user variable
	(setq point (point)))

       ((and sym			;Is this variable defined?
	     (or
	      (and (not (string-match "defun" type))
		   (boundp sym)		;Then check is easy
		   (user-variable-p sym))
	      (and (string-match "defun" type)
		   (fboundp sym)
		   (commandp sym))
	      ))
	(setq point (point))
	)

       (t
        ;; ................................................ not loaded ...
	;; package is not loaded into memory, we may be looking at
	;; varible or function. Determine var/func region first.
	;;
	(setq beg (point))
	(setq end (save-excursion
		    (beginning-of-line)
		    (forward-sexp 1) (point)))
	(beginning-of-line)

;;;	(d! (point) end (looking-at "^(defun"))

	;;  This fails only if variable docs at flushed left, but
	;;  then you don't follow guidelines...
	;;
	;;  (defvar nil
	;;  "*docs"
	;;
	(if (if (looking-at "^(defun")
		(re-search-forward "(interactive[) ]" end t)
	      (re-search-forward "^[ \t]+\"\\*" end t)) ;It's variable
	    (setq point beg))
	))

      ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. . continue  ..
      (if beg
	  (goto-char beg))

      ;;  So that next search starts.
      ;;  We put cursot over 2nd character, because from there we
      ;;  can continue either forward or backward
      ;;
      ;;  (defvar
      ;;    *
      ;;
      ;;  If the point were on the begining of line of in the beginning
      ;;  of word, the search wouldn't start right. Search forward would
      ;;  then already match!
      ;;
      ;;  (defvar
      ;;  *
      ;;
      (beginning-of-line) (forward-char 2)

      (setq sym nil  type nil  beg nil  end nil) ;Reset round
      )

    (unless point
      (goto-char opoint)
      (if verb (message "tili: no more user variables or functions."))
      )

    point
    ))

;;; ----------------------------------------------------------------------
;;;
;;;  You can also do this in program code like this.
;;;
;;;  (fset 'test
;;;    (byte-compile-sexp
;;;      (lambda () nil)
;;;      ))
;;;
(defun tili-byte-compile-sexp  (&optional disassemble verb)
  "Byte compile function around point.
If you give prefig argument DISASSEMBLE, then the function is also
disassembled to byte code format. VERB."
  (interactive "P")
  (let* ((debug-on-error    t)
	 name name-sym
         )
    (VERB)
    (tili-defun-macro
      (setq name (strmatget "def[a-zA-Z]+ +\\([^() \t\n\]+\\)" 1 str))
      (compile-defun)
      (if verb (message "tili: byte compiled [%s]" name))
      (sit-for 1)
      (cond
       (disassemble
	;;   did the regexp find function name? Be cautious
	;;
	(if (setq name-sym (intern-soft name))
	    (disassemble name-sym)
	  (error "Can't find funtion name to disassemble?")))
       (t
	(if verb (message "tili: byte compiled [%s]" name))
	))
      )))


;;; ----------------------------------------------------------------------
;;;
(defun tili-byte-compile-display-call-tree  ()
  "See bytecomp.el `display-call-tree'."
  (interactive)
  (let* ((byte-compile-generate-call-tree  t)
	 (file	(buffer-file-name))
         )
    (if (null file)
	(message "You're not visiting file.")
      (call-interactively 'byte-compile-file)
      (job-buffer "*Call-Tree*"
	(save-excursion
	  (PMAX)
	  (insert "\
tili remark: 'Noninteractive functions not known to be called' usually
means that the functions were declared defsubst.\n")
	  ))
      )))


;;; ----------------------------------------------------------------------
;;;
(defun tili-set-value-at-point  ()
  "Read word under point and if it's variable, ask new value for it."
  (interactive)
  (let* ((var  (tili-read-word))
         val
         )
    (if (var-nil-p var)
        (message "tili: Couldn't read variable at point")
      (tili-symbol-do-macro var 'noerr
        (if (not (boundp var))
            (message "tili: There is no %s variable" (symbol-name var))
          (setq val
                (read-from-minibuffer
                 (format "Set %s to lisp expression: " (symbol-name var))
                 (prin1-to-string (symbol-value var))
                 ))

          (setq val (read val))         ;Convert to lisp
          (set var val)
          )))))

;;; ----------------------------------------------------------------------
;;;
(defun tili-call-at-point (&optional record)
  "Call object at point.
if prefix arg RECORD is given, the content of the variable
is appended to record buffer.

o   If read object is not in obarray, do nothing.
o   If it is function; ask what to do
    Show symol-function, so that you can tell if it is bytecompiled.
    Call it, possibly interactively
o   If it's variable, eval it, possibly yielding the content."
  (interactive "P")
  (let* ((str       (tili-read-word))
         sym
         type
         )
    (if (or (var-nil-p str)
            (null (setq sym (intern-soft str))))
        (message "Can't use word to eval (void?): %s "
		 (or str "<no word read>" ))

      (if (and (fboundp sym)
               (boundp sym))
	  (if (y-or-n-p (format
			 "Which %s eval: Y = variable, N = function " str))
              (setq type 'var) (setq type 'func)))

      ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. . func type ..
      (cond
       ((or (eq type 'func)
	    (and (eq type nil)
		 (fboundp sym)))

	(unless (y-or-n-p "Y = skip , N = see symbol-function ")
	  (read-char-safe-until
	   (concat "[press]"
		   (prin1-to-string (symbol-function sym)))))

	(if (and (commandp sym)
		 (y-or-n-p (format "Call interactively '%s' " str)))
	    (call-interactively sym)
	  (setq str (function-args-p 'tili-find-func-list)))
	(cond
	 ((or (var-nil-p str)
	      (y-or-n-p
	       (format "Seems to need args %s; call anyway? " str))
	      )
	  (setq str (funcall sym))
	  (message "Returned: %s" (prin1-to-string str))
	  )))

       ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. . var type  ..
       ((or (eq type 'var)
	    (and (eq type nil)
		 (boundp sym)))
	(setq str (prin1-to-string (eval sym)))
	(tili-record-macro record
	  (insert "\n" (symbol-name sym) ":\n" str)
	  (message-quick "Content of variable recorded.")
	  )

	(if (< (length str) 73)
	    (message (format "Content: %s" str))
	  (read-from-minibuffer "[press]" str))
	))
      )
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tili-eval-at-point ()
  "Evaluate variable or function around point.

Note:

  The definition must be written like this

        (defvar , (defconst , (defun ..

  And there must be no spaces after the opening paren. The following
  statement is not recognized

        (  defvar

defcustom note:

  When we evaluate defcustom variable, we don't actually evaluate statement
  as is, but pretend that the defcustom is read like 'defconst'. this has the
  effect of setting new value for the variable. If you really want to
  evaluate variable as it stand there: as defcustom, you have to put cursor
  manually behind the definition and call \\[eval-last-sexp]. In this case
  defcustom treats the variable as `defvar' and only defcustom properties are
  touched.

  DANGER:

  When you evaluate `defcustom` variable with this function, be very carefull
  that you have written it correctly, so that you won't get thown out to
  error. If this happens, you're in BIG TROUBLE; repeat ; BE ON YOUR TOES
  and think carefully your next move.

  An error condition prevented from restoring advice that was enabled for
  special form 'defconst' during the evaluation of `defcustom' definition.
  The advice is still in effect and you should immediately disable it
  before you do anything else.

  Call \\[tili-emergency] NOW! After that things are back to normal.
  and you can continue as usual."
  (interactive)
  (let* ((debug-on-error t)		;Make sure this is on!
	 )

    (tili-defun-macro
;;;      (d! (buffer-substring beg end))

    ;;  We handle defvar as defconst so that new value takes in
    ;;  effect.
    ;;
    (cond
     ((string-match "defcustom" str)
      (tili-defcustom-macro
        (tili-eval "defcustom" "defconst" nil buffer beg end)))

     ((string-match "defvar" str)
      (tili-eval "defvar" "defconst" nil buffer beg end))

     (t
      (eval-last-sexp nil)))

    (message (concat "evaled at point: " (or str "<nothing>")))
    )))

;;; ----------------------------------------------------------------------
;;;
(defun tili-eval-current-buffer-defconst  ()
  "Eval buffer as defconst and print message."
  (interactive)
  (let* ((obuffer   (current-buffer))
         (beg       (point-min))        ;maybe narrowed?
         (end       (point-max))
         )

  (job-buffer (use-tmp-buffer tili-:buffer-tmp 'clear)
    (insert-buffer-substring obuffer beg end)
    (PMIN)
    (while (re-search-forward "^(defvar \\|^(defcustom " nil t)
      (replace-match "(defconst "))
    (tili-defcustom-macro (tili-eval-current-buffer))
    )
  (message "ok, evaled whole buffer as defconst.")
  ))


;;; ----------------------------------------------------------------------
;;;
(defun tili-eval-current-buffer-from-file  ()
  "Evaluate buffer by doing `load-file' from disk.
This effectively stored the function and variable definitions
to `load-history'.

If current buffer has no file, call `tili-eval-current-buffer'."
  (interactive)
  (cond
   ((null buffer-file-name)
    (tili-eval-current-buffer))
   (t
    (if (and (buffer-modified-p)
	     (y-or-n-p "Save before loading? "))
	(save-buffer))
    (load buffer-file-name)  ;; `load' prints message for user
    )))



;;; ----------------------------------------------------------------------
;;;
(defun tili-eval-current-buffer  ()
  "Eval buffer and print message."
  (interactive)

  ;;  This silences byte compiler whinings
  ;;
  (cond
   ((fboundp 'eval-buffer)  (sfuncall 'eval-buffer)) ;XEmacs
   ((fboundp 'eval-current-buffer) (sfuncall 'eval-current-buffer))
   )

  (message "ok, evaled whole buffer")
  )


;;; ----------------------------------------------------------------------
;;;
(defun tili-eval-reverse ()
  "Search backward for opening paren and Reverse the statement's intention.
See variable `tili-:table-reverse-eval-alist'"
  (interactive)
  (let* ((stat  (tili-read-symbol-around-point))
         (table tili-:table-reverse-eval-alist)
         func
         str1 str2
	 statement
         )
    (if (or (null stat)
            (var-nil-p (setq func (nth 1 stat))))
        (message "Can't find command around point.")

      (tili-symbol-do-macro func 'noerr
        (setq str1 (symbol-name func))
        (if (null (setq func (cdr-safe (assq func table))))
            (message "Can't find reverse command for %s" str1)
          (setq str2 (symbol-name func)
		statement (nth 2 stat))

	  ;; Do some special handling, eg. add hook may have
	  ;; additional argument 'add , remove it.
	  ;;
	  (cond
	   ((string-match "add-hook +[^ ]+ +[^ ]+\\( +[^ )]+\\))"
			  statement)
	    (setq statement (kill-match 1 "" statement))
	    ))

          (tili-eval str1 str2 'string statement)
          (message "evaled as %s" str2)
          )))
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tili-eval-and-edit ()
  "Read current line and allow editing the staement before evaling it."
  (interactive)
  (let* ((line (ti::s-sqz (read-full-line)))
         ret
         )
    (setq ret (eval (read (read-from-minibuffer "tili-Eval: " line))))
    (message "Returned: %s" (prin1-to-string ret))
    ))


;;}}}
;;{{{ code help: functions and variables

;;; ........................................................ &function ...

;;; ----------------------------------------------------------------------
;;;
(defun tili-find-func-list-occur ()
  "Run occur to find functions from whole buffer."
  (interactive)
  (occur-do-macro tili-:regexp-function nil
    (tima-re-search-forward "(defmacro")
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tili-find-func-list (&optional no-show mode)
  "Find functions from buffer (macros too).

Output line format:

  [DEF][!?]   FUNCTION-NAME  INTERACTIVE-SPEC

  The DEF can de defmacro, defun and defsubst.


  [!] If you see exclamation mark then it means that you have mixed
  defsubst and interactive function, which is very dangerous situation,
  becasue when function is inlined the (interactive-p) tests from functions
  are inlined too. Check that you really want to do inlining for
  interactive functions.

  [?]Question mark means that the function does not exist in obarray
  and the possible interactive property is unknown.

Input:

  NO-SHOW   if no-show is non-nil, then the result buffer is not shown.
  MODE	    if 'alternative then if there are no left flushed functions then
	    try finding indented ones.


return:
 '((type-string . name) (t-s . n) ..)"
  (interactive)
  (let* ((re        tili-:regexp-function)
         (buffer    tili-:buffer-functions)
         (loop      t)
         list
         type
         var
	 str
	 func
         )
    (while loop
      (setq loop nil)
      (save-excursion
	(PMIN)
	(while (re-search-forward re nil t)
	  (setq type (read-match 1)
		var  (read-match 2))
	  (if (and type var)
	      (nconcc list (cons type var)))
	  )

	(if (and (null list)
		 (eq mode 'alternative))
	    (setq loop t		;try again
		  ;;   remove anchor
		  re (substring re 1)
		  ))
	))

    (if (and list (null no-show))
	(job-buffer (use-tmp-buffer buffer 'clear)
	  (list-loop list var
	    (setq str nil)		;Clear this


	    ;;  Is it symbol? Yes; okay is there really such function?
	    ;;  Okay, read the interactive arguments the, OTW
	    ;;  it was not a function.
	    ;;
	    (if (setq func (intern-soft (cdr var)))
		(if (fboundp func)
		    (setq str (commandp func))
		  (setq func nil)))

	    (insert (format "%-8s%s%s %-40s %s\n"
			    (car var)
			    ;;  Interactive and defsubst? this is dangerous!
			    ;;
			    (if (and str
				     (string= "defsubst" (car var)))
				" !" "")
			    (if (null func) " ?" "")

			    (cdr var)
			    (or str "")
			    ))
	    )
	  (pop-to-buffer (current-buffer)) (PMIN)
	  (run-hooks 'tili-:find-func-list-hook)
	  ))
    list
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tili-find-var-list-occur ()
  "Run occur to find variables from whole buffer."
  (interactive)
  (occur-do-macro tili-:regexp-variable nil
    (tima-re-search-forward "(defconst")
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tili-find-var-list  (&optional show-type)
  "Get all defvars and defconst from current buffer.
you can eg. call this function to get all variables and update them
to your M - x xxx-submit-bug-report function's variable list.

the appearing list will vave defvar's first, then defconst.

input:
  SHOW-TYPE	if non-nil, then show `user-variable-p' and
		`defcustom' information too."
  (interactive "p")
  (let* ((re        tili-:regexp-variable)
         (buffer    tili-:buffer-variables)
	 str
	 sym
         type var
         vl				;def(v)ar   (l)ist
         cl				;def(c)onst (l)ist
	 list elt
         )
    (save-excursion
      (PMIN)
      (while (re-search-forward re nil t)
        (setq type (read-match 1)
              var  (read-match 2))
        (if (string-match "defvar\\|defcustom" type)
	    (push (cons type var) vl)
	  (push (cons type var) cl))
	))

    (if (not (or vl cl))
	(message "tili: Can't' find any variables.")
      (pop-to-buffer (use-tmp-buffer buffer 'clear))

      ;;  Preserve order with reverse
      ;;
      (setq vl (reverse vl)  cl (reverse cl))
      (setq list (list vl cl))

      (list-loop list elt		;loop both lists

	(list-loop elt var
	  (setq type (car var)
		sym  (cdr var)
		str  ";; #symbol not found"
		)

	  (tili-symbol-do-macro sym 'noerr
	    (setq str "")
	    (if (user-variable-p sym)
		(setq str "user variable"))

	    (if (string= type "defcustom")
		(setq str (concat str " defcustom")))

	    (if (not (var-nil-p str))	;Add comment prefix if not empty
		(setq str (concat ";; " str)))
	    )

	  (if (null show-type)
	      (insert (cdr var) "\n")
	    (insert (format "%-40s%s\n" (cdr var) str)))
	  )
	(insert "\n")
	(PMIN)
	(run-hooks 'tili-:find-var-list-hook)
	))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tili-narrow-to-function  ()
  "Narrow to current lisp function around point."
  (interactive)
  (let* ((re   tili-:regexp-function)
         beg
         end
         )
    (save-excursion
      (beginning-of-line)
      (if (not (looking-at re))
          (re-search-backward tili-:regexp-function))

      ;;  find first empty line
      ;;
      (re-search-backward "^[ \t]*$" nil t)
      (setq beg (point))

      (forward-sexp 1)
      (setq end (point))
      )
    (narrow-to-region beg end)
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tili-find-library-info   (&optional verb)
  "Examine load history and print libraries loaded into emacs.
The summary for each library is in folowing format:

  NN [*]xxx.el  /usr/local/emacs/lisp tinylibm tinylib
  |   |	      |			    |
  |   |	      |			    |  What it `required'
  |   |         Where it is according to `load-path' order.
  |   If star, then the load history had full path name for library
  How many symbols it defined

If VERB parameter is nil, then the buffer is not shown and no
messages are displayed.

Return
 buffer	    `tili-:buffer-functions'"
  (interactive)
  (let* ((list	    load-history)
	 (max	    (length load-history ))
	 (buffer    (use-tmp-buffer tili-:buffer-functions 'clear))
	 (i	    0)
	 dep-list
	 pkg
	 name path
         )
    (VERB)
    (job-buffer buffer
      (list-loop list pkg

	(when (stringp (setq name (car pkg)))
	  (setq path (tili-where-is-package name)))

	;;  Go to next element, these will have dependency information
	;;  ("tinycom" (require . tinylibm) byte-compile-dynamic ...
	;;			|
	;;			Get these
	;;
	(NEXTP pkg)
	(while  (conspp (car pkg))
	  (push (cdr (car pkg)) dep-list)
	  (NEXTP pkg)
	  )

	(insert
	 (format
	  "%3d %s%-15s %-35s %s\n"
	  (length pkg)
	  (if (string-match "^/" name) "*" "")
	  (file-name-nondirectory name)
	  (if path (file-name-directory path) "<no path>")
	  (mapconcat (function (lambda (x) (symbol-name x))) dep-list  " ")
	  ))

	(if verb
	    (message "tili: lib info %d/%d %s" i max name))

	(inc i)
	(setq dep-list  nil pkg nil)
	))
    (if verb
	(message "Done."))
    buffer
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tili-find-symbol-load-info  ()
  "Try to look up load history to determine from where functions was defined.
read current line from point forward.

displayed message format:

  [AD] symbol-xxx: package.el (~/elisp/mime/)
  [AD] symbol-xxx: ~/elisp/xxx.el

Description:

  the first one says that the load history entry contains only
  \"package.e\" and according to `load-path' information the package was
  found from directory ~/elisp/mime/.

  in the latter; `load-history' contained full path for the package

Note:

  The additiona 'AD' String apepears on the line of the function has
  any advice code attached to it. You may look with C - h f for
  the explation what the advice code does. BUT, this flag only tells
  if there is advice code, it does not tell whether the acvice is
  active or not (If you don't see advice mentioned in C - h f, then
  the aadvice is instrumented, but latent, and not working currently)."
  (interactive)
  (let* ((str
	  (save-excursion
	    (if (looking-at "[ \t\n]+") ;only spaces ahead?
		(read-full-line)

	      ;;  go backward until space(word) or function call
	      ;;
	      (if (not (char= (following-char) ?\( ))
		  (re-search-backward "[( \t]" nil t))
	      (buffer-substring (point) (EOLP))
	      )))
	 (sym	  (tili-get-symbol str))
	 (ad-info "")
	 package
	 path
	 msg

         )
    (if (null sym)
	(message "\
Can't find _defined_ variable or function on the line (eval buffer first).")

      (if (memq 'ad-advice-info (symbol-plist sym))
	  (setq ad-info "AD "))

      (cond
       ((or (subrp sym)
	    (string-match
	     "built-in"
	     (or (documentation-property sym 'variable-documentation)  ""))
	    )
	(setq msg "<Built-in function>"))

       ((setq package (car-safe (ti::y-lh-where sym)))
	(if (setq path (tili-where-is-package package))
	    (setq msg
		  (format "%s (%s)"
			  (file-name-nondirectory path)
			  (file-name-directory path)))
	  (setq msg "<no path found>")
	  ))
       (t
	(setq msg "no load-history entry; maybe evaled locally?")
	))
      (message "%s%s: %s" ad-info (symbol-name sym) msg)
      )))

;;}}}
;;{{{ code help: misc

;;; ----------------------------------------------------------------------
;;;
(defun tili-autoload-generate  ()
  "Generate autolad from current lisp package buffer."
  (interactive)
  (let* (
         )
    (cond
     ((null (buffer-file-name))
      (message "tili: sorry, autoloads can only be generated from file."))
     (t
      (ti::m-pkg-autoload-create-on-file
       (buffer-file-name) tili-:buffer-autoload)
      ))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tili-indent-around-point  (&optional verb )
  "Indent current lisp statement around the point. typically a function.
VERB."
  (interactive)
  (let* (;;  remove anchor ^
         (re-f  (substring tili-:regexp-function
                           1 (length tili-:regexp-function)))

         (re-v  (substring tili-:regexp-variable
                           1 (length tili-:regexp-variable)))

         beg
         end
         )

    (VERB)
    (save-excursion
      (setq beg
            (or (setq beg (re-search-backward re-f nil t))
                (setq beg (re-search-backward re-v nil t)))
            end
            (and beg
                 (progn
                   (forward-sexp 1) (forward-line 1)
                   (setq end (point))
                   )))
      (if (not (and beg end))
          (if verb (message "tili: can't find anything to indent here."))
        (lisp-indent-region beg end))

      )))


;;; ----------------------------------------------------------------------
;;;
(defun tili-defmacro-surround-word ()
  "Surround current word with (, )  defmacro statement."
  (interactive)
  (let* (
         )
    (unless (char-in-list-case (preceding-char) '(?\ ?\t ?\n))
      (backward-word 1))
    (insert "(, ")
    (forward-word 1)
    (insert ")")
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tili-macroexpand (&optional expand-function)
  "Expand macro call with EXPAND-FUNCTION which is string.
If point is sitting inside call to macro, expand it.
in the following example the cursor is at point [*].

  (macro-function-call arg1 * arg2 arg3)

references:
  `tili-:buffer-macro'
  `tili-:macroexpand-function-list'"

  (interactive
   (list
    (intern-soft
     (completing-read
      "Expand with function: "
      (list-to-assoc-menu tili-:macroexpand-function-list)
      nil nil (car tili-:macroexpand-function-list)
      ))))

  (let* ((mac-re tili-:regexp-macro-definition)
	 (opoint (point))
	 point
	 symbol
	 sym
	 to-buffer
         )
    (when (not (and (symbolp expand-function)
		    (fboundp expand-function)
		    ))
      (error "Not a function %s" expand-function))

    (tili-defun-macro
      (if (setq symbol (strmatget "[^() \t\n\]+" 0 str))
	  (setq sym (intern-soft symbol)))

      (cond
       ((and (stringp symbol)
	     ;;
	     ;; This are exeptions, they are macros
	     ;;
	     (not (string-match mac-re symbol))

	     ;;  Others are supposed to be function definitions
	     ;;
	     (string-match "^def" symbol)
	     )
	(message "grabbed %s, but it is not believed to be a macro " symbol)
	)

       ((and (stringp symbol)
	     sym
	     (defmacro-p sym)
	     )
	(setq to-buffer (use-tmp-buffer tili-:buffer-macro 'clear))
	(append-to-buffer to-buffer beg end)
	(goto-char opoint)		;restore position
	(pop-to-buffer to-buffer) (PMIN)

	(insert "(" (symbol-name expand-function) " '\n"  )
	(PMAX) (insert ")")
	(setq point (point))

	(eval-last-sexp 'output)
	(delete-region (point-min) point)
	(PMIN)
	)

       ((and (stringp symbol)
	     (fboundp sym))
	(message "macroexpand, sexp was function: %s" symbol))

       (t
	(message "macroexpand, skipped: %s" (or str "<can't read>"))
	))
      )))

;;}}}

;;{{{ properties display

;;; ........................................................ &property ...



;;; ----------------------------------------------------------------------
;;;
(defun tili-syntax-post-command ()
  "Show syntax information for current point."
  (message "[syntax info] %s: %s"
	   (char-to-string (following-char))
	   (ti::s-syntax-info (following-char))))

;;; ----------------------------------------------------------------------
;;;
(defun tili-syntax-show-mode (&optional arg verb)
  "Constantly show character syntax info, ARG behaves like mode arg. VERB."
  (interactive "p")
  (VERB)

  (var-toggle tili-:syntax-show-mode arg)

  (cond
   (tili-:syntax-show-mode
    (add-hook-local 'post-command-hook 'tili-syntax-post-command)
    )
   (t
    (remove-hook    'post-command-hook 'tili-syntax-post-command)
    ))

  (if verb
      (message
       "tili: syntax show mode is %s"
	(if tili-:syntax-show-mode "on" "off")
	))
  )


;;; ----------------------------------------------------------------------
;;;
(defun tili-property-info  (&optional arg)
  "See `tili-property-show' and ARG. Return string 'face-info ov-info'."
  (let ((count      0)
        (face-str   "")
        (ov-str     "")
        ovl
        elt
        )

    (if (member arg '(1 (16) (64)))
        (setq face-str
              (format
               "%s"
               (prin1-to-string (text-properties-at (point))))))

    (when (member arg '((4) (16) (64)))
      (setq ovl (ti::xe-overlays-at (point)))
      (list-loop ovl elt
	(inc count)
	(setq ov-str
	      (format
	       "%sov%d%s "
	       ov-str
	       count
	       (prin1-to-string (ti::xe-overlay-properties elt))
	       ))))
    (concat face-str " " ov-str)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tili-property-post-command ()
  "Display property info according to `tili-:property-show-mode'.
This is post command."
  (let* ((record (equal '(64) tili-:property-show-mode))
         (ch     (char-to-string (following-char)))
         str
         )

    (setq str
          (format
           "%s:%s"
           (point)
           (tili-property-info tili-:property-show-mode)
           ))
    (tili-record-macro record (insert ch str "\n"))
    (message "%s%s" (if record "r" "") str)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tili-property-show-mode  (arg &optional verb)
  "Toggle permanent text property info mode with ARG. VERB.
Utilizes local `post-command-hook'.

The echo-area will show following message; definition first, then example.
It says that there is one face and two overlays in current position 12.
The little 'r' appeard at the beginning if the record mode is selected.

   [r]POINT:FACE-PROPERTIES[ovN:OVERLAY-PROPERTIES]
   12:(face highlight)ov1(face region)ov2(owner my)

Input ARG:
   nil			toggle between 0 and '(16)
   0			off
   1			show face properties
   '(4)	 C -u		show overlay properties.
   '(16) C -u C -u	show both text properties and overlays.
   '(64) C -u C -u C -u show both text properties and overlays AND
                        recond info in buffer `tili-:buffer-record'.
VERB			verbose flag"
  (interactive "P")
  (VERB)

  (cond
   ((null arg)
    (if (null tili-:property-show-mode)
        (setq tili-:property-show-mode '(16))
      (setq tili-:property-show-mode nil)))
   ((member arg '((4) (16) (64)))
    (setq tili-:property-show-mode arg)
    ))

  (cond
   (tili-:property-show-mode
    (if verb
        (message-quick
         "tili: Property show mode is on %s"
         (if (equal arg '(64)) "(RECORDING)" "")
         ))
    (add-hook-local 'post-command-hook 'tili-property-post-command)
    )
   (t
    (remove-hook    'post-command-hook 'tili-property-post-command)
    (if verb (message  "tili: Property show mode is off"))
    ))
  )

;;}}}
;;{{{ Snooping

;;; ........................................................... &shoop ...


;;; ----------------------------------------------------------------------
;;;
(defun tili-snoop-variables-i-args  ()
  "Ask arguments to `tili-snoop-variables'."
  (list
   current-prefix-arg
   (nth
    1
    (assoc
     (completing-read
      "Name of variable snoop list: "
      (list-to-assoc-menu (mapcar 'car tili-:table-snoop-variables))
      nil 'match-it)
     tili-:table-snoop-variables))
   ))

;;; ----------------------------------------------------------------------
;;;
(defun tili-snoop-variables (&optional arg list)
  "Display contents of hooks. See `tili-:table-snoop-variables'.

ARG can be
   1           With prefix arg, variables values are recorded to
	       to buffer `tili-:buffer-record' and

   0           Save variables values.
   9           Restore variables values from the saved copies.
   8           Kill saved variable state
   5           Set all variables to nil in list

   C -u        allows editing the variables.

LIST	       list of variables.

Flags when viewing, editing echo-area:

+w   Is shown when your're actually modifying the contents.
!    is shown if the variable's state has been saved and is non-nil."
  (interactive (tili-snoop-variables-i-args))
  (let* ((write	    (equal arg '(4)))
	 (read	    (eq arg nil))
	 (record    (eq arg 1))
	 (save      (eq arg 0))
	 (restore   (eq arg 9))
	 (kill      (eq arg 8))
	 (reset     (eq arg 5))
	 (msg       (format
		     "(%s) %s"
		     (length list)
		     (mapconcat 'symbol-name list " ")))
	 (prop      'tili-original)
         elt
         str
         val
         ok
         )
    (list-loop list elt
      (setq ok  (boundp elt))

      (cond
       ((or read write record)

	(if ok
	    (setq val (prin1-to-string (symbol-value elt)))
	  (setq val "<variable does not exist>"))

	(tili-record-macro record
	  (insert (format "%s %s\n" (symbol-name elt)  val)))

	;; Using rsz-mini we can show whole content.
	;;
	(setq str (read-from-minibuffer
		   (format "%s%s%s: "
			   (if write "+w " "")
			   (if (get elt prop)
			       "! " "")
			   (symbol-name elt)
			   )
		   val))
	(if write                         ;replace content?
	    (set elt (read str))))

       (save
	(put elt prop (symbol-value elt)))
       (kill
	(remprop elt prop))
       (restore
	(set elt (get elt prop)))
       (reset
	(set elt nil))
       (t
	(message "Unknown arg %s" (prin1-to-string arg)))
       ))

    (cond
     (save  (message "tili: Saved %s" msg))
     (save  (message "tili: Restored %s" msg))
     (kill  (message "tili: Killed saved value copies %s" msg))
     (kill  (message "tili: Set to nil %s" msg))
     )

    ))

;;}}}
;;{{{ Occur

;;; ........................................................... &occur ...


;;; ----------------------------------------------------------------------
;;;
(defun tili-occur-i-args ()
  "Ask arg1 to `tili-occur'."
  (read-from-minibuffer
   "(tili) occur: "
   (nth 1 (tili-read-symbol-around-point))
   nil nil 'tili-:history-occur
   ))

;;; ----------------------------------------------------------------------
;;;
(defun tili-occur-verbose (arg)
  "Call `tili-occur' as user would with ARG."
  (interactive "P")
  (tili-occur (tili-occur-i-args) arg))


;;; ----------------------------------------------------------------------
;;;
(defun tili-occur  (regexp &optional arg)
  "Run occur on REGEXP for whole buffer and Highlight matched text.
ARG  non-nil says _not_ to filter 'comment only' lines out.
VERB verbose flag."
  (interactive (list (tili-occur-i-args) current-prefix-arg))

  (let* ((obuffer (current-buffer))
	 )
    (occur-do-macro regexp nil
      (tima-re-search-forward regexp)
      (PMIN)
      (unless arg
	(while (re-search-forward "^ *+[0-9]+:\\([ \t]*;.*\\)" nil t)
	  (kill-match 1))))

    ;;  Keep cursor in original buffer
    (pop-to-buffer obuffer)
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tili-occur-select-forward  (&optional back)
  "Select next line from the occur buffer. You must first run `tili-occur'.
Optionally goes BACK."
  (interactive "P")
  (let* ((buffer (get-buffer tili-:occur-buffer-name))
	 line
	 file
	 str
	 go-buffer
	 )
    (if (null buffer)
	(message "tili: No occur buffer exist.")
      (job-buffer buffer

	;; This is ugly, but I don't know other way to move
	;; point permanently in the buffer. The select-window
	;; is the crucial command to make the point move.
	;;
	(save-window-excursion
	  (pop-to-buffer (current-buffer))
	  (select-window (selected-window))
	  (if back
	      (forward-line -1)
	    (forward-line 1)))

	(setq str (read-full-line))
	(if (null (setq line (bufmatget "^\\([0-9]+\\):" 1)))
	    (message "Can't find line number from occur buffer.")
	  (setq line (string-to-int line))


	  ;;  first line in occur buffer has
	  ;;  "Lines matching "tipgpd" in buffer xxx.el"
	  ;;
	  (if (null (setq file
			  (re-search-check "^Lines matching.* \\(.*\\).$"
					   1 nil 'matched)))
	      (message "tili: Can't find file name from occur buffer.")
	      )))
      (if (and file
	       (null (setq go-buffer (get-buffer file))))
	  (message "tili: buffer not exist %s" file)
	(pop-to-buffer go-buffer)
	(goto-line line)
	(message str)
	))
    ))

;;}}}

;;{{{ Additional support functions

;;; ......................................................... &support ...


;;; ----------------------------------------------------------------------
;;;
(defun tili-emergency (&optional verb)
  "Restore any dangerously adviced functions. See `tili-eval-at-point'. VERB."
  (interactive)
  (VERB)
  (ad-disable-advice 'defconst 'around 'tili)
  (ad-activate 'defconst)

  (if verb
      (message "tili: Function states restored; you can continue as usual."))
  )



(defun tili-checkdoc  ()
  "IOnteractively check document from current point forward.
See `checkdoc-interactive'."
  (interactive)
  (checkdoc-interactive (point))
  )


;;; ----------------------------------------------------------------------
;;;
(defun tili-checkdoc-notes  ()
  "See `checkdoc'."
  (interactive)
  (checkdoc 'take-notes))

;;; ----------------------------------------------------------------------
;;;
(defun tili-checkdoc-comment-notes  ()
  "See `checkdoc'."
  (interactive)
  (checkdoc-comments t))

;;; ----------------------------------------------------------------------
;;;
(defun tili-lm-verify  ()
  "Check package layout.
The latest Emacs distribution has improved lisp-mnt.el which has
function `lm-verify', which you should run in your package. It helps
ensuring that you have all the proper keywords in place. Here is rough
vallid layout format:

    ;; @(#) XXX.el -- proper first line

    ;; Author
    ;; Maintainer
    ;; Created:
    ;; Keywords:

    ;;; Commentary:
    ;;; Change Log:
    ;;; Code:

    ;;; XXX.el ends here

See unix what(1) and GNU RCS indent(1) why you should adopt a style where
you use @(#) and $Keywords$."
  (interactive)
  (require 'lisp-mnt)
  (if (not (string= (symbol-value 'lm-history-header)
		    "Change Log\\|History"))
      (message "\
Sorry, your lisp-mnt.el is too old to have improved checking. Get newer.")
    (call-interactively 'lm-verify)
    ))

;;}}}

(add-hooks '(lisp-mode-hook
	     emacs-lisp-mode-hook
	     lisp-interaction-mode-hook
	     )
	   'turn-on-tili-mode
	   )

(add-hook' tili-:mode-define-keys-hook 'tili-mode-define-keys)

;; tili-elp-summary-install-mode
;;
(add-hook' tili-:elp-summary-mode-define-keys-hook
	   'tili-elp-summary-mode-define-keys)

(provide   'tinylisp)
(run-hooks 'tili-:load-hook)

;;; tinylisp.el ends here
