;; @(#) tinyload.el -- Load set of packages when emacs is idle

;; @(#) $Id: tinyload.el,v 1.8 1997/07/04 18:42:41 jaalto NotReleased $
;; @(#) $Keywords: extensions $
;;
;; $KnownCompatibility: 19.28 - 34+, XEmacs19.14+ $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1997 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Jul 2 1997
;;
;; To get information on this program use ident(1) or do M-x tilo-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinyload|Jari Aalto|jari.aalto@poboxes.com|
;; Queue list of packaes to load until emacs is idle. Kinda lazy load/require|
;; 04-Jul-1997|$Revision: 1.8 $|~/misc/|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ....................................................... &t-install ...
;; Put this file on your Emacs-Lisp load path, add following into your
;; ~/.emacs startup file. Rip code with tinylib.el / ti::m-pkg-rip-magic
;; Move all your `require' commands into the Lazy load list.
;;
;;* _
;;*     (setq tili-:load-list '("package" "package" ...))
;;*     (require 'tinyload)
;;
;; This package can't be autoloaded, but you can delay loading the
;; Lazy load setup until Emacs is idle. You will find the file content
;; of the 'rc-lazy from the documentation.
;;
;;* _
;;* 	(run-with-idle-timer 15 nil 'my-tilo-idle-timer)  ;; Emacs 19.34+
;;* 	(defun my-tilo-idle-timer () (require 'rc-lazy "~/.emacs.lazy"))
;;
;; If you have any questions, contact maintainer with this function
;;
;;      M-x tilo-submit-bug-report

;;}}}

;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "tilo-" in front of
;;   every function & variable. It stands for '(ti)ny lazy package (lo)ader'
;; - variable names contain letters 'tilo-:'.

;;  Preface, Jul 1997
;;
;;	While I could manage to setup my .emacs 70% to use autolods, there
;;	still were packages that couldn't be autoloaded due to their setup
;;	nature or behavior. I still needed to put `require' commands into
;;	my .emacs to have them available. It also meant that for every
;;	`require' command, the emacs startup time increased.
;;
;;	Fortunately new emacs releases have idle timers where I could use
;;	to load all the remaining packages. I could therefore put the
;;	`load' effect in some later pending time, istead of having to wait
;;	them to load at Emacs startup.
;;
;;	The benefit is that you Emacs starts up instantly, and when it is
;;	idle it loads all the remaining packages that you want to be
;;	available in your Emacs session.
;;
;;  Overview of features
;;
;;	o   Emacs and XEmacs compatible
;;	o   Lazy loading of packages (in some later time); eg after
;;	    15 seconds idle, you see message in the echo area
;;	    "Lazy load XXXX ok (75% 12/16)", where
;;	    the loaded count/length of load list is displayed.
;;	o   You no longer have to use `require' in your .emacs, instead,
;;	    you can put the package to `tilo-:load-list' and have it loaded
;;	    when Emacs is idle.
;;	o   Your .emacs starts faster when the `require' commands are out.
;;
;;  Lazy load messages in *Message* buffer
;;
;;	There are two type of messages
;;
;;	    Lazy load fdb ok (100% 18/18)
;;	    Lazy load fdb error! (100% 18/18)
;;
;;	The first one just tells that package was loaded and show some
;;	statistics. The latter "error!" means that you have defined the
;;	'noerr parameter for `load' command and the recent load of that
;;	package failed. The package did not exist.
;;	
;;	If you see this message, then there the file XX couldn't be evaled at
;;	all and you should check the syntax of that file.
;;	
;;	    Lazy load: fatal load error, XX
;;
;;  How to use Lazy loading
;;
;;	You better understand you would use this package, if I demonstrate
;;	the ussage with examples. Lets supposes your emacs startup consists of
;;	following files.
;;
;;	    .emacs		-- the main load controller
;;	    .emacs.path.el	-- settings paths/1st line mode list...
;;	    .emacs.bup.el	-- Backup settings
;;	    .emacs.set.el	-- Std Emacs variable settings
;;
;;	    .emacs.keys.el	-- Keyboard settings
;;	    .emacs.font.el	-- Fonts and Font lock; faces settings
;;	    .emacs.hooks.el	-- All add-hook commands and mode settings.
;;	    .emacs.gnus5.el	-- GNUS 5 customizations (could be .gnus too)
;;	    .emacs.std-pkg.el	-- Loading stad EMacs packages and their seutp
;;	    .emacs.pkg.el	-- Non-std distrib, additional packages
;;	    .emacs.tips.el	-- Tips(code samples) from the Usenet
;;	    .emacs.mail.el	-- mail agent; mode; settings
;;
;;	Now suppose your .emacs has loaded all those files in the
;;	following manner
;;
;;	    ;; .emacs -- Emacs startup controller
;;	    ;;
;;	    (require 'rc-path	"~/.emacs.path")
;;	    (require 'rc-bup	"~/.emacs.bup")
;;	    (require 'rc-set	"~/.emacs.set")
;;	    (load "~/.emacs.keys.el")
;;	    (require 'rc-font	"~/.emacs.font")
;;	    (load "~/.emacs.hooks")
;;	    (load "~/.emacs.gnus5")
;;	    (load "~/.emacs.std-pkg")
;;	    (load "~/.emacs.pkg")
;;	    (load "~/.emacs.tips")
;;	    (add-hook 'mail-mode-hook
;;                    '(lambda () (require 'rc-mail "~/.emacs.mail")))
;;
;;	Both `require' and `load' commnds have been used: you use the
;;	`require', if you don't expect that the file will change during
;;	your emacs session and you use `load' if you expect it to
;;	change. If you would do command
;;
;;	    M-x load-file ~/.emacs
;;
;;	Only the `load' command's files would be loaded again. All the
;;	`require' files would have been skipped, because the `rc' resource
;;	features had already been defined.
;;
;;	Okay, loading all those files, either with `require' or `load',
;;	takes too much time when you start Emacs; it time to activate the
;;	lazy loading. Your emacs looks like this after rearranging
;;
;;	    ;; .emacs -- Emacs startup controller
;;	    ;; WE want to have these minimum features immediately available
;;	    ;;
;;	    (require 'rc-path	"~/.emacs.path")
;;	    (require 'rc-bup	"~/.emacs.bup")
;;	    (require 'rc-set	"~/.emacs.set")
;;	    (load "~/.emacs.keys.el")
;;
;;	    ;;  Load this setup when I start composing mail.
;;	    ;;
;;	    (add-hook 'mail-mode-hook
;;                    '(lambda () (require 'rc-mail "~/.emacs.mail")))
;;	    ;;
;;	    ;;  But we can afford loading these later
;;	    ;;
;;	    (defconst tilo-:load-list
;;	      '(("~/.emacs.font"    'rc-font)
;;		("~/.emacs.hooks"   'rc-hooks)
;;		("~/.emacs.gnus5"   'rc-gnus5)
;;		("~/.emacs.std-pkg" 'rc-std-pkg)
;;		("~/.emacs.pkg"	    'rc-pkg)
;;		("~/.emacs.tips"    'rc-tips)
;;		("~/.emacs.mail"    'rc-mail)
;;		))
;;	    (require 'tinyload)
;;
;;	Whe you load this .emacs, only the most necessary files are loaded
;;	and your Emacs is up very fast. When the `tinyload' finds that
;;	your Emacs is idle it loads all the rest packages you defined in
;;	the `tilo-:load-list'. As simple as that.
;;
;;	  NOTE: Please pay attention to one detail above. The ~/.emacs.mail
;;	  will be loaded from Lazy load list _and_ it will be loaded
;;	  when you call M-x `mail'. Do you believe there is redundancy?
;;	  No :-) The idea is that you may call M-x `mail' way before
;;	  the Lazy load reaches that file in its load list and the hook
;;	  guarrantees that you get the setup at mail invoke time.
;;
;;	  But it may be the other way round: Lazy load has already loaded
;;	  the mail setup for you and thus invoking M-x `mail' is fast,
;;	  because there is nothing to load any more.
;;
;;	  Similar things you should do to GNUS; VM; RMAIL and others that
;;	  you call and whose setup you want to have immediately available
;;
;;	For more complex setup, refer to end of this source file, where I
;;	have attached my lazy load setup.
;;
;;  Lazy loading, require and autoload
;;
;;	Above you saw how to load your Emacs `rc' files in full. But the
;;	Lazy loading is not only suitable for those. It also helps you
;;	to load files, that can't be autoloaded.
;;
;;	  If you can arrange loading a packages with `autoload' command,
;;	  do that. Never put `require' or direct `load' command into your
;;	  Emacs `rc' file.
;;
;;	Well defined package explain in the *installtion* section two ways
;;	how to load it: here is example from tinytab.el package
;;
;;	    (require 'tinytab)
;;
;;	    or use this; your .emacs loads up a bit quicker
;;
;;	    (autoload 'tit-mode              "tinytab" t t)
;;	    (autoload 'tit-return-key-toggle "tinytab" t t)
;;
;;	The first way forces loading the file (takes time); and the latter
;;	only tells that the package's function `tit-return-key-toggle' and
;;	`tit-mode' exist. If you happen to call those functions, _only_
;;	then the package gets loaded. The big difference here is that
;;	when you put the latter in your Emacs rc file, Emacs reads it 20X
;;	faster than the `require' command.
;;
;;	It is not always possible arrange loading package with autoloads,
;;	because the package may behave so that in order to get the
;;	features installed, package must do the setup by itself: you can't
;;	do it yourself as in autoload case. Here are few packages that
;;	can't be autoloaded:
;;
;;	    tinymy	-- collection of utilities
;;	    fdb		-- Emacs Debug error filter
;;	    fa-extras	-- Filling extras
;;	    suggbind	-- Command reminder
;;
;;	When you would normally include a `require' command into your
;;	Emacs `rc' file, you can now move the package to lazy load
;;	list and keep only autolaods for other packagws in the `rc' files.
;;
;;	    ;; Old .emacs.pkg.el rc file
;;	    (autoload ....
;;	    (autoload ....
;;	    (require 'fa-extras)
;;	    (autoload ....
;;
;;	    ;; New .emacs.pkg.el rc file
;;	    (autoload ....
;;	    (autoload ....
;;	    (autoload ....
;;
;;	And the mising entry has been moved to `tilo-:load-list'. When the
;;	rc file only contains autoloads, it will be loaded much quicker.
;;
;;  Use separate Lazy load rc file
;;
;;	It may be good idea to set up Lazy load rc file that only has
;;	the load list definition and call to tinyload.el: like this
;;
;;	    ;; .emacs.lazy.el -- Lazy load definitions
;;	    ;;
;;	    ;; If you compile this file, defconsts shuts up ByteCompiler
;;	    ;;
;;	    (defconst tilo-:load-list
;;	      '(...
;;	        ...
;;	        ))
;;	    (require 'tinyload)
;;	    (provide 'rc-lazy)
;;	    ;; End of file
;;
;;	Ans then you add following call to your .emacs
;;
;;	    (require 'rc-lazy "~/.emacs.lazy")
;;
;;  Old Emacs releases
;;
;;	In Emacs 19.29 or lower, there is no way to detect if Emacs has
;;	been idle or not. In these Emacs versions a regular timer
;;	process is used to load the packeges in the load list. The
;;	timer runs in regular intervals and loads at least one package
;;	at a time: more packages are not loaded if there was input
;;	pending at the time of previous load.
;;
;;	In these Emacs releases, the *Messages* buffer that exist in later
;;	releases, is created and the load messages are put there.
;;
;;  Intermediate Emacs releases
;;
;;	In Emacs releases that have `post-command-idle-hook' (19.30-19.33),
;;	the idle process runs after each keypress. The laoding is activated
;;	only if the you keep a pause between typing.
;;
;;  New Emacs releases
;;
;;	Starting from 19.34, there is function `run-with-idle-timer' that
;;	truly runs after NNsec idle priod.
;;
;;  About implementation
;;
;;	When you define `tili-:load-list', the value of variable if saved
;;	under property 'original. When the idle timer runs, the list is
;;	read from the beginning and each package at a time is loaded.  The
;;	last unloaded package position is saved under property 'pos.
;;
;;	The situation looks like this:
;;
;;	    tili-:load-list 'original	--> (list) original contents
;;	    tili-:load-list 'pos	--> (nth nbr) last package loaded.
;;
;;	If your do something in your emacs while the list is beeing
;;	looped, or when the loader function is about to be called, that
;;	interrupts the work. Next time the timer funtions run runs:
;;
;;	o   It checks if the current list matches 'original. Yes, means that
;;	    the list hasn't been modified. No, means that it should examine
;;	    the list all aver again, starting from the beginning.
;;	o   If the list was original, it picks the 'pos point and
;;	    loads all the remaining packages.
;;	o   If there is nothing to load, the 'pos points to the end
;;	    of list. Function returns immediately and does nothing.

;;}}}
;;{{{ history

;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; Jul	4	1997	[jari]		19.28	v1.6-1.8	NotReleased
;; - Wrote complete documentations how to use this package.
;; - Tested in XEmacs19.14 and in Emacs 19.28,30,34
;;
;; Jul	3	1997	[jari]		19.28	v1.2-1.5	NotReleased
;; - Small corrections and fine tuning. Not tested in 19.34 yet.
;;
;; Jul	2	1997	[jari]		19.28	v1.1		Created


;;}}}

;;; Code:

;;{{{ require

;;; ......................................................... &require ...
(require 'tinylibm)

(eval-when-compile (use-dynamic-compilation))

(defgroup-tiny TinyLoad tilo-: extensions
  "Overview of features
	o  Lazy loading of packages (in some later time)
	o  You no longer have to use `require' in your .emacs, instead,
	   you can put the package to `tilo-:load-list' and have it loaded
	   when Emacs is idle.
	o  Your .emacs starts faster when the `require' commands are out."
  )


;;}}}
;;{{{ variables

;;; ......................................................... &v-hooks ...

(defcustom tilo-:load-hook nil
  "*Hook that is run when package is loaded."
  :type 'hook :group 'TinyLoad
  )


;;; ....................................................... &v-private ...
;;; Private variables

(defvar tilo-:timer-elt  nil
  "The timer elt process.")


;;; ........................................................ &v-public ...
;;; User configurable


(defcustom tilo-:idle-time  15
  "*When emacs is this many seconds idle, load next package."
  :type 'integer :group 'TinyLoad
  )


(defcustom tilo-:load-list  nil
  "*List of packages to load when emacs has been idle.
The idle time is on `tilo-:idle-time'.

Format:

  '((PACKAGE-OF-FILE [FEATURE-SYM] [NOERR] [NOMSG] [FORM-BEFORE] [FORM-AFTER])
     PACKAGE-OR-FILE
    )

  PACKAGE-OR-FILE can be any valid `load' command filename parameter.
  Eg. \"package\" \"package.el\" \"~/elisp/package.el\"

  You use (PACKAGE FEATURE) if the package provides different feature than
  the package name; eg. entry (\"~/.emacs.my\" 'rc-my) says; that you
  want to do (load \"~/.emacs.my\") only if (featurep 'rc-my) returns false.

  [NOERR] is optional and parameter for `load' command
  [NOMSG] is optional and parameter for `load' command

  [FORM-BEFORE] is evaluated just _before_ laod command.
  [FORM-AFTER]  is evaluated _after_ laod command.

Note:

  Nil entries in this table are skipped. This allows you to contruct
  conditional load list in the format

  (setq tilo-:load-list
    (list
     (if (= 28 emacs-minor-version)
	 (list \"~/.emacs.19-28\" 'rc-28))
     ))
  --> '(nil)   in non-19.28 Emacs


Example:

  '(\"ffap.el\"
    \"tinylibmt.el\"
    )"
  :type '(repeat sexp)  :group 'TinyLoad
  )


;;}}}
;;{{{ version

;;; ....................................................... &v-version ...


(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinyload.el"
 "tilo"
 tilo-:version-id
 "$Id: tinyload.el,v 1.8 1997/07/04 18:42:41 jaalto NotReleased $"
 '(tilo-:version-id
   tilo-:load-hook
   )))

;;}}}

;;; ########################################################### &Funcs ###

;;{{{ installation

;;; --------------------------------------------------------- &install ---
;;;
;;;###autoload
(defun tilo-install  (&optional remove)
  "Install package or REMOVE."
  (interactive "P")
  (cond
   ;; ...................................................... new Emacs ...
   ((fboundp 'run-with-idle-timer)
    ;; 19.34+ (run-with-idle-timer SECS REPEAT FUNCTION &rest ARGS)
    ;;
    (cond
     ((and (emacs-p) tilo-:timer-elt)
      (sfuncall 'cancel-timer tilo-:timer-elt))
     ((and (xemacs-p) tilo-:timer-elt)
      ;; I don't know xe equivalent for remove
      ))

    (unless remove
      (setq tilo-:timer-elt
	    (sfuncall
	     'run-with-idle-timer
	     tilo-:idle-time
	     t
	     'tilo-load-process
	     )))
    )

   ;; ..................................... intermediate Emacs release ...
   ((boundp 'post-command-idle-hook)	;19.30+
    (if remove
	(remove-hook 'post-command-idle-hook  'tilo-load-process)
      (add-hook 'post-command-idle-hook  'tilo-load-process)
      ))

   ;; ...................................................... old Emacs ...
   (t					;19.28+
    (when tilo-:timer-elt
      (ti::xe-timer-key-delete tilo-:timer-elt)
      (setq tilo-:timer-elt nil)
      )

    (when (and (null remove) (null  tilo-:timer-elt))
      (setq tilo-:timer-elt
	    (run-at-time
	     (format "%d sec" tilo-:idle-time)
	     tilo-:idle-time
	     'tilo-load-process
	     )))
    ))

  )

;;}}}
;;{{{ macros

;;; ----------------------------------------------------------- &macro ---
;;;
(defmacro input-pending-throw-macro (&optional force)
  "Do [FORCE] `throw' if input is pending. (action depends on emacs version)."
  (`
   ;;  Varibales, that are already defined by toplevel call
   ;;
   ;;    timer-load
   ;;    input-p
   ;;
   (when (input-pending-p)
     (setq input-p t)
     (if (or (, force)
	     (not timer-load)
	     input-p
	     )
	 (throw 'exit t)))
   ))

;;}}}
;;{{{ support functions

;;; ----------------------------------------------------------------------
;;;
(defun tilo-feature-p  (pkg &optional feature)
  "Check if feture has been loaded.
See PKG and FEATURE from `tilo-:load-list'"
  ;;  User didn't give us separate feature name, cmonstruct
  ;;  one from package name ~/elisp/test.el --> "test"
  ;;
  (when (and (null feature) pkg)
    (setq feature (file-name-nondirectory pkg))
    (if (and (string-match "^\\(.*\\)\\.el" feature)
	     (match-end 1))
	(setq feature (read-match 1)))
    )

  (setq
   pkg					;Reuse this variable
   (cond
    ((and (symbolp feature) (featurep feature)) 'symbol)
    ((and (stringp feature) (intern-soft feature) 'intern))
    (t nil)))

  pkg
  )

;;; ----------------------------------------------------------------------
;;;
(defun tilo-load-allowed-p  ()
  "Check if can load package now."
  (or ;;
      ;;  run-at-time check
      ;;
      (and (not (input-pending-p))
	   (not (fboundp 'run-with-idle-timer))
	   (not (boundp  'post-command-idle-hook))
	   tilo-:timer-elt
	   (no-action-in-progress-p 'timer))

      (and (not (input-pending-p))
	   (boundp 'post-command-idle-hook)
	   (no-action-in-progress-p 'post-command))
      (not (input-pending-p))
      ))


;;; ----------------------------------------------------------------------
;;;
(defun tilo-check-time (prev-time)
  "Check if time lag `tilo-:idle-time' compared to PREV-TIME."
  ;;  See if we invoke loading in this emacs
  ;;  Was it less than 30secs between commands
  ;;
  (cond
   ((not (boundp 'post-command-idle-hook))
    ;;   ok; we don't have to check the timestamp
    t)
   ((< (- (nth 1 (current-time)) (nth 1 prev-time))
       tilo-:idle-time)
    ;; not enough time elapsed since last call
    ;;
    nil)
   (t
    (- (nth 1 (current-time)) (nth 1 prev-time))
    )))

;;; ----------------------------------------------------------------------
;;;
(defun tilo-check-time-idle-hook  ()
  "Check if idle hook shuld run."
  (let* ((time      (get 'tilo-:load-list 'time))
	 )
    (unless time			;No prev time? Update.
      (setq time (current-time))
      (put 'tilo-:load-list 'time time))

    (tilo-check-time time)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tilo-message  (msg)
  "Display MSG and put it to *Messages* Buffer."
  (message (subst-char-with-string msg ?% "%%"))
  (let* (buffer
	 )
    (when (and (emacs-p)
	       (< emacs-minor-version 30))
      (setq buffer (get-buffer-create "*Messages*"))
      )

    (if buffer (job-buffer buffer (PMAX) (insert msg "\n")))
    ))


;;}}}
;;{{{ main

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tilo-load-process  (&optional force)
  "Load packages defined in `tilo-:load-list'.
If called interactively, FORCE loading all packaes in the list."
  (interactive (list 'force))
  (let* ((continue   (or force (inline (tilo-load-allowed-p))))
	 ;;  make sure we catch errors here (for bug report)
	 ;;
	 (debug-on-error t)
	 ;;
	 ;;  In Emacs 19.28, the timer process loads the files.
	 ;;  We don't see any idle time there --> we must load
	 ;;  at least one package at a time when timer triggers
	 ;;
	 (timer-load (and (emacs-p) (< emacs-minor-version 30)))
	 input-p
	 orig
	 list pos
	 pkg feature noerr nomsg form-b form-a
	 len
	 stat
	 )
    (if continue			;make one more check
	(setq continue (tilo-check-time-idle-hook)))

    (catch 'exit
      (when continue
        ;; ..................................................... plist ...
	;;  No original value available
	;;
	(unless (setq orig (get 'tilo-:load-list 'original))
	  (put 'tilo-:load-list 'original tilo-:load-list))

	(unless (get 'tilo-:load-list 'pos)
	  (put 'tilo-:load-list 'pos 0))

	;; List has changed. Update.
	;;
	(unless (eq orig tilo-:load-list)
	  (put 'tilo-:load-list 'original tilo-:load-list)
	  (put 'tilo-:load-list 'pos 0)
	  )

        ;; ................................................. load list ...
	(setq pos  (get 'tilo-:load-list 'pos)
	      len  (length tilo-:load-list)
	      list (nthcdr pos tilo-:load-list)
	      )
;;;	(d!! pos len (car list) "\n")

	(dolist (elt list)
	  (setq pkg	    nil
		feature	    nil
		noerr	    nil
		nomsg	    nil
		form-b      nil
		form-a	    nil
		)

          ;; ............................................... read args ...
	  (cond
	   ((listpp elt)
	    (setq pkg	    (nth 0 elt)
		  feature   (nth 1 elt)
		  noerr     (nth 2 elt)
		  nomsg     (nth 3 elt)
		  form-b    (nth 4 elt)
		  form-a    (nth 5 elt)
		  ))
	   ((stringp elt)
	    (setq pkg elt)
	    ))


;;;	  (d!! elt pkg feature "\n")


          ;; .................................................... load ...
	  (when (and elt pkg)
	    ;;  Try to sit for some time before preceeding, otherwise
	    ;;  if we can't sit still that long, user is
	    ;;  doing something..
	    ;;
	    (or (sit-for 1)
		(input-pending-throw-macro 'throw-it))

	    (setq stat (tilo-feature-p pkg feature))

;;;	    (d!! "stat" stat input-p form-b form-a "\n")

	    (when (not stat)		;Not already in Emacs
	      (if form-b (eval form-b))

	      ;; Load the package.
	      ;; If noerr flag is not given, trap the errors manually.
	      ;;
	      (if noerr
		  (setq stat (load pkg noerr nomsg))
		(or (safe-do (setq stat (load pkg noerr nomsg)))
		    (progn 
		      (tilo-message 
		       (format "Lazy load: fatal load error, %s"
			       pkg))
		      (throw 'exit)
		      )))
		      
			
	      (if form-a (eval form-a))

	      (tilo-message
	       (format "Lazy load %s %s (%d%% %d/%d)"
		       pkg
		       (if stat "ok" "error!")
		       (/ (* 100 (1+ pos)) len)
		       (1+ pos) len
		       ))
	      (put 'tilo-:load-list 'pos pos)
	      (input-pending-throw-macro 'throw-if-pending)
	      ))
	  (inc pos)
	  (put 'tilo-:load-list 'pos pos)
	  )))

    ))

;;}}}

;;{{{ example

;;; ......................................................... &example ...
;;; Here is my own setup example at the time of writing v1.6
;;; Hope you get some ideas from this.
;;; Rip code with tinylib.el / ti::m-pkg-rip-magic
;;;
;;; o  The files ~/.emacs.xxxx are part of my emacs initialization
;;; o  Function (uta) checks if I'm in University account
;;; o  {{{ tag is from folding.el

;;* ;; @(#) ~/.emacs.lazy.el -- Lazy loading of files.
;;* ;;
;;* _
;;* _
;;* ;;{{{ Lazy loading
;;* _
;;* ;;; ............................................................ &load ...

;;* (defconst tilo-:load-list
;;*   (list
;;*    ;;  startup files
;;*    ;;
;;*    (list "~/.emacs.tiny"	'rc-tiny)
;;*    (list "~/.emacs.19"		'rc-19)
;;*    (list "~/.emacs.hooks"	'rc-hooks 'noerr)
;;*    ;;  Has faces, load non-compiled version for XEmacs
;;*    (list (if (emacs-p)  "~/.emacs.mail"  "~/.emacs.mail.el") 'rc-mail)
;;*    (list "~/.emacs.tips"	'rc-tips)
;;*    (list "~/.emacs.tm-mime"	'rc-tm-mime 'noerr)
;;*    (list "~/.emacs.news-load"	'rc-news)
;;* _
;;*    (if (memq emacs-minor-version  '(28))
;;*        (list "~/.emacs.o-28" 'rc-28))
;;* _
;;*    ;;  Windowed emacs
;;*    ;;  - Load non-compiled rc file in XEmacs, because the compiled faces
;;*    ;;    are not compatible with XEmacs.
;;*    ;;
;;*    (when window-system
;;*      (list (if (emacs-p) "~/.emacs.font"  "~/.emacs.font.el")
;;* 	   'rc-font nil nil nil
;;* 	   '(progn (my-face-change 'def))))
;;* _
;;*    (when (and window-system (emacs-p))
;;*      (list "~/.emacs.x-menu" 'rc-xmenu))
;;* _
;;*    (when (and window-system (emacs-p))
;;*      (list "scroll-bar" nil 'noerr))
;;* _
;;* _
;;* ;;;   (if (not (uta))
;;* ;;;       (list "~/.emacs.time" 'rc-time))
;;* _
;;* _
;;* _
;;*    (list "suggbind" nil 'noerr nil
;;* 	'(progn (defconst suggest-key-bindings nil))
;;* 	'(progn
;;* 	   (if (fboundp 'suggbind-install) (suggbind-install))
;;* 	   ))
;;* _
;;* _
;;*    (if (emacs-p)
;;*        (list "mldrag" nil 'noerr nil
;;* 	     '(progn (setq mldrag-load-hook 'mldrag-default-keys))))
;;* _
;;*    (let ((file (my-e-file-p "tinyad")))
;;*      (when (emacs-p)
;;*        (list "tinyad" nil 'noerr nil
;;* 	     '(progn (add-hook 'tiad-compile-internal-hook
;;* 			       'my-compile-font-lock))
;;* 	     )))
;;* _
;;*    ;;  only needed for Emacs 28,30
;;*    ;;
;;*    (let ((file (my-e-file-p "tinypatch")))
;;*      (when (and (emacs-p)
;;* 		(memq emacs-minor-version '(28 30)))
;;*        (list "tinypatch")))
;;* _
;;*    (list "tinymy" nil 'noerr)
;;* _
;;*    (list "fa-extras" nil 'noerr nil nil '(progn (my-fa-setup)))
;;* _
;;*    (if (emacs-p)			;Don't know if works in XEmacs
;;*        (list "fdb" nil 'noerr))
;;* _
;;*    ))
;;* _
;;* (require 'tinyload)

;;* ;;}}}
;;* _
;;* _
;;* ;;{{{ auto byte compile
;;* _
;;* ;;; This piece of code
;;* ;;; a) saves the buffer
;;* ;;; b) byte compiles it after save; only IF
;;* ;;;    o  there is no variable 'my-no-elc defined. Use makunbound
;;* ;;;       function to abolish any such variable if you want byte
;;* ;;;       compilation.
;;* ;;;    o  there is no .gz file which would indicates that you use
;;* ;;;       compressed initialize files with tinyezip.el to save disk space.
;;* ;;;
;;* ;;; Local Variables:
;;* ;;; eval: (defun byte-compile-this-file () (write-region (point-min) (point-max) buffer-file-name nil 't) (if (and (not (boundp 'my-no-elc)) (not (file-exists-p (concat buffer-file-name ".gz")))) (byte-compile-file buffer-file-name)) t)
;;* ;;; write-file-hooks: (byte-compile-this-file)
;;* ;;; End:
;;* _
;;* ;;}}}
;;* _
;;* (provide 'rc-lazy)
;;* _
;;* ;; End of file


;;}}}
;;{{{ final setup

;;; ........................................................... &final ...

;;  Put startup info into *Messages*" buffer
;;
(tilo-message
 (format "[TinyLoad] started with %d elts in load list"
	 (length tilo-:load-list)))




(tilo-install)
(provide   'tinyload)
(run-hooks 'tilo-:load-hook)

;;}}}

;;; tinyload.el ends here
