;; @(#) tinymacro.el -- Fast way to assign newly created macro to key

;; @(#) $Id: tinymacro.el,v 1.16 1997/05/14 08:15:27 jaalto NotReleased $
;; @(#) $Keywords: tools macro key binding $
;;
;; $KnownCompatibility: 19.28-30, XE 19.14 $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1995,96,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Mar 23 1995
;;
;; To get information on this program use ident(1) or do M-x tim-version
;; Look at the code with folding.el, tinyfold.el


;; LCD Archive Entry:
;; tinymacro|Jari Aalto|jari.aalto@poboxes.com|
;; Fast way to assign newly created macros to keys. Record-Set and that's it!|
;; 14-May-1997|$Revision: 1.16 $|~/misc/tinymacro.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install
;; ........................................................ &t-install ...
;;   Put this file on your Emacs-Lisp load path, add following into your
;;   ~/.emacs startup file. Rip code with tinylib.el/ti::m-pkg-rip-magic
;;
;;* _
;;*      (require 'tinymacro)
;;
;;   or use autoload and your .emacs starts faster
;;
;;* _
;;*      (global-set-key [C-f9] 'tim-end-kbd-macro-and-assign)
;;*      (autoload 'tim-end-kbd-macro-and-assign  "tinymacro" t t)
;;
;;   Loading this file automatically defines key f9
;;   to start kbd macro recording and Control-f9 to terminate
;;   recording and assigning the macro to key. If you don't want those
;;   bindings, add this prior the require or autoload commands:
;;
;;* _
;;* 	(setq tim-:load-hook nil)
;;
;;   If you have any questions, feedback, use this function
;;
;;      M-x tim-submit-bug-report


;;}}}
;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "tim-" in front of
;;   every function & variable. It stands for '(ti)ny (m)acro
;; - variable names contain letters 'tim-:'

;;  Preface, 1995
;;
;;	Started as a very little project when mosh@ramanujan.cs.albany.edu
;;	(Mohsin-Ahmed) Mar 17 -95 sent a post to g.e.h asking for easy way
;;	to assign newly created macro to some key.
;;
;;	I sent a reply, where I had made simple function to do such thing,
;;	unfortunately Mohsin bounced me and complained that it didn't
;;	work right :') and that he didn't want only one macro, which was
;;	recycled every time. I had coded it so that it used only _one_,
;;	and the same function name every time it was called, effectively
;;	erasing the old macro..
;;
;;	I started twisting the code little bit, and so I finally came up with
;;	this, which uses user configurable macro-name-functions.
;;
;; Description
;;
;;	o   Two keystrokes to make a macro: one to record, one to
;;	    assign it to key.
;;	o   To see the macro assignments to keys, just call tim-macro-info


;;}}}
;;{{{ History

;; ........................................................ &t-history ...
;;; Change log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; Apr	21	1997	[jari]		19.28	v1.15		NotReleased
;; - Internal changes. Loading this file now automatically occupies f9
;;   and C-f9 keys.
;; - Changed all non standard boolean tests to use when and unless
;;
;; Mar	18	1997	[jari]		19.28	v1.13-1.14	NotReleased
;; - XE20 CHAR changes.
;;
;; Feb	21	1997	[jari]		19.28	v1.10-1.12	NotReleased
;; - Macros can't be autoloaded. Added require lib 'm' and changed
;;   few library functions (name change)
;;
;; Sep	25	1996	[jari]		19.28	v1.9		NotReleased
;; - Rearranged loading. Corrected all byte compiler errors.
;;
;; Aug	16	1996	[jari]		19.28	v1.8		NotReleased
;; - Corrected code so that it works in XE 19.14.
;;
;; Jun	10	1996	[jari]		19.28	v1.7		NotReleased
;; - Correct slight but serious problem in the main function. Without
;;   this correction, you couldn't make macros :-(
;;
;; May	28	1996	[jari]		19.28	v1.5-1.6	NotReleased
;; - Made slight changes to layout and moved genral functions to
;;   tinylib. Correct problem in tim-assign, since it didn't warn
;;   about already occupied keys.
;;
;; Feb	12	1996	[jari]		19.28	v1.4		NotReleased
;; - stephene@cogs.susx.ac.uk (Stephen Eglen) Suggested that the macro
;;   binding should be temporary, and the original bindings should be
;;   restored somehow.
;; - Now you can call function tim-restore if you want to get back
;;   all bindings that are occupied by macros.
;;
;; Feb	8	1996	[jari]		19.28	v1.3		NotReleased
;; - lisp-mnt's lm-verify checked. Rewrote feedback funcs.
;; - Removed stack++ function, which was sedined inside function. Don't
;;   know how much this affects whole program --> copied the body
;;   to points where it was called.
;;
;; Mar 24	1995	[jari]		19.28	v1.2		Release_1
;; - More goodies: funcs tim-macro-info , tim-func2key-desc , hooks ...
;;
;; Mar 23	1995	[jari]		19.28	v1.1		Created
;; - First implementation sent to Mohsin.


;;}}}


;;; Code:

;;{{{ setup: require

(require 'tinylibm)

(eval-when-compile (use-dynamic-compilation))

;;}}}
;;{{{ setup: hooks, private

;;; ........................................................... &hooks ...

(defvar tim-:macro-assigned-hook nil
  "*If new macro were asiigned, this hook will be run. The function
SYMBOL that was used is in variable tim-:last-macro-func")

(defvar tim-:load-hook  'tim-install-default-bindings
  "*Hook run when file has been loaded.")


;;; ....................................................... &v-private ...


(defvar tim-:stack-ptr 0
  "Keeps record of available stack  space.")

(defvar tim-:last-macro-func nil
  "holds last function SYMBOL that were used in assignment.")

(defvar tim-:last-macro-key nil
  "holds last key STRING or VECTOR that were used in assignment.")

(defvar tim-:function-list nil
  "List of original KEY -- FUNCTION pairs, whic are currently occupied
by macros")

;;}}}
;;{{{ setup: user

;;; ........................................................ &v-public ...

(defvar tim-:general-name "tim--macro"
  "*The lisp name to use, when assigning name to last kbd macro")

(defvar tim-:wrap-query nil
  "*If set to non-nil, then the user is asked before the macro stack
if recycled when it got full. This means that macro numbering, starts
from 0 agaon.")

(defvar tim-:replace-key-query t
  "*When NON-NIL; If user selects keys that already hold some functionality,
eg. function or previous macro he will be asked for confirmation before
new macro to it assigned to these keys.")

(defvar tim-:stack-max 10
  "*Maximum stack depth of unique macronames. The name run from 0..max,
and wraps to 0 after max.")

(defvar tim-:tmp-buffer "*temp*"
  "*Temporary buffer. Eg. displaying the macro bindings to keys.")

;;}}}
;;{{{ setup: version

;;; ....................................................... &v-version ...


(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinymacro.el"
 "tim"
 tim-:version-id
 "$Id: tinymacro.el,v 1.16 1997/05/14 08:15:27 jaalto NotReleased $"
 '(tim-:version-id
   tim-:stack-ptr
   tim-:last-macro-func
   tim-:last-macro-key
   tim-:function-list
   tim-:macro-assigned-hook
   tim-:load-hook
   tim-:general-name
   tim-:wrap-query
   tim-:replace-key-query
   tim-:stack-max
   tim-:tmp-buffer
   )))

;;}}}

;;; ########################################################### &funcs ###

;;{{{ code: misc


;;; ----------------------------------------------------------------------
;;;
(defun tim-install-default-bindings  ()
  "Install keys f10 and C-f10 to start and (assign) macro."
  (interactive)

  (tinyck-maybe-activate 'xemacs)

  ;; These bindings work in every emacs when the tinyck is on.
  ;;
  (global-set-key [(f9)] 'start-kbd-macro)
  (global-set-key [(control f9)] 'tim-end-kbd-macro-and-assign)

  (tinyck-maybe-activate 'xemacs 'disable)
  )

;;; ----------------------------------------------------------------------
;;;
(defun tim-yn ()
  "Return t if [yY ] pressed, otherwise returns nil."
  (char-in-list-case (read-char-safe) '(?y ?Y ?\ )))



;;; ----------------------------------------------------------------------
;;;
(defun tim-restore  ()
  "Restores all macro bindings, so that keys that occupy macros
are restored to original functions.

References:
  tim-:function-list     ,list is cleared too.
"
  (interactive)
  (let* ((verb  (interactive-p))
	 (list  tim-:function-list)
	 key func
	 )
    (cond
     ((null list)
      (if verb (message "No macros active.")))
     (t
      (while list
	(setq key (nth 0 (car list)) func (nth 1 (car list)))
	(global-set-key key func)
	(setq list (cdr list)))
      (setq  tim-:function-list nil)
      ))
    ))

;;}}}
;;{{{ code: symbol

;;; ----------------------------------------------------------------------
;;;
(defun tim-create-symbol()
  "Creates macro variable. Returns NEW or EXISTING SYMBOL."
  (let* ((max	tim-:stack-max)
	 (sp	tim-:stack-ptr)
	 (q	tim-:wrap-query)
	 (n	tim-:general-name)
	 sym2
	 new
	 ret
	 )

    (if (or (null q) (< sp max))	; yes, go ahead with new
	(setq new (concat n (if (< sp max)		; 0..max
				(setq sp (1+ sp))
			      (setq sp 0))))

      (message "Macro stack full, wrap y/[n] ?")
      (if (tim-yn)
	  (setq new
		(if (< sp max)		; 0..max
		    (setq sp (1+ sp))
		  (setq sp 0)))
	))

    (if (null new) nil			;  Must update stack
      (setq tim-:stack-ptr sp)
      (setq ret (intern-soft new))	; returns symbol

      (if ret nil			; Already exist
	;;   Seem stricky ?, just a)make it b)set to nil c)put into ret val
	(setq sym2 (intern new))
	(set sym2 nil)
	(setq ret sym2)
	))
    ret
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tim-create-name()
  "Creates macro name."
  (let* ((max	tim-:stack-max)
	 (sp	tim-:stack-ptr)
	 (q	tim-:wrap-query)
	 (n	tim-:general-name)
	 new
	 )

    (if (or q (< sp max))		; yes, go ahead with new
	(setq new
	      (concat n (if (< sp max)		; 0..max
			    (setq sp (1+ sp))
			  (setq sp 0))))

      (message "Macro stack full, wrap y/[n] ?")
      (if (tim-yn)
	  (setq new (concat n (if (< sp max)		; 0..max
				  (setq sp (1+ sp))
				(setq sp 0))))
	))

    (if new				; Must update stack
	(setq tim-:stack-ptr sp))
    new
    ))

;;}}}
;;{{{ code: main

;;; ----------------------------------------------------------------------
;;;
(defun tim-macro-info ()
  "Shows which macros are assigned to which keys."
  (interactive)
  (let* (
	 (sp	tim-:stack-ptr)
	 (max	tim-:stack-max)
	 (buf	tim-:tmp-buffer)
	 (base	tim-:general-name)
	 (i	0)
	 (round 0)
	 bp				;buffer pointer
	 name
	 key
	 )
    (while (< i (1+ max))
      (setq name (concat base i)   i (1+ i)   key "")

      (if (null (intern-soft name)) nil	;not use yet

	(if (> round 0) nil		;do only once
	  (setq bp (get-buffer-create buf))
	  (set-buffer bp) (erase-buffer)
	  (insert (format "Stack pointer : %d\n" sp )))

	(if (null (setq key (ti::k-function-bind-info (intern name))))
	    (setq key "[??]"))		;should never happen

	(insert (format "%-10s %s\n" key name))
	(setq round (1+ round))
	))

    (if (and (interactive-p) (eq 0 round))
	(message "No macros bound or set."))

    (switch-to-buffer-other-window bp)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tim-end-kbd-macro-and-assign  ()
  "Terminate reading macro and assign it to key."
  (interactive)
  (end-kbd-macro)
  (call-interactively 'tim-assign)
  )

;;; ----------------------------------------------------------------------
;;;
(defun tim-assign (&optional key)
  "Names last macro and assigns it to user defined key.
Runs tim-:macro-assigned-hook if key macro gets installed.
The query options should be turned off if you call this within
function, since it always return nil if the options are on.

Input:
  key  Should be valid emacs key-bind-sequence/key-vector

Return:
  t    is assigned
  nil  not assigned keyboard-quit

"
  (interactive)

  (let* ((query	    tim-:replace-key-query)	;permission to assign
	 (do-it	    (not query))		;when query=nil, do-it=t
	 (verb      (interactive-p))
	 (f-name    "")			;func name
	 macro-name			;our new macro !
	 lookup				;what was found

	 ;; --- 1 ---
 	 ;; The bullet proof way to find key bind for abort
	 ;; (abort-ch (ti::k-function-bind-info 'keyboard-quit 'raw))

	 ;; --- 2 --
	 ;; - Or we just say where it is... Nobody relocates it anyway
	 ;; - We use this because function2key does not work in XE
	 ;;
	 (abort-ch (char-to-string ?\007))
	 )

    (if (null key)
	(setq key (read-key-sequence "Set last macro to key(s): ")))


    (if (equal key abort-ch)
	(progn
	  (if (interactive-p) (message "Skipping abort key. Not assigned."))
	  nil)

      ;;  Search the key, if it's already assigned
      ;;
      (setq lookup
	    (or (and (current-local-map) ;in fundamental-mode this is nil.
		     (lookup-key (current-local-map) key))
		(lookup-key global-map key) key))

      ;; ................................................... occupied? ...
      (when
	lookup
	(cond				;what we found ?
	 ((and (symbolp lookup)
	       (fboundp lookup))	;just plain function
	  (setq f-name (symbol-name lookup)))

	 ((vectorp lookup)
	  (setq f-name "[..vector]"))

	 ((stringp lookup)
	  (setq f-name "[...string"))

	 (t
	  (setq f-name "[unidentified]")
	  )))

      ;; ............................................. ask permission? ...
      (when
	(and verb
	     query
	     (not (null lookup)))
	(setq
	 do-it
	 (y-or-n-p
	  (concat "Key already occupied by ("
		  f-name ") , remove it y/[n]? ")))
	)

      ;; ................................................ assign macro ...
      (cond

       ((and verb (null do-it))
	(message
	 (substitute-command-keys
	  "Cancelled. Use \\[tim-assign] to rebind.")))

       (t
	(setq macro-name (tim-create-symbol))
	(name-last-kbd-macro macro-name)

        ;; ............................................. save previous ...
	(cond
	 ((and (symbolp lookup) (fboundp lookup)
	       (not (string-match "^tim" f-name))
	       (not (assoc key tim-:function-list))
	       )
	  (setq tim-:function-list
		(cons
		 (list
		  key
		  lookup
		  )
		  tim-:function-list
		  ))
	  ))

	(global-set-key key macro-name)

	(setq tim-:last-macro-func macro-name ;set >> GLOBALS <<
	      tim-:last-macro-key key)

	(if verb
	    (message
	     (concat "Assigned to macro name : " (symbol-name macro-name))))
	(run-hooks 'tim-:macro-assigned-hook)
	t
	))
      );; ...  ...  ...  ...  ...  ...  ...  ...  ...  ...  ...  ... abort-key
    ))

;;}}}



(provide   'tinymacro)
(run-hooks 'tim-:load-hook)

;;; tinymacro.el ends here
