;; @(#) tinymy.el -- Collection of 'my' functions. Simple solutions.

;; @(#) $Id: tinymy.el,v 1.51 1997/06/26 22:59:34 jaalto NotReleased $
;; @(#) $Keywords: tools extensions $
;;
;; $KnownCompatibility: 19.28+ XEmacs19.14+ $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1995,96,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Nov 8 1995
;;
;; To get information on this program use ident(1) or do M-x timy-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinymy|Jari Aalto|jari.aalto@poboxes.com|
;; Collection of 'my' functions. Simple shar/tar/uu/compilation, periodic save|
;; 27-Jun-1997|$Revision: 1.51 $|~/misc/tinymy.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ....................................................... &t-install ...
;; Put this file on your Emacs-Lisp load path, add following into your
;; ~/.emacs startup file. Rip code with with tinylib.el/ti::m-pkg-rip-magic
;;
;;* _
;;*     (require 'tinymy)
;;* _
;;
;; This package can't be autoloaded easily, because it installs timers
;; and many global bindings. For the first time you should install the
;; attached sripts. You need 'pgp' executable for this (Pretty Good Privacy)
;;
;;      M-x load-library RET tinymy RET
;;      M-x timy-install-files
;;
;; While you can't autoload this, you can delay the loading to later
;; time so that your .emacs is loaded faster by using this command
;; instead of the 'require' call
;;
;;*     ;; Delayed load, after 5 minutes...
;;*     (run-at-time "5 min" nil '(lambda () (require 'tinymy)))
;;* _
;;
;; Modify the following variable and put your own installation there
;; if the default setting doesn't suit you.
;;
;;      timy-:define-key-table
;;
;; You can force to use the default keys by setting this prior
;; package load call statements
;;
;;* _
;;*     (setq timy-:define-key-force t)
;;* _
;;
;; If you have any questions, use this function
;;
;;      M-x timy-submit-bug-report      ,send bug report or feedback

;;}}}

;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "timy-" in front of
;;   every function & variable. It stands for '(ti)ny

;;  Preface, Nov 1995
;;
;;      My .emacs files started to look quite interesting:
;;
;;          .ema.tips       .emacrs.el      .emacrs.el.b    .emacrs.el.sun
;;          .emacrs.el~     .emacs          .emacs.18       .emacs.19
;;          .emacs.19~      .emacs.abb      .emacs.c-fl     .emacs.compile
;;          .emacs.debug    .emacs.default  .emacs.font     .emacs.ding
;;          .emacs.font.b   .emacs.gnus     .emacs.gnus.o   .emacs.hooks
;;          .emacs.init     .emacs.init2    .emacs.ja       .emacs.mail
;;          .emacs.o        .emacs.o-19     .emacs.out      .emacs.out~
;;          .emacs.path     .emacs.pc       .emacs.prog     .emacs.set
;;          .emacs.test     .emacs.time     .emacs.tips     .emacs.vc
;;          .emacs.w3       .emacs.x-menu   .emacs.xx       .emacs_context
;;          .emacs.ding.hilit    .emacs.ding.hilit2         .emacs.dired
;;          .emacs.mode-line     .emacs.news-load           .emacs.news-set
;;
;;      My private functions seemed to grow all the time, most of them
;;      were turned into packages, but sometimes it was just
;;      function or two that made my life with emacs easier.
;;
;;      What you see here is a selection of there *my* functions
;;      I developed for my private use. The consept *my* does not
;;      refer to *mine*, but has backround in function naming.
;;      Remember? all user functions are recommended to be named so,
;;      that the first word is =my-=, like my-FUNC-NAME-HERE
;;
;;      If I have used/found someone else's function/idea, I have recorded
;;      the author.
;;
;;  Overview of features
;;
;;      Handy everyman's 'my' functions that solve many tasks.
;;      Contains gems that are not "packaged"
;;
;;      additional timer processes
;;
;;      o   RMAIL/other buffer save in regular intervals.
;;      o   Reverting of buffers in background, if file
;;          has changed on disk. By <duthen@cegelec-red.fr>
;;          (Jacques Prestataire)
;;      o   Mail lock watchdog. If you have this lock in your
;;          filesystem, you cannot receive mail!
;;
;;      buffer
;;
;;      o   Renaming *shell* or *mail* buffer easily to get another
;;          *shell* or *mail* buffer.
;;      o   Adding up numbers in rectangle area
;;      o   Scrolling jumps to window end/beginning and does not scroll
;;          immediately. See variable timy-:scroll-mode for more.
;;      o   Trim trailing whitespaces from the buffer when file
;;          is saved.
;;      o   Chmod'ing buffer to w+ or w- on disk.
;;      o   Stamping Docid: and Contactid: and LCD pacakge entry fields.
;;          when you save the buffer.
;;
;;      compile
;;
;;      o   guesses compile command by looking at the buffer content
;;          Includes shell scripts "emc" and "mak" to do emaacs
;;          byte compilation and C/C++ make file compilation.
;;
;;      gnus, mail
;;
;;      o   save lisp package in mail buffer to file
;;      o   Copy current buffer's contents to new mail buffer and
;;	    set subject line. You can send diff buffers and file buffers
;;	    conveniently this way.
;;
;;      keys
;;
;;      o   "%" jumps to matching {(["
;;      o   Better word movement: LikeThisInC++Mode.
;;          Moving forward/backward always keeps cursor at the
;;          beginning of word.
;;      o   Repeat last command you did: this is for simple
;;          commands which do not take arguments. By Fabio Somenzi
;;      o   PgUp and PgDown behave differently; they jump to
;;          window's beg/end first and only next key hit scrolls.
;;
;;      shell
;;
;;      o   Easy shar/tar/UU commands
;;
;;      window
;;
;;	o   Window auto-select when you move mouse-cursor over it.
;;      o   Flipping the order of two windows
;;      o   In X, pointing at window swithes to it automatically
;;          without clicking.
;;
;;      misc
;;
;;	o   Sort minor mode: to sort columns 1-9 easily
;;
;;
;;  What commands are defined when you load this file?
;;
;;      It's better to look at the code of this file, than to explain all the
;;      key definitions here, because I may not remember update this
;;      text section every time I add new interactive commands to the file.
;;
;;      All the new interactive comnmands can be found from these two
;;      functions:
;;
;;          timy-define-keys
;;          timy-mail-common-keys
;;
;;      See their description, or alternatively hit 'C-h m' to view
;;      all bindings in  (r)mail/gnus or global map and search for string
;;      "timy".
;;
;;  How to use this package
;;
;;      I would suggest just loading this package right away and then
;;      printing out the whole file and reading the comments about
;;      individual functions: what they do.
;;
;;  Warning
;;
;;      When you load this package, it install global keybindings
;;      that *may* replace your private bindings. Please see if you
;;      need to rearrange timy-:load-hook which now runs functions
;;
;;           '(timy-install timy-define-keys timy-alias)
;;
;;      Please use this if you want to use your own bindings.
;;
;;          (setq timy-:load-hook '(timy-install my-timy-keys timy-alias)
;;                                               ^^^^^^^^^^^^
;;          (defun my-timy-keys ()
;;            ;; .. different global key remappings
;;            )
;;
;;      There is table of global bindings which you can modify if the
;;      bindings clash: the auto installl will warn you about
;;      this automatically and your own bindings are not replaced by
;;      default. See variable
;;
;;          timy-:define-key-table

;;}}}
;;{{{ history

;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; Jun	 26	1997	[jari]		19.28	v1.47-1.50	NotReleased
;; - 19.34 has binding f10 tmm-menubar, moved repeat command to shift-f10.
;; - Library functions changes.
;; - Rick Flower <flower@ms31.sp.trw.com> reported compiler error,
;;   (function that a didn't exist). Corrected.
;;   timy-scroll	:! Small adjustment
;; - fixed timy-view function
;;
;; May	 27	1997	[jari]		19.28	v1.44-1.46	NotReleased
;; - Change byte compiling: now m library defines more autoloads for XEmacs.
;; - Corrected timy-lisp-package-save-to-file. (doesn't require .el any more)
;; - Defcustom.el support, checkdoc 1.42 clear.
;;
;; May	 14	1997	[jari]		19.28	v1.42-1.43	NotReleased
;; - Yigal Hochberg <hochberg@cisco.com> suggested that the VI paren match
;;   would insert literal if the parens were closed. Changed the code
;;   accordingly and wrote new doc string for the function.
;; - Changed all non standard boolean tests to use when and unless
;;
;; May	 12	1997	[jari]		19.28	v1.40-1.41	NotReleased
;; - timy-:stamp-lcd-regexp was errorneous.
;; - Adjusted scroll command a little: if line wraps then the
;;   emacs scroll command is used, because it is too hard to calculate
;;   window pmin/pmax jump points.
;;
;; Apr	 21	1997	[jari]		19.28	v1.35-1.39	NotReleased
;; - New user function timy-mail-buffer, bound to C-z m.
;; - New user function timy-buffer-file-chmod, bound to C-x q
;; - new functions timy-file-stamp, timy-file-stamp-elisp-lcd
;; - new minor mode interface used.
;;
;; Mar	 19	1997	[jari]		19.28	v1.30-1.34	NotReleased
;; - Some siple library call changes. Trim function does not touch
;;   read-only-buffers any more. (bug)
;; - Changed the mouse window auto-select code so that ByteComp does not
;;   give warnings any more.
;; - XEmacs changes. The default keysettings was changed. THe automatic
;;   buffer selection is now hooked for XEmacs, but it doesn't work for
;;   some reason.
;; - ESC ` binding added. timy-mouse-select-window-auto now works in XEmacs.
;; - Added sort minor mode.
;;
;; Mar	 4	1997	[jari]		19.28	v1.28-1.29	NotReleased
;; - Corrected timy-view function to take care of turning off expand-mode
;;   because it occupied the SPC key that should be available for view.
;;
;; Feb	 28	1997	[jari]		19.28	v1.25-1.27	NotReleased
;; - Corrected timy-view.Track mouse is notinstalled for window system
;;   only
;; - Added more features to timy-copy-file, now C-u deletes safe copies,
;;   M-1 restores safe copy over the regular file.
;;
;; Feb	 21	1997	[jari]		19.28	v1.24		NotReleased
;; - New user functon timy-read-only installed into C-xC-q. It will
;;   turn on view when buffer is read only. VC confirm asked.
;;
;; Feb	 21	1997	[jari]		19.28	v1.22-1.23	NotReleased
;; - Rewrote progrma and keybinding installation. The package is now
;;   tar/PGP base64 sealed. New user table timy-:define-key-table
;;   helps moving keys according to one's preference.
;; - All keybindings are now in global map and there is no special
;;   mal mode bindings as before.
;; - Moved the tar listing command to mail library. It is now generalized.
;; - New user commands C-up and C-down to go over functions. Very handy
;;   in lisp and C/C++
;; - new process: watchdog for your mail lock, so that you know if you
;;   canb receive mail. It happened to me that there was lock and I
;;   didn't get mail for days!
;; - Trims blanks from the file when it is saved.
;;
;; Feb   16     1997    [jari]          19.28   v1.21           NotReleased
;; - Corrected timy-lisp-package-save-to-file, now detects proper
;;   end of file better.
;;
;; Feb   12     1997    [jari]          19.28   v1.20           NotReleased
;; - Corrected some byteCompiler errors. Added note; that scrolling
;;   doesn't work right in long line buffers.
;;
;; Jan   27     1997    [jari]          19.28   v1.19           NotReleased
;; - Jacques Prestataire <duthen@cegelec-red.fr> set new function
;;   how to do the reverting of the buffers. I replaced the old with this
;;   new one. Thanks.
;;
;; Jan   23     1997    [jari]          19.28   v1.14-18        NotReleased
;; - Added C-right and C-left word movement commands from my keyboard
;;   settings. autoloads are now in 'm' lib.
;; - Added better scroll up/down movement.
;; - Added Fabio's timy-repeat-last-simple-command bound to f10
;;
;; Nov   8      1996    [jari]          19.28   v1.13           NotReleased
;; - Corrct timy-lisp-package-save-to-file (regexp error, premeature
;;   cut sliced file in half.)
;; - New user function timy-copy-file in "C-zwfc"
;;
;; Nov   3      1996    [jari]          19.28   v1.12           NotReleased
;; - Changed makefile to match "\\.mak|makefile|Makefile"
;;
;; Oct   11     1996    [jari]          19.28   v1.10-11        NotReleased
;; - Used exec-to-string, because it will be part of emacs distribution.
;; - Used (function and deleted '(lambda quotations.
;; - Fixed some documentation errors.
;;   timy-lisp-package-save-to-file     :! now writes "ends here" tag
;;
;; Sep   25     1996    [jari]          19.28   v1.8-1.9        NotReleased
;; - Noticed that automatic window selection and menubar cannot be mixed.
;;   (I dind't use menubar so I didn't notice this before). Now the
;;   installation of this feature is special.
;; - Rearranged installation: now uses autoloads.
;;   timy-:install-select-window-auto   :+ new user variable
;;   timy-scroll-up                     :+ new user func
;;   timy-scroll                        :+ new user func
;;
;; Sep   19     1996    [jari]          19.28   v1.7            NotReleased
;; - Added timy-install-files so that user can easily get this package
;;   working in 1 minute. Shell scripts changed in this file.
;;   timy-toggle-buffer-name            :+ new user func "C-zn"
;;   timy-add-rectangle                 :+ new user func "C-xr+"
;;
;; Sep   11     1996    [jari]          19.28   v1.6            NotReleased
;; - Rewrote patrt of the installation functions and moved one of
;;   my general compiling function into this package from my private
;;   function library.
;; - Added shell scripts "emc" and "mak". Wrote section
;;   'What commands are defined when you load this file?'
;;   timy-:compile-command-c-code       :+
;;   timy-:compile-table                :+
;;   timy-define-keys           :+
;;   timy-compile-run-command           :+
;;
;; Sep   10     1996    [jari]          19.28   v1.4-1.5        NotReleased
;; - Rewrote most of the package and payed more attention to installation.
;; - Thanks to Rune.Juntti@pjl.frontec.se corrected the order of loading
;;   the libraries. Checked code with XEmacs 19.14 bytecompiler.
;; - The mouse code doesn't work in XEmacs.
;; - Added message how big the inserted UU block is in the message
;;   when user runs timy-uu-make.
;;
;; Jun  28      1996    [jari]          19.28   v1.3            NotReleased
;; - XEmacs compatibility (timer functions)
;;
;; Jun  13      1996    [jari]          19.28   v1.2            NotReleased
;; - Wow, I haven't looked at this package for some tim. Now added
;;   nice buffer reverting function in the background.
;;
;; Nov  8       1995    [jari]          19.28   v1.1            Created



;;}}}

;;; Code:

;;{{{ setup: require

;;; ......................................................... &require ...

(require 'tinylibm)

(eval-and-compile
  (use-timer)
  (autoload 'compile-internal               "compile")
  (autoload 'operate-on-rectangle           "rect")

  (defvar   view-mode nil)
  (autoload 'view-mode			    "view" t t)
  )

(eval-when-compile (use-dynamic-compilation))

(defgroup-tiny TinyMy timy-: tools
  "Collection of small so called 'my' utility functions.
The full feature list is in the source code documentation, read it well."
  )


;;}}}
;;{{{ setup: variables

;;; .......................................................... &v-bind ...
;;; Change this table if you have conflicting bindings.
;;;

(defcustom timy-:define-key-force nil
  "*If non-nil; assign keys without any check."
  :type 'boolean :group 'TinyMy
  )


(defcustom timy-:define-key-table
  '(
    ;;  This must be in key that can be hit directly.
    ;;  Otherwise it's not much use
    ;;
    ( [(shift f10)]   . timy-repeat-last-simple-command)

    ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. C-x . .
    ;;  The 'rectangle' map. This sould be fee for us.
    ;;
    ("\C-xr+"	. timy-add-rectangle)

    ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. C-c . .
    ;; minor modes in 'm' map
    ;;
    ("\C-cmv"   . timy-view)
    ("\C-cmS"   . timy-sort-mode)

    ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. C-z . .
    ;; Pick "c" for all (c)ompile commads, now define additional
    ;; "c" for this particular command
    ;;
    ("\C-zcc"   . timy-compile-run-command)

    ("\C-zm"    . timy-mail-buffer)
    ("\C-zn"    . timy-toggle-buffer-name)
    ("\C-zS"	. ti::m-surround-with-char)

    ;;  Oher miscellaneout to "x" extra map
    ;;
    ("\C-zxc"   . timy-copy-file)
    ("\C-zxf"   . timy-flip-windows)
    ("\C-zxl"   . timy-lisp-package-save-to-file)
    ("\C-zxt"   . timy-trim-blanks)


    ;;  's' for shell commands
    ("\C-zxsu"  . timy-uu-make)
    ("\C-zxsU"  . ti::p-uu-save-buffer) ;; Already in tinylib
    ("\C-zxss"  . timy-shar)
    ("\C-zxst"  . timy-tar)
    )
  "*Define command to `global-map' keys.
Format '((KEY . FUNCTION) (K . F) ..)"

  :type '(repeat
	  (cons
	   (string :tag "Key Bind sequence")
	   function
	   ))
  :group 'TinyMy
  )



;;; ......................................................... &v-hooks ...

(defcustom timy-:load-hook '(timy-install timy-define-keys timy-alias)
  "*Hook that is run when package is loaded."
  :type 'hook :group 'TinyMy
  )


(defcustom timy-:mail-buffer-hook  nil
  "*This hook run last in `timy-mail-buffer' function."
  :type 'hook :group 'TinyMy
  )

;;; ....................................................... &v-private ...

(defvar timy-:package-exist-mview (ti::f-get-load-path "mview.el" load-path)
  "Location of mview.el if exists -- view minor mode.")


;;; ....................................................... &vu-config ...
;;; all "vu" -- "variable user" sections are meant for user configurable

(defcustom timy-:install-select-window-auto 'yes
  "*Variable is used only in window system.
The automatic window selection function selects window by pointing
at it with mouse . No clicking is needed. However if you use menu bar, it is
a bit difficult to use this automatic selection feature, because the
menu bar reflects the current window: When you are at lower window and reach
for the menu bar, the upper window gets selected and the menu bar reflects
that window. You never get menu bar for the other windows but for the
topmost one. (Well, you can go round of emacs, and then reach for
menu-bar, but that's a bit awkward)

Values in this variable:
  'yes
  'no
  'ask

To turn off the menu bar use: (menu-bar-mode -1) in your ~/.emacs."
  :type '(choice
	  (const yes)
	  (const no)
	  (const ask)
	  )
  :group 'TinyMy
  )


(defcustom timy-:register  ?r
  "*An Emacs register where to put results of commands.
User can then afterwards yank the result into desired buffer."
  :type 'character :group 'TinyMy
  )


(defcustom timy-:scroll-mode 'window
  "*If non-nil, then `timy-scroll' does not immediately scroll.
The following happen if variable is non-nil.
o  up: if the cursor is not at the window's start line, go there
o  up: if cursor is at window's top, line, now scroll
o  down: --''-- behaves same as up"
  :type 'boolean :group 'TinyMy
  )


(defcustom timy-:vi-type-paren-match-special-list '( ?\" ?\' ?\$ )
  "*List of special character to matched in \\[timy-vi-type-paren-match].
If the sentence delimited by these chars spread multiple lines,
the missing part is searched backward.

If you call \\[timy-vi-type-paren-match] with optional arg, then
the search is forced FORWARD."
    :type '(repeat character) :group 'TinyMy
    )

;;  This could have been (CH . CH) list but because XEmacs20
;;  has different character handling that emacs; we prefer to check
;;  strings.
;;
;;  This is not configurable variable right now, because the match
;;  function uses hard coded regexps.
;;
(defconst  timy-:vi-type-paren-match-list
  '( ( "(" . ")" )
     ( "{" . "}" )
     ( "[" . "]" )
;;; NOPE, DON'T add these. It won't work and I haven't have time to
;;; investigate why.
;;;
;;;     ( "<" . ">" )
     )
  "List of character string pairs to match.
Format:

  ((BEGIN-CHARACTER-PAIR-STR . END-CHARACTER-PAIR-STR)
   (B . E)
   ..
   )

Example:

  '( ( \"(\" . \")\" )
     ( \"{\" . \"}\" )
     ( \"[\" . \"]\" )
     ))"
  )


;;; ......................................................... &vu-word ...

(defcustom timy-:move-word-set  "-[]$%@#&*\":,.{}()<>/\\ \t\n"
  "*How to move forward/backward word. This is character set."
  :type '(string :tag "Charset") :group 'TinyMy
  )

(defcustom timy-:move-word-case-set  "-[]$%@#&*\":{}()<>/\\ \t\na-z"
  "*How to move forward/backward word. This is character set.
used only over mixed case words."
  :type '(string :tag "Charset") :group 'TinyMy
  )

(defcustom timy-:move-word-case-modes '(c-mode c++-mode cc-mode)
  "*Modes that should trigger using `timy-:move-word-case-set'."
  :type '(repeat function) :group 'TinyMy
  )


;;; ........................................................ &vu-shell ...

;;; -c      ,Create a new archive
;;;
(defcustom timy-:tar-command "tar -cf"
  "*Tar create command, eg. used in `timy-tar'."
  :type '(string :tag "Shell command") :group 'TinyMy
  )


(defcustom timy-:shar-command "shar -a -c -C -e -t -u"
  "*Shar command used by `timy-shar'.
In HP-UX:

 -a       ,do not protect them specially (uu)
 -c       ,data-integrity check using wc
 -C       ,Insert a line of the form --- cut here ---
 -e       ,code that prevents shar .. overwrite existing files.
 -t       ,Write diagnostics to stdout
 -u       ,Assume that the remote site has uudecode"
  :type '(string :tag "Shell command") :group 'TinyMy
  )


;;; ...................................................... &vu-compile ...

(defcustom timy-:compile-command-c-code "CC -g -z"
  "*C/C++ code compile command.
Do not add option -o In HP-UX default flags for C++ compiler:
 -g  ,make debuggable
 -z  ,trace null pointers errors"
  :type '(string :tag "Shell command") :group 'TinyMy
  )

(defcustom timy-:compile-table
  '(("perl"   "perl -w")
    ("^sh-"   "sh -x")
    ("^csh-"  "csh -x")
    ("^ksh-"  "ksh -x")
    ("awk"    "awk -f")

    ;;  The "emc",(em)acs (c)ompile, is a shell script to batch
    ;;  compile a file. We use XEmacs, the -x switch,
    ;;  if we can because it's byte compiler reports errors better
    ;;  than Emacs.
    ;;
    ("lisp"   "emc -x")
    ("java"   "javac")
    )
  "*Compilation table, how to run the code through interpreters.
format of table is
  '((REGEXP-for-buffer-type  COMPILE-COMMAND))
    (R C)
    ..
   )
The regexp is like 'code-c' 'code-pascal' or alternatively a
`mode-name' if buffer content can't be identified. See tinybufid.el
for more."
  :type '(retpeat
	  (string :tag "Regexp")
	  (string :tag "Shell command")
	  )

  :group 'TinyMy
  )

;;; ................................................ &vu-periodic-save ...

(defcustom timy-:save-buffer-modes  '(rmail-mode)
  "*Buffers that have this mode on are periodically saved."
  :type '(repeat function) :group 'TinyMy
  )

(defcustom timy-:save-buffer-regexp  nil
  "*Buffers whose `buffer-file-name' match this regexp are periodically saved."
  :type '(string :tag "Regexp") :group 'TinyMy
  )


;;; ........................................................ &vu-stamp ...

(defcustom timy-:stamp-lcd-regexp (concat "[|,]" user-mail-address "[|]")
  "*When you save package.el, the LCD entry's time is stamped too.
But only if this regexp is found from the file. This regexp should
match the first line after line

  ;; LCD Archive Entry:

See `timy-file-stamp-elisp-lcd'."
  :type '(string :tag "Regexp") :group 'TinyMy
  )


(defvar timy-:stamp-user-mail-address  user-mail-address
  "Email address that is inserted into Contactid: field.
See `timy-file-stamp'.")


;;}}}
;;{{{ setup: version

;;; ......................................................... &version ...

(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinymy.el"
 "timy"
 timy-:version-id
 "$Id: tinymy.el,v 1.51 1997/06/26 22:59:34 jaalto NotReleased $"
 '(timy-version-id
   timy-:register
   timy-:tar-command
   timy-:shar-command
   timy-:save-buffer-modes
   timy-:save-buffer-regexp
   )))


;;}}}

;;; ########################################################### &Funcs ###

;;{{{ install: main


;;; ----------------------------------------------------------------------
;;;
(defun timy-define-keys  ()
  "Install keys."

  (tinyck-maybe-activate 'xemacs)

  (define-key global-map                "%" 'timy-vi-type-paren-match)
  (define-key shared-lisp-mode-map      "%" 'timy-vi-type-paren-match)
  (define-key emacs-lisp-mode-map       "%" 'timy-vi-type-paren-match)
  (define-key lisp-mode-map             "%" 'timy-vi-type-paren-match)

  ;;  was kbd-macro-query, but I don't believe people use that much
  ;;  At least I have never needed it.
  ;;
  (global-set-key "\C-xq"    'timy-buffer-file-chmod)

  ;;  Redefine scroll keys, we don't confirm these...
  ;;
  (global-set-key [(prior)]             'timy-scroll-up)
  (global-set-key [(next)]              'timy-scroll)


  ;;  In XEmacs these already have default bindings, but we override them.
  ;;
  (global-set-key [(control right)]     'timy-forward-word)
  (global-set-key [(control left)]      'timy-backward-word)
  (global-set-key [(control up)]	'timy-beginning-of-defun)
  (global-set-key [(control down)]	'timy-end-of-defun)

  (global-set-key "\C-x\C-q"		'timy-read-only)


  ;; Use C-z prefix because it is most user friendly to pinky
  ;; Pretty useless in X-windowed emacs, and in windowed
  ;; emacs you seldom use suspend-emacs: emacs has shell mode
  ;;
  (use-prefix-key global-map "\C-z")

  ;;  Set global keys, confirm these
  ;;
  (mapcar
   (function
    (lambda (x)
      (if timy-:define-key-force
          (define-key global-map (car x) (cdr x))
        (define-key-if-free global-map
          (car x) (cdr x) 'timy-define-key-error))))
   timy-:define-key-table)

  (tinyck-maybe-activate 'xemacs 'disable)
  )


;;; ----------------------------------------------------------------------
;;;
;;;###autoload (autoload  'timy-install-files "tinymy" t t)
(ti::m-fmacro-install-pgp-tar timy-install-files "tinymy.el")

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun timy-install ()
  "Configures emacs variables and add/changed bindings."
  (interactive)


  ;;  Needed by mail lock watchdog.
  ;;
  (or (getenv "MAIL")
      (error "\
Don't know where your MAIL comes. Please setenv MAIL to directory name")
      )

  (when (ti::xe-window-system)
    (let ((ok
	   (or (eq timy-:install-select-window-auto 'yes)
	       (and
		 (eq timy-:install-select-window-auto 'ask)
		 (null
		  (y-or-n-p
		   (concat "timy: Do you use menubar? "
			   "Auto win selec. not available if you use it")))
		 )))
	   )
      (cond
       ((and ok (emacs-p))
	(defvar track-mouse nil)	;ByteComp silencer in XEmacs
	(setq track-mouse t)		;This is essential
        (global-set-key [mouse-movement] 'timy-mouse-select-window-auto)
        )
       ((and ok (xemacs-p))
	;; See also package mode-motion+.el
	;;
	;; setting mode-motion-hook it buffer local. Hm. And it is called from
	;; default-mouse-motion-handler inside save-window-excursion.
	;; Not good. I can't change window with that hook.
	;;
	;; --> we have to pull out bigger hammer.
	;;
	;;
	(defvar mouse-motion-handler nil) ;ByteComp silencer in Emacs

	(if  (eq mouse-motion-handler
		 'default-mouse-motion-handler)
	    (setq mouse-motion-handler  'timy-default-mouse-motion-handler)
	  (read-char-safe-until
	   "\
[press]Can't install auto-select: mouse-motion-handler changed."))

	(defun timy-default-mouse-motion-handler (event)
	  (prog1 (default-mouse-motion-handler event)
	    (timy-mouse-select-window-auto event)
	    ))

	))))

  ;; save buffers every 20 minutes.
  ;; To delete timer, use this command in XEmacs and Emacs
  ;;
  ;;  (ti::xe-timer-key-delete
  ;;    (ti::xe-timer-elt 'timy-save-buffers-periodically))
  ;;
  ;;
  (if (not (ti::xe-timer-elt 'timy-save-buffers-periodically))
      (run-at-time "60 sec" (* 20 60) 'timy-save-buffers-periodically))

  (if (not (ti::xe-timer-elt 'timy-revert-some-buffers))
      (run-at-time "60 sec" (* 5 60) 'timy-revert-some-buffers))


  (if (not (ti::xe-timer-elt 'timy-mail-lock-check))
      (run-at-time "60 sec" (* 5 60) 'timy-mail-lock-check))


  (add-hook 'write-file-hooks 'timy-write-file-hook)
  )


;;; ----------------------------------------------------------------------
;;;
(defun timy-define-key-error (key def)
  "Call back function. Warn about conflicting key binding for KEY and DEF."
  (error "tinymy: Cannot auto-install, key already occupied: %s %s"
         key def))


;;}}}
;;{{{ alias definitions

;;; ........................................................... &alias ...

(defun timy-alias ()
  "Install some aliases."
  ;;  I want always to use y-or-n-p.
  (defalias 'yes-or-no-p 'y-or-n-p)
  )

;;}}}

;;{{{ file: stamp

;;; ............................................................ &file ...

;;; ----------------------------------------------------------------------
;;;
(defun timy-write-file-hook  ()
  "Stamp file."
  (safe-widen
    (timy-trim-blanks)
    (timy-file-stamp)			;any file
    (timy-file-stamp-elisp-lcd)		;.el
    ))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun timy-buffer-file-chmod (&optional verb)
  "Toggle current buffer's Read-Write permission permanently on disk. VERB.
Does nothing if buffer is not visiting a file or file is not owned by us."
  (interactive)
  (let* ((file  (buffer-file-name))
	 stat
	 )
    (VERB)
    (when file
      (setq stat (ti::f-chmod-w-toggle file))
      (when verb
	(cond
	 ((eq stat 'w+)
	  (message "chmod w+")
	  (setq buffer-read-only nil)
	  )
	 ((eq stat 'w-)
	  (message "chmod w-")
	  (setq buffer-read-only t)
	  )
	 (t
	  (message "couldn't chmod")
	  ))
	(ti::xe-modeline-update)
	))
    ))



;;; ----------------------------------------------------------------------
;;;
(defun timy-file-stamp ()
  "Help tracking document change.
Your `save-buffer' function now stamps following
fields every time you save your document. If they don't exist, it
behaves exactly like normal `save-buffer'.

  $Docid\:$ 		the date/time and full user name.
  $Contactid\:$		the tis-contactid-info string.

If there are multiple stamps, only the first one met, starting from
the beginning of file, will be touched. This approach has been chosen
so that it would take least time possible to do stamping before file
is saved.
  You might wonder why there is $ marks surrounding the words. The
format for revision control system 'GNU rcs' requires $ marks if you
run their program ident\(1\) over the file. Ident detects only rigidly
formatted Directives  so that any keyword of the following form can be
printed out.

  $keyword:<space>.......desc text<space>$

You cannot modify contents of these fields by hand, because function
kills old stamping and writes new ones every time it's called. It is
also recommended that you _put_ directive '@(#)' at the beginning
of these stamps; that way the UNIX 'what'-command can display these
fields.
  If there were any errors during stamping the user will be notified.
More information is available in tinystamp.el where this function
belongs."
  (interactive)
  (let* ((case-fold-search  nil)

	 (re1"[$]Docid")		;optimize search
	 (re1-1  ":\\([^'\"\\$]+\\)[$]")

	 (re2    "[$]Contactid" )
	 (re2-1  ":\\([^'\"\\$]+\\)[$]")

	 (time   (ti::d-standard-date))
	 (user   (user-full-name))
	 (doc1   (concat " " time " " user " "))
	 (doc2   (concat " <" timy-:stamp-user-mail-address "> "))

	 (no-re  "RMAIL")		;do not touch this
	 )
;;;    (setq R1 re1 R11 re1-1 R2 re2 R22 re2-1)
    (cond
     ((not (string-match no-re (buffer-name)))
      (safe-widen
       (save-excursion
	 (PMIN)				; case 1


	 (if (and (re-search-forward re1 nil t)
		  (looking-at re1-1))
	     (kill-match 1 doc1))

	 (PMIN)				; case 2
	 (if (and (re-search-forward re2 nil t)
		  (looking-at re2-1))
	     (kill-match 1 doc2))
	 ))))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun timy-file-stamp-elisp-lcd ()
  "Updated .el file's LCD entry for file if it's owned by user.
See `timy-:stamp-lcd-regexp'."
  (let* ((email-re    timy-:stamp-lcd-regexp)
	 (lcd-re      "^;; +LCD.*entry:")
	 list
	 date
	 )
    (cond
     ((string-match "\\.el" (or (buffer-name) ""))
      (save-excursion
	(PMIN)

	;;  Try to find LCD indication and my mail address.
	;;
	(cond
	 ((and (re-search-forward lcd-re nil t)
	       (re-search-forward email-re nil t))
	  (setq list (ti::c-time-list 'zero-form))
	  (setq date (concat (nth 0 list) "-" (nth 5 list) "-" (nth 6 list)))

	  ;; Now replace the contents
	  (and (re-search-forward "; \\([0-9].*9[4-9]\\)[|]" nil t)
	       (kill-match 1 date))
	  )))))
    ))



;;}}}

;;{{{ buffers

;;; ......................................................... &buffers ...


;;; ----------------------------------------------------------------------
;;;
(defun timy-read-only ()
  "Puts buffer in `view-mode' if read-only is turned on.

If file is vc controlled; then confirm vc call.
Normally \\[toggle-read-only] would do CheckOut if the file was read-only,
but sometimes we just want to put buffer to read-only
state to prevent changing anything in there for a while."
  (interactive)
  (let* ((key-func (if (featurep 'vc)
		       'vc-toggle-read-only
		     'toggle-read-only))
	 state
	 call
	 turn-mode
	 )
    (cond
     ((memq major-mode '(dired-mode))	;plain C-x C-q for these modes...
      (toggle-read-only))
     (t
      (cond
       ((and (eq key-func 'vc-toggle-read-only)
	    buffer-file-name		;maybe *temp* buffer ?
	    (vc-name buffer-file-name)) ;is file registered ?

	(if (y-or-n-p "Really call vc ?")
	    (call-interactively 'vc-toggle-read-only)
	      (toggle-read-only))
	)
      (t
       (call-interactively key-func)
       ))

      (setq state buffer-read-only)	;what happened ?
      (setq turn-mode			;can't use nil, because it toggles
	   (if  state 1 0))

      (setq call (timy-view turn-mode))

      (if (and (eq 'mview-mode call)
	      (eq state nil))
	 ;;  It's too smart.. the vc-toggle-read-only already turned on
	 ;;  the read-only and now mview turned itself off, while retaining
	 ;;  the *original* read-only status ... force it back to basics
	 (setq buffer-read-only nil)
       )))))



;;; ----------------------------------------------------------------------
;;;
(defun timy-view (&optional arg)
  "Pick the right view mode using mode ARG. Let it be a minor one if possible.
This is emacs version independent function.

Note:

    Some minor modes occupy view key, eg. expand.el Those interfering modes
    are also turned off.

Returns;
  SYMBOL     ,which mode turned on."
  (interactive "P")
  (let* (ret
	 )

    ;;  In emacs 19.30+ the std view.el is minor mode, but not in lower ones..
    ;;
    (cond
     ((assq 'view-mode minor-mode-alist);; is this minor mode?

      (view-mode arg)			;use the new minor mode
      (when (and timy-:package-exist-mview ; turn this off.
		 (featurep 'mview)
		 (symbol-value 'mview-mode))
	;;  No autoloads allowed
	(require 'mview)
	(sfuncall 'mview-mode 0)
	)

      (setq ret 'view-mode-minor)
      )

     (timy-:package-exist-mview	;get this from OHIO mview.el
      (require 'mview)
      (sfuncall 'mview-mode arg)
      (when view-mode			;make sure this is off
	(setq view-mode 0))

      (when (eq major-mode 'view-mode)
	(view-exit))

      (setq ret 'mview-mode)
      )
     (t					;no other choice, use the major-mode
      (if (eq major-mode 'view-mode)
	  (view-exit)
	(view-buffer))
      (setq ret 'view-mode-major)
      ))

    (if (and (featurep 'expand)		;Turn this mod off, occupies SPC key
	     (symbol-value 'expand-mode))
	(sfuncall 'expand-mode 0))

    ret
    ))


;;; ----------------------------------------------------------------------
;;;  07 Jun 1996, Bart Robinson <lomew@cs.utah.edu>, gnu.emacs.sources
;;;  13 Jun 1996, Slightly modified by [jari]
;;;  27 Jan 1997  duthen@cegelec-red.fr (Jacques Prestataire),
;;;               wrote the current 3 buffer revert functions.
;;;
(defvar timy-:force-revert nil "No docs.")
(defun timy-revert-buffer-prompter (buffer)
  "Called from `timy-revert-some-buffers' with BUFFER. Confirm revert."
  (save-window-excursion
    (if (and (if (stringp buffer)       ;Handle buffer STRING or PTR
                 (get-buffer buffer)
               (and buffer
                    (buffer-name buffer)))

	     ;; Skip this time if user has prompt in minibuffer
	     ;; 3.2.1997 , this is not a good idea, user uses too much
	     ;; minibuffer...
	     ;;
;;;	     (no-action-in-progress-p)

             (progn                     ;Buffer FILE check
               (set-buffer buffer)
               (and (buffer-file-name)
                    (file-exists-p (buffer-file-name))
                    (not (verify-visited-file-modtime buffer))
                    ))

             ;;  Make sure buffer in emacs is not modified, We don't
             ;;  want to wipe out changes user did
             ;;
             (not (buffer-modified-p))
             )

        (if timy-:force-revert
            ;;  dynamic var bound in `timy-revert-some-buffers'
            ;;  non-nil means act on that object without asking the user
            ;;
            'timy-force-revert
          ;; a string means use it as the prompt for that object
          ;;
          (format "Revert buffer %s " (buffer-name buffer)))
      ;;  nil means ignore that object
      ;;
      nil
      )))

;;; ----------------------------------------------------------------------
;;;
(defun timy-revert-buffer-actor (buffer)
  "Called from `timy-revert-some-buffers' with BUFFER."
  (save-window-excursion
    (set-buffer buffer)
    (revert-buffer nil 'no-confirm)
    (set-visited-file-modtime)))


;;; ----------------------------------------------------------------------
;;;
(defun timy-revert-some-buffers (&optional force-revert)
  "Revert some buffers whose underlying file has changed underneath them.
Asks user about each one.  Optional arg FORCE-REVERT (interactive prefix
arg) means to suppress any questions.

Does not touch changed buffers."
  (interactive "P")

  (setq timy-:force-revert force-revert)

  (map-y-or-n-p
   'timy-revert-buffer-prompter
   'timy-revert-buffer-actor
   (buffer-list)
   '("buffer" "buffers" "revert")
   ))



;;}}}
;;{{{ Mouse

;;; ........................................................... &mouse ...


;;; ----------------------------------------------------------------------
;;;
;;; 4 Mar 1996, kevinr@ihs.com (Kevin Rodgers), gnu.emacs.help
;;; 8 Mar 1996, jaalto@tre.tele.nokia.com (Jari Aalto) -- rewritten
;;;             and Added XEmacs support
;;;
;;;
;;; Terry E Eminhizer (teminhi@hubcap.clemson.edu) wrote:
;;; >How can I get the selected window to change as I move the mouse cursor
;;; >into that window?  In other words, I don't want to have to click the
;;; >mouse in the new window every time I move between windows (windows, not
;;; >frames, this is not a click-to-focus window manager question).
;;;
;;;

(defvar timy-:window-previous nil
  "Used in `timy-mouse-select-window-auto'.")

(defun timy-mouse-select-window-auto (event)
  "*Like `mouse-select-window' with EVENT but don't generate an error."
  (interactive "e")
  (let* ((case-fold-search      t)
	 win
	 mini
	 )
    (if (emacs-p)
	(setq win (posn-window (event-start event)))
      ;;  XEmacs calls us from motion hook
      ;;  #<motion-event 644, 221>
      ;;
      (setq win (sfuncall 'event-window event)))

    ;;  The WIN could be frame pointer too, that's why we check it.
    ;;
    (cond
     ((and (windowp win)
           (window-live-p win)
           ;;  Motion in same window as previously, forget this event
           ;;
           (not (eq timy-:window-previous win))
           )
      (setq timy-:window-previous win
	    mini  (window-minibuffer-p win))

      ;;    1. We select window if it's not minibuffer
      ;;    2. if it's minibuffer, select it _only_ if it's active
      ;;       otherwise emacs cries.
      ;;
      (when (or (not mini)
		(and mini
		     (minibuffer-window-active-p win)
		     ))
	(select-window win)
	)))
    ))

;;}}}

;;{{{ buffers; revert buffers

;;; ----------------------------------------------------------------------
;;;
(defun timy-flip-windows  ()
  "Switch window order. There must be only 2 windows."
  (interactive)
  (let* ((list (ti::w-list))
         w1 w2 b1 b2
         )
    (if (not (eq 2 (length list)))
        (message "timy: There is not exactly 2 windows.")
      (setq w1 (nth 0 list)
            w2 (nth 1 list)
            b1 (window-buffer w1)
            b2 (window-buffer w2)
            )
      (select-window w1)
      (switch-to-buffer b2)
      (select-window w2)
      (switch-to-buffer b1)
      )
    ))


;;; ----------------------------------------------------------------------
;;; 25 Mar 1995, Michael Cook <mcook@cognex.com>
;;; - This is very handy function, consider the following:
;;;    M-x shell     --> *shell*
;;;    M-x shell     --> doesn't give you new shell, but puts you in *shell*
;;;
;;;    Instead you should have done this:
;;;    M-x shell                        --> *shell*
;;;    M-x my-toggle-buffer-name        --> *shell*<2>
;;;    M-x shell                        --> *shell*
;;;
;;;    Now you have 2 shell buffers and you can very easily get more.
;;;    The same is true for mail sending, just rename the buffer with
;;;    this command if you decide you need to send another
;;;    mail message while you're in *mail*
;;;
(defun timy-toggle-buffer-name ()
  "If the current buffer is named like `foo<2>', rename it to `foo'.
If there is already a buffer `foo', first rename it like `foo<3>'."
  (interactive)
  (let ((i (string-match ".<[1-9][0-9]*>$" (buffer-name))))
    (if (null i)
        (rename-uniquely)
      (let* ((name (substring (buffer-name) 0 (1+ i)))
             (buf (get-buffer name)))
        (if buf
            (save-excursion
              (set-buffer buf)
              (rename-uniquely)))
        (rename-buffer name)))))



;;}}}
;;{{{ compilation

;;; ..................................................... &compilation ...

;;; ----------------------------------------------------------------------
;;;
(defun timy-compile-run-command ()
  "Run current buffer through compile buffer.
For C/C++ code the default command suggested if 'mak', but if there
is no makefile in the directory, then a normal compile command
is proposed.

References:
  `timy-:compile-table'
  `timy-:compile-command-c-code'"
  (interactive)

  (let* (;;  We change this so that compile goes to right dir
         (file      (buffer-file-name))

         (default-directory
           (if file
               (file-name-directory file)
             default-directory))

         (table     timy-:compile-table)
         (cc-cmd    timy-:compile-command-c-code)
         (type      (or (tib-info nil 'variable)
                        (symbol-name major-mode)))


         elt
         run-it
         mak-files
         filename                       ;without directory part
         cmd
         )
    (if (null file)
        (message "Sorry, no file in this buffer.")

      (setq filename (file-name-nondirectory file))
      (setq elt (list-find table type))
      (if elt
          (setq cmd (format "%s %s" (nth 1 elt) file))
        ;;  Some special compile commands for C/C++, which usually
        ;;  have .mak files
        ;;
        (cond
         ((string-match "c[+]+\\|^cc?-\\|code-c" type)
          ;;  "mak" is csh script to run all make files in the directory.
          ;;
          (setq cmd "mak")

          ;;  Check if there are any .mak files in directory ?
          (setq mak-files
                (ti::f-dir-files
                 (file-name-directory file)
                 "\\.mak|makefile|Makefile"
                 ))

          (if (null mak-files)          ;None found ?
              (setq cmd
                    (format "%s %s -o %s "
                            cc-cmd
                            filename
                            ;;  Drop extension
                            (strmatget "^[^.]+" 0 filename)
                            )))
          ))
        ))


    (setq run-it (read-from-minibuffer
                  "Compile: " (or cmd "mak")
                  nil nil 'compile-history
                  ))


    (unless (var-nil-p run-it)  ;Command given ?
      (compile-internal run-it "No more errors.")
      (pop-to-buffer "*compilation*")
      )

    ))

;;}}}
;;{{{ elisp: package saving from mail, gnus

;;; ----------------------------------------------------------------------
;;; - Imagine that you're reading gnu.emacs.sources and want to get
;;;   that package in the post.
;;; - Or you receive a package in private mail message...
;;; - This does the job of saving that package to file very easily!
;;;
(defun timy-lisp-package-save-to-file (dir &optional zip)
  "Save attached lisp code package in current buffer to DIR.
If  DIR is nil nil, use ~/elisp/ by default. Remember to have ending slash
in DIR name.

The file is supposed to have special heading + file's end is at `point-max'.

    ;; file.el -- description

If function can't find

   'End of file' or '&eof' or 'XXX ends here'

identifier, it'll add one, before writing a file.


If optional ZIP argument is string \".Z\" or \".gz\" the file is
compressed when saved."
  (interactive "DSave to dir: \nsZip extension: ")
  (let* ((dir           (or dir
                            (expand-file-name "~/elisp/")))
         (to-buf        (get-buffer-create "*tmp*"))
         (code-buf      (current-buffer))

         ;;  - the file start and it's name
         ;;  - The regexp will jump until there is a-zA-Z0-9
         ;;
         (re1           "^;;+[ \t]+[^a-zA-Z0-9_]+\\(.*\\.el\\) +[-][-]+")
         (re2           "^;;+[ \t]+\\(.*\\.el\\) +[-][-]+")
         (re3           "^;;+[ \t]+\\(.*\\) +[-][-]+")
         (end           (concat
                         "^;+[ \t]+\\(end[ \t]+of[ \t]file\\|^;.*&eof\\)"
                         "\\|^;;+[ \t]+.*ends here"
                         ))
         end-str
         END
         file
         p1 p2
         do-it
         dfile
         ans
         )

    (if (not (file-directory-p dir))
        (error "Not a valid directory."))


    (save-excursion
      (PMIN)

      ;;  See if we can detect the package name in this buffer
      ;;
      (if (null (or (re-search-forward re1 nil t)
                    (re-search-forward re2 nil t)
                    (re-search-forward re3 nil t)
                    ))
          (message "No proper header found.")


        (setq file (read-match 1))
        (setq file (ti::s-verify-ends file "\\.el" ".el"))


        (setq end-str (format "\n;;; %s ends here\n"  file))

        (setq p1 (BOLP)  p2 (point-max) )

        ;;  Confirm that we got right file name.
        (setq file (read-from-minibuffer "File grabbed: " file))
        (setq do-it (not (var-nil-p file)))

        (setq dfile (concat dir file))


        ;; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ existing file ^^^
        (when do-it			;handle file with same name
          (cond
           ((file-exists-p dfile)
            (setq ans (read-from-minibuffer "overwrite?: " dfile))

            (cond
             ((string= ans dfile)
              (delete-file dfile))
             ((var-nil-p ans)           ;clear all, cancell operation
              (setq do-it nil))
             (t
              (setq dfile ans)))

            )))

        (when do-it
          ;; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ end check ^^^
          ;;  Check if it has proper ending ?
          (if (re-search-forward end nil t)
              (progn
                (setq p2 (1+ (EOLP)))   ;include newline
                )
            (message "Hm, No proper ending here..using point-max ")
            (sit-for 1)
            (setq END t)
	    )

          ;; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ moving ^^^
          (set-buffer to-buf)
          (erase-buffer)
          (insert-buffer-substring code-buf p1 p2)
          (when END                         ;should we end the file ?
            (PMAX)
            (insert end-str))

          (PMIN)

          ;; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ zip file ^^^
          (when (stringp zip)
            (require 'jka-compr)
            (setq dfile (concat dfile zip)))

          ;; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ writing ^^^
          (write-file dfile)            ;jka handles compressing
          (not-modified)
          (kill-buffer (current-buffer))
          (message (concat "Wrote: " dfile))
          )))
    ))

;;}}}
;;{{{ file

;;; ............................................................ &file ...

;;; ----------------------------------------------------------------------
;;;
(defun timy-trim-blanks ()
  "Delete trailing blanks from all lines; including lines from end of buffer."
  (interactive)
  (save-excursion
    (unless buffer-read-only
      (ti::b-trim-blanks (point-min) (point-max))

      ;;  Now delete extra lines from the end of buffer
      ;;
      (goto-char (point-max))
      (when (re-search-backward "^.*[^ \t\n].*$" nil t)
	(forward-line 1)
	(if (not (eq (point) (point-max)))
	    (ti::b-cnv-empty-lines (point) (point-max)))
	)))
  (if (interactive-p)
      (message "Blanks trimmed"))
  nil					;Clean return code
  )

;;; ----------------------------------------------------------------------
;;; - Especially when I'm making diff to the Author I find this
;;;   very usefull.
;;;
;;;
(defun timy-copy-file  (file1 file2 &optional arg)
  "Make copy of current buffer FILE1 to FILE2 (FILE1.orig or FILE1.VER).
Function tries to find possible RCS version.
You usually make backup if you make a change and send diff to author.

If you supply PREFIX ARG, then

  C - u	remove the copy files; namely,     (buffer-file-name).*
  nbr   Copy back: this like doing
        FILE.VER  --> FILE
        FILE.orig --> FILE

        If you had made a safe copy previously, this restores
	the safe copy to original file."
  (interactive
   (let* ((ver   (or (ti::m-rcs-buffer-version)
		     ;;  No rcs string found, then try Regular lisp pacakge
		     ;;  syntax.
		     ;;
		     ;;  Version:       2.37
		     (re-search-check
		      "^;+[ \t]+Version:[ \t]*\\([0-9.]+\\)" 1 nil 'read)
		     ))

          (file1 (or (buffer-file-name)
		     (error "Buffer has no filename")
		     ))
          (ext   (if ver (concat "." ver) ".orig"))
          file2
          )
     (if current-prefix-arg
	 (list file1 nil current-prefix-arg)
       (setq file2 (read-from-minibuffer "Make copy to: " (concat file1 ext)))
       (list file1 file2)
       )))

  (let* ((re	    (format "^%s\\." (file-name-nondirectory file1)))
	 (file-list (ti::f-dir-files
		    (file-name-directory file1)
		    re 'abs t))
	 )
    (cond
     ((null arg)
      (cond
       ((or (not (file-exists-p file2))
	    (and (file-exists-p file2)
		 (y-or-n-p (format "%s exists. Remove? " file2))
		 (progn (delete-file file2) t)
		 ))
	(file-delete-safe file2)
	(copy-file file1 file2)
	(message "Safe copy done.")
	)
       (t
	(message "timy: Sorry; can decide how to do the copying.")
	)))

     ((equal arg '(4))
      (if (null file-list)
	  (message "timy: There is no safe copy files matching %s" re)
	(list-loop file-list file1
	  (if (y-or-n-p (format "Delete %s ? " file1))
	      (delete-file file1)))
	))

     ((integerp arg)
      (cond
       ((null file-list)
	(message "timy: There is no safe copy for %s" file1))

       ((eq 1 (length file-list))
	(when (y-or-n-p
	       (message "Found safe copy %s; copy it over original? "))
	  (delete-file file1)		;copy-file barfs otherwise
	  (copy-file (car file-list) file1)
	  (message "Safe copy restored.")
	  ))

       ((> (length file-list) 1)
	(setq file2
	      (completing-read
	       "Don't know which one to use as source, complete: "
	       (list-to-assoc-menu (mapcar 'file-name-nondirectory file-list))
	       nil 'must-match))
	(setq file2 (concat (file-name-directory file1) file2))
	(delete-file file1)
	(copy-file file2 file1)
	(message "Safe copy restored: %s --> %s"
		 (file-name-nondirectory file2 )
		 (file-name-nondirectory file1)
		 )
	)))
     ) ;; cond end

    ))

;;}}}

;;{{{ key: % matching


;;; ----------------------------------------------------------------------
;;; idea                by michael@pbinfo.UUCP  or  michael@uni-paderborn.de
;;; Totally rewritten   by Jari Aalto <jalto@tre.tele.nokia.fi> Feb 1996
;;;
;;; All the posts so far in the internet to make the "%" match parens
;;; right in every possible _mode_ failed. That's why I started writing
;;; my own function, which you see here.
;;;
(defun timy-vi-type-paren-match (&optional arg)
  "Match engine: find {[( or )]} pairs. ARG is character repeat count.
See also 'timy-:vi-type-paren-match-special-list

Note:

  If the parens are 'closed', ie. there is nothing between them
  them, act like self insert command.

    ()  ->          (%)
    *   pressing % inserts literal, does not try to match.

   The characters must be pairs; eg. (] does not full fill similar condition.


'"
  (interactive "P")
  (let* ((p	    (point))
	 (ch	    (following-char))
	 (ch-next   (ti::b-read-char nil   1))
	 (ch-prev   (ti::b-read-char 'back 0))
	 (ptable    (syntax-table))	;previous, the original
	 (table	    (copy-syntax-table lisp-mode-syntax-table))
	 (pairs	    timy-:vi-type-paren-match-list)
         (spread-limit   (* 10 60))	;approx 10 lines of code.
         (m-list    timy-:vi-type-paren-match-special-list) ;match list

         s-func add-func max-func bigger-func
	 elt
         go
         max
         )

    (catch 'terminate

      ;;  check if the parens are "closed", ie. there is nothing beween them
      (when (or
	     (and (setq elt (assoc (char-to-string ch) pairs)) ;; Left
		  (string=  (char-to-string ch-next) (cdr elt)))

	     (and (setq elt (rassoc (char-to-string ch) pairs))  ;; Right
		  (string=  (char-to-string ch-prev) (car elt)))

	     )
	(self-insert-command (or arg 1))
	(throw 'terminate t)
	)


      (mapcar
       (function
	(lambda (x)
	  (modify-syntax-entry (string-to-char (car x)) "(" table)
	  (modify-syntax-entry (string-to-char (cdr x)) ")" table)
	  ))
       pairs)

      (set-syntax-table table)

      (unwind-protect
	  (condition-case nil
	      (cond

	       ;; ........................................ handle quotes ...
	       ((char-in-list-case ch m-list)
		(setq ch (regexp-quote (char-to-string ch)))

		(setq s-func   're-search-forward
		      add-func '+
		      max-func 'point-max
		      bigger-func '>
		      )

		(cond
		 ((looking-at (concat ch "[ \t]*$"))
		  ;; Only search backward if no ARG given.
		  ;; if the " char is at the end of line,
		  ;; then it propably is the 'closing' one.
		  ;;
		  (if (null arg)
		      (setq s-func   're-search-backward
			    add-func '-
			    max-func 'point-min
			    bigger-func '<
			    ))
		  ))

		;; Do not go too far away....
		;;
		(setq go (funcall add-func p  spread-limit))
		(setq max (funcall max-func))
		(setq max
		      (if (funcall bigger-func go max)
			  max go))

		;; ... ... ... ... ... ... ... ... ... ... ... do search . .
		(if (eq s-func 're-search-forward)
		    (forward-char 1))	;move out of way

		(funcall s-func ch go t)

		;; This is funny, it both a) restores the position
		;; if search failed, b) adjusts the "after" search
		;; point back to char.
		;;
		(if (eq s-func 're-search-forward)
		    (forward-char -1))
		)

	       ;; ..................................... handle BEG pairs ...
	       ((looking-at "[[({<]")
		(forward-sexp 1)
		(backward-char)

		(cond ((not (looking-at "[])}>]"))
		       (error "..booomerang")))
		)

	       ;; ..................................... handle END pairs ...
	       ((looking-at "[])}>]")
		(forward-char 1)
		(forward-sexp -1)

		(when (not (eq p (point))) ;moved ?
		  ;;  In lisp, jumping from closing ) to starting
		  ;;  "'(lambda" puts cursor at "'"?? Correct it.
		  (if (and (not (eq (following-char) ?\( ))
			   (looking-at ".[]({<]"))
		      (forward-char 1))
		  ))

	       ;; ...................................... no special char ...
	       (t
		(self-insert-command (or arg 1))
		))
	    (error
	     (goto-char p)		;restore position
	     (message "No match."))
	    )
        ;; make sure we restore this
        (set-syntax-table ptable))
      )))



;;; ----------------------------------------------------------------------
;;;
;;; 22 Jan 1997, Fabio Somenzi <Fabio@colorado.edu>, gnu.emacs.help
;;; The use of "my-last-command" can probably be avoided by using
;;; "this-command", but I haven't had the time to test the modification.
;;;

(defvar timy-:last-command 'ignore
  "Saved last command.
Equals `last-commad' unless last-command's value is
'repeat-last-simple-command.  Used by repeat-last-simple-command to
avoid infinite recursion. The initial value is just a no-op.")

(defun timy-repeat-last-simple-command (arg)
  "Repeat the last simple command. See ARG below.

Simple here means that the COMMAND TAKES NO ARGUMENTS OR PREFIX ARGUMENTS.
Useful for cases like C - c @ C - n (next section in outline minor mode).

Two cases must be distinguished for functions
that take a prefix argument: the argument may be optional or not. If the
argument is 1 (default) repeat-last-simple-command first tries not to pass
it to the command to be executed. This works if the argument is optional.
Otherwise an error is generated. The error is handled by calling the command
again---this time with the argument."
  (interactive "p")
  (if (not (or
            (equal last-command 'timy-repeat-last-simple-command)
            (equal last-command nil)))
      (setq timy-:last-command last-command))
  (if (/= arg 1)
      (funcall timy-:last-command arg)
    (condition-case nil
        (funcall timy-:last-command)
      (error (funcall timy-:last-command arg)))))



;;}}}
;;{{{ mail

;;; ............................................................ &mail ...


;;; ----------------------------------------------------------------------
;;;
(defun timy-mail-subject-get  ()
  "Look buffer content and return subject for mail message.

diff buffer:

  'context diff file.txt'

rcsdiff buffer:

  'diff 1.23 --> 1.25 file.txt'

Regular buffer:

   '1.25 file.txt'       ;; possibly without version information


buffer with no filename:

   nil"
  (let* ((rcs-re    "retrieving revision +\\(.*\\)")
	 (v1	    "")
	 (v2	    "")
	 type
	 ver
	 file
	 message
         )
    (save-excursion
      (PMIN)
      (cond
       ;;  See if this is rcsdiff
       ;;
       ;;  RCS file: RCS/tinylib.el,v
       ;;  retrieving revision 1.95
       ;;  retrieving revision 1.97
       ;;  diff -c -r1.95 -r1.97
       ;;  *** 1.95	1997/03/22 12:26:59
       ;;  --- 1.97	1997/03/22 15:17:22
       ;;
       ((re-search-forward "^RCS file:[^/]*/?\\(.*\\),v" nil t)
	(setq file (read-match 1))
	(and (re-search-forward rcs-re nil t)
	     (setq v1 (read-match 1))
	     (re-search-forward rcs-re nil t)
	     (setq v2 (read-match 1))
	     )
	(setq message (format "diff %s --> %s %s" v1 v2 file))
	)
       (buffer-file-name
	;;  Regular file, see if this one has RCS version information
	;;
	(if (setq ver (ti::m-rcs-buffer-version))
	    (setq ver (concat " " ver " ")))

	(setq message (concat
		       (or ver "")
		       (file-name-nondirectory buffer-file-name)
		       ))
	)

       ((setq type (ti::b-diff-type-p))
	;; *** /users/jaalto/T.orig  Sun Mar 23 16:37:43 1997
	;; --- /users/jaalto/T	     Sat Mar 22 14:44:34 1997
	;;
	(save-excursion
	  (PMIN)
	  (if (or (re-search-forward "^--- \\([^ \t\n]+\\)" nil t)
		  (re-search-forward "^\\*\\*\\* \\([^ \t\n]+\\)" nil t)
		  (re-search-forward "^\\+\\+\\+ \\([^ \t\n]+\\)" nil t)
		  )
	      (setq file (read-match 1))))
	(setq message (format "%s diff %s"
			      (prin1-to-string (car type))
			      (if file
				  (file-name-nondirectory file )
				"")))
	)

       ))
    message
    ))


;;; ----------------------------------------------------------------------
;;; Sep 9 1995 <jari.aalto@poboxes.com> (Finland) Public domain.
;;;
(defun timy-mail-buffer ()
  "Mail current buffer.
The subject line is constructed by looking at the buffer content:
eg if buffer contains rcsdiff of diff,
The subject line will tell the versions."
  (interactive)
  (let* ((data-buffer   (current-buffer))
	 subj
	 )
    (setq subj (timy-mail-subject-get))

    (mail)
    (pop-to-buffer "*mail*")


    ;;  This package gives nice alias expansion
    ;;
    (cond
     ((emacs-p)
      (require 'mailabbrev)
      (sfuncall 'mail-abbrevs-setup))
     ((xemacs-p)
      (require 'mail-abbrevs)
      ))

    (ti::mt-text-start 'move)

    (insert "\n\n\n\n")
    (save-excursion (insert-buffer data-buffer))

    ;; Make sure the outlline/folding is opened first
    ;;
    (ti::b-xtra-open-outline)

    (if subj
	(ti::mt-kill-field "Subject:" subj))

    (PMIN)
    (end-of-line)			;"TO:" field
    (run-hooks 'timy-:mail-buffer-hook)
    ))


;;}}}

;;{{{ Programming: function bounds

;;; ----------------------------------------------------------------------
;;;
(defun timy-function-bounds  (&optional forward)
  "Find function area. Return (beg . end).
The search is first done backward, unless FORWARD is given,
to find function beginning.

Notes:
 All function start lines must be left flushed, ie. no empty spaces before
 function name declaration. Functions must start/end with left flushed
 \"{\" and \"}\".

Supported modes:
  C/C++
  perl
  awk
  lisp"
  (let* ((mode                (or (tib-info) (symbol-name major-mode)))
         (max-lines           1500)        ;rows, function cannot be bigger
         (skip-lines          1300)         ;maximum skip lines backward
         (start               (point))
         beg end
         range point
         fwd-flag
         )

    (cond
     ((string-match "lisp" mode)
      ;;  Only lisp has decent find functions
      (save-excursion
        (if forward
            (safe-do
             (end-of-defun)
             (forward-line 2)
             (setq fwd-flag t)
             ))

        (safe-do
         (beginning-of-defun)
         (setq beg (point)))

        (if (or (and beg (null forward))
                (and beg forward fwd-flag))
            (safe-do
             (end-of-defun)
             (setq end (point))))
         ))

     ((string-match "perl\\|awk" mode)
      (save-excursion
        (cond
         (forward
          (if (re-search-forward "^sub\\|^function" nil t)
              (setq fwd-flag t))))

        (cond
         ((and (or (null forward)
                   (and forward fwd-flag))
               (re-search-backward "^sub\\|^function" nil t))
          (beginning-of-line)
          (setq beg (point))

          (if (re-search-forward "^}" nil t)
              (setq end (point)))
          ))))

     ((string-match "code-c\\|c-\\|cc-\\|c[+]" mode)
      ;;  The opening block says where is function start, this is only
      ;;  possible for NEW styled programming, not K&R styled 'hanging'
      ;;
      ;;  C++: int funtion() {             ;; nope, too diffucult to detect.
      ;;  perl sub funtion   {             ;; allowed
      ;;
      (save-excursion

        (cond
         (forward
          (if (re-search-forward "^{" nil t)
              (setq fwd-flag t))))

        (cond
         ((and (or (null forward)
                   (and forward fwd-flag))
               (re-search-backward "^{" nil t)
               (re-search-backward "(" nil t) ;find parameter list beginning
               )
          ;;  There is a problem in writing the C++ funcs:
          ;;
          ;;  // Comment
          ;;  /* Comment
          ;;  */
          ;;  int
          ;;  functionName
          ;;  ( parameters
          ;;

          (beginning-of-line)
          (setq point (point))
          ;; We just search line by line backward until no comment,
          ;; or empty line
          (while (not (looking-at " *//+\\| *[*]+/\\| *[*]+ \\|^[ \t]*$"))
            (backward-line 1))

          (if (not (eq point (point)))  ;if the while loop moved.
              (forward-line 1))         ;go to func beginning.

          (setq beg (point))

          (if (re-search-forward "^}" nil t)      ;;  This is easy.
              (setq end (point)))

          ))))
     )

;;;    (d! (count-lines beg start))

    (if (and beg end
             ;; must not be too far away from current point
             (< (count-lines beg start) skip-lines))
        (setq range (count-lines beg end)))

    ;;  The return value
    (if (and range (< range max-lines))
        (cons beg end)
      nil)
    ))



;;; ----------------------------------------------------------------------
;;;
(defun timy-beginning-of-defun (&optional end-of-fun)
  "See `timy-function-bounds'. END-OF-FUN must be nil or t."
  (interactive)
  (let* ((bounds (timy-function-bounds  end-of-fun))
         (beg    (car-safe bounds))
         (end    (cdr-safe bounds))
         (point  (if end-of-fun end beg))
         )

    (if (null bounds)
        (message "Sorry, can't find function.")
      (goto-char point))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun timy-end-of-defun ()
  "See `timy-function-bounds'."
  (interactive)
  (timy-beginning-of-defun t)
  )

;;}}}

;;{{{ rectangle

;;; ............................................................ &rect ...

;;; ----------------------------------------------------------------------
;;; From Ralf Fassel <ralf@prl.philips.nl>, 22 Jun 95.
;;;
(defun timy-add-rectangle (START END &optional insert)
  "Add or Multiply columns in rectangle in START END.
With optional arg INSERT, insert the sum and product to
the current point."

  (interactive "r\nP")
  (require 'rect)
  (let ((sum 0)(rownum 0)(prod 1)(rowval 0))
    (operate-on-rectangle
        (lambda (POS BEFORE AFTER)
          (setq rownum (1+ rownum))
          (setq rowval (string-to-number (buffer-substring POS (point))))
          (setq sum  (+ sum  rowval))
          (setq prod (* prod rowval)))
        START END 't)
    (if (interactive-p)
        (message "For %d rows, sum=%f, product=%f" rownum sum prod))
    (if insert
        (insert (format "%0.2f %0.2f" sum  prod)))

   ))


;;}}}
;;{{{ scrolling


;;; ----------------------------------------------------------------------
;;;
(defun timy-scroll-up ()
  "Call `timy-scroll'."
  (interactive)
  (timy-scroll 'up))


;;; ----------------------------------------------------------------------
;;;
(defun timy-scroll (&optional up)
  "Scrolls down, optionally UP. No errors generated.
This function behaves like DOS/windows scroll commands, where cursor jumps
to the end or beginning of window first and only next scrolls. It also
keeps the cursor in the bottom or top of window according to the direction
of scroll.

Note:

  This function does not work properly if the lines in the window
  exceed the length of the window. If the current line is longer than
  window length, then normal emacs scroll command will be called."
  ;;  - the 'error' call is most disturbing if you have
  ;;  - debug-on-error t
  ;;  - This is for *interactive* only! Lisp manual forbids using scroll
  ;;    command in normal lisp code.
  (interactive "P")
  (let* ((mode   timy-:scroll-mode)
         lines
         )

    (cond
     ((line-wrap-p)
      ;; ............................................. wrapping line ...
      ;; Let emacs handle this
      ;;
      (cond
       (up
	(if (window-pmin-visible-p)
	    (PMIN)
	  (scroll-down)
	  ))
       (t
	(if (window-pmax-visible-p)
	    (PMAX)
	  (scroll-up)
	  )))
      )
     (t
      ;; ........................................... non wrapping line ...
      (cond
       (up
	(if (bobp)
	    (message "beg of buffer.")
	  (if (window-pmin-visible-p)	;if top is visible
	      (goto-char (point-min))
	    (if (or (null mode)
		    (and mode
			 (window-pmin-line-p)))
		(scroll-down))
	    (goto-char (window-start))
	    (beginning-of-line)
	    )))
       (t
	(if (eobp)
	    (message "end of buffer.")
	  (if (window-pmax-visible-p)
	      (goto-char (point-max))

	    (if (or (null mode)
		    (and mode
			 (window-pmax-line-p)))
		(progn
		  ;;   - if outline/folding mode is on, we can't determine
		  ;;     line count with count-lines function
		  ;;   - The count gives 1 extra line, check with M-x =
		  ;;     around the window region
		  ;;
		  (setq lines
			(1-
			 (count-char-in-region
			  (window-start) (window-end) ?\n)))

		  ;; the scroll command does not update window points
		  ;; in 19.28! That means that the function window-end
		  ;; can't be trusted. only when this fucntion ends,
		  ;; the window is updated.
		  ;;
		  (scroll-up)

		  ;;  We must manually go to the end line
		  ;;  - The cursor is always left in the line 2, after
		  ;;    scrolling in window. We have to go N lines downward
		  ;;    to put cursor at window end line
		  ;;  - next-line is used, because it hanbdles folding/outline.
		  ;;    forward-line can't be used.
		  ;;
		  ;;  Note:
		  ;;  - If the lines are longer than window-width; then this
		  ;;    whole next-line call may end anywhere...can't help
		  ;;    that
		  ;;
		  (next-line (- lines 1))
		  )
	      (goto-char (window-pmax-line-bol))
	      )))
	))))
    ))


;;}}}
;;{{{ shell -- shar, tar, uu

;;; ----------------------------------------------------------------------
;;; - You usually
;;;   a. Make a tar file
;;;   b. Possibly compress it
;;;   c. --> Now you call this function to UU he package and send it via mail
;;;
(defun timy-uu-make  (file compress-ext)
  "Make uu FILE and insert it into current point.
COMPRESS-EXT can be any valid jka-compr recognised compression
extension (like .gz)"
  (interactive "ffile to uu: \nsCompress extension: ")
  (let* ((buffer (current-buffer))
         point count
         )

    (setq point (point))
    (insert "\n\n")

    (ti::p-uu-make
     file compress-ext
     'insert buffer 'verb)

    (setq count (- (point-max) point))
    (save-excursion
      (goto-char point)
      (insert (format "%sUU package approx. %d k"
                      (if (var-nil-p compress-ext)
                          "" "Compressed ")
                      (/ count 1000))))
    (if (ti::p-tar-view-command file)
        (insert "\n"))

    ))



;;; ----------------------------------------------------------------------
;;;
(defun timy-shar (single-or-list)
  "Generate SHAR file using SINGLE-OR-LIST.
List of  files can include shell regexps. The result is put into
`timy-:register'."
  (interactive
   (let* (arg1
          )
     (setq
      arg1
      (ti::f-complete-filename-minibuffer
        (read-from-minibuffer
         (format "[%s] Shar files: " default-directory)
         nil map
         )))
     (list arg1)
     ))

  (let* ((cmd         (concat timy-:shar-command " "))
         (register    timy-:register)
         (verb        (interactive-p))
         out
         )
    (if (var-nil-p single-or-list)
        (error "Missing args")
      (setq out
            (exec-to-string
             (format "cd %s; %s %s"
                     default-directory
                     cmd
                     (list-to-string (list-make single-or-list))
                     )))
      (set-register register out)
      (if verb
          (message (format "Registed %s has shar"
                           (char-to-string register))))
      )))


;;; ----------------------------------------------------------------------
;;; - This is handy if you want to send 2-5 files easily with email.
;;;
;;; 1. Create tar with this
;;; 2. Read the tar (UU + .gz ) into buffer with command timy-uu-make
;;;
(defun timy-tar (tar-file file-list)
  "Generate TAR-FILE using FILE-LIST.
Return:
 t or nil       ,if tar created."
  (interactive
   (let* ((default-directory    default-directory)
          (default-tar-name     "pkg.tar")
          (default-tar          (concat default-directory default-tar-name))
          arg1 arg2
          tar-dir
          )

     (setq
      arg1
      (ti::f-complete-filename-minibuffer
        (read-from-minibuffer
         (format "[%s] Tar name: " default-tar)
         nil map
         )))

     (cond
      ((var-nil-p arg1)
       (setq arg1 default-tar))

     ((file-directory-p arg1)
      (setq arg1 (concat arg1 default-tar))
      ))

     (setq tar-dir (or (file-name-directory arg1)
                       default-directory))
     (setq default-directory tar-dir)

     (setq
      arg2
      (ti::f-complete-filename-minibuffer
        (read-from-minibuffer
         (format "[%s] Files: " arg1)
         nil map
         )))

     (list arg1 arg2)
     ))

  ;; ... ... ... ... ... ... ... ... ... ... ... ... ... ... ...  main . .
  ;;
  (if (or (var-nil-p tar-file)
          (var-nil-p file-list))
      (error "Missing args"))

  (let* ((tar-cmd       (concat timy-:tar-command " "))
         (edir          (file-name-directory
                         (expand-file-name tar-file)))
         (cmd           (concat "cd " edir "; "
                                tar-cmd
                                (file-name-nondirectory tar-file)
                                " " file-list))
         ret
         )
    ;; ................................................... then case ...
    (if (and (file-exists-p edir)                  ;; must exist
             (or (not  (file-exists-p tar-file))   ;; good if not exist
                 (and  (file-exists-p tar-file)    ;; we have to remove it
                       (y-or-n-p "Tar exists, remove ? ")
                       (progn (delete-file tar-file) t)
                       )))
               (progn
                 (setq cmd (read-from-minibuffer "cmd: " cmd))
                 (shell-execute cmd)
                 (setq ret (file-exists-p tar-file))
                 )
        (message "Aborted"))

    (when (and (interactive-p)
               (y-or-n-p
                (format "Insert uuencoded %s to point " tar-file))
               (setq cmd (read-from-minibuffer "Compress ext: "))
               )
      (timy-uu-make tar-file cmd))

    ret
    ))

;;}}}

;;{{{ timers, saving buffer

;;; ........................................................... &timer ...

;;; ----------------------------------------------------------------------
;;;
(defun timy-mail-lock-check  ()
  "Update mode line whenever there is a mail lock.
If there is a lock, then you can't receive mail. Please contact your
sysadm if this happens often and your lock file doesn't go a way within short
period of time."
  (interactive)
  (let* ((file (format
                "%s.lock"
                (or (getenv "MAIL")
                    (error "\
Don't know where your MAIL comes. Please setenv MAIL to directory name")
                     )))
         (lock (file-exists-p file))
         )
    (when lock
      (ding)
      (read-char-safe-until "[press] timy: There is mail Lock!")
      )
    (if (interactive-p)
        (message (if lock "Lock" "no lock")))
    ))


;;; ----------------------------------------------------------------------
;;; Idea by 14 Mar 1996, kevinr@ihs.com (Kevin Rodgers), gnu.emacs.help
;;; Adjusted by [jari]
;;;
(defun timy-save-buffers-periodically ()
  "*Periodically save selected buffers.
References:
  `timy-:save-buffer-modes'
  `timy-:save-buffer-regexp'"
  (interactive "P")
  (let ((buffers (buffer-list))
        buffer
        )
    (list-loop buffers buffer
      (save-excursion
        (when (and (set-buffer-safe buffer)
                   (buffer-modified-p)
                   (or (memq major-mode timy-:save-buffer-modes)
                       (string-match (or (buffer-file-name) "fooBarQuu")
                                     (or timy-:save-buffer-regexp
                                         ;; this won't ever match
                                         "#@123#"))
                       ))
	  (message "Periodic save:  %s" (buffer-name))
          (save-buffer)
          )))
    ))

;;}}}
;;{{{ word movement

;;; --------------------------------------------------- &word-movement ---
;;;
(defun timy-word-move-1 (&optional back)
  "Low level word movement control. Optionally move BACK."
  (let* ((up-case (memq major-mode timy-:move-word-case-modes))

         (case-fold-search (not up-case))

         (charset
          (cond
           ((and up-case
                 (looking-at "[a-z0-9]*[A-Z]+[a-z0-9]+[A-Z]+"))
            ;;  This Upcase charset is only used if the cursor is
            ;;  within AnUpCaseWord.
            timy-:move-word-case-set)
           (t
            timy-:move-word-set
            )))
         )
    (cond
     ((and (null back)
           (looking-at "\\([A-Za-z]+\\)[^a-zA-Z \t]*$"))
      ;;  Skip to the end of word if at EOL
      ;;  (this-he-is-word sse-it-now?)
      ;;                          *     cursor here
      ;;                            *   after
      ;;  otherwise it would skip to next line's word begin.
      (goto-char (match-end 1))
      )
     (t
      (ti::m-word-move charset back)
      ))))

(defun timy-backward-word ()
  "Word backward See `timy-:move-word-case-set'."
  (interactive)

  (timy-word-move-1 'back))


(defun timy-forward-word ()
  "Word forward. See `timy-:move-word-case-set'."
  (interactive)
  (timy-word-move-1))

;;}}}
;;{{{ minor mode: sort

;;; ...................................................... &sort-minor ...

(defvar timy-sort-mode  nil
  "Minor mode on or off variable.")
(make-variable-buffer-local 'timy-sort-mode)

(defvar timy-:sort-mode-map  nil
  "Keymap.")

(defvar timy-:sort-mode-name  " Tsort"
  "Mode name.")

(defcustom timy-:sort-mode-prefix-key  "`"
  "*Mode prefix key."
  :type '(string :tag "Key sequence") :group 'TinyMy
  )


(defcustom timy-:sort-mode-define-keys-hook
  'timy-sort-mode-define-keys-default
  "*Functions to call to define keys for the mode."
  :type 'hook :group 'TinyMy
  )

(add-hook 'timy-:sort-mode-define-keys-hook ;To be sure
	  'timy-sort-mode-define-keys-default)

(defcustom timy-:sort-mode-hook  nil
  "*Hook run when mode is called."
  :type 'hook :group 'TinyMy
  )


;;; ----------------------------------------------------------------------
;;;
(defun timy-sort-mode-define-keys-default  ()
  "Define keys."
  (interactive)
  (let* ((map timy-:sort-mode-map)
	 (p   timy-:sort-mode-prefix-key)
         )
    (if (eq 1 (length p))
	(define-key map (concat p p) 'self-insert-command))

    (define-key map (concat p "1")  'timy-sort-column-1)
    (define-key map (concat p "2")  'timy-sort-column-2)
    (define-key map (concat p "3")  'timy-sort-column-3)
    (define-key map (concat p "4")  'timy-sort-column-4)
    (define-key map (concat p "5")  'timy-sort-column-5)
    (define-key map (concat p "6")  'timy-sort-column-6)
    (define-key map (concat p "7")  'timy-sort-column-7)
    (define-key map (concat p "8")  'timy-sort-column-8)
    (define-key map (concat p "9")  'timy-sort-column-9)
    ))

;;; ----------------------------------------------------------------------
;;;
(ti::m-fmacro-minor-mode-install
 timy-sort-install
 timy-sort-mode
 timy-:sort-mode-map
 timy-:sort-mode-name
 timy-:sort-mode-define-keys-hook
 )

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(ti::m-fmacro-minor-mode
 timy-sort-mode

  "Very simple minor mode for soring lines in the buffer easily.

Mode description:
\\{timy-:sort-mode-map}

"
    timy-sort-install
    timy-sort-mode
    timy-:sort-mode-name

    timy-:sort-mode-prefix-key
    nil
    nil
    "TinySort"

    timy-:sort-mode-hook
    )



;; Create functions
;;
;; (defun timy-sort-column-0 (beg end)
;;    (interactive "*r") (timy-sort-column beg end 0))
;;
(mapcar
 (function
  (lambda (x)
    (let ((sym (intern (format "timy-sort-column-%d" x)))
          def
          )
      (setq def
            (` (defun (, sym) (beg end)
                 (interactive "*r")
                 (timy-sort-column beg end (, x) ))))
      (eval def)
      )))
 '(1 2 3 4 5 6 7 8 9))


;;; ----------------------------------------------------------------------
;;;
(defun timy-sort-column  (beg end nbr)
  "Sort region BEG END according to column NBR."
  (interactive "r\np")
  (untabify (point-min) (point-max))
  (sort-fields nbr beg end)
  )



;;}}}

;;{{{ examples

;;; ........................................................ &examples ...
;; Here is 2 files that you propably want to use in addition to this
;; elisp package.
;;
;; mak      Bourne sh-script, run all .mak files in a directory
;; emc      Bourne sh-script, easy emacs pacakge compile command
;;
;; These files have been tarred and signed with PGP. The PGP is used
;; because it allows ASCII armor and ZIP compression. To unpack this,
;; call M-x timy-install-files
;;
;; file pkg.tar:
;; r--r--r-- 240/222    593 Feb 20 21:20 1997 mak.sh
;; r-xr-xr-x 240/222   1725 Feb 20 20:56 1997 emc
;;
;; -----BEGIN PGP MESSAGE-----
;; Version: 2.6.3ia
;; Comment: base64 signed file pkg.tar. Uncompresses to approx. 10K
;;
;; owHtVkFoHFUY3ioVdlT0YG8t/E62Jk0zszO72cSsxjZNY1MxSXG7WImBvp15m3nN
;; zLztvNlkF8Sr1oMIonjyKIh48KIgPfQmFC8ePAp6FBS8FAUV/N97s5tNTZCiQQrz
;; kezMvPf/7//+//3//96NQvPBo4XqI1+9devLn7+98MSJ2pEj39058uex4x+tffbi
;; 7z98/NTjb7//4a9ff3I7Xm5svDu3/NMfK5M/Tn7wyw3z0cfmLn1z+70T5265nx9/
;; 4847N7+f/e3mQ+yL5gPzbx77dOLh1lGR+iwuICKyZYugcKhwENPT0+pZnXH0szqj
;; ngjXrbgFZ6bqVN1KxXGrOO5WZqcrBedwaWl0RUoSNHmNkDDlB8vxyItIfPB85szw
;; eZ+gbhhjAHAp4ZsJiYD5xpgakDg7MXYKFsuLp0/DCtmiQMIQMF9om4VUgI2vwGLw
;; uklC4xR8luzVLF3066Dza2obXNsFd25utuxUyhUHX+sVpz5dBR13WOp1oDRiu3Se
;; ewz1n6ctqDhpoHSh4tZdF14gCYMFpVbaVVjkcUq8VCo9ew0lbLXw2Q5v8R4Vtsej
;; 54YWGv2YdwQT8rN4UpKE9YWXLjRXllYvb6jBywH62+LbFDzp9g7DnzaLfRWESeW7
;; DsPeCFAv5Ukf9QmKslTrqRUo8QLgMcUh9IZ3UsZjEsLAqo1WjcWV8/Ml1zCoF3Aw
;; pVfLa43LdbgacJHGJKJXTdwu655hhCLj/Bq8auwESBwSSnxp3fC5IS3t2pRbDFYb
;; SjiLYV05bxr7DqNmTA2ZPxSd5W0VECPf//tq/2nkHXqPkf2wVqsd1P+rldla1v/d
;; WfxT/b8q5Q+dWSHv//v1/6Kq36WVhcUGtEiKiYvV08GsmYJOsikTEdat3oaqAP1j
;; 2/ZATZU9ppWq+dqemn+6Xpupu7W/1XzxgGp3UP6uai/eU52v8pQqAy9TiHCnISBY
;; 0SSFkBL8ohHxBNqyq87pKWhRj3QFhZDv0CSbS6iUpNgoikpV0ITxroBWd1NWvkW3
;; sYYFlrQX2CB7Bg99VMY6F1j28XiKij6HVj+lOoRE1r3uAcT3qQ8dkgbCViSbuPq4
;; wBCzbZIiDyY6uz0FmADR7XS4QCWMRYsiAaX2eplKUfX+Cu9qH1HCC0i8iW/IypwQ
;; NL2OnhHfkgZhwuMxHuG2Kb1IA7m2l7COpMva0OfdZD/zPKI7AU0o0FBQ21C9H5qC
;; xZvYKwUPu8haLY/CbQywnB5lFHGftfvaXock6RQMjLVYjFuJQSO4esxTTUtQucKA
;; A07baI3KZMRS3L2pnARsaZikassstZ16Dk2qnu3xRC6hYy0brZxs6GBCo7lqrTW0
;; QQrNmG3TRLAUm7hxZQl5zZe3eVjGl3JPGTCWqs6eYT0ql1yI4CIk3TiWEVFe4l5H
;; fdjhyRYeMl7AsP8vX7KaV8B1zqABdH8dLB/KeB6UQ44HBWw8I3nERnasQcZhKDDC
;; RLrqTg8lB7z2io4Gpc0M49zF1fmSFs3iQKnchh0Kmxh3GQOrl51PsqztAU2z5Jow
;; D6bVMyXJAUsRsHaKT+0KXg8U4YEbxTHwuTwlsdh6DEvujKarWChJ/JbZNDKsyRVR
;; tYH71p8apo8sThXVzDS6Iz3S9EpjSM4Bi8OTigYuNRLLwYmqz+yICZm0U8ATTDZY
;; l8IbWHdU4DFum8aQEIZnbYsgBYyOZhEM0w/kv/xWudvX/omBmjxV8EbAdSZQGKl+
;; W8uM3LoU2VEJLExDzWeRx3kVex2yPfFXdx+pb+lODVlTGt+/5M2sW5i7M6dOjcuI
;; KXVLdior6/ZQmtQsVDz2sWVdB8uKuYXlQi11DgyNw93W4SDz8M/2Bxs9dtddM9//
;; f73///8NJEeOHDly5MiRI0eOHDly5MiRI0eOHDly5MjxX+Mv
;; =UTL4
;; -----END PGP MESSAGE-----

;;}}}

(provide   'tinymy)
(run-hooks 'timy-:load-hook)

;;; tinymy.el ends here
