;; @(#) tinypatch.el --  Emacs 19.28-19.xx std distrib. corrections

;; @(#) $Id: tinypatch.el,v 1.27 1997/06/19 10:10:20 jaalto NotReleased $
;; @(#) $Keywords: emacs distribution, corrections $
;; $KnownCompatibility: 19.28+ $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $packageinstallre: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1995,96,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      May 12 1995
;;
;; To get information on this program use ident(1) or do M-x tipa-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el


;; LCD Archive Entry:
;; tinypatch|Jari Aalto|jari.aalto@poboxes.com|
;; Emacs 19.28 - 19.xx std distrib. corrections |
;; 19-Jun-1997|$Revision: 1.27 $|~/misc/tinypatch.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ....................................................... &t-install ...
;;  Put this file on your Emacs-Lisp load path, add following into your
;;  ~/.emacs startup file, The patches take in effect immediately.
;;
;;     (require 'tinypatch)
;;
;;  This package will handle also autoloaded functions as well as
;;  already loaded functions right. Patched function will have
;;  text 'Patched.' in their documentation string.
;;
;;  If you wish to restore the default functions use command:
;;
;;	(tipa-unpatch)
;;
;;  Warning:
;;
;;      Do not byte compile this file !
;;
;; If you have any questions, use this function
;;
;;      M-x tipa-submit-feedback        ,send feedback, impressions, query


;;}}}
;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVERD prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "tipa-" in front of
;;   every function & variable. It stands for '(ti)ny (pa)tches
;; - variable names contain letters 'tipa-:', except version vars.

;;  Preface, 1995,  Overview of features
;;
;;        [This package is only for Emacs 19.28 and 19.30 and is bit obsolete]
;;
;;	o   Minor corrections to emacs 19.xx distribution.
;;	o   Serious completion error fix in 19.30
;;	o   serious vc.el bugs corrected. Eg rcsdiff,
;;	o   VC symlinked file support for 19.28 and 30.
;;	o   RMAIL "cannot convert to babyl" correction.
;;	o   When Ange created Dires for VAX file, the "f" find-file
;;	    command is fixed.
;;
;;  About patches
;;
;;	In front of every function is explained what the patch corrects or
;;	how does it change the original behaviour. Propably you don't
;;	even realize that the function is changed, until you go back to
;;	old ones.
;;
;;	These patches are ment to be _little_corrections_ only, not total
;;	rewrite of funcs or so.
;;
;;	If you find something similar that should be corrected, let me
;;	hear about it. I only use one side of the emacs, so I propably
;;	have corrected only those things that I use frequently.
;; 	Your name and patch description will be included of cource.
;;
;;  How it works
;;
;;	Emacs has variable after-load-alist, which activates whenever
;;	file beeing loaded is found from there. Autoload functions cause
;;	library to be loaded, so we have a chance to override the
;;	functions by loading our patches immediately after the functions
;;	get defined.
;;
;;	This file hooks itself to that list and after the original library
;;	has been loaded it loads this file again and appropriate
;;	function(s) are replaced with patched ones.If the function is
;;	already in memory, loading this file will immediately replace any
;;	unpatched function.
;;
;;	The unload function does nothing else, than removing the forms
;;	from after-load-alist that cause this file to be loaded.
;;	Then it loads original files again, so that patched functions are
;;	wiped out.
;;
;;  Which functions are patched ?
;;
;;	You can check in emacs which functions are patched by
;;	looking at the variables patched-LIBRARY, where library is some
;;	emacs .el file. The variable holds all functions that are patched
;;	for that library.
;;
;;	There is also function tipa-patched-p for programmers, which
;;	returns non-nil if the function is patched.

;;}}}
;;{{{ History

;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; Jun	19	1997	[jari]		19.28	v1.26-1.27	NotReleased
;; - Better byte compilation message.
;;
;; Jan	15	1997	[jari]		19.28	v1.24-25	NotReleased
;; - Textual changes. Moved dired/ange patches to tinydired.el
;;
;; Jun	13	1996	[jari]		19.28	v1.22		NotReleased
;; - Added more characters to tipa-:completion-1930-disable-list.
;;   How many we must add to there  to prevent nasty errors?
;;
;; Jun	12	1996	[jari]		19.28	v1.21		NotReleased
;; - RMAIL "cannot convert to babyl" in 19.28, corrected.
;;
;; Jun	6	1996	[jari]		19.28	v1.20		NotReleased
;; - in 19.28 vc-do-command there was serious error that caused
;;   "*** empty log message ***" to be written while Checking In.
;; - The "-m" message option was mysteriously deleted from the ci
;;   args list. Now it's ok again.
;;
;; May	31	1996	[jari]		19.28	v1.19		NotReleased
;; - Uses ti::m-ad-control now for advice control.
;;   use-completion-under-or-before-point	:+ 19.30 fix
;;
;; May	28	1996	[jari]		19.28	v1.18		NotReleased
;; - Started getting weird ange-ftp errors, which seemed to be
;;   due to 'nil' values in some ange variables. Added advice to
;;   fix things.
;;   ange-ftp-set-binary-mode		:+
;;
;; May	25	1996	[jari]		19.28	v1.17		NotReleased
;; - Added 19.30 completion error fixes.
;;   tipa-:advice-table			:+
;;   tipa-:advice-regexp		:+
;;   tipa-advice-control		:+
;;   tipa-completion-1930-disable-p	:+
;;   use-completion-before-separator	:+
;;   completion-separator-self-insert-autofilling	:+
;;
;; May	8	1996	[jari]		19.28	v1.15		NotReleased
;; - Added 19.30 vc symlink support. 19.31 is reported to have it
;;   already, but at the time of writing it's not realeased yet.
;;   tipa-vc-enable-symlinks		:+
;;   vc-do-command			:+
;;   vc-simple-command			:+
;;   vc-registered			:+
;;
;; Mar	8	1996	[jari]		19.28	v1.14		NotReleased
;; - Corrected VC.EL patches.
;;
;; Feb	18	1996	[jari]		19.28	v1.11-1.13	NotReleased
;; - Added corrections to vc.el. some variables didn't have 'tipa-:'
;;   prefix. Corrected. Removed '19.30 fdb' notes.
;; - Now patches are emacs version specific. See function
;;   vc-next-action-dired
;; - eval-when-compile now generates error if tried.
;;
;; Feb	10	1996	[jari]		19.28	v1.10		NotReleased
;; - lisp-mnt.el's lm-verify checked. New Feedback. ByteComp almost ok.
;;
;; Sep	10	1995	[jari]		19.28	v1.9		NotReleased
;; - Dired + Ange ftp error corrections, now can load from vax
;;   and used can pick VAX file from ange's dired output with "f".
;; - Added documentation for the package fdb.el, removed patched that
;;   can be corrected with it [files.el, simple.el]
;;
;; Jun	8	1995	[jari]		19.28	v1.8		Release_2
;; - Added completion.el, mouse.el  patches.
;;
;; May	12	1995	[jari]		19.28	v1.3-1.7	NotReleased
;; - Byte compiler warnings corrected, now compiles nicely.
;; - Corrected form variables: because of copying the variable that got
;;   defined was 'patched-dabbrev' , when it should have been
;;   patched-LIBRARY --> The patches didn't take effect properly.
;; - overload allow function had extra test for variable definition,
;;   removed, because a) if funtion was in memory b) the variable never
;;   got defined, unless the file were loaded again.
;;
;; May	12	1995	[jari]		19.28	v1.2		Release_1
;; - Made complete package after testing.
;;
;; May	12	1995	[jari]		19.28	v1.1		Created



;;}}}


;;; Code:

;;{{{ private

;;; ......................................................... &require ...

(require 'advice)
(autoload 'ange-ftp-chase-symlinks "ange-ftp")

(eval-and-compile
  (if (boundp 'xemacs-logo)
      (error "tinypatch: ** This package is only for Emacs. Abort."))
  )

(eval-when-compile
  (message "\n\
tinypatch: ** You will see ~100 messages from byteCompiler
           ** about non defined functions and variables, just ignore
	   ** those safely.
           **
           ** In general, you shouldn't compile this package.")
  )


;;; .......................................................... &v-bind ...

;;; ......................................................... &v-hooks ...

(defvar tipa-load-hook nil
  "*Hook run when file has been loaded.")


;;; ... private variables ................................. &v-private ...

(defconst tipa-:completion-1930-disable-list
  '(?\( ?\) ?\[ ?\] ?\{ ?\} ?\" ?\' ?/ ?\\ )
  "*Disable 10.30 competion just before these character. This prevents
fatal error to happen.")



(defconst  tipa-:forms-ispell
      '(progn
	 (defconst tipa-:patched-ispell "ispell-command-loop")
	 (load "tinypatch")))
(eval-after-load "ispell"  tipa-:forms-ispell)

(defconst  tipa-:forms-dabbrev
      '(progn
	 (defconst tipa-:patched-dabbrev "dabbrev-expand")
	 (load "tinypatch")))
(eval-after-load "dabbrev"  tipa-:forms-dabbrev)


(defconst  tipa-:forms-mouse
      '(progn
	 (defconst tipa-:patched-mouse "mouse-drag-region")
	 (load "tinypatch")))
(eval-after-load "mouse"  tipa-:forms-mouse)


(defconst  tipa-:forms-completion
      '(progn
	 (defconst tipa-:patched-completion
	   (concat "check-completion-length "
		   "add-completion "
		   "kill-completionext-history-element "
		   "completion-separator-self-insert-autofilling "
		   "use-completion-before-separator "
		   "use-completion-under-or-before-point "
		   ))
	 (load "tinypatch")))
(eval-after-load "completion"  tipa-:forms-completion)


(defconst  tipa-:forms-rmail
      '(progn
	 (defconst tipa-:patched-rmail
	   "rmail-convert-to-babyl-format")
	 (load "tinypatch")))
(eval-after-load "rmail"  tipa-:forms-rmail)


(defconst  tipa-:forms-vc
      '(progn
	 (defconst tipa-:patched-vc
	   (concat
	    "vc-do-command"
	    " vc-next-action-dired"
	    " vc-start-entry"
	    ))
	 (load "tinypatch")))
(eval-after-load "vc"  tipa-:forms-vc)


(defconst  tipa-:forms-vc-hooks
      '(progn
	 (defconst tipa-:patched-vc-hooks
	   (concat
	    "vc-registered"
	    " vc-simple-command"
	    ))
	 (load "tinypatch")))
(eval-after-load "vc-hooks"  tipa-:forms-vc-hooks)


(defconst tipa-:advice-regexp "^tipa"
  "*Advices of tinypatch.el")

(defconst tipa-:advice-table
  '(
    completion-separator-self-insert-autofilling
    use-completion-under-or-before-point
    use-completion-before-separator
    )
  "tinypatch.el advices.")



;;}}}

;;; ... user configurable .................................. &v-config ...

(defvar tipa-vc-enable-symlinks t
 "*VC, Follow symlinks to find true filename.")


;;{{{ version

;;; ... version info ...................................... &v-version ...

(defconst tipa-version-id
  "$Id: tinypatch.el,v 1.27 1997/06/19 10:10:20 jaalto NotReleased $"
  "Latest modification time and version number.")


;;; ----------------------------------------------------------------------
;;;

(defun tipa-version (&optional arg)
  "version information."
  (interactive "P")
  (require 'tinylib)
  (ti::m-pkg-version-info "tinypatch.el" arg))

;;; ----------------------------------------------------------------------
;;;
(defun tipa-submit-feedback ()
  "Submit suggestions, error corrections, impressions, anything..."
  (interactive)
  (require 'tinylib)
  (ti::m-pkg-submit-feedback "tinypatch.el"))

;;}}}

;;; ########################################################### &Funcs ###

(eval-when-compile
  ;;  Prevent accidental M-x byte-compile...
  ;;
;;;  (if (string-match "byte\\|minibuffer" (prin1-to-string this-command))
;;;      (error "tinypatch.el is not designed to be compiled.")


  )


;;{{{ misc


;;; ----------------------------------------------------------------------
;;;
(defmacro tipa-autoload-p (fun)
  "Tests if function is in its autoload form."
  (list
   'and
   (list 'fboundp fun)
   (list 'listp fun)
   (list 'eq ''autoload (list 'car (list 'symbol-function fun)))))

;;; ----------------------------------------------------------------------
;;;
(defun tipa-patched-p (func)
  "Checks if fuction is patched. Arg must be SYMBOL."
  (and (fboundp func)
       (string-match
	"Patched"
	(or (documentation func) ""))))

;;; ----------------------------------------------------------------------
;;;
(defun tipa-overload-allow (func func-patch-var &optional version-re)
  "Determines if FUNCTION should be overloaded NOW."
  (if (integerp version-re)
      (setq version-re (int-to-string version-re)))

  (and (or (eq version-re nil)		;version does not count
	   (string-match
	    version-re
	    (int-to-string emacs-minor-version)))
       (fboundp func)
       (not (tipa-autoload-p func))	;not in autoload state ?
;;;       (boundp func-patch-var)		;it's loaded
;;;       (null (tipa-patched-p func))	;already patched...
       ))

;;}}}
;;{{{ patching


;;; ----------------------------------------------------------------------
;;;
(defun tipa-advice-control (&optional disable)
  "Lowlevel patch control. If called with arg UNadvices all needed funcs."
  (ti::m-ad-control tipa-:advice-table tipa-:advice-regexp disable 'verb)
  )

;;; ----------------------------------------------------------------------
;;;
(defun tipa-patch-1 (&optional disable)
  "Lowlevel patch control. If called with arg UNpatches all funcs."
  (cond
   (disable
    (tipa-cleanup  tipa-:forms-ispell)	    (load "ispell")
    (tipa-cleanup  tipa-:forms-dabbrev)	    (load "dabbrev")
    (tipa-cleanup  tipa-:forms-mouse)	    (load "mouse")
    (tipa-cleanup  tipa-:forms-completion)  (load "completion")
    (tipa-cleanup  tipa-:forms-vc)          (load "vc")
    (tipa-cleanup  tipa-:forms-vc-hooks)    (load "vc-hooks")
    (tipa-cleanup  tipa-:forms-rmail)       (load "rmail")
    (tipa-advice-control disable)
    )
   (t
    (load "tinypatch")			;puts the forms back
    )))


;;; ----------------------------------------------------------------------
;;;
(defun tipa-cleanup (ELT)
  "Removes load form ELT from after-load-alist."
  (let* (ptr
	 fn
	 forms
	 frm
	 el
	 prev
	 base
	 )
    (setq ptr after-load-alist)

    (while ptr
      (setq el (car ptr)  fn (car el) forms (cdr el))
;;;      (d! ">>" fn (prin1-to-string forms))
      (setq prev forms base forms)	;here is the FORMS list

      (while forms
	(setq frm (car forms))
	;; change form to nil
	(if (equal frm ELT)
	    (setcar forms nil))
	(setq prev forms)
	(setq forms (cdr forms))
	)
      (setq ptr (cdr ptr))
      )
    ))




;;; ----------------------------------------------------------------------
;;;
(defun tipa-unpatch ()
  "Will discard all patches, and restore default functions."
  (interactive)
  (tipa-patch-1 'disable))

;;; ----------------------------------------------------------------------
;;;
(defun tipa-patch ()
  "Will patch the functions again."
  (interactive)
  (tipa-patch-1 nil))

;;}}}


;;; ......................................................... &patches ...

;;{{{ ispell.el

;;; .......................................................... &ispell ...
;;; - If you accidentally happen to touch mouse, this breaks,
;;;   because it uses read-char command --> fix it GNU ISPELL version 4.
;;;
;;; - If you load file "levents.el" you can now use mouse with ispell
;;;    Mouse-2                  ,No
;;;    C-mouse2 and mouse-3     ,Abort

;;; ----------------------------------------------------------------------
;;;
;;; See ispell.el, this is almost the same as version check in there.
(defun tipa-ispell-version ()
  "Patched."
  (interactive)
  (let* ((case-fold-search t)
	 (bp (get-buffer-create " *tmp*"))
	 status
	 ret
	 )
    (if (not (boundp 'ispell-program-name))
	nil				;ispell not loaded
      (save-excursion
	(set-buffer bp)
	(erase-buffer)
	(call-process ispell-program-name nil t nil "-v")

	(cond
	 ((not (memq status '(0 nil)))
	  (error "%s exited with %s %s" ispell-program-name
		 (if (stringp status) "signal" "code") status))
	 (t
	  (goto-char (point-min))
	  (if (re-search-forward "version[ \t]+\\([0-9.]+\\)" nil t)
	      (setq ret
		    (buffer-substring
		     (match-beginning 1) (match-end 1))))
	  (if (and ret (interactive-p)) (message ret))
	  ))))
      ret
      ))


(defconst tipa-ispell 	 (or (tipa-ispell-version) "")
  "Ispell version in use.")

;;; ----------------------------------------------------------------------
;;;
(if (and (tipa-overload-allow 'ispell-command-loop 'tipa-:patched-ispell 28)
	 (string-match "^4" tipa-ispell))
    (progn
(defun ispell-command-loop (word start end message)
  "Patched."
  (let ((flag t)
        (rescan nil)
        first-line)
    (if (null message)
        (setq first-line (concat "No near misses for '" word "'"))
      (setq first-line (concat "Near misses for '" word "'")))
    (ispell-highlight start end)
    (while flag
      (ispell-show-choices word message first-line)
      (message "Ispell command: ")
      (undo-boundary)
      (let (c                           g;was (downcase (read-char)))
	    event
	    (event-ch-ignore ?Z)	;pass thru command, mouse movement ...
	    ev-ch
            replacement
	    (EVENT (fboundp 'event-to-character)) ;can events be used ?
	    )
	(if (not EVENT)
	    (setq c (downcase (read-char)))
	  (discard-input)
	  (setq event (read-event))

	  ;;  Mouse generates lists...
	 (setq ev-ch
	       (if (not (listp event))
		   nil			;regular keypress
		 ;; Handle mouse
		 (sit-for 0.1)		;slow down a bit!
		 (discard-input)
		 (setq ev-ch (car event))))

	  (cond
	   ((memq ev-ch '(down-mouse-2 mouse-2)) ;for NO
	    (setq c ?\ ))
	   ((or (memq ev-ch '(down-mouse-3 mouse-3)) ;for ABORT
		(memq ev-ch '(C-down-mouse-2 mouse-3))) ; in case 3 not exist
	    (setq c ?q))
	   (t
	    (setq c (or (event-to-character event) event-ch-ignore))))
	  )

        (cond ((and (>= c ?0)
                    (<= c ?9)
                    (setq replacement (nth (- c ?0) message)))
               (ispell-replace start end replacement)
               (setq flag nil))
              ((= c ?q)
               (throw 'ispell-quit nil))
              ((= c (nth 3 (current-input-mode)))
               (keyboard-quit))
              ((= c ? )
               (setq flag nil))
              ((= c ?r)
               (ispell-replace start end (read-string "Replacement: "))
               (setq rescan t)
               (setq flag nil))
              ((= c ?i)
               (ispell-insert word)
               (setq flag nil))
              ((= c ?a)
               (ispell-accept word)
               (setq flag nil))
              ((= c ?l)
               (let ((val (ispell-do-look word)))
                 (setq first-line (car val))
                 (setq message (cdr val))))
              ((= c ??)
               (message
                "Type 'C-h d ispell' to the emacs main loop for more help")
               (sit-for 2))
	      ((= c event-ch-ignore) nil)
              (t
               (message "Bad ispell command")
               (sit-for 2)))))
    rescan)
  (sit-for 0.4)
  (discard-input)			;mouse-2 paste! No!
  )
;;; --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++--
))



;;}}}
;;{{{ dabbrev.el

;;; ......................................................... &dabbrev ...
;;; 19.28 std distribution
;;; dabbrev.el
;;; - I hate when I have debug on error ON , and it says "No dynamic abbrevs"
;;;   --> that is _not_ error condition
;;;
;;; - Man.. , this function should have been much shorter..
;;;


(if (tipa-overload-allow 'dabbrev-expand 'tipa-:patched-dabbrev 28)
    (progn
(defun dabbrev-expand (arg)
  "Patched. Expand previous word \"dynamically\".
Expands to the most recent, preceding word for which this is a prefix.
If no suitable preceding word is found, words following point are considered.

If `case-fold-search' and `case-replace' are non-nil (usually true)
then the substituted word may be case-adjusted to match the abbreviation
that you had typed.  This takes place if the substituted word, as found,
is all lower case, or if it is at the beginning of a sentence and only
its first letter was upper case.

A positive prefix arg N says to take the Nth backward DISTINCT
possibility.  A negative argument says search forward.  The variable
`dabbrev-backward-only' may be used to limit the direction of search to
backward if set non-nil.

If the cursor has not moved from the end of the previous expansion and
no argument is given, replace the previously-made expansion
with the next possible expansion not yet tried."
  (interactive "*P")
  (let (abbrev expansion old which loc n pattern
        (do-case (and case-fold-search case-replace))
	first
	)
    ;; abbrev -- the abbrev to expand
    ;; expansion -- the expansion found (eventually) or nil until then
    ;; old -- the text currently in the buffer
    ;;    (the abbrev, or the previously-made expansion)
    ;; loc -- place where expansion is found
    ;;    (to start search there for next expansion if requested later)
    ;; do-case -- non-nil if should transform case when substituting.
    (catch 'giveup
      (save-excursion
	(if (and (null arg)
		 (eq last-command this-command)
		 last-dabbrevs-abbrev-location)
	    (progn
	      (setq abbrev last-dabbrevs-abbreviation)
	      (setq old last-dabbrevs-expansion)
	      (setq which last-dabbrevs-direction))
	  (setq which (if (null arg)
			  (if dabbrevs-backward-only 1 0)
                        (prefix-numeric-value arg)))
	  (setq loc (point))
	  (forward-word -1)
	  (setq last-dabbrevs-abbrev-location (point)) ; Original location.
	  (setq abbrev (buffer-substring (point) loc))
	  (setq old abbrev)
	  (setq last-dabbrevs-expansion-location nil)
	  (setq last-dabbrev-table nil))          ; Clear table of things seen.

	(setq pattern (concat "\\b" (regexp-quote abbrev) "\\(\\sw\\|\\s_\\)+"))
	;; Try looking backward unless inhibited.
	(if (>= which 0)
          (progn
            (setq n (max 1 which))
            (if last-dabbrevs-expansion-location
                (goto-char last-dabbrevs-expansion-location))
            (while (and (> n 0)
                        (setq expansion (dabbrevs-search pattern t do-case)))
              (setq loc (point-marker))
              (setq last-dabbrev-table (cons expansion last-dabbrev-table))
              (setq n (1- n)))
            (or expansion
                (setq last-dabbrevs-expansion-location nil))
            (setq last-dabbrevs-direction (min 1 which))))

      (if (and (<= which 0) (not expansion)) ; Then look forward.
          (progn
            (setq n (max 1 (- which)))
            (if last-dabbrevs-expansion-location
                (goto-char last-dabbrevs-expansion-location))
            (while (and (> n 0)
                        (setq expansion (dabbrevs-search pattern nil do-case)))
              (setq loc (point-marker))
              (setq last-dabbrev-table (cons expansion last-dabbrev-table))
              (setq n (1- n)))
            (setq last-dabbrevs-direction -1))))

    (if expansion nil
      (setq first (string= abbrev old))
      (setq last-dabbrevs-abbrev-location nil)
      (if (not first)
	  (progn (undo-boundary)
		 (search-backward old)
		 (if (eq major-mode 'picture-mode)
		     (picture-replace-match abbrev t 'literal)
		   (replace-match abbrev t 'literal))))


      (message
       (format
	(if first
	    "No dynamic expansion for \"%s\" found."
	  "No further dynamic expansions for \"%s\" found.")
	abbrev))
      (throw 'giveup t)
      )

    (if (not expansion) nil
      ;; Success: stick it in and return.
      (undo-boundary)
      (search-backward old)
      ;; Make case of replacement conform to case of abbreviation
      ;; provided (1) that kind of thing is enabled in this buffer
      ;; and (2) the replacement itself is all lower case.
      ;; First put back the original abbreviation with its original
      ;; case pattern.
      (save-excursion
        (if (eq major-mode 'picture-mode)
            (picture-replace-match abbrev t 'literal)
          (replace-match abbrev t 'literal)))
      (search-forward abbrev)
      (let ((do-case (and do-case
                          (string= (substring expansion 1)
                                   (downcase (substring expansion 1))))))
        ;; First put back the original abbreviation with its original
        ;; case pattern.
        (save-excursion
          (replace-match abbrev t 'literal))
;;; This used to be necessary, but no longer,
;;; because now point is preserved correctly above.
;;;     (search-forward abbrev)
        (if (eq major-mode 'picture-mode)
            (picture-replace-match (if do-case (downcase expansion) expansion)
                                   (not do-case)
                                   'literal)
          (replace-match (if do-case (downcase expansion) expansion)
                         (not do-case)
                         'literal)))
      ;; Save state for re-expand.
      (setq last-dabbrevs-abbreviation abbrev)
      (setq last-dabbrevs-expansion expansion)
      (setq last-dabbrevs-expansion-location loc)
      )

    )))
;;; --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++--
))


;;}}}

;;{{{ completion.el
;;; ...................................................... &completion ...
;;; - I don't want to see errors when the completion string is too short,
;;;   messag is sufficent


(if (tipa-overload-allow 'check-completion-length 'tipa-:patched-completion 28)
    (progn
(defun check-completion-length (string)
  "Patched. 19.28"
  (if (< (length string) completion-min-length)
      (progn
	(message "The string \"%s\" is too short to be saved as a completion."
		 string)
	nil)
      (list string)))
;;; --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++--
))

(if (tipa-overload-allow 'add-completion 'tipa-:patched-completion 28)
    (progn
(defun add-completion (string &optional num-uses last-use-time)
  "Patched. 19.28
Add STRING to completion list, or move it to head of list.
The completion is altered appropriately if num-uses and/or last-use-time is
specified."
  (interactive (interactive-completion-string-reader "Completion to add"))
  (if (null (check-completion-length string))
      nil
    (let* ((current-completion-source (if (interactive-p)
					  cmpl-source-interactive
                                        current-completion-source))
	   (entry (add-completion-to-head string)))

      (if num-uses (set-completion-num-uses entry num-uses))
      (if last-use-time
	  (set-completion-last-use-time entry last-use-time))
      )))
;;; --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++--
))


(if (tipa-overload-allow 'kill-completion 'tipa-:patched-completion 28)
    (progn
(defun kill-completion (string)
  "Patched. 19.28"
  (interactive (interactive-completion-string-reader "Completion to kill"))
  (if (check-completion-length string)
      (delete-completion string))
  )
;;; --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++--
))


;;; ----------------------------------------------------------------------
;;;
(if (tipa-overload-allow 'complete 'tipa-:patched-completion 28)
    (progn
(defun complete (&optional arg)
  "Patched. 19.28
Fill out a completion of the word before point.
Point is left at end.  Consecutive calls rotate through all possibilities.
Prefix args ::
  control-u :: leave the point at the beginning of the completion rather
               than at the end.
  a number  :: rotate through the possible completions by that amount
  `-'       :: same as -1 (insert previous completion)
 {See the comments at the top of `completion.el' for more info.}"
  (interactive "*p")
  ;;; Set up variables
  (catch 'quit
    (cond ((eq last-command this-command)
	   ;; Undo last one
	   (delete-region cmpl-last-insert-location (point))
	   ;; get next completion
	   (setq cmpl-current-index (+ cmpl-current-index (or arg 1)))
	   )
	  (t
	   (if (not cmpl-initialized-p)
	       (initialize-completions));; make sure everything's loaded
	   (cond ((consp current-prefix-arg);; control-u
		  (setq arg 0)
		  (setq cmpl-leave-point-at-start t)
		  )
		 (t
		  (setq cmpl-leave-point-at-start nil)
		  ))
	   ;; get string
	   (setq cmpl-original-string (symbol-before-point-for-complete))
	   (cond ((not cmpl-original-string)
		  (setq this-command 'failed-complete)
		  (message
		   (format
		    (concat
		     "To complete, the point must"
		     "be after a symbol at least %d "
		     "character long.")
		     completion-prefix-min-length))
		  (throw 'quit t)
		  ))
	   ;; get index
	   (setq cmpl-current-index (if current-prefix-arg arg 0))
	   ;; statistics
	   (cmpl-statistics-block
	    (note-complete-entered-afresh cmpl-original-string))
	   ;; reset database
	   (completion-search-reset cmpl-original-string)
	   ;; erase what we've got
	   (delete-region cmpl-symbol-start cmpl-symbol-end)
	   ))

    ;; point is at the point to insert the new symbol
    ;; Get the next completion
    (let* ((print-status-p
	    (and (>= baud-rate completion-prompt-speed-threshold)
		 (not (minibuffer-window-selected-p))))
	   (insert-point (point))
	   (entry (completion-search-next cmpl-current-index))
	   string
	   )
      ;; entry is either a completion entry or a string (if cdabbrev)

      ;; If found, insert
      (cond (entry
	     ;; Setup for proper case
	     (setq string (if (stringp entry)
			      entry (completion-string entry)))
	     (setq string (cmpl-merge-string-cases
			   string cmpl-original-string))
	     ;; insert
	     (insert string)
	     ;; accept it
	     (setq completion-to-accept string)
	     ;; fixup and cache point
	     (cond (cmpl-leave-point-at-start
		    (setq cmpl-last-insert-location (point))
		    (goto-char insert-point))
		   (t;; point at end,
		    (setq cmpl-last-insert-location insert-point))
		   )
	     ;; statistics
	     (cmpl-statistics-block
	      (note-complete-inserted entry cmpl-current-index))
	     ;; Done ! cmpl-stat-complete-successful
	     ;;display the next completion
	     (cond
	      ((and print-status-p
		    ;; This updates the display and only prints if there
		    ;; is no typeahead
		    (sit-for 0)
		    (setq entry
			  (completion-search-peek
			   completion-cdabbrev-prompt-flag)))
	       (setq string (if (stringp entry)
				entry (completion-string entry)))
	       (setq string (cmpl-merge-string-cases
			     string cmpl-original-string))
	       (message "Next completion: %s" string)
	       ))
	     )
	    (t;; none found, insert old
	     (insert cmpl-original-string)
	     ;; Don't accept completions
	     (setq completion-to-accept nil)
	     ;; print message
	     ;; This used to call cmpl19-sit-for, an undefined function.
	     ;; I hope that sit-for does the right thing;don't know, rms.
	     (if (and print-status-p (sit-for 0))
		 (message "No %scompletions."
			  (if (eq this-command last-command) "more " "")))
	     ;; statistics
	     (cmpl-statistics-block
	      (record-complete-failed cmpl-current-index))
	     ;; Pretend that we were never here
	     (setq this-command 'failed-complete)
	     )))))
;;; --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++--
))






;;; ----------------------------------------------------------------------
;;;
(defun tipa-completion-1930-disable-p ()
  "Check if completion should be disabled to prevent error."
;;;  (d! (point) (char-to-string (char-after 1)))
  (or (memq (following-char)    tipa-:completion-1930-disable-list)
      (memq (char-after 1)	tipa-:completion-1930-disable-list)
      ))



;;; ----------------------------------------------------------------------
;;; RMS reported a fix in 19.31
;;;
(if (tipa-overload-allow 'completion-separator-self-insert-autofilling
			 'tipa-:patched-completion 30)
    (progn
(defadvice completion-separator-self-insert-autofilling
  (around tipa-fault-fix act)
  "Patched. 19.30.
Prevents calling function if error is likely to happen."
  (cond
   ((tipa-completion-1930-disable-p)
    (self-insert-command arg))
   (t
    ad-do-it
    )))
))

;;; ----------------------------------------------------------------------
;;; RMS reported a fix in 19.31
;;;
(if (tipa-overload-allow 'use-completion-before-separator
			 'tipa-:patched-completion 30)
    (progn
(defadvice use-completion-before-separator (around tipa-fault-fix act)
  "Correct function. Prevents calling function if error is likely to happen."
  (cond
   ((tipa-completion-1930-disable-p)
    nil)
   (t
    ad-do-it
    )))
))

;;; ----------------------------------------------------------------------
;;; RMS reported a fix in 19.31
;;;
(if (tipa-overload-allow 'use-completion-under-or-before-point
			 'tipa-:patched-completion 30)
    (progn
(defadvice use-completion-under-or-before-point  (around tipa-fault-fix  act)
  "Correct function. Prevents calling function if error is likely to happen."
  (cond
   ((tipa-completion-1930-disable-p)
    nil)
   (t
    ad-do-it
    )))

))


;;}}}
;;{{{ mouse

;;; ........................................................... &mouse ...
;;; There is one major flaw in this function: it doesn't allow us to move
;;; point when dragging up! Let's correct it...
;;;

(if (tipa-overload-allow 'mouse-drag-region 'tipa-:patched-mouse-drag-region 28)
    (progn
(defun mouse-drag-region (start-event)
  "Patched. Set the region to the text that the mouse is dragged over.
Highlight the drag area as you move the mouse.
This must be bound to a button-down mouse event.
In Transient Mark mode, the highlighting remains once you
release the mouse button.  Otherwise, it does not."
  (interactive "e")
  (mouse-minibuffer-check start-event)
  (let* (sp
	 (start-posn (event-start start-event))
         (start-point (posn-point start-posn))
         (start-window (posn-window start-posn))
         (start-frame (window-frame start-window))
         (bounds (window-edges start-window))
         (top (nth 1 bounds))
         (bottom (if (window-minibuffer-p start-window)
                     (nth 3 bounds)
                   ;; Don't count the mode line.
                   (1- (nth 3 bounds))))
         (click-count (1- (event-click-count start-event))))
    (setq mouse-selection-click-count click-count)
    (mouse-set-point start-event)
    (let ((range (mouse-start-end start-point start-point click-count)))
      (move-overlay mouse-drag-overlay (car range) (nth 1 range)
                    (window-buffer start-window)))
    (deactivate-mark)
    (let (event end end-point)
      (track-mouse
        (while (progn
                 (setq event (read-event))
                 (or (mouse-movement-p event)
                     (eq (car-safe event) 'switch-frame)))
          (if (eq (car-safe event) 'switch-frame)
              nil
            (setq end (event-end event)
                  end-point (posn-point end))

            (cond
              ;; Are we moving within the original window?
             ((and (eq (posn-window end) start-window)
                   (integer-or-marker-p end-point))
              (goto-char end-point)
              (let ((range (mouse-start-end start-point (point) click-count)))
                (move-overlay mouse-drag-overlay (car range) (nth 1 range))))

             (t
              (let ((mouse-row (cdr (cdr (mouse-position)))))
                (cond
                 ((null mouse-row))
                 ((< mouse-row top)
                  (mouse-scroll-subr start-window (- mouse-row top)
                                     mouse-drag-overlay start-point))
                 ((>= mouse-row bottom)
                  (mouse-scroll-subr start-window (1+ (- mouse-row bottom))
                                     mouse-drag-overlay start-point)))))))))


      (if (consp event)
;;; When we scroll into the mode line or menu bar, or out of the window,
;;; we get events that don't fit these criteria.
;;;            (eq (get (event-basic-type event) 'event-kind) 'mouse-click)
;;;            (eq (posn-window (event-end event)) start-window)
;;;            (numberp (posn-point (event-end event)))
          (let ((fun (key-binding (vector (car event))))
		(ovs (overlay-start mouse-drag-overlay))
		(ove (overlay-end mouse-drag-overlay))
		)
            (if (not (= ovs ove))
                (let (last-command this-command)
		  (if (= ovs start-point)
		      (progn (push-mark ovs t t)
			     (goto-char ove))
		    (push-mark ove t t)
		    (goto-char ovs))
                  (copy-region-as-kill (point) (mark t)))
              (goto-char ove)
              (setq this-command 'mouse-set-point))))
      (delete-overlay mouse-drag-overlay))))
;;; --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++-- --++--
))

;;}}}


;;{{{ vc.el, vc-hooks.el

;;; .............................................................. &vc ...
;;; - One important thing:
;;;   put following entries into your .emacs if you're guing to
;;;   use vc at all.

;;* ;;  So that files are detected as VC when C-x C-f
;;* (require 'vc)
;;* (require 'vc-hooks)
;;* _
;;* ;;;  For some funny reason, the above require does not always work,
;;* ;;;  check that modeline is ok --> vc is active ?
;;* (cond
;;*  ((not (assoc 'vc-mode minor-mode-alist))
;;*   (load "vc")
;;*   (load "vc-hooks")))

;;; ----------------------------------------------------------------------
;;; vc.el
;;; - Updated the file RWX state after action.
;;; - C-x v v fails to registed new file in dired...problem in
;;;   vc-next-action-on-file --> vc-register
;;;   where buffer-file-name is nil... don't know how to fix it.
;;;
(if (tipa-overload-allow 'vc-next-action-dired 'tipa-:patched-vc 28)
    (progn
(defun vc-next-action-dired (file rev comment)
  "Patched."
  ;; We've accepted a log comment, now do a vc-next-action using it on all
  ;; marked files.
  (set-buffer vc-parent-buffer)

  (dired-map-over-marks
   (save-window-excursion
     (let ((file (dired-get-filename)))
       (message "Processing %s..." file)
       (vc-next-action-on-file file nil comment)
       (message "Processing %s...done" file)))
   nil t)

  ;; ++
  (dired-do-redisplay)
  )
))



;;; ----------------------------------------------------------------------
;;; vc.el
;;; - There is serious error in 19.28, that prevents using rcsdiff
;;;   in many cases. See ++ note in code
;;;
(if (tipa-overload-allow 'vc-do-command 'tipa-:patched-vc 28)
    (progn
(defun vc-do-command (okstatus command file &rest flags)
  "Patched. 19.28
Execute a version-control command, notifying user and checking for errors.
The command is successful if its exit status does not exceed OKSTATUS.
Output from COMMAND goes to buffer *vc*.  The last argument of the command is
the master name of FILE; this is appended to an optional list of FLAGS."
  (setq file (expand-file-name file))
  (if vc-command-messages
      (message "Running %s on %s..." command file))
  (let ((obuf (current-buffer)) (camefrom (current-buffer))
        (squeezed nil)
        (vc-file (and file (vc-name file)))
        status)
    (set-buffer (get-buffer-create "*vc*"))
    (set (make-local-variable 'vc-parent-buffer) camefrom)
    (set (make-local-variable 'vc-parent-buffer-name)
         (concat " from " (buffer-name camefrom)))

    (erase-buffer)

    ;; This is so that command arguments typed in the *vc* buffer will
    ;; have reasonable defaults.
    (setq default-directory (file-name-directory file))


    ;; ++ The list gets "" element which drives RCSdiff nuts,
    ;; ++ don't accept empty "" string to option list
    ;;
    (mapcar
     (function
      (lambda (s)
	(and  (stringp s)
	      (not (string= "" s))
	      (setq squeezed (append squeezed (list s))))
;;;	(d! "map" s squeezed)
	))
     flags)

;;;    (mapcar
;;;     (function (lambda (s) (and s (setq squeezed (append squeezed (list s))))))
;;;     flags)
;;;



;;;    (d! "SQZ" squeezed)
;;;    (d! "VC F" file vc-file)

    ;; ++ Sometimes rcsdiff gets RCS/file,v
    ;; ++ --> do not allow that.
    (if (and vc-file
	     (string-match "rcsdiff" command)
	     (string-match ",v$" vc-file))
	(setq vc-file file))		;restore

    (if vc-file
        (setq squeezed (append squeezed (list vc-file))))




    (let ((default-directory (file-name-directory (or file "./")))
          (exec-path (if vc-path (append exec-path vc-path) exec-path))
          ;; Add vc-path to PATH for the execution of this command.
          (process-environment
           (cons (concat "PATH=" (getenv "PATH")
                         ":" (mapconcat 'identity vc-path ":"))
                 process-environment)))
;;;      (d! "VC-C" command squeezed)
      (setq status (apply 'call-process command nil t nil squeezed)))


    (goto-char (point-max))
    (forward-line -1)

    (if (or (not (integerp status)) (< okstatus status))
        (progn
          (pop-to-buffer "*vc*")
          (goto-char (point-min))
          (shrink-window-if-larger-than-buffer)
          (error "Running %s...FAILED (%s)" command
                 (if (integerp status)
                     (format "status %d" status)
                   status))
          )
      (if vc-command-messages
          (message "Running %s...OK" command))
      )
    (set-buffer obuf)
    status)
  )


))


;;; ----------------------------------------------------------------------
;;; vc.el, enable symlinks
;;;
(if (tipa-overload-allow 'vc-do-command 'tipa-:patched-vc-hooks 30)
    (progn
(defun vc-do-command (buffer okstatus command file last &rest flags)
  "Patched. 19.30. Execute a version-control command, notifying user and
checking for errors.
Output from COMMAND goes to BUFFER, or *vc* if BUFFER is nil.
The command is successful if its exit status does not exceed OKSTATUS.
The last argument of the command is the master name of FILE if LAST is
`MASTER', or the workfile of FILE if LAST is `WORKFILE'; this is appended
to an optional list of FLAGS."

  (cond
   (file
    (setq file
          (if tipa-vc-enable-symlinks
              (file-chase-links file)
            (expand-file-name file)))
    ))

  (if (not buffer) (setq buffer "*vc*"))
  (if vc-command-messages
      (message "Running %s on %s..." command file))

  (let ((obuf (current-buffer)) (camefrom (current-buffer))
        (squeezed nil)
        (vc-file (and file (vc-name file)))
        (olddir  (if file		;symlink's real dir perhaps?
		     (file-name-directory file)
		   default-directory
		   ))
        status
        )

    (set-buffer (get-buffer-create buffer))
    (set (make-local-variable 'vc-parent-buffer) camefrom)
    (set (make-local-variable 'vc-parent-buffer-name)
         (concat " from " (buffer-name camefrom)))
    (setq default-directory olddir)

    (erase-buffer)

    (mapcar
     (function (lambda (s) (and s (setq squeezed (append squeezed (list s))))))
     flags)

    (if (and vc-file (eq last 'MASTER))
        (setq squeezed (append squeezed (list vc-file))))

    (if (eq last 'WORKFILE)
        (progn
          (let* ((pwd (expand-file-name
                       default-directory))
                 (preflen (length pwd))
                 )
            (if (string= (substring file 0 preflen) pwd)
                (setq file (substring file preflen))))
          (setq squeezed (append squeezed (list file)))))

    (let ((exec-path (append vc-path exec-path))
          ;; Add vc-path to PATH for the execution of this command.
          (process-environment
           (cons (concat "PATH=" (getenv "PATH")
                         path-separator
                         (mapconcat 'identity vc-path path-separator))
                 process-environment))
          )
      (setq status (apply 'call-process command nil t nil squeezed)))

    (goto-char (point-max))
    (set-buffer-modified-p nil)
    (forward-line -1)

    (if (or (not (integerp status)) (< okstatus status))
        (progn
          (pop-to-buffer buffer)
          (goto-char (point-min))
          (shrink-window-if-larger-than-buffer)
          (error "Running %s...FAILED (%s)" command
                 (if (integerp status)
                     (format "status %d" status)
                   status))
          )
      (if vc-command-messages
          (message "Running %s...OK" command))
      )
    (set-buffer obuf)
    status)
  )
))


;;; ----------------------------------------------------------------------
;;; vc-hooks.el
;;; - Follow symlinks too. 19.28
;;; - In newer emacs version you can use (find-file-visit-truename t)
;;;
;;;  vc-next-action-on-file
;;;     ;; if there is no master file corresponding, create one
;;;  (vc-file (vc-name file))     <-- vc-hooks.el , vc-registered
;;;  ((not vc-file)
;;;   (vc-register verbose comment))
;;;
(if (tipa-overload-allow 'vc-registered 'tipa-:patched-vc-hooks 28)
    (progn
(defun vc-registered (file)
  "Patched. 19.28."
  (let (handler handlers)
    (if (boundp 'file-name-handler-alist)
        (setq handler (find-file-name-handler file 'vc-registered)))
    (if handler
        (funcall handler 'vc-registered file)

      ;; Search for a master corresponding to the given file
      ;; ++ Search for a master corresponding to the given file
      ;; ++ chase symlinks

      (let* ((f		(ange-ftp-chase-symlinks file))
	    (dirname	(or (file-name-directory f) ""))
            (basename	(file-name-nondirectory f)))
;;;	(d! dirname basename file )

        (catch 'found
          (mapcar
           (function (lambda (s)
              (let ((trial (format (car s) dirname basename)))
                (if (and (file-exists-p trial)
                         ;; Make sure the file we found with name
                         ;; TRIAL is not the source file itself.
                         ;; That can happen with RCS-style names
                         ;; if the file name is truncated
                         ;; (e.g. to 14 chars).  See if either
                         ;; directory or attributes differ.
;;;				  (d! (car s) trial)

                         (or (not (string= dirname
                                           (file-name-directory trial)))
                             (not (equal
                                   (file-attributes file)
                                   (file-attributes trial)))))
                    (throw 'found (cons trial (cdr s)))))))
           vc-master-templates)
;;;	  (d! "reg. nil")
          nil
	  )))))
))


;;; ----------------------------------------------------------------------
;;; vc-hooks.el, 19.30, enable symlinks
;;;

(if (tipa-overload-allow 'vc-simple-command 'tipa-:patched-vc-hooks 30)
    (progn
(defun vc-simple-command (okstatus command file &rest args)
  "Patched. 19.30."
  ;; Simple version of vc-do-command, for use in vc-hooks only.
  ;; Don't switch to the *vc-info* buffer before running the
  ;; command, because that would change its default directory
  (save-excursion (set-buffer (get-buffer-create "*vc-info*"))
                  (erase-buffer))
  (let* ((ffile    (if tipa-vc-enable-symlinks
                       (file-chase-links file)
                     file))
         (exec-path (append vc-path exec-path))
         exec-status
         ;; Add vc-path to PATH for the execution of this command.
         (process-environment
          (cons (concat "PATH=" (getenv "PATH")
                        path-separator
                        (mapconcat 'identity vc-path path-separator))
                process-environment))
         )

    (setq exec-status
          (apply 'call-process command nil "*vc-info*" nil
                 (append args (list ffile))))

    (cond ((> exec-status okstatus)
           (switch-to-buffer (get-file-buffer ffile))
           (shrink-window-if-larger-than-buffer
            (display-buffer "*vc-info*"))
           (error "Couldn't find version control information")))
    (setq ad-return-value exec-status)
    ))
))


;;; ----------------------------------------------------------------------
;;; vc-hooks.el,19.30 enable symlinks
;;;
(if (tipa-overload-allow 'vc-registered 'tipa-:patched-vc-hooks 30)
    (progn
(defun vc-registered (file)
  "Patched. 19.30."
  (let (handler handlers)
    (if (boundp 'file-name-handler-alist)
        (setq handler (find-file-name-handler file 'vc-registered)))
    (if handler
        (funcall handler 'vc-registered file)
      ;; Search for a master corresponding to the given file
      (let* ((ffile     (if tipa-vc-enable-symlinks
                            (file-chase-links file)
                            file))
             (dirname   (or (file-name-directory ffile) ""))
             (basename  (file-name-nondirectory ffile))
             )

        ;; Update true filename
        ;;
        (vc-file-setprop file 'vc-filename ffile)

        (catch 'found
          (mapcar
           (function (lambda (s)
              (if (atom s)
                  (funcall s dirname basename)
                (let ((trial (format (car s) dirname basename)))
                  (if (and (file-exists-p trial)
                           ;; Make sure the file we found with name
                           ;; TRIAL is not the source file itself.
                           ;; That can happen with RCS-style names
                           ;; if the file name is truncated
                           ;; (e.g. to 14 chars).  See if either
                           ;; directory or attributes differ.
                           (or (not (string= dirname
                                             (file-name-directory trial)))
                               (not (equal
                                     (file-attributes ffile)
                                     (file-attributes trial)))))
                      (throw 'found (cons trial (cdr s))))))))
           vc-master-templates)
          nil)))))

))




;;}}}
;;{{{ RMAIL


;;; ----------------------------------------------------------------------
;;; - Have you seen message "Cannot convert to Babyl?"
;;;   after you have hit the "g" to get new mail.
;;; - Well, i had a hard time to figure out what was wrong with
;;;   the RMAIL. This function is the core of detecting "mail,
;;;   and the bad thing is that it only looked:
;;;
;;;	"^From"
;;;
;;;   as the first header. But My mail had fiedl "^Date" most of the
;;;   time as header. So I just added regexp
;;;
;;;	"^From\\|^Date"
;;;
;;;   and that fixed the problem.
;;;
;;; - I really don't know why this is hardcoded in the function...
;;;
;;;
(if (tipa-overload-allow 'rmail-convert-to-babyl-format
			 'tipa-:patched-rmail "28")
    (progn
(defun rmail-convert-to-babyl-format ()
  "Patched."
  (let ((count 0) start
        (case-fold-search nil)
        (invalid-input-resync
         (function (lambda ()
                     (message "Invalid Babyl format in inbox!")
                     (sit-for 1)
                     ;; Try to get back in sync with a real message.
                     (if (re-search-forward
                          (concat mmdf-delim1 "\\|^From") nil t)
                         (beginning-of-line)
                       (goto-char (point-max)))))))

    (goto-char (point-min))
    (save-restriction
      (while (not (eobp))
	(setq P (point) PM (point-min) PX (point-max))
        (cond ((looking-at "BABYL OPTIONS:");Babyl header

               (if (search-forward "\n\^_" nil t)
                   ;; If we find the proper terminator, delete through there.
                   (delete-region (point-min) (point))
                 (funcall invalid-input-resync)
                 (delete-region (point-min) (point))))

              ;; Babyl format message
              ((looking-at "\^L")
               (or (search-forward "\n\^_" nil t)
                   (funcall invalid-input-resync))
               (setq count (1+ count))
               ;; Make sure there is no extra white space after the ^_
               ;; at the end of the message.
               ;; Narrowing will make sure that whatever follows the junk
               ;; will be treated properly.
               (delete-region (point)
                              (save-excursion
                                (skip-chars-forward " \t\n")
                                (point)))
               (narrow-to-region (point) (point-max)))
              ;;*** MMDF format
              ((let ((case-fold-search t))
                 (looking-at mmdf-delim1))
               (let ((case-fold-search t))
                 (replace-match "\^L\n0, unseen,,\n*** EOOH ***\n")
                 (setq start (point))
                 (re-search-forward mmdf-delim2 nil t)
                 (replace-match "\^_"))
               (save-excursion
                 (save-restriction
                   (narrow-to-region start (1- (point)))
                   (goto-char (point-min))
                   (while (search-forward "\n\^_" nil t); single char "\^_"
                     (replace-match "\n^_")))); 2 chars: "^" and "_"
               (narrow-to-region (point) (point-max))
               (setq count (1+ count)))

              ;;*** Mail format
              ;; ++ modification done here. [jari]
              ;;
              ((looking-at "^From\\|^Date")
               (setq start (point))
               (insert "\^L\n0, unseen,,\n*** EOOH ***\n")
               (rmail-nuke-pinhead-header)
               ;; If this message has a Content-Length field,
               ;; skip to the end of the contents.
               (let* ((header-end (save-excursion
                                    (and (re-search-forward "\n\n" nil t)
                                         (1- (point)))))
                      (case-fold-search t)
                      (size
                       ;; Get the numeric value from the Content-Length field.
                       (save-excursion
                         ;; Back up to end of prev line,
                         ;; in case the Content-Length field comes first.
                         (forward-char -1)
                         (and (search-forward "\ncontent-length: "
                                              header-end t)
                              (let ((beg (point))
                                    (eol (progn (end-of-line) (point))))
                                (string-to-int (buffer-substring beg eol)))))))
                 (and size
                      (if (and (natnump size)
                               (<= (+ header-end size) (point-max))
                               ;; Make sure this would put us at a position
                               ;; that we could continue from.
                               (save-excursion
                                 (goto-char (+ header-end size))
                                 (skip-chars-forward "\n")
                                 (or (eobp)
                                     (and (looking-at "BABYL OPTIONS:")
                                          (search-forward "\n\^_" nil t))
                                     (and (looking-at "\^L")
                                          (search-forward "\n\^_" nil t))
                                     (let ((case-fold-search t))
                                       (looking-at mmdf-delim1))
                                     (looking-at "From "))))
                          (goto-char (+ header-end size))
                        (message "Ignoring invalid Content-Length field")
                        (sit-for 1 0 t))))

               (if (re-search-forward
                    (concat "^[\^_]?\\("
                            rmail-unix-mail-delimiter
                            "\\|"
                            mmdf-delim1 "\\|"
                            "^BABYL OPTIONS:\\|"
                            "\^L\n[01],\\)") nil t)
                   (goto-char (match-beginning 1))
                 (goto-char (point-max)))
               (setq count (1+ count))
               (save-excursion
                 (save-restriction
                   (narrow-to-region start (point))
                   (goto-char (point-min))
                   (while (search-forward "\n\^_" nil t); single char
                     (replace-match "\n^_")))); 2 chars: "^" and "_"
               (insert ?\^_)
               (narrow-to-region (point) (point-max)))
              ;;
              ;; This kludge is because some versions of sendmail.el
              ;; insert an extra newline at the beginning that shouldn't
              ;; be there.  sendmail.el has been fixed, but old versions
              ;; may still be in use.  -- rms, 7 May 1993.
              ((eolp)
	       (delete-char 1))
              (t
	       (error "Cannot convert to babyl format")
	       ))))
    count))


))
;;}}}

(provide 'tinypatch)
(run-hooks 'tipa-load-hook)

;;; tinypatch.el ends here
